# -*- encoding:utf-8; warn-indent:false; frozen_string_literal: true  -*-

# line 1 "lib/parser/lexer.rl"

# line 3 "lib/parser/lexer.rl"
#
# === BEFORE YOU START ===
#
# Read the Ruby Hacking Guide chapter 11, available in English at
# http://whitequark.org/blog/2013/04/01/ruby-hacking-guide-ch-11-finite-state-lexer/
#
# Remember two things about Ragel scanners:
#
#   1) Longest match wins.
#
#   2) If two matches have the same length, the first
#      in source code wins.
#
# General rules of making Ragel and Bison happy:
#
#  * `p` (position) and `@te` contain the index of the character
#    they're pointing to ("current"), plus one. `@ts` contains the index
#    of the corresponding character. The code for extracting matched token is:
#
#       @source_buffer.slice(@ts...@te)
#
#  * If your input is `foooooooobar` and the rule is:
#
#       'f' 'o'+
#
#    the result will be:
#
#       foooooooobar
#       ^ ts=0   ^ p=te=9
#
#  * A Ragel lexer action should not emit more than one token, unless
#    you know what you are doing.
#
#  * All Ragel commands (fnext, fgoto, ...) end with a semicolon.
#
#  * If an action emits the token and transitions to another state, use
#    these Ragel commands:
#
#       emit($whatever)
#       fnext $next_state; fbreak;
#
#    If you perform `fgoto` in an action which does not emit a token nor
#    rewinds the stream pointer, the parser's side-effectful,
#    context-sensitive lookahead actions will break in a hard to detect
#    and debug way.
#
#  * If an action does not emit a token:
#
#       fgoto $next_state;
#
#  * If an action features lookbehind, i.e. matches characters with the
#    intent of passing them to another action:
#
#       p = @ts - 1
#       fgoto $next_state;
#
#    or, if the lookbehind consists of a single character:
#
#       fhold; fgoto $next_state;
#
#  * Ragel merges actions. So, if you have `e_lparen = '(' %act` and
#    `c_lparen = '('` and a lexer action `e_lparen | c_lparen`, the result
#    _will_ invoke the action `act`.
#
#    e_something stands for "something with **e**mbedded action".
#
#  * EOF is explicit and is matched by `c_eof`. If you want to introspect
#    the state of the lexer, add this rule to the state:
#
#       c_eof => do_eof;
#
#  * If you proceed past EOF, the lexer will complain:
#
#       NoMethodError: undefined method `ord' for nil:NilClass
#

class Parser::Lexer

  
# line 85 "lib/parser/lexer.rb"
class << self
	attr_accessor :_lex_trans_keys
	private :_lex_trans_keys, :_lex_trans_keys=
end
self._lex_trans_keys = [
	0, 0, 101, 101, 103, 103, 
	105, 105, 110, 110, 69, 
	69, 78, 78, 68, 68, 
	95, 95, 95, 95, 0, 26, 
	0, 127, 0, 127, 0, 
	127, 0, 127, 0, 45, 
	0, 77, 0, 77, 0, 92, 
	0, 26, 0, 26, 0, 
	45, 0, 99, 0, 26, 
	67, 99, 45, 45, 0, 92, 
	0, 77, 0, 102, 0, 
	127, 0, 127, 0, 127, 
	0, 127, 0, 45, 0, 77, 
	0, 77, 0, 92, 0, 
	26, 0, 26, 0, 45, 
	0, 99, 0, 26, 67, 99, 
	45, 45, 0, 92, 0, 
	77, 0, 102, 0, 127, 
	0, 127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 
	127, 0, 127, 0, 127, 
	0, 26, 0, 127, 58, 58, 
	0, 127, 58, 58, 60, 
	60, 62, 62, 10, 10, 
	0, 127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 
	127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 
	127, 0, 127, 0, 127, 
	0, 127, 0, 127, 115, 115, 
	99, 99, 117, 117, 101, 
	101, 108, 116, 101, 101, 
	115, 115, 115, 115, 105, 105, 
	108, 108, 105, 105, 108, 
	108, 58, 58, 0, 127, 
	10, 10, 0, 127, 58, 58, 
	98, 98, 101, 101, 103, 
	103, 105, 105, 110, 110, 
	0, 122, 61, 61, 0, 127, 
	0, 127, 61, 126, 0, 
	127, 0, 127, 93, 93, 
	0, 127, 0, 127, 10, 10, 
	10, 34, 10, 10, 10, 
	39, 0, 127, 10, 96, 
	0, 45, 0, 77, 0, 77, 
	0, 92, 0, 26, 0, 
	26, 0, 45, 0, 99, 
	0, 26, 67, 99, 45, 45, 
	0, 92, 0, 77, 0, 
	102, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 127, 
	0, 127, 58, 58, 58, 
	58, 0, 127, 43, 57, 
	48, 57, 48, 57, 48, 57, 
	48, 57, 0, 127, 58, 
	58, 9, 92, 9, 92, 
	9, 92, 9, 92, 9, 92, 
	9, 92, 60, 60, 10, 
	10, 9, 46, 46, 46, 
	0, 95, 9, 32, 0, 0, 
	10, 10, 10, 10, 98, 
	98, 9, 32, 10, 10, 
	95, 95, 0, 92, 9, 32, 
	36, 123, 0, 127, 48, 
	57, 0, 120, 0, 0, 
	0, 0, 48, 55, 48, 55, 
	0, 0, 0, 0, 0, 
	92, 0, 0, 0, 0, 
	0, 0, 0, 92, 45, 45, 
	0, 0, 0, 0, 0, 
	0, 0, 92, 0, 45, 
	0, 92, 0, 92, 0, 0, 
	0, 0, 0, 92, 0, 
	45, 10, 10, 0, 92, 
	0, 123, 0, 26, 0, 26, 
	0, 26, 0, 0, 0, 
	102, 0, 102, 0, 102, 
	0, 0, 0, 125, 0, 125, 
	0, 125, 0, 125, 0, 
	125, 0, 0, 0, 125, 
	0, 125, 0, 0, 0, 125, 
	0, 26, 0, 125, 0, 
	125, 0, 125, 0, 125, 
	0, 125, 0, 125, 0, 125, 
	0, 125, 0, 125, 0, 
	0, 0, 125, 0, 0, 
	48, 102, 0, 0, 0, 92, 
	36, 123, 0, 127, 48, 
	57, 0, 120, 0, 0, 
	0, 0, 48, 55, 48, 55, 
	0, 0, 0, 0, 0, 
	92, 0, 0, 0, 0, 
	0, 0, 0, 92, 45, 45, 
	0, 0, 0, 0, 0, 
	0, 0, 92, 0, 45, 
	0, 92, 0, 92, 0, 0, 
	0, 0, 0, 92, 0, 
	45, 10, 10, 0, 92, 
	0, 123, 0, 26, 0, 26, 
	0, 26, 0, 0, 0, 
	102, 0, 102, 0, 102, 
	0, 0, 0, 125, 0, 125, 
	0, 125, 0, 125, 0, 
	125, 0, 0, 0, 125, 
	0, 125, 0, 0, 0, 125, 
	0, 26, 0, 125, 0, 
	125, 0, 125, 0, 125, 
	0, 125, 0, 125, 0, 125, 
	0, 125, 0, 125, 0, 
	0, 0, 125, 0, 0, 
	48, 102, 0, 0, 0, 92, 
	9, 32, 0, 26, 0, 
	92, 0, 26, 0, 35, 
	36, 123, 0, 127, 48, 57, 
	0, 26, 65, 122, 65, 
	122, 36, 64, 0, 127, 
	48, 57, 0, 127, 0, 127, 
	0, 127, 0, 127, 9, 
	32, 0, 0, 61, 126, 
	10, 10, 10, 10, 0, 127, 
	0, 127, 48, 57, 115, 
	115, 38, 38, 42, 42, 
	64, 64, 58, 58, 60, 61, 
	62, 62, 61, 126, 61, 
	61, 61, 62, 0, 127, 
	0, 127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 
	127, 93, 93, 10, 10, 
	0, 127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 
	127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 
	127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 
	127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 
	127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 
	127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 
	127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 
	127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 
	127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 
	127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 
	127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 127, 
	124, 124, 0, 127, 0, 
	127, 9, 32, 10, 10, 
	10, 10, 10, 10, 0, 0, 
	0, 127, 0, 127, 61, 
	61, 0, 0, 9, 32, 
	0, 0, 61, 126, 10, 10, 
	10, 10, 38, 38, 42, 
	42, 64, 64, 60, 61, 
	62, 62, 61, 126, 61, 61, 
	61, 62, 0, 127, 93, 
	93, 10, 10, 124, 124, 
	0, 126, 0, 127, 0, 61, 
	9, 61, 9, 61, 0, 
	0, 9, 61, 9, 62, 
	46, 46, 46, 46, 58, 58, 
	9, 32, 0, 0, 0, 
	127, 0, 0, 9, 124, 
	0, 0, 10, 10, 10, 10, 
	0, 0, 9, 61, 58, 
	58, 60, 60, 62, 62, 
	9, 32, 10, 10, 0, 127, 
	102, 102, 101, 101, 110, 
	110, 104, 104, 0, 127, 
	0, 127, 0, 127, 0, 0, 
	0, 127, 10, 10, 0, 
	123, 9, 32, 10, 10, 
	10, 10, 10, 10, 0, 0, 
	111, 111, 0, 0, 0, 
	127, 0, 127, 9, 32, 
	0, 0, 10, 10, 10, 10, 
	10, 10, 0, 0, 0, 
	127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 
	127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 
	127, 0, 127, 58, 61, 
	0, 0, 61, 126, 61, 61, 
	0, 0, 0, 0, 0, 
	0, 9, 32, 61, 61, 
	9, 32, 61, 126, 10, 10, 
	10, 10, 65, 122, 0, 
	122, 38, 61, 0, 0, 
	42, 61, 61, 61, 48, 61, 
	48, 62, 46, 46, 46, 
	46, 0, 26, 0, 127, 
	0, 127, 61, 61, 0, 0, 
	61, 126, 61, 62, 0, 
	0, 0, 0, 0, 0, 
	0, 0, 61, 126, 0, 127, 
	48, 57, 38, 38, 42, 
	42, 64, 64, 60, 61, 
	62, 62, 61, 61, 61, 62, 
	124, 124, 60, 61, 0, 
	0, 62, 62, 61, 126, 
	61, 62, 0, 122, 0, 0, 
	0, 127, 0, 127, 0, 
	120, 0, 0, 0, 0, 
	48, 55, 48, 55, 0, 0, 
	0, 0, 0, 92, 0, 
	0, 0, 0, 0, 0, 
	0, 92, 45, 45, 0, 0, 
	0, 0, 0, 0, 0, 
	92, 0, 45, 0, 92, 
	0, 92, 0, 0, 0, 0, 
	0, 92, 0, 45, 10, 
	10, 0, 92, 0, 123, 
	0, 26, 0, 26, 0, 26, 
	0, 0, 0, 102, 0, 
	102, 0, 102, 0, 0, 
	0, 125, 0, 125, 0, 125, 
	0, 125, 0, 125, 0, 
	0, 0, 125, 0, 125, 
	0, 0, 0, 125, 0, 26, 
	0, 125, 0, 125, 0, 
	125, 0, 125, 0, 125, 
	0, 125, 0, 125, 0, 125, 
	0, 125, 0, 125, 0, 
	125, 0, 125, 0, 125, 
	0, 125, 0, 0, 0, 0, 
	48, 102, 0, 0, 0, 
	127, 0, 127, 0, 127, 
	0, 0, 10, 10, 0, 0, 
	0, 127, 0, 127, 0, 
	127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 
	127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 
	127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 
	127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 
	127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 
	127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 
	127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 
	127, 0, 127, 61, 126, 
	0, 127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 
	127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 
	127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 
	127, 0, 127, 0, 127, 
	0, 127, 0, 0, 61, 124, 
	0, 92, 9, 32, 0, 
	0, 10, 10, 10, 10, 
	10, 10, 0, 0, 0, 127, 
	0, 127, 9, 32, 0, 
	0, 10, 10, 10, 10, 
	10, 10, 0, 0, 0, 127, 
	0, 127, 61, 61, 0, 
	0, 9, 32, 0, 0, 
	61, 126, 10, 10, 10, 10, 
	0, 127, 0, 127, 48, 
	57, 61, 61, 38, 61, 
	0, 0, 0, 0, 42, 61, 
	61, 62, 46, 57, 46, 
	46, 48, 101, 48, 95, 
	46, 120, 48, 114, 43, 57, 
	48, 105, 0, 0, 105, 
	105, 0, 0, 48, 114, 
	48, 114, 48, 114, 48, 114, 
	105, 114, 0, 0, 105, 
	105, 0, 0, 48, 114, 
	48, 114, 48, 114, 48, 114, 
	48, 114, 48, 114, 48, 
	114, 48, 114, 46, 114, 
	48, 114, 46, 114, 48, 114, 
	58, 58, 60, 61, 62, 
	62, 61, 126, 61, 61, 
	61, 62, 0, 127, 0, 127, 
	0, 0, 0, 127, 0, 
	127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 0, 
	10, 10, 0, 0, 0, 
	127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 
	127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 
	127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 
	127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 
	127, 0, 127, 0, 127, 
	0, 127, 0, 127, 9, 92, 
	0, 127, 0, 127, 0, 
	127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 
	127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 
	127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 
	127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 
	127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 
	127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 
	127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 
	127, 0, 127, 0, 127, 
	0, 127, 0, 127, 0, 127, 
	0, 127, 0, 0, 61, 
	124, 0, 0, 9, 46, 
	9, 46, 46, 46, 10, 61, 
	10, 10, 10, 101, 10, 
	110, 10, 100, 10, 10, 
	0
]

class << self
	attr_accessor :_lex_key_spans
	private :_lex_key_spans, :_lex_key_spans=
end
self._lex_key_spans = [
	0, 1, 1, 1, 1, 1, 1, 1, 
	1, 1, 27, 128, 128, 128, 128, 46, 
	78, 78, 93, 27, 27, 46, 100, 27, 
	33, 1, 93, 78, 103, 128, 128, 128, 
	128, 46, 78, 78, 93, 27, 27, 46, 
	100, 27, 33, 1, 93, 78, 103, 128, 
	128, 128, 128, 128, 128, 128, 128, 128, 
	27, 128, 1, 128, 1, 1, 1, 1, 
	128, 128, 128, 128, 128, 128, 128, 128, 
	128, 128, 128, 128, 128, 128, 128, 128, 
	128, 128, 1, 1, 1, 1, 9, 1, 
	1, 1, 1, 1, 1, 1, 1, 128, 
	1, 128, 1, 1, 1, 1, 1, 1, 
	123, 1, 128, 128, 66, 128, 128, 1, 
	128, 128, 1, 25, 1, 30, 128, 87, 
	46, 78, 78, 93, 27, 27, 46, 100, 
	27, 33, 1, 93, 78, 103, 128, 128, 
	128, 128, 128, 128, 1, 1, 128, 15, 
	10, 10, 10, 10, 128, 1, 84, 84, 
	84, 84, 84, 84, 1, 1, 38, 1, 
	96, 24, 0, 1, 1, 1, 24, 1, 
	1, 93, 24, 88, 128, 10, 121, 0, 
	0, 8, 8, 0, 0, 93, 0, 0, 
	0, 93, 1, 0, 0, 0, 93, 46, 
	93, 93, 0, 0, 93, 46, 1, 93, 
	124, 27, 27, 27, 0, 103, 103, 103, 
	0, 126, 126, 126, 126, 126, 0, 126, 
	126, 0, 126, 27, 126, 126, 126, 126, 
	126, 126, 126, 126, 126, 0, 126, 0, 
	55, 0, 93, 88, 128, 10, 121, 0, 
	0, 8, 8, 0, 0, 93, 0, 0, 
	0, 93, 1, 0, 0, 0, 93, 46, 
	93, 93, 0, 0, 93, 46, 1, 93, 
	124, 27, 27, 27, 0, 103, 103, 103, 
	0, 126, 126, 126, 126, 126, 0, 126, 
	126, 0, 126, 27, 126, 126, 126, 126, 
	126, 126, 126, 126, 126, 0, 126, 0, 
	55, 0, 93, 24, 27, 93, 27, 36, 
	88, 128, 10, 27, 58, 58, 29, 128, 
	10, 128, 128, 128, 128, 24, 0, 66, 
	1, 1, 128, 128, 10, 1, 1, 1, 
	1, 1, 2, 1, 66, 1, 2, 128, 
	128, 128, 128, 128, 128, 128, 1, 1, 
	128, 128, 128, 128, 128, 128, 128, 128, 
	128, 128, 128, 128, 128, 128, 128, 128, 
	128, 128, 128, 128, 128, 128, 128, 128, 
	128, 128, 128, 128, 128, 128, 128, 128, 
	128, 128, 128, 128, 128, 128, 128, 128, 
	128, 128, 128, 128, 128, 128, 128, 128, 
	128, 128, 128, 128, 128, 128, 128, 128, 
	128, 128, 128, 128, 128, 128, 128, 128, 
	128, 128, 128, 128, 128, 128, 128, 128, 
	128, 128, 128, 128, 128, 128, 128, 128, 
	128, 128, 128, 1, 128, 128, 24, 1, 
	1, 1, 0, 128, 128, 1, 0, 24, 
	0, 66, 1, 1, 1, 1, 1, 2, 
	1, 66, 1, 2, 128, 1, 1, 1, 
	127, 128, 62, 53, 53, 0, 53, 54, 
	1, 1, 1, 24, 0, 128, 0, 116, 
	0, 1, 1, 0, 53, 1, 1, 1, 
	24, 1, 128, 1, 1, 1, 1, 128, 
	128, 128, 0, 128, 1, 124, 24, 1, 
	1, 1, 0, 1, 0, 128, 128, 24, 
	0, 1, 1, 1, 0, 128, 128, 128, 
	128, 128, 128, 128, 128, 128, 128, 128, 
	128, 128, 128, 128, 128, 128, 128, 4, 
	0, 66, 1, 0, 0, 0, 24, 1, 
	24, 66, 1, 1, 58, 123, 24, 0, 
	20, 1, 14, 15, 1, 1, 27, 128, 
	128, 1, 0, 66, 2, 0, 0, 0, 
	0, 66, 128, 10, 1, 1, 1, 2, 
	1, 1, 2, 1, 2, 0, 1, 66, 
	2, 123, 0, 128, 128, 121, 0, 0, 
	8, 8, 0, 0, 93, 0, 0, 0, 
	93, 1, 0, 0, 0, 93, 46, 93, 
	93, 0, 0, 93, 46, 1, 93, 124, 
	27, 27, 27, 0, 103, 103, 103, 0, 
	126, 126, 126, 126, 126, 0, 126, 126, 
	0, 126, 27, 126, 126, 126, 126, 126, 
	126, 126, 126, 126, 126, 126, 126, 126, 
	126, 0, 0, 55, 0, 128, 128, 128, 
	0, 1, 0, 128, 128, 128, 128, 128, 
	128, 128, 128, 128, 128, 128, 128, 128, 
	128, 128, 128, 128, 128, 128, 128, 128, 
	128, 128, 128, 128, 128, 128, 128, 128, 
	128, 128, 128, 128, 128, 128, 128, 128, 
	128, 128, 128, 128, 128, 128, 128, 128, 
	128, 128, 128, 128, 128, 128, 128, 128, 
	128, 128, 128, 128, 128, 128, 128, 66, 
	128, 128, 128, 128, 128, 128, 128, 128, 
	128, 128, 128, 128, 128, 128, 128, 128, 
	128, 128, 128, 128, 128, 128, 128, 128, 
	128, 0, 64, 93, 24, 0, 1, 1, 
	1, 0, 128, 128, 24, 0, 1, 1, 
	1, 0, 128, 128, 1, 0, 24, 0, 
	66, 1, 1, 128, 128, 10, 1, 24, 
	0, 0, 20, 2, 12, 1, 54, 48, 
	75, 67, 15, 58, 0, 1, 0, 67, 
	67, 67, 67, 10, 0, 1, 0, 67, 
	67, 67, 67, 67, 67, 67, 67, 69, 
	67, 69, 67, 1, 2, 1, 66, 1, 
	2, 128, 128, 0, 128, 128, 128, 128, 
	128, 128, 0, 1, 0, 128, 128, 128, 
	128, 128, 128, 128, 128, 128, 128, 128, 
	128, 128, 128, 128, 128, 128, 128, 128, 
	128, 128, 128, 128, 128, 128, 128, 128, 
	128, 128, 128, 128, 128, 128, 128, 128, 
	128, 128, 84, 128, 128, 128, 128, 128, 
	128, 128, 128, 128, 128, 128, 128, 128, 
	128, 128, 128, 128, 128, 128, 128, 128, 
	128, 128, 128, 128, 128, 128, 128, 128, 
	128, 128, 128, 128, 128, 128, 128, 128, 
	128, 128, 128, 128, 128, 128, 128, 128, 
	128, 128, 128, 128, 128, 128, 128, 128, 
	128, 128, 128, 128, 128, 128, 128, 128, 
	128, 128, 128, 128, 0, 64, 0, 38, 
	38, 1, 52, 1, 92, 101, 91, 1
]

class << self
	attr_accessor :_lex_index_offsets
	private :_lex_index_offsets, :_lex_index_offsets=
end
self._lex_index_offsets = [
	0, 0, 2, 4, 6, 8, 10, 12, 
	14, 16, 18, 46, 175, 304, 433, 562, 
	609, 688, 767, 861, 889, 917, 964, 1065, 
	1093, 1127, 1129, 1223, 1302, 1406, 1535, 1664, 
	1793, 1922, 1969, 2048, 2127, 2221, 2249, 2277, 
	2324, 2425, 2453, 2487, 2489, 2583, 2662, 2766, 
	2895, 3024, 3153, 3282, 3411, 3540, 3669, 3798, 
	3927, 3955, 4084, 4086, 4215, 4217, 4219, 4221, 
	4223, 4352, 4481, 4610, 4739, 4868, 4997, 5126, 
	5255, 5384, 5513, 5642, 5771, 5900, 6029, 6158, 
	6287, 6416, 6545, 6547, 6549, 6551, 6553, 6563, 
	6565, 6567, 6569, 6571, 6573, 6575, 6577, 6579, 
	6708, 6710, 6839, 6841, 6843, 6845, 6847, 6849, 
	6851, 6975, 6977, 7106, 7235, 7302, 7431, 7560, 
	7562, 7691, 7820, 7822, 7848, 7850, 7881, 8010, 
	8098, 8145, 8224, 8303, 8397, 8425, 8453, 8500, 
	8601, 8629, 8663, 8665, 8759, 8838, 8942, 9071, 
	9200, 9329, 9458, 9587, 9716, 9718, 9720, 9849, 
	9865, 9876, 9887, 9898, 9909, 10038, 10040, 10125, 
	10210, 10295, 10380, 10465, 10550, 10552, 10554, 10593, 
	10595, 10692, 10717, 10718, 10720, 10722, 10724, 10749, 
	10751, 10753, 10847, 10872, 10961, 11090, 11101, 11223, 
	11224, 11225, 11234, 11243, 11244, 11245, 11339, 11340, 
	11341, 11342, 11436, 11438, 11439, 11440, 11441, 11535, 
	11582, 11676, 11770, 11771, 11772, 11866, 11913, 11915, 
	12009, 12134, 12162, 12190, 12218, 12219, 12323, 12427, 
	12531, 12532, 12659, 12786, 12913, 13040, 13167, 13168, 
	13295, 13422, 13423, 13550, 13578, 13705, 13832, 13959, 
	14086, 14213, 14340, 14467, 14594, 14721, 14722, 14849, 
	14850, 14906, 14907, 15001, 15090, 15219, 15230, 15352, 
	15353, 15354, 15363, 15372, 15373, 15374, 15468, 15469, 
	15470, 15471, 15565, 15567, 15568, 15569, 15570, 15664, 
	15711, 15805, 15899, 15900, 15901, 15995, 16042, 16044, 
	16138, 16263, 16291, 16319, 16347, 16348, 16452, 16556, 
	16660, 16661, 16788, 16915, 17042, 17169, 17296, 17297, 
	17424, 17551, 17552, 17679, 17707, 17834, 17961, 18088, 
	18215, 18342, 18469, 18596, 18723, 18850, 18851, 18978, 
	18979, 19035, 19036, 19130, 19155, 19183, 19277, 19305, 
	19342, 19431, 19560, 19571, 19599, 19658, 19717, 19747, 
	19876, 19887, 20016, 20145, 20274, 20403, 20428, 20429, 
	20496, 20498, 20500, 20629, 20758, 20769, 20771, 20773, 
	20775, 20777, 20779, 20782, 20784, 20851, 20853, 20856, 
	20985, 21114, 21243, 21372, 21501, 21630, 21759, 21761, 
	21763, 21892, 22021, 22150, 22279, 22408, 22537, 22666, 
	22795, 22924, 23053, 23182, 23311, 23440, 23569, 23698, 
	23827, 23956, 24085, 24214, 24343, 24472, 24601, 24730, 
	24859, 24988, 25117, 25246, 25375, 25504, 25633, 25762, 
	25891, 26020, 26149, 26278, 26407, 26536, 26665, 26794, 
	26923, 27052, 27181, 27310, 27439, 27568, 27697, 27826, 
	27955, 28084, 28213, 28342, 28471, 28600, 28729, 28858, 
	28987, 29116, 29245, 29374, 29503, 29632, 29761, 29890, 
	30019, 30148, 30277, 30406, 30535, 30664, 30793, 30922, 
	31051, 31180, 31309, 31438, 31567, 31696, 31825, 31954, 
	32083, 32212, 32341, 32470, 32472, 32601, 32730, 32755, 
	32757, 32759, 32761, 32762, 32891, 33020, 33022, 33023, 
	33048, 33049, 33116, 33118, 33120, 33122, 33124, 33126, 
	33129, 33131, 33198, 33200, 33203, 33332, 33334, 33336, 
	33338, 33466, 33595, 33658, 33712, 33766, 33767, 33821, 
	33876, 33878, 33880, 33882, 33907, 33908, 34037, 34038, 
	34155, 34156, 34158, 34160, 34161, 34215, 34217, 34219, 
	34221, 34246, 34248, 34377, 34379, 34381, 34383, 34385, 
	34514, 34643, 34772, 34773, 34902, 34904, 35029, 35054, 
	35056, 35058, 35060, 35061, 35063, 35064, 35193, 35322, 
	35347, 35348, 35350, 35352, 35354, 35355, 35484, 35613, 
	35742, 35871, 36000, 36129, 36258, 36387, 36516, 36645, 
	36774, 36903, 37032, 37161, 37290, 37419, 37548, 37677, 
	37682, 37683, 37750, 37752, 37753, 37754, 37755, 37780, 
	37782, 37807, 37874, 37876, 37878, 37937, 38061, 38086, 
	38087, 38108, 38110, 38125, 38141, 38143, 38145, 38173, 
	38302, 38431, 38433, 38434, 38501, 38504, 38505, 38506, 
	38507, 38508, 38575, 38704, 38715, 38717, 38719, 38721, 
	38724, 38726, 38728, 38731, 38733, 38736, 38737, 38739, 
	38806, 38809, 38933, 38934, 39063, 39192, 39314, 39315, 
	39316, 39325, 39334, 39335, 39336, 39430, 39431, 39432, 
	39433, 39527, 39529, 39530, 39531, 39532, 39626, 39673, 
	39767, 39861, 39862, 39863, 39957, 40004, 40006, 40100, 
	40225, 40253, 40281, 40309, 40310, 40414, 40518, 40622, 
	40623, 40750, 40877, 41004, 41131, 41258, 41259, 41386, 
	41513, 41514, 41641, 41669, 41796, 41923, 42050, 42177, 
	42304, 42431, 42558, 42685, 42812, 42939, 43066, 43193, 
	43320, 43447, 43448, 43449, 43505, 43506, 43635, 43764, 
	43893, 43894, 43896, 43897, 44026, 44155, 44284, 44413, 
	44542, 44671, 44800, 44929, 45058, 45187, 45316, 45445, 
	45574, 45703, 45832, 45961, 46090, 46219, 46348, 46477, 
	46606, 46735, 46864, 46993, 47122, 47251, 47380, 47509, 
	47638, 47767, 47896, 48025, 48154, 48283, 48412, 48541, 
	48670, 48799, 48928, 49057, 49186, 49315, 49444, 49573, 
	49702, 49831, 49960, 50089, 50218, 50347, 50476, 50605, 
	50734, 50863, 50992, 51121, 51250, 51379, 51508, 51637, 
	51704, 51833, 51962, 52091, 52220, 52349, 52478, 52607, 
	52736, 52865, 52994, 53123, 53252, 53381, 53510, 53639, 
	53768, 53897, 54026, 54155, 54284, 54413, 54542, 54671, 
	54800, 54929, 54930, 54995, 55089, 55114, 55115, 55117, 
	55119, 55121, 55122, 55251, 55380, 55405, 55406, 55408, 
	55410, 55412, 55413, 55542, 55671, 55673, 55674, 55699, 
	55700, 55767, 55769, 55771, 55900, 56029, 56040, 56042, 
	56067, 56068, 56069, 56090, 56093, 56106, 56108, 56163, 
	56212, 56288, 56356, 56372, 56431, 56432, 56434, 56435, 
	56503, 56571, 56639, 56707, 56718, 56719, 56721, 56722, 
	56790, 56858, 56926, 56994, 57062, 57130, 57198, 57266, 
	57336, 57404, 57474, 57542, 57544, 57547, 57549, 57616, 
	57618, 57621, 57750, 57879, 57880, 58009, 58138, 58267, 
	58396, 58525, 58654, 58655, 58657, 58658, 58787, 58916, 
	59045, 59174, 59303, 59432, 59561, 59690, 59819, 59948, 
	60077, 60206, 60335, 60464, 60593, 60722, 60851, 60980, 
	61109, 61238, 61367, 61496, 61625, 61754, 61883, 62012, 
	62141, 62270, 62399, 62528, 62657, 62786, 62915, 63044, 
	63173, 63302, 63431, 63516, 63645, 63774, 63903, 64032, 
	64161, 64290, 64419, 64548, 64677, 64806, 64935, 65064, 
	65193, 65322, 65451, 65580, 65709, 65838, 65967, 66096, 
	66225, 66354, 66483, 66612, 66741, 66870, 66999, 67128, 
	67257, 67386, 67515, 67644, 67773, 67902, 68031, 68160, 
	68289, 68418, 68547, 68676, 68805, 68934, 69063, 69192, 
	69321, 69450, 69579, 69708, 69837, 69966, 70095, 70224, 
	70353, 70482, 70611, 70740, 70869, 70998, 71127, 71256, 
	71385, 71514, 71643, 71772, 71901, 71902, 71967, 71968, 
	72007, 72046, 72048, 72101, 72103, 72196, 72298, 72390
]

class << self
	attr_accessor :_lex_indicies
	private :_lex_indicies, :_lex_indicies=
end
self._lex_indicies = [
	1, 0, 2, 0, 3, 0, 4, 0, 
	5, 0, 6, 0, 7, 0, 8, 0, 
	9, 0, 10, 0, 0, 0, 10, 0, 
	0, 0, 0, 0, 11, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 10, 0, 12, 12, 
	12, 12, 12, 12, 12, 12, 12, 12, 
	12, 12, 12, 12, 12, 12, 12, 12, 
	12, 12, 12, 12, 12, 12, 12, 12, 
	12, 12, 12, 12, 12, 12, 12, 14, 
	14, 12, 14, 12, 14, 14, 12, 12, 
	14, 14, 14, 15, 14, 14, 16, 16, 
	16, 16, 16, 16, 16, 16, 16, 16, 
	14, 14, 14, 14, 14, 14, 14, 13, 
	13, 13, 13, 13, 13, 13, 13, 13, 
	13, 13, 13, 13, 13, 13, 13, 13, 
	13, 13, 13, 13, 13, 13, 13, 13, 
	13, 12, 14, 12, 12, 13, 14, 13, 
	13, 13, 13, 13, 13, 13, 13, 13, 
	13, 13, 13, 13, 13, 13, 13, 13, 
	13, 13, 13, 13, 13, 13, 13, 13, 
	13, 12, 12, 12, 14, 12, 13, 12, 
	12, 12, 12, 12, 12, 12, 12, 12, 
	12, 12, 12, 12, 12, 12, 12, 12, 
	12, 12, 12, 12, 12, 12, 12, 12, 
	12, 12, 12, 12, 12, 12, 12, 12, 
	12, 12, 12, 12, 12, 12, 12, 12, 
	12, 12, 12, 12, 12, 12, 12, 14, 
	14, 14, 14, 14, 14, 14, 14, 14, 
	14, 12, 12, 12, 12, 12, 12, 12, 
	14, 14, 14, 14, 14, 14, 14, 14, 
	14, 14, 14, 14, 14, 14, 14, 14, 
	14, 14, 14, 14, 14, 14, 14, 14, 
	14, 14, 12, 12, 12, 12, 14, 12, 
	14, 14, 14, 14, 14, 14, 14, 14, 
	14, 14, 14, 14, 14, 14, 14, 14, 
	14, 14, 14, 14, 14, 14, 14, 14, 
	14, 14, 12, 12, 12, 12, 12, 14, 
	12, 12, 12, 12, 12, 12, 12, 12, 
	12, 12, 12, 12, 12, 12, 12, 12, 
	12, 12, 12, 12, 12, 12, 12, 12, 
	12, 12, 12, 12, 12, 12, 12, 12, 
	12, 12, 12, 12, 12, 12, 12, 12, 
	12, 12, 12, 12, 12, 12, 12, 12, 
	13, 13, 13, 13, 13, 13, 13, 13, 
	13, 13, 12, 12, 12, 12, 12, 12, 
	17, 13, 13, 13, 13, 13, 13, 13, 
	13, 13, 13, 13, 13, 13, 13, 13, 
	13, 13, 13, 13, 13, 13, 13, 13, 
	13, 13, 13, 12, 12, 12, 12, 13, 
	12, 13, 13, 13, 13, 13, 13, 13, 
	13, 13, 13, 13, 13, 13, 13, 13, 
	13, 13, 13, 13, 13, 13, 13, 13, 
	13, 13, 13, 12, 12, 12, 12, 12, 
	13, 12, 12, 12, 12, 12, 12, 12, 
	12, 12, 12, 12, 12, 12, 12, 12, 
	12, 12, 12, 12, 12, 12, 12, 12, 
	12, 12, 12, 12, 12, 12, 12, 12, 
	12, 12, 12, 12, 12, 12, 12, 12, 
	12, 12, 12, 12, 12, 12, 12, 12, 
	12, 13, 13, 13, 13, 13, 13, 13, 
	13, 13, 13, 12, 12, 12, 12, 12, 
	12, 12, 13, 13, 13, 13, 13, 13, 
	13, 13, 13, 13, 13, 13, 13, 13, 
	13, 13, 13, 13, 13, 13, 13, 13, 
	13, 13, 13, 13, 12, 12, 12, 12, 
	13, 12, 13, 13, 13, 13, 13, 13, 
	13, 13, 13, 13, 13, 13, 13, 13, 
	13, 13, 13, 13, 13, 13, 13, 13, 
	13, 13, 13, 13, 12, 12, 12, 12, 
	12, 13, 18, 19, 19, 19, 18, 19, 
	19, 19, 19, 19, 19, 19, 19, 19, 
	19, 19, 19, 19, 19, 19, 19, 19, 
	19, 19, 19, 19, 18, 19, 19, 19, 
	19, 19, 19, 19, 19, 19, 19, 19, 
	19, 19, 19, 19, 19, 19, 19, 20, 
	19, 18, 21, 21, 21, 18, 21, 21, 
	21, 21, 21, 22, 21, 21, 21, 21, 
	21, 21, 21, 21, 21, 21, 21, 21, 
	21, 21, 21, 18, 21, 21, 21, 21, 
	21, 21, 21, 21, 21, 21, 21, 21, 
	21, 21, 21, 21, 21, 21, 21, 21, 
	21, 21, 21, 21, 21, 21, 21, 21, 
	21, 21, 21, 21, 21, 21, 21, 21, 
	21, 21, 21, 21, 21, 21, 21, 21, 
	21, 21, 21, 21, 21, 21, 23, 21, 
	18, 21, 21, 21, 18, 21, 21, 21, 
	21, 21, 21, 21, 21, 21, 21, 21, 
	21, 21, 21, 21, 21, 21, 21, 21, 
	21, 21, 18, 21, 21, 21, 21, 21, 
	21, 21, 21, 21, 21, 21, 21, 21, 
	21, 21, 21, 21, 21, 21, 21, 21, 
	21, 21, 21, 21, 21, 21, 21, 21, 
	21, 21, 21, 21, 21, 21, 21, 21, 
	21, 21, 21, 21, 21, 21, 21, 21, 
	21, 21, 21, 21, 21, 24, 21, 18, 
	25, 25, 25, 18, 25, 25, 25, 25, 
	25, 25, 25, 25, 25, 25, 25, 25, 
	25, 25, 25, 25, 25, 25, 25, 25, 
	25, 18, 25, 25, 25, 25, 25, 25, 
	25, 25, 25, 25, 25, 25, 25, 25, 
	25, 25, 25, 25, 25, 25, 25, 25, 
	25, 25, 25, 25, 25, 25, 25, 25, 
	25, 25, 25, 25, 25, 25, 26, 25, 
	25, 25, 25, 25, 25, 25, 25, 25, 
	25, 25, 25, 25, 25, 25, 25, 25, 
	25, 25, 25, 25, 25, 25, 25, 25, 
	25, 25, 25, 27, 25, 18, 28, 28, 
	28, 18, 28, 28, 28, 28, 28, 29, 
	28, 28, 28, 28, 28, 28, 28, 28, 
	28, 28, 28, 28, 28, 28, 28, 18, 
	28, 18, 28, 28, 28, 18, 28, 28, 
	28, 28, 28, 28, 28, 28, 28, 28, 
	28, 28, 28, 28, 28, 28, 28, 28, 
	28, 28, 28, 18, 28, 18, 19, 19, 
	19, 18, 19, 19, 19, 19, 19, 19, 
	19, 19, 19, 19, 19, 19, 19, 19, 
	19, 19, 19, 19, 19, 19, 19, 18, 
	19, 19, 19, 19, 19, 19, 19, 19, 
	19, 19, 19, 19, 19, 19, 19, 19, 
	19, 19, 30, 19, 18, 31, 31, 31, 
	18, 31, 31, 31, 31, 31, 32, 31, 
	31, 31, 31, 31, 31, 31, 31, 31, 
	31, 31, 31, 31, 31, 31, 18, 31, 
	31, 31, 31, 31, 31, 31, 31, 31, 
	31, 31, 31, 31, 31, 31, 31, 31, 
	31, 31, 31, 31, 31, 31, 31, 31, 
	31, 31, 31, 31, 31, 31, 31, 31, 
	31, 31, 31, 31, 31, 31, 31, 33, 
	31, 31, 31, 31, 31, 31, 31, 31, 
	31, 31, 31, 31, 31, 31, 31, 31, 
	31, 31, 31, 31, 31, 31, 31, 31, 
	34, 31, 31, 31, 31, 31, 31, 35, 
	31, 18, 31, 31, 31, 18, 31, 31, 
	31, 31, 31, 31, 31, 31, 31, 31, 
	31, 31, 31, 31, 31, 31, 31, 31, 
	31, 31, 31, 18, 31, 36, 18, 18, 
	18, 18, 18, 18, 18, 18, 18, 18, 
	18, 18, 18, 18, 18, 18, 18, 18, 
	18, 18, 18, 18, 18, 18, 18, 18, 
	18, 18, 18, 18, 18, 37, 18, 37, 
	18, 18, 38, 38, 38, 18, 38, 38, 
	38, 38, 38, 38, 38, 38, 38, 38, 
	38, 38, 38, 38, 38, 38, 38, 38, 
	38, 38, 38, 18, 38, 38, 38, 38, 
	38, 38, 38, 38, 38, 38, 38, 38, 
	38, 38, 38, 38, 38, 38, 38, 38, 
	38, 38, 38, 38, 38, 38, 38, 38, 
	38, 38, 38, 38, 38, 38, 38, 38, 
	39, 38, 38, 38, 38, 38, 38, 38, 
	38, 38, 38, 38, 38, 38, 38, 38, 
	38, 38, 38, 38, 38, 38, 38, 38, 
	38, 38, 38, 38, 38, 40, 38, 18, 
	21, 21, 21, 18, 21, 21, 21, 21, 
	21, 22, 21, 21, 21, 21, 21, 21, 
	21, 21, 21, 21, 21, 21, 21, 21, 
	21, 18, 21, 21, 21, 21, 21, 21, 
	21, 21, 21, 21, 21, 21, 21, 21, 
	21, 21, 21, 21, 21, 21, 21, 21, 
	21, 21, 21, 21, 21, 21, 21, 21, 
	21, 21, 21, 21, 21, 21, 21, 21, 
	21, 21, 21, 21, 21, 21, 21, 21, 
	21, 21, 21, 21, 24, 21, 18, 41, 
	41, 41, 18, 41, 41, 41, 41, 41, 
	41, 41, 41, 41, 41, 41, 41, 41, 
	41, 41, 41, 41, 41, 41, 41, 41, 
	18, 41, 41, 41, 41, 41, 41, 41, 
	41, 41, 41, 41, 41, 41, 41, 41, 
	41, 41, 41, 41, 41, 41, 42, 42, 
	42, 42, 42, 42, 42, 42, 42, 42, 
	41, 41, 41, 41, 41, 41, 41, 42, 
	42, 42, 42, 42, 42, 41, 41, 41, 
	41, 41, 41, 41, 41, 41, 41, 41, 
	41, 41, 41, 41, 41, 41, 41, 41, 
	41, 41, 41, 41, 41, 41, 41, 42, 
	42, 42, 42, 42, 42, 41, 43, 43, 
	43, 43, 43, 43, 43, 43, 43, 43, 
	43, 43, 43, 43, 43, 43, 43, 43, 
	43, 43, 43, 43, 43, 43, 43, 43, 
	43, 43, 43, 43, 43, 43, 43, 45, 
	45, 43, 45, 43, 45, 45, 43, 43, 
	45, 45, 45, 46, 45, 45, 47, 47, 
	47, 47, 47, 47, 47, 47, 47, 47, 
	45, 45, 45, 45, 45, 45, 45, 44, 
	44, 44, 44, 44, 44, 44, 44, 44, 
	44, 44, 44, 44, 44, 44, 44, 44, 
	44, 44, 44, 44, 44, 44, 44, 44, 
	44, 43, 45, 43, 43, 44, 45, 44, 
	44, 44, 44, 44, 44, 44, 44, 44, 
	44, 44, 44, 44, 44, 44, 44, 44, 
	44, 44, 44, 44, 44, 44, 44, 44, 
	44, 43, 43, 43, 45, 43, 44, 43, 
	43, 43, 43, 43, 43, 43, 43, 43, 
	43, 43, 43, 43, 43, 43, 43, 43, 
	43, 43, 43, 43, 43, 43, 43, 43, 
	43, 43, 43, 43, 43, 43, 43, 43, 
	43, 43, 43, 43, 43, 43, 43, 43, 
	43, 43, 43, 43, 43, 43, 43, 45, 
	45, 45, 45, 45, 45, 45, 45, 45, 
	45, 43, 43, 43, 43, 43, 43, 43, 
	45, 45, 45, 45, 45, 45, 45, 45, 
	45, 45, 45, 45, 45, 45, 45, 45, 
	45, 45, 45, 45, 45, 45, 45, 45, 
	45, 45, 43, 43, 43, 43, 45, 43, 
	45, 45, 45, 45, 45, 45, 45, 45, 
	45, 45, 45, 45, 45, 45, 45, 45, 
	45, 45, 45, 45, 45, 45, 45, 45, 
	45, 45, 43, 43, 43, 43, 43, 45, 
	43, 43, 43, 43, 43, 43, 43, 43, 
	43, 43, 43, 43, 43, 43, 43, 43, 
	43, 43, 43, 43, 43, 43, 43, 43, 
	43, 43, 43, 43, 43, 43, 43, 43, 
	43, 43, 43, 43, 43, 43, 43, 43, 
	43, 43, 43, 43, 43, 43, 43, 43, 
	44, 44, 44, 44, 44, 44, 44, 44, 
	44, 44, 43, 43, 43, 43, 43, 43, 
	48, 44, 44, 44, 44, 44, 44, 44, 
	44, 44, 44, 44, 44, 44, 44, 44, 
	44, 44, 44, 44, 44, 44, 44, 44, 
	44, 44, 44, 43, 43, 43, 43, 44, 
	43, 44, 44, 44, 44, 44, 44, 44, 
	44, 44, 44, 44, 44, 44, 44, 44, 
	44, 44, 44, 44, 44, 44, 44, 44, 
	44, 44, 44, 43, 43, 43, 43, 43, 
	44, 43, 43, 43, 43, 43, 43, 43, 
	43, 43, 43, 43, 43, 43, 43, 43, 
	43, 43, 43, 43, 43, 43, 43, 43, 
	43, 43, 43, 43, 43, 43, 43, 43, 
	43, 43, 43, 43, 43, 43, 43, 43, 
	43, 43, 43, 43, 43, 43, 43, 43, 
	43, 44, 44, 44, 44, 44, 44, 44, 
	44, 44, 44, 43, 43, 43, 43, 43, 
	43, 43, 44, 44, 44, 44, 44, 44, 
	44, 44, 44, 44, 44, 44, 44, 44, 
	44, 44, 44, 44, 44, 44, 44, 44, 
	44, 44, 44, 44, 43, 43, 43, 43, 
	44, 43, 44, 44, 44, 44, 44, 44, 
	44, 44, 44, 44, 44, 44, 44, 44, 
	44, 44, 44, 44, 44, 44, 44, 44, 
	44, 44, 44, 44, 43, 43, 43, 43, 
	43, 44, 49, 50, 50, 50, 49, 50, 
	50, 50, 50, 50, 50, 50, 50, 50, 
	50, 50, 50, 50, 50, 50, 50, 50, 
	50, 50, 50, 50, 49, 50, 50, 50, 
	50, 50, 50, 50, 50, 50, 50, 50, 
	50, 50, 50, 50, 50, 50, 50, 51, 
	50, 49, 52, 52, 52, 49, 52, 52, 
	52, 52, 52, 53, 52, 52, 52, 52, 
	52, 52, 52, 52, 52, 52, 52, 52, 
	52, 52, 52, 49, 52, 52, 52, 52, 
	52, 52, 52, 52, 52, 52, 52, 52, 
	52, 52, 52, 52, 52, 52, 52, 52, 
	52, 52, 52, 52, 52, 52, 52, 52, 
	52, 52, 52, 52, 52, 52, 52, 52, 
	52, 52, 52, 52, 52, 52, 52, 52, 
	52, 52, 52, 52, 52, 52, 54, 52, 
	49, 52, 52, 52, 49, 52, 52, 52, 
	52, 52, 52, 52, 52, 52, 52, 52, 
	52, 52, 52, 52, 52, 52, 52, 52, 
	52, 52, 49, 52, 52, 52, 52, 52, 
	52, 52, 52, 52, 52, 52, 52, 52, 
	52, 52, 52, 52, 52, 52, 52, 52, 
	52, 52, 52, 52, 52, 52, 52, 52, 
	52, 52, 52, 52, 52, 52, 52, 52, 
	52, 52, 52, 52, 52, 52, 52, 52, 
	52, 52, 52, 52, 52, 55, 52, 49, 
	56, 56, 56, 49, 56, 56, 56, 56, 
	56, 56, 56, 56, 56, 56, 56, 56, 
	56, 56, 56, 56, 56, 56, 56, 56, 
	56, 49, 56, 56, 56, 56, 56, 56, 
	56, 56, 56, 56, 56, 56, 56, 56, 
	56, 56, 56, 56, 56, 56, 56, 56, 
	56, 56, 56, 56, 56, 56, 56, 56, 
	56, 56, 56, 56, 56, 56, 57, 56, 
	56, 56, 56, 56, 56, 56, 56, 56, 
	56, 56, 56, 56, 56, 56, 56, 56, 
	56, 56, 56, 56, 56, 56, 56, 56, 
	56, 56, 56, 58, 56, 49, 59, 59, 
	59, 49, 59, 59, 59, 59, 59, 60, 
	59, 59, 59, 59, 59, 59, 59, 59, 
	59, 59, 59, 59, 59, 59, 59, 49, 
	59, 49, 59, 59, 59, 49, 59, 59, 
	59, 59, 59, 59, 59, 59, 59, 59, 
	59, 59, 59, 59, 59, 59, 59, 59, 
	59, 59, 59, 49, 59, 49, 50, 50, 
	50, 49, 50, 50, 50, 50, 50, 50, 
	50, 50, 50, 50, 50, 50, 50, 50, 
	50, 50, 50, 50, 50, 50, 50, 49, 
	50, 50, 50, 50, 50, 50, 50, 50, 
	50, 50, 50, 50, 50, 50, 50, 50, 
	50, 50, 61, 50, 49, 62, 62, 62, 
	49, 62, 62, 62, 62, 62, 63, 62, 
	62, 62, 62, 62, 62, 62, 62, 62, 
	62, 62, 62, 62, 62, 62, 49, 62, 
	62, 62, 62, 62, 62, 62, 62, 62, 
	62, 62, 62, 62, 62, 62, 62, 62, 
	62, 62, 62, 62, 62, 62, 62, 62, 
	62, 62, 62, 62, 62, 62, 62, 62, 
	62, 62, 62, 62, 62, 62, 62, 64, 
	62, 62, 62, 62, 62, 62, 62, 62, 
	62, 62, 62, 62, 62, 62, 62, 62, 
	62, 62, 62, 62, 62, 62, 62, 62, 
	65, 62, 62, 62, 62, 62, 62, 66, 
	62, 49, 62, 62, 62, 49, 62, 62, 
	62, 62, 62, 62, 62, 62, 62, 62, 
	62, 62, 62, 62, 62, 62, 62, 62, 
	62, 62, 62, 49, 62, 67, 49, 49, 
	49, 49, 49, 49, 49, 49, 49, 49, 
	49, 49, 49, 49, 49, 49, 49, 49, 
	49, 49, 49, 49, 49, 49, 49, 49, 
	49, 49, 49, 49, 49, 68, 49, 68, 
	49, 49, 69, 69, 69, 49, 69, 69, 
	69, 69, 69, 69, 69, 69, 69, 69, 
	69, 69, 69, 69, 69, 69, 69, 69, 
	69, 69, 69, 49, 69, 69, 69, 69, 
	69, 69, 69, 69, 69, 69, 69, 69, 
	69, 69, 69, 69, 69, 69, 69, 69, 
	69, 69, 69, 69, 69, 69, 69, 69, 
	69, 69, 69, 69, 69, 69, 69, 69, 
	70, 69, 69, 69, 69, 69, 69, 69, 
	69, 69, 69, 69, 69, 69, 69, 69, 
	69, 69, 69, 69, 69, 69, 69, 69, 
	69, 69, 69, 69, 69, 71, 69, 49, 
	52, 52, 52, 49, 52, 52, 52, 52, 
	52, 53, 52, 52, 52, 52, 52, 52, 
	52, 52, 52, 52, 52, 52, 52, 52, 
	52, 49, 52, 52, 52, 52, 52, 52, 
	52, 52, 52, 52, 52, 52, 52, 52, 
	52, 52, 52, 52, 52, 52, 52, 52, 
	52, 52, 52, 52, 52, 52, 52, 52, 
	52, 52, 52, 52, 52, 52, 52, 52, 
	52, 52, 52, 52, 52, 52, 52, 52, 
	52, 52, 52, 52, 55, 52, 49, 72, 
	72, 72, 49, 72, 72, 72, 72, 72, 
	72, 72, 72, 72, 72, 72, 72, 72, 
	72, 72, 72, 72, 72, 72, 72, 72, 
	49, 72, 72, 72, 72, 72, 72, 72, 
	72, 72, 72, 72, 72, 72, 72, 72, 
	72, 72, 72, 72, 72, 72, 73, 73, 
	73, 73, 73, 73, 73, 73, 73, 73, 
	72, 72, 72, 72, 72, 72, 72, 73, 
	73, 73, 73, 73, 73, 72, 72, 72, 
	72, 72, 72, 72, 72, 72, 72, 72, 
	72, 72, 72, 72, 72, 72, 72, 72, 
	72, 72, 72, 72, 72, 72, 72, 73, 
	73, 73, 73, 73, 73, 72, 74, 74, 
	74, 74, 74, 74, 74, 74, 74, 74, 
	74, 74, 74, 74, 74, 74, 74, 74, 
	74, 74, 74, 74, 74, 74, 74, 74, 
	74, 74, 74, 74, 74, 74, 74, 76, 
	76, 74, 76, 74, 76, 76, 74, 74, 
	76, 76, 76, 77, 76, 76, 78, 78, 
	78, 78, 78, 78, 78, 78, 78, 78, 
	76, 76, 76, 76, 76, 76, 76, 75, 
	75, 75, 75, 75, 75, 75, 75, 75, 
	75, 75, 75, 75, 75, 75, 75, 75, 
	75, 75, 75, 75, 75, 75, 75, 75, 
	75, 74, 76, 74, 74, 75, 76, 75, 
	75, 75, 75, 75, 75, 75, 75, 75, 
	75, 75, 75, 75, 75, 75, 75, 75, 
	75, 75, 75, 75, 75, 75, 75, 75, 
	75, 74, 74, 74, 76, 74, 75, 74, 
	74, 74, 74, 74, 74, 74, 74, 74, 
	74, 74, 74, 74, 74, 74, 74, 74, 
	74, 74, 74, 74, 74, 74, 74, 74, 
	74, 74, 74, 74, 74, 74, 74, 74, 
	74, 74, 74, 74, 74, 74, 74, 74, 
	74, 74, 74, 74, 74, 74, 74, 76, 
	76, 76, 76, 76, 76, 76, 76, 76, 
	76, 74, 74, 74, 74, 74, 74, 74, 
	76, 76, 76, 76, 76, 76, 76, 76, 
	76, 76, 76, 76, 76, 76, 76, 76, 
	76, 76, 76, 76, 76, 76, 76, 76, 
	76, 76, 74, 74, 74, 74, 76, 74, 
	76, 76, 76, 76, 76, 76, 76, 76, 
	76, 76, 76, 76, 76, 76, 76, 76, 
	76, 76, 76, 76, 76, 76, 76, 76, 
	76, 76, 74, 74, 74, 74, 74, 76, 
	74, 74, 74, 74, 74, 74, 74, 74, 
	74, 74, 74, 74, 74, 74, 74, 74, 
	74, 74, 74, 74, 74, 74, 74, 74, 
	74, 74, 74, 74, 74, 74, 74, 74, 
	74, 74, 74, 74, 74, 74, 74, 74, 
	74, 74, 74, 74, 74, 74, 74, 74, 
	75, 75, 75, 75, 75, 75, 75, 75, 
	75, 75, 74, 74, 74, 74, 74, 74, 
	79, 75, 75, 75, 75, 75, 75, 75, 
	75, 75, 75, 75, 75, 75, 75, 75, 
	75, 75, 75, 75, 75, 75, 75, 75, 
	75, 75, 75, 74, 74, 74, 74, 75, 
	74, 75, 75, 75, 75, 75, 75, 75, 
	75, 75, 75, 75, 75, 75, 75, 75, 
	75, 75, 75, 75, 75, 75, 75, 75, 
	75, 75, 75, 74, 74, 74, 74, 74, 
	75, 74, 74, 74, 74, 74, 74, 74, 
	74, 74, 74, 74, 74, 74, 74, 74, 
	74, 74, 74, 74, 74, 74, 74, 74, 
	74, 74, 74, 74, 74, 74, 74, 74, 
	74, 74, 74, 74, 74, 74, 74, 74, 
	74, 74, 74, 74, 74, 74, 74, 74, 
	74, 75, 75, 75, 75, 75, 75, 75, 
	75, 75, 75, 74, 74, 74, 74, 74, 
	74, 74, 75, 75, 75, 75, 75, 75, 
	75, 75, 75, 75, 75, 75, 75, 75, 
	75, 75, 75, 75, 75, 75, 75, 75, 
	75, 75, 75, 75, 74, 74, 74, 74, 
	75, 74, 75, 75, 75, 75, 75, 75, 
	75, 75, 75, 75, 75, 75, 75, 75, 
	75, 75, 75, 75, 75, 75, 75, 75, 
	75, 75, 75, 75, 74, 74, 74, 74, 
	74, 75, 82, 82, 82, 82, 82, 82, 
	82, 82, 82, 82, 82, 82, 82, 82, 
	82, 82, 82, 82, 82, 82, 82, 82, 
	82, 82, 82, 82, 82, 82, 82, 82, 
	82, 82, 82, 81, 81, 82, 81, 82, 
	81, 81, 82, 82, 81, 81, 81, 83, 
	81, 81, 84, 84, 84, 84, 84, 84, 
	84, 84, 84, 84, 81, 81, 81, 81, 
	81, 81, 81, 80, 80, 80, 80, 80, 
	80, 80, 80, 80, 80, 80, 80, 80, 
	80, 80, 80, 80, 80, 80, 80, 80, 
	80, 80, 80, 80, 80, 82, 81, 82, 
	82, 80, 81, 80, 80, 80, 80, 80, 
	80, 80, 80, 80, 80, 80, 80, 80, 
	80, 80, 80, 80, 80, 80, 80, 80, 
	80, 80, 80, 80, 80, 82, 82, 82, 
	81, 82, 80, 82, 82, 82, 82, 82, 
	82, 82, 82, 82, 82, 82, 82, 82, 
	82, 82, 82, 82, 82, 82, 82, 82, 
	82, 82, 82, 82, 82, 82, 82, 82, 
	82, 82, 82, 82, 82, 82, 82, 82, 
	82, 82, 82, 82, 82, 82, 82, 82, 
	82, 82, 82, 81, 81, 81, 81, 81, 
	81, 81, 81, 81, 81, 82, 82, 82, 
	82, 82, 82, 82, 81, 81, 81, 81, 
	81, 81, 81, 81, 81, 81, 81, 81, 
	81, 81, 81, 81, 81, 81, 81, 81, 
	81, 81, 81, 81, 81, 81, 82, 82, 
	82, 82, 81, 82, 81, 81, 81, 81, 
	81, 81, 81, 81, 81, 81, 81, 81, 
	81, 81, 81, 81, 81, 81, 81, 81, 
	81, 81, 81, 81, 81, 81, 82, 82, 
	82, 82, 82, 81, 82, 82, 82, 82, 
	82, 82, 82, 82, 82, 82, 82, 82, 
	82, 82, 82, 82, 82, 82, 82, 82, 
	82, 82, 82, 82, 82, 82, 82, 82, 
	82, 82, 82, 82, 82, 82, 82, 82, 
	82, 82, 82, 82, 82, 82, 82, 82, 
	82, 82, 82, 82, 85, 85, 85, 85, 
	85, 85, 85, 85, 85, 85, 82, 82, 
	82, 82, 82, 82, 86, 85, 85, 85, 
	85, 85, 85, 85, 85, 85, 85, 85, 
	85, 85, 85, 85, 85, 85, 85, 85, 
	85, 85, 85, 85, 85, 85, 85, 82, 
	82, 82, 82, 85, 82, 85, 85, 85, 
	85, 85, 85, 85, 85, 85, 85, 85, 
	85, 85, 85, 85, 85, 85, 85, 85, 
	85, 85, 85, 85, 85, 85, 85, 82, 
	82, 82, 82, 82, 85, 82, 82, 82, 
	82, 82, 82, 82, 82, 82, 82, 82, 
	82, 82, 82, 82, 82, 82, 82, 82, 
	82, 82, 82, 82, 82, 82, 82, 82, 
	82, 82, 82, 82, 82, 82, 82, 82, 
	82, 82, 82, 82, 82, 82, 82, 82, 
	82, 82, 82, 82, 82, 87, 87, 87, 
	87, 87, 87, 87, 87, 87, 87, 82, 
	82, 82, 82, 82, 82, 82, 87, 87, 
	87, 87, 87, 87, 87, 87, 87, 87, 
	87, 87, 87, 87, 87, 87, 87, 87, 
	87, 87, 87, 87, 87, 87, 87, 87, 
	82, 82, 82, 82, 87, 82, 87, 87, 
	87, 87, 87, 87, 87, 87, 87, 87, 
	87, 87, 87, 87, 87, 87, 87, 87, 
	87, 87, 87, 87, 87, 87, 87, 87, 
	82, 82, 82, 82, 82, 87, 88, 88, 
	88, 88, 88, 88, 88, 88, 88, 88, 
	88, 88, 88, 88, 88, 88, 88, 88, 
	88, 88, 88, 88, 88, 88, 88, 88, 
	88, 88, 88, 88, 88, 88, 88, 88, 
	88, 88, 88, 88, 88, 88, 88, 88, 
	88, 88, 88, 88, 88, 88, 89, 89, 
	89, 89, 89, 89, 89, 89, 89, 89, 
	88, 88, 88, 88, 88, 88, 88, 89, 
	89, 89, 89, 89, 89, 89, 89, 89, 
	89, 89, 89, 89, 89, 89, 89, 89, 
	89, 89, 89, 89, 89, 89, 89, 89, 
	89, 88, 88, 88, 88, 89, 88, 89, 
	89, 89, 89, 89, 89, 89, 89, 89, 
	89, 89, 89, 89, 89, 89, 89, 89, 
	89, 89, 89, 89, 89, 89, 89, 89, 
	89, 88, 88, 88, 88, 88, 89, 90, 
	91, 91, 91, 90, 91, 91, 91, 91, 
	91, 91, 91, 91, 91, 91, 91, 91, 
	91, 91, 91, 91, 91, 91, 91, 91, 
	91, 90, 91, 92, 92, 92, 92, 92, 
	92, 92, 92, 92, 92, 92, 92, 92, 
	92, 92, 92, 92, 92, 92, 92, 92, 
	92, 92, 92, 92, 92, 92, 92, 92, 
	92, 92, 92, 92, 94, 92, 92, 92, 
	92, 92, 92, 92, 92, 92, 92, 92, 
	92, 92, 92, 93, 93, 93, 93, 93, 
	93, 93, 93, 93, 93, 95, 92, 92, 
	92, 92, 94, 92, 93, 93, 93, 93, 
	93, 93, 93, 93, 93, 93, 93, 93, 
	93, 93, 93, 93, 93, 93, 93, 93, 
	93, 93, 93, 93, 93, 93, 92, 92, 
	92, 92, 93, 92, 93, 93, 93, 93, 
	93, 93, 93, 93, 93, 93, 93, 93, 
	93, 93, 93, 93, 93, 93, 93, 93, 
	93, 93, 93, 93, 93, 93, 92, 92, 
	92, 92, 92, 93, 95, 92, 96, 96, 
	96, 96, 96, 96, 96, 96, 96, 96, 
	96, 96, 96, 96, 96, 96, 96, 96, 
	96, 96, 96, 96, 96, 96, 96, 96, 
	96, 96, 96, 96, 96, 96, 96, 98, 
	96, 96, 96, 96, 96, 96, 96, 96, 
	96, 96, 96, 96, 96, 96, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	99, 96, 96, 96, 96, 98, 96, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 96, 96, 96, 96, 97, 96, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 96, 96, 96, 96, 96, 97, 99, 
	96, 101, 100, 102, 100, 103, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	105, 100, 105, 105, 105, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 105, 
	100, 100, 100, 100, 106, 107, 100, 108, 
	100, 109, 110, 111, 112, 113, 106, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 114, 100, 115, 111, 116, 117, 100, 
	104, 104, 104, 104, 104, 104, 104, 104, 
	104, 104, 104, 104, 104, 104, 104, 104, 
	104, 104, 104, 104, 104, 104, 104, 104, 
	104, 104, 118, 119, 111, 120, 104, 100, 
	104, 104, 104, 104, 104, 104, 104, 104, 
	121, 104, 104, 104, 104, 104, 104, 104, 
	104, 122, 104, 104, 123, 104, 124, 104, 
	104, 104, 125, 126, 100, 120, 100, 104, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 127, 100, 127, 127, 127, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	127, 100, 100, 100, 100, 128, 129, 100, 
	130, 100, 131, 132, 133, 134, 135, 128, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 136, 100, 137, 133, 138, 139, 
	100, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 140, 141, 133, 102, 97, 
	100, 97, 97, 97, 97, 97, 97, 97, 
	97, 142, 97, 97, 97, 97, 97, 97, 
	97, 97, 143, 97, 97, 144, 97, 145, 
	97, 97, 97, 146, 147, 100, 102, 100, 
	97, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 98, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 99, 100, 100, 100, 100, 
	98, 100, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 100, 100, 100, 100, 
	97, 100, 97, 97, 97, 97, 97, 148, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 100, 100, 100, 100, 
	100, 97, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 98, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 99, 100, 100, 100, 
	100, 98, 100, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 100, 100, 100, 
	100, 97, 100, 97, 97, 97, 97, 149, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 100, 100, 100, 
	100, 100, 97, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 98, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 99, 100, 100, 
	100, 100, 98, 100, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 100, 100, 
	100, 100, 97, 100, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 150, 97, 
	97, 97, 97, 97, 97, 97, 100, 100, 
	100, 100, 100, 97, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 98, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 99, 100, 
	100, 100, 100, 98, 100, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 100, 
	100, 100, 100, 97, 100, 97, 97, 151, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 100, 
	100, 100, 100, 100, 97, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 98, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 99, 
	100, 100, 100, 100, 98, 100, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	100, 100, 100, 100, 97, 100, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 152, 97, 97, 97, 97, 97, 
	100, 100, 100, 100, 100, 97, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 98, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	99, 100, 100, 100, 100, 98, 100, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 100, 100, 100, 100, 97, 100, 97, 
	97, 97, 97, 148, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 100, 100, 100, 100, 100, 97, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	98, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 99, 100, 100, 100, 100, 98, 100, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 100, 100, 100, 100, 97, 100, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 153, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 100, 100, 100, 100, 100, 97, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 98, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 99, 100, 100, 100, 100, 98, 
	100, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 100, 100, 100, 100, 97, 
	100, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 154, 97, 97, 97, 
	97, 97, 97, 97, 155, 97, 97, 97, 
	97, 97, 97, 100, 100, 100, 100, 100, 
	97, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 98, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 99, 100, 100, 100, 100, 
	98, 100, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 100, 100, 100, 100, 
	97, 100, 97, 97, 97, 97, 156, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 100, 100, 100, 100, 
	100, 97, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 98, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 99, 100, 100, 100, 
	100, 98, 100, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 100, 100, 100, 
	100, 97, 100, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 157, 97, 97, 
	97, 97, 97, 97, 97, 100, 100, 100, 
	100, 100, 97, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 98, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 99, 100, 100, 
	100, 100, 98, 100, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 100, 100, 
	100, 100, 97, 100, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 148, 97, 
	97, 97, 97, 97, 97, 97, 100, 100, 
	100, 100, 100, 97, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 98, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 99, 100, 
	100, 100, 100, 98, 100, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 100, 
	100, 100, 100, 97, 100, 97, 97, 97, 
	97, 97, 97, 97, 97, 158, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 100, 
	100, 100, 100, 100, 97, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 98, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 99, 
	100, 100, 100, 100, 98, 100, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	100, 100, 100, 100, 97, 100, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 148, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	100, 100, 100, 100, 100, 97, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 98, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	99, 100, 100, 100, 100, 98, 100, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 100, 100, 100, 100, 97, 100, 97, 
	97, 97, 97, 97, 97, 97, 159, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 100, 100, 100, 100, 100, 97, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	98, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 99, 100, 100, 100, 100, 98, 100, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 100, 100, 100, 100, 97, 100, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	160, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 100, 100, 100, 100, 100, 97, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	100, 98, 100, 100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100, 100, 100, 100, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 99, 100, 100, 100, 100, 98, 
	100, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 100, 100, 100, 100, 97, 
	100, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 152, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 100, 100, 100, 100, 100, 
	97, 162, 161, 163, 161, 164, 161, 133, 
	161, 165, 161, 161, 161, 161, 161, 161, 
	161, 166, 161, 167, 161, 168, 161, 133, 
	161, 169, 161, 133, 161, 170, 161, 164, 
	161, 172, 171, 173, 173, 173, 173, 173, 
	173, 173, 173, 173, 175, 173, 175, 175, 
	175, 173, 173, 173, 173, 173, 173, 173, 
	173, 173, 173, 173, 173, 173, 173, 173, 
	173, 173, 173, 175, 173, 173, 173, 173, 
	173, 173, 173, 176, 173, 173, 173, 173, 
	173, 173, 173, 173, 173, 173, 173, 173, 
	173, 173, 173, 173, 173, 173, 173, 173, 
	173, 173, 173, 173, 174, 174, 174, 174, 
	174, 174, 174, 174, 174, 174, 174, 174, 
	174, 174, 174, 174, 174, 174, 174, 174, 
	174, 174, 174, 174, 174, 174, 173, 177, 
	173, 173, 174, 173, 174, 174, 174, 178, 
	174, 174, 174, 174, 174, 174, 174, 174, 
	174, 174, 174, 174, 174, 174, 174, 174, 
	174, 174, 174, 174, 174, 174, 173, 173, 
	173, 173, 173, 174, 179, 173, 173, 173, 
	173, 173, 173, 173, 173, 173, 173, 181, 
	173, 181, 181, 181, 173, 173, 173, 173, 
	173, 173, 173, 173, 173, 173, 173, 173, 
	173, 173, 173, 173, 173, 173, 181, 173, 
	173, 173, 173, 173, 173, 173, 182, 173, 
	173, 173, 173, 173, 173, 173, 173, 173, 
	173, 173, 173, 173, 173, 173, 173, 173, 
	173, 173, 173, 173, 173, 173, 173, 180, 
	180, 180, 180, 180, 180, 180, 180, 180, 
	180, 180, 180, 180, 180, 180, 180, 180, 
	180, 180, 180, 180, 180, 180, 180, 180, 
	180, 173, 183, 173, 173, 180, 173, 180, 
	180, 180, 184, 180, 180, 180, 180, 180, 
	180, 180, 180, 180, 180, 180, 180, 180, 
	180, 180, 180, 180, 180, 180, 180, 180, 
	180, 173, 173, 173, 173, 173, 180, 185, 
	186, 188, 187, 189, 187, 190, 187, 191, 
	187, 192, 187, 193, 194, 194, 194, 193, 
	194, 194, 194, 194, 194, 194, 194, 194, 
	194, 194, 194, 194, 194, 194, 194, 194, 
	194, 194, 194, 194, 194, 193, 194, 194, 
	194, 194, 194, 194, 194, 194, 194, 194, 
	194, 194, 194, 194, 194, 194, 194, 194, 
	194, 194, 194, 194, 194, 194, 194, 194, 
	194, 194, 194, 194, 194, 194, 194, 194, 
	194, 194, 194, 194, 195, 195, 195, 195, 
	195, 195, 195, 195, 195, 195, 195, 195, 
	195, 195, 195, 195, 195, 195, 195, 195, 
	195, 195, 195, 195, 195, 195, 194, 194, 
	194, 194, 194, 194, 195, 195, 195, 195, 
	195, 195, 195, 195, 195, 195, 195, 195, 
	195, 195, 195, 195, 195, 195, 195, 195, 
	195, 195, 195, 195, 195, 195, 194, 196, 
	185, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 199, 199, 197, 199, 197, 199, 
	199, 197, 197, 199, 199, 199, 200, 199, 
	199, 201, 201, 201, 201, 201, 201, 201, 
	201, 201, 201, 199, 199, 199, 199, 199, 
	199, 199, 198, 198, 198, 198, 198, 198, 
	198, 198, 198, 198, 198, 198, 198, 198, 
	198, 198, 198, 198, 198, 198, 198, 198, 
	198, 198, 198, 198, 197, 199, 197, 197, 
	198, 199, 198, 198, 198, 198, 198, 198, 
	198, 198, 198, 198, 198, 198, 198, 198, 
	198, 198, 198, 198, 198, 198, 198, 198, 
	198, 198, 198, 198, 197, 197, 197, 199, 
	197, 198, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 199, 199, 199, 199, 199, 199, 
	199, 199, 199, 199, 197, 197, 197, 197, 
	197, 197, 197, 199, 199, 199, 199, 199, 
	199, 199, 199, 199, 199, 199, 199, 199, 
	199, 199, 199, 199, 199, 199, 199, 199, 
	199, 199, 199, 199, 199, 197, 197, 197, 
	197, 199, 197, 199, 199, 199, 199, 199, 
	199, 199, 199, 199, 199, 199, 199, 199, 
	199, 199, 199, 199, 199, 199, 199, 199, 
	199, 199, 199, 199, 199, 197, 197, 197, 
	197, 197, 199, 202, 199, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 199, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 203, 198, 
	198, 198, 198, 198, 198, 198, 198, 198, 
	198, 198, 198, 198, 198, 198, 198, 198, 
	198, 198, 198, 198, 198, 198, 198, 198, 
	198, 197, 197, 197, 197, 198, 197, 198, 
	198, 198, 198, 198, 198, 198, 198, 198, 
	198, 198, 198, 198, 198, 198, 198, 198, 
	198, 198, 198, 198, 198, 198, 198, 198, 
	198, 197, 197, 197, 197, 197, 198, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	198, 198, 198, 198, 198, 198, 198, 198, 
	198, 198, 198, 198, 198, 198, 198, 198, 
	198, 198, 198, 198, 198, 198, 198, 198, 
	198, 198, 197, 197, 197, 197, 198, 197, 
	198, 198, 198, 198, 198, 198, 198, 198, 
	198, 198, 198, 198, 198, 198, 198, 198, 
	198, 198, 198, 198, 198, 198, 198, 198, 
	198, 198, 197, 197, 197, 197, 197, 198, 
	202, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 205, 197, 197, 197, 
	197, 206, 197, 197, 197, 197, 197, 207, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 196, 
	197, 197, 197, 204, 204, 204, 204, 204, 
	204, 204, 204, 204, 204, 204, 204, 204, 
	204, 204, 204, 204, 204, 204, 204, 204, 
	204, 204, 204, 204, 204, 197, 197, 197, 
	197, 204, 208, 204, 204, 204, 204, 204, 
	204, 204, 204, 204, 204, 204, 204, 204, 
	204, 204, 204, 204, 204, 204, 204, 204, 
	204, 204, 204, 204, 204, 197, 197, 197, 
	207, 197, 204, 210, 210, 210, 210, 210, 
	210, 210, 210, 210, 210, 211, 210, 210, 
	210, 210, 210, 210, 210, 210, 210, 210, 
	210, 210, 210, 210, 210, 210, 210, 210, 
	210, 210, 210, 210, 210, 210, 210, 210, 
	210, 210, 210, 210, 210, 210, 210, 210, 
	210, 210, 210, 209, 209, 209, 209, 209, 
	209, 209, 209, 209, 209, 210, 210, 210, 
	210, 210, 210, 210, 209, 209, 209, 209, 
	209, 209, 209, 209, 209, 209, 209, 209, 
	209, 209, 209, 209, 209, 209, 209, 209, 
	209, 209, 209, 209, 209, 209, 210, 210, 
	210, 210, 209, 210, 209, 209, 209, 209, 
	209, 209, 209, 209, 209, 209, 209, 209, 
	209, 209, 209, 209, 209, 209, 209, 209, 
	209, 209, 209, 209, 209, 209, 210, 210, 
	210, 210, 210, 209, 213, 212, 197, 205, 
	205, 205, 205, 205, 205, 205, 205, 205, 
	205, 205, 205, 205, 205, 205, 205, 205, 
	205, 205, 205, 205, 205, 205, 214, 205, 
	211, 210, 197, 206, 206, 206, 206, 206, 
	206, 206, 206, 206, 206, 206, 206, 206, 
	206, 206, 206, 206, 206, 206, 206, 206, 
	206, 206, 206, 206, 206, 206, 206, 214, 
	206, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 205, 197, 197, 197, 197, 
	206, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 204, 204, 204, 204, 204, 204, 
	204, 204, 204, 204, 204, 204, 204, 204, 
	204, 204, 204, 204, 204, 204, 204, 204, 
	204, 204, 204, 204, 197, 197, 197, 197, 
	204, 208, 204, 204, 204, 204, 204, 204, 
	204, 204, 204, 204, 204, 204, 204, 204, 
	204, 204, 204, 204, 204, 204, 204, 204, 
	204, 204, 204, 204, 197, 197, 197, 197, 
	197, 204, 197, 208, 208, 208, 208, 208, 
	208, 208, 208, 208, 208, 208, 208, 208, 
	208, 208, 208, 208, 208, 208, 208, 208, 
	208, 208, 208, 208, 208, 208, 208, 208, 
	208, 208, 208, 208, 208, 208, 208, 208, 
	208, 208, 208, 208, 208, 208, 208, 208, 
	208, 208, 208, 208, 208, 208, 208, 208, 
	208, 208, 208, 208, 208, 208, 208, 208, 
	208, 208, 208, 208, 208, 208, 208, 208, 
	208, 208, 208, 208, 208, 208, 208, 208, 
	208, 208, 208, 208, 208, 208, 208, 208, 
	214, 208, 215, 216, 216, 216, 215, 216, 
	216, 216, 216, 216, 216, 216, 216, 216, 
	216, 216, 216, 216, 216, 216, 216, 216, 
	216, 216, 216, 216, 215, 216, 216, 216, 
	216, 216, 216, 216, 216, 216, 216, 216, 
	216, 216, 216, 216, 216, 216, 216, 217, 
	216, 215, 218, 218, 218, 215, 218, 218, 
	218, 218, 218, 219, 218, 218, 218, 218, 
	218, 218, 218, 218, 218, 218, 218, 218, 
	218, 218, 218, 215, 218, 218, 218, 218, 
	218, 218, 218, 218, 218, 218, 218, 218, 
	218, 218, 218, 218, 218, 218, 218, 218, 
	218, 218, 218, 218, 218, 218, 218, 218, 
	218, 218, 218, 218, 218, 218, 218, 218, 
	218, 218, 218, 218, 218, 218, 218, 218, 
	218, 218, 218, 218, 218, 218, 220, 218, 
	215, 218, 218, 218, 215, 218, 218, 218, 
	218, 218, 218, 218, 218, 218, 218, 218, 
	218, 218, 218, 218, 218, 218, 218, 218, 
	218, 218, 215, 218, 218, 218, 218, 218, 
	218, 218, 218, 218, 218, 218, 218, 218, 
	218, 218, 218, 218, 218, 218, 218, 218, 
	218, 218, 218, 218, 218, 218, 218, 218, 
	218, 218, 218, 218, 218, 218, 218, 218, 
	218, 218, 218, 218, 218, 218, 218, 218, 
	218, 218, 218, 218, 218, 221, 218, 215, 
	222, 222, 222, 215, 222, 222, 222, 222, 
	222, 222, 222, 222, 222, 222, 222, 222, 
	222, 222, 222, 222, 222, 222, 222, 222, 
	222, 215, 222, 222, 222, 222, 222, 222, 
	222, 222, 222, 222, 222, 222, 222, 222, 
	222, 222, 222, 222, 222, 222, 222, 222, 
	222, 222, 222, 222, 222, 222, 222, 222, 
	222, 222, 222, 222, 222, 222, 223, 222, 
	222, 222, 222, 222, 222, 222, 222, 222, 
	222, 222, 222, 222, 222, 222, 222, 222, 
	222, 222, 222, 222, 222, 222, 222, 222, 
	222, 222, 222, 224, 222, 215, 225, 225, 
	225, 215, 225, 225, 225, 225, 225, 226, 
	225, 225, 225, 225, 225, 225, 225, 225, 
	225, 225, 225, 225, 225, 225, 225, 215, 
	225, 215, 225, 225, 225, 215, 225, 225, 
	225, 225, 225, 225, 225, 225, 225, 225, 
	225, 225, 225, 225, 225, 225, 225, 225, 
	225, 225, 225, 215, 225, 215, 216, 216, 
	216, 215, 216, 216, 216, 216, 216, 216, 
	216, 216, 216, 216, 216, 216, 216, 216, 
	216, 216, 216, 216, 216, 216, 216, 215, 
	216, 216, 216, 216, 216, 216, 216, 216, 
	216, 216, 216, 216, 216, 216, 216, 216, 
	216, 216, 227, 216, 215, 228, 228, 228, 
	215, 228, 228, 228, 228, 228, 229, 228, 
	228, 228, 228, 228, 228, 228, 228, 228, 
	228, 228, 228, 228, 228, 228, 215, 228, 
	228, 228, 228, 228, 228, 228, 228, 228, 
	228, 228, 228, 228, 228, 228, 228, 228, 
	228, 228, 228, 228, 228, 228, 228, 228, 
	228, 228, 228, 228, 228, 228, 228, 228, 
	228, 228, 228, 228, 228, 228, 228, 230, 
	228, 228, 228, 228, 228, 228, 228, 228, 
	228, 228, 228, 228, 228, 228, 228, 228, 
	228, 228, 228, 228, 228, 228, 228, 228, 
	231, 228, 228, 228, 228, 228, 228, 232, 
	228, 215, 228, 228, 228, 215, 228, 228, 
	228, 228, 228, 228, 228, 228, 228, 228, 
	228, 228, 228, 228, 228, 228, 228, 228, 
	228, 228, 228, 215, 228, 233, 215, 215, 
	215, 215, 215, 215, 215, 215, 215, 215, 
	215, 215, 215, 215, 215, 215, 215, 215, 
	215, 215, 215, 215, 215, 215, 215, 215, 
	215, 215, 215, 215, 215, 234, 215, 234, 
	215, 215, 235, 235, 235, 215, 235, 235, 
	235, 235, 235, 235, 235, 235, 235, 235, 
	235, 235, 235, 235, 235, 235, 235, 235, 
	235, 235, 235, 215, 235, 235, 235, 235, 
	235, 235, 235, 235, 235, 235, 235, 235, 
	235, 235, 235, 235, 235, 235, 235, 235, 
	235, 235, 235, 235, 235, 235, 235, 235, 
	235, 235, 235, 235, 235, 235, 235, 235, 
	236, 235, 235, 235, 235, 235, 235, 235, 
	235, 235, 235, 235, 235, 235, 235, 235, 
	235, 235, 235, 235, 235, 235, 235, 235, 
	235, 235, 235, 235, 235, 237, 235, 215, 
	218, 218, 218, 215, 218, 218, 218, 218, 
	218, 219, 218, 218, 218, 218, 218, 218, 
	218, 218, 218, 218, 218, 218, 218, 218, 
	218, 215, 218, 218, 218, 218, 218, 218, 
	218, 218, 218, 218, 218, 218, 218, 218, 
	218, 218, 218, 218, 218, 218, 218, 218, 
	218, 218, 218, 218, 218, 218, 218, 218, 
	218, 218, 218, 218, 218, 218, 218, 218, 
	218, 218, 218, 218, 218, 218, 218, 218, 
	218, 218, 218, 218, 221, 218, 215, 238, 
	238, 238, 215, 238, 238, 238, 238, 238, 
	238, 238, 238, 238, 238, 238, 238, 238, 
	238, 238, 238, 238, 238, 238, 238, 238, 
	215, 238, 238, 238, 238, 238, 238, 238, 
	238, 238, 238, 238, 238, 238, 238, 238, 
	238, 238, 238, 238, 238, 238, 239, 239, 
	239, 239, 239, 239, 239, 239, 239, 239, 
	238, 238, 238, 238, 238, 238, 238, 239, 
	239, 239, 239, 239, 239, 238, 238, 238, 
	238, 238, 238, 238, 238, 238, 238, 238, 
	238, 238, 238, 238, 238, 238, 238, 238, 
	238, 238, 238, 238, 238, 238, 238, 239, 
	239, 239, 239, 239, 239, 238, 185, 185, 
	185, 185, 185, 185, 185, 185, 185, 185, 
	185, 185, 185, 185, 185, 185, 185, 185, 
	185, 185, 185, 185, 185, 185, 185, 185, 
	185, 185, 185, 185, 185, 185, 185, 241, 
	185, 185, 185, 185, 185, 185, 185, 185, 
	185, 185, 185, 185, 185, 185, 240, 240, 
	240, 240, 240, 240, 240, 240, 240, 240, 
	242, 185, 185, 243, 185, 241, 185, 240, 
	240, 240, 240, 240, 240, 240, 240, 240, 
	240, 240, 240, 240, 240, 240, 240, 240, 
	240, 240, 240, 240, 240, 240, 240, 240, 
	240, 185, 185, 185, 185, 240, 185, 240, 
	240, 240, 240, 240, 240, 240, 240, 240, 
	240, 240, 240, 240, 240, 240, 240, 240, 
	240, 240, 240, 240, 240, 240, 240, 240, 
	240, 185, 185, 185, 185, 185, 240, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	241, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 240, 
	240, 240, 240, 240, 240, 240, 240, 240, 
	240, 242, 197, 197, 243, 197, 241, 197, 
	240, 240, 240, 240, 240, 240, 244, 240, 
	240, 240, 240, 240, 240, 240, 240, 240, 
	240, 240, 240, 240, 240, 240, 240, 240, 
	240, 240, 197, 197, 197, 197, 240, 197, 
	240, 240, 240, 240, 240, 240, 240, 240, 
	240, 240, 240, 240, 240, 240, 240, 240, 
	240, 240, 240, 240, 240, 240, 240, 240, 
	240, 240, 197, 197, 197, 197, 197, 240, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 241, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	240, 240, 240, 240, 240, 240, 240, 240, 
	240, 240, 242, 197, 197, 243, 197, 241, 
	197, 240, 240, 240, 240, 240, 240, 240, 
	240, 245, 240, 240, 240, 240, 240, 240, 
	240, 240, 240, 240, 240, 240, 240, 240, 
	240, 240, 240, 197, 197, 197, 197, 240, 
	197, 240, 240, 240, 240, 240, 240, 240, 
	240, 240, 240, 240, 240, 240, 240, 240, 
	240, 240, 240, 240, 240, 240, 240, 240, 
	240, 240, 240, 197, 197, 197, 197, 197, 
	240, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 241, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 240, 240, 240, 240, 240, 240, 240, 
	240, 240, 240, 242, 197, 197, 243, 197, 
	241, 197, 240, 240, 240, 240, 240, 240, 
	240, 240, 240, 240, 240, 240, 240, 246, 
	240, 240, 240, 240, 240, 240, 240, 240, 
	240, 240, 240, 240, 197, 197, 197, 197, 
	240, 197, 240, 240, 240, 240, 240, 240, 
	240, 240, 240, 240, 240, 240, 240, 240, 
	240, 240, 240, 240, 240, 240, 240, 240, 
	240, 240, 240, 240, 197, 197, 197, 197, 
	197, 240, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 197, 241, 197, 197, 197, 197, 
	197, 197, 197, 197, 197, 197, 197, 197, 
	197, 197, 240, 240, 240, 240, 240, 240, 
	240, 240, 240, 240, 242, 197, 197, 243, 
	197, 241, 197, 240, 240, 240, 246, 240, 
	240, 240, 240, 240, 240, 240, 240, 240, 
	240, 240, 240, 240, 240, 240, 240, 240, 
	240, 240, 240, 240, 240, 197, 197, 197, 
	197, 240, 197, 240, 240, 240, 240, 240, 
	240, 240, 240, 240, 240, 240, 240, 240, 
	240, 240, 240, 240, 240, 240, 240, 240, 
	240, 240, 240, 240, 240, 197, 197, 197, 
	197, 197, 240, 247, 247, 247, 247, 247, 
	247, 247, 247, 247, 247, 247, 247, 247, 
	247, 247, 247, 247, 247, 247, 247, 247, 
	247, 247, 247, 247, 247, 247, 247, 247, 
	247, 247, 247, 247, 249, 247, 247, 247, 
	247, 247, 247, 247, 247, 247, 247, 247, 
	247, 247, 247, 248, 248, 248, 248, 248, 
	248, 248, 248, 248, 248, 250, 247, 247, 
	247, 247, 249, 247, 248, 248, 248, 248, 
	248, 248, 248, 248, 248, 248, 248, 248, 
	248, 248, 248, 248, 248, 248, 248, 248, 
	248, 248, 248, 248, 248, 248, 247, 247, 
	247, 247, 248, 247, 248, 248, 248, 248, 
	248, 248, 248, 248, 248, 248, 248, 248, 
	248, 248, 248, 248, 248, 248, 248, 248, 
	248, 248, 248, 248, 248, 248, 247, 247, 
	247, 247, 247, 248, 250, 247, 247, 251, 
	252, 252, 252, 252, 252, 252, 252, 252, 
	252, 252, 252, 252, 252, 252, 252, 252, 
	252, 252, 252, 252, 252, 252, 252, 252, 
	252, 252, 252, 252, 252, 252, 252, 252, 
	252, 252, 252, 252, 252, 252, 252, 252, 
	252, 252, 252, 252, 252, 252, 252, 252, 
	253, 253, 253, 253, 253, 253, 253, 253, 
	253, 253, 252, 252, 252, 252, 252, 252, 
	252, 253, 253, 253, 253, 253, 253, 253, 
	253, 253, 253, 253, 253, 253, 253, 253, 
	253, 253, 253, 253, 253, 253, 253, 253, 
	253, 253, 253, 252, 252, 252, 252, 253, 
	252, 253, 253, 253, 253, 253, 253, 253, 
	253, 253, 253, 253, 253, 253, 253, 253, 
	253, 253, 253, 253, 253, 253, 253, 253, 
	253, 253, 253, 252, 252, 252, 252, 252, 
	253, 255, 254, 255, 254, 254, 256, 256, 
	256, 256, 256, 256, 256, 256, 256, 256, 
	254, 256, 256, 256, 256, 256, 256, 256, 
	256, 256, 256, 254, 257, 257, 257, 257, 
	257, 257, 257, 257, 257, 257, 254, 259, 
	259, 259, 259, 259, 259, 259, 259, 259, 
	259, 258, 260, 260, 260, 260, 260, 260, 
	260, 260, 260, 260, 258, 252, 252, 252, 
	252, 252, 252, 252, 252, 252, 252, 252, 
	252, 252, 252, 252, 252, 252, 252, 252, 
	252, 252, 252, 252, 252, 252, 252, 252, 
	252, 252, 252, 252, 252, 252, 252, 252, 
	252, 252, 252, 252, 252, 252, 252, 252, 
	252, 252, 252, 252, 252, 261, 261, 261, 
	261, 261, 261, 261, 261, 261, 261, 252, 
	252, 252, 252, 252, 252, 252, 261, 261, 
	261, 261, 261, 261, 261, 261, 261, 261, 
	261, 261, 261, 261, 261, 261, 261, 261, 
	261, 261, 261, 261, 261, 261, 261, 261, 
	252, 252, 252, 252, 261, 252, 261, 261, 
	261, 261, 261, 261, 261, 261, 261, 261, 
	261, 261, 261, 261, 261, 261, 261, 261, 
	261, 261, 261, 261, 261, 261, 261, 261, 
	252, 252, 252, 252, 252, 261, 262, 258, 
	263, 264, 263, 263, 263, 258, 258, 258, 
	258, 258, 258, 258, 258, 258, 258, 258, 
	258, 258, 258, 258, 258, 258, 258, 263, 
	258, 258, 265, 258, 258, 258, 258, 258, 
	258, 258, 258, 258, 258, 258, 258, 258, 
	258, 258, 258, 258, 258, 258, 258, 258, 
	258, 258, 258, 266, 258, 258, 258, 258, 
	258, 258, 258, 258, 258, 258, 258, 258, 
	258, 258, 258, 258, 258, 258, 258, 258, 
	258, 258, 258, 258, 258, 258, 258, 258, 
	258, 258, 258, 267, 258, 268, 269, 268, 
	268, 268, 258, 258, 258, 258, 258, 258, 
	258, 258, 258, 258, 258, 258, 258, 258, 
	258, 258, 258, 258, 268, 258, 258, 270, 
	258, 258, 258, 258, 258, 258, 258, 258, 
	258, 258, 258, 258, 258, 258, 258, 258, 
	258, 258, 258, 258, 258, 258, 258, 258, 
	271, 258, 258, 258, 258, 258, 258, 258, 
	258, 258, 258, 258, 258, 258, 258, 258, 
	258, 258, 258, 258, 258, 258, 258, 258, 
	258, 258, 258, 258, 258, 258, 258, 258, 
	272, 258, 274, 275, 274, 274, 274, 273, 
	273, 273, 273, 273, 273, 273, 273, 273, 
	273, 273, 273, 273, 273, 273, 273, 273, 
	273, 274, 273, 273, 276, 273, 273, 273, 
	273, 273, 273, 273, 273, 273, 273, 273, 
	273, 273, 273, 273, 273, 273, 273, 273, 
	273, 273, 273, 273, 273, 277, 273, 273, 
	273, 273, 273, 273, 273, 273, 273, 273, 
	273, 273, 273, 273, 273, 273, 273, 273, 
	273, 273, 273, 273, 273, 273, 273, 273, 
	273, 273, 273, 273, 273, 278, 273, 280, 
	281, 280, 280, 280, 279, 279, 279, 279, 
	279, 279, 279, 279, 279, 279, 279, 279, 
	279, 279, 279, 279, 279, 279, 280, 279, 
	279, 282, 279, 279, 279, 279, 279, 279, 
	279, 279, 279, 279, 279, 279, 279, 279, 
	279, 279, 279, 279, 279, 279, 279, 279, 
	279, 279, 283, 279, 279, 279, 279, 279, 
	279, 279, 279, 279, 279, 279, 279, 279, 
	279, 279, 279, 279, 279, 279, 279, 279, 
	279, 279, 279, 279, 279, 279, 279, 279, 
	279, 279, 284, 279, 280, 281, 280, 280, 
	280, 279, 279, 279, 279, 279, 279, 279, 
	279, 279, 279, 279, 279, 279, 279, 279, 
	279, 279, 279, 280, 279, 279, 282, 279, 
	279, 279, 279, 279, 279, 279, 279, 279, 
	279, 279, 279, 279, 279, 279, 279, 279, 
	279, 279, 279, 279, 279, 279, 279, 285, 
	279, 279, 279, 279, 279, 279, 279, 279, 
	279, 279, 279, 279, 279, 279, 279, 279, 
	279, 279, 279, 279, 279, 279, 279, 279, 
	279, 279, 279, 279, 279, 279, 279, 284, 
	279, 280, 286, 280, 280, 280, 279, 279, 
	279, 279, 279, 279, 279, 279, 279, 279, 
	279, 279, 279, 279, 279, 279, 279, 279, 
	280, 279, 279, 282, 279, 279, 279, 279, 
	279, 279, 279, 279, 279, 279, 279, 279, 
	279, 279, 279, 279, 279, 279, 279, 279, 
	279, 279, 279, 279, 283, 279, 279, 279, 
	279, 279, 279, 279, 279, 279, 279, 279, 
	279, 279, 279, 279, 279, 279, 279, 279, 
	279, 279, 279, 279, 279, 279, 279, 279, 
	279, 279, 279, 279, 284, 279, 287, 258, 
	264, 258, 289, 288, 289, 289, 289, 288, 
	288, 288, 288, 288, 288, 288, 288, 288, 
	288, 288, 288, 288, 288, 288, 288, 288, 
	288, 289, 288, 288, 288, 288, 288, 290, 
	288, 288, 288, 288, 288, 288, 288, 291, 
	288, 292, 288, 294, 293, 293, 293, 294, 
	293, 293, 293, 293, 295, 296, 295, 295, 
	295, 293, 293, 293, 293, 293, 293, 293, 
	293, 293, 293, 293, 293, 294, 293, 293, 
	293, 293, 293, 295, 293, 293, 297, 293, 
	293, 293, 293, 293, 293, 293, 293, 293, 
	293, 293, 293, 293, 293, 293, 293, 293, 
	293, 293, 293, 293, 293, 293, 293, 293, 
	298, 293, 293, 293, 293, 293, 293, 293, 
	293, 293, 293, 293, 293, 293, 293, 293, 
	293, 293, 293, 293, 293, 293, 293, 293, 
	293, 293, 293, 293, 293, 293, 293, 299, 
	293, 293, 300, 293, 295, 301, 295, 295, 
	295, 301, 301, 301, 301, 301, 301, 301, 
	301, 301, 301, 301, 301, 301, 301, 301, 
	301, 301, 301, 295, 301, 302, 303, 304, 
	305, 306, 308, 307, 310, 311, 310, 310, 
	310, 309, 309, 309, 309, 309, 309, 309, 
	309, 309, 309, 309, 309, 309, 309, 309, 
	309, 309, 309, 310, 309, 296, 307, 312, 
	307, 314, 313, 313, 313, 314, 313, 313, 
	313, 313, 315, 316, 315, 315, 315, 313, 
	313, 313, 313, 313, 313, 313, 313, 313, 
	313, 313, 313, 314, 313, 313, 313, 313, 
	313, 315, 313, 313, 317, 313, 313, 313, 
	313, 313, 313, 313, 313, 313, 313, 313, 
	313, 313, 313, 313, 313, 313, 313, 313, 
	313, 313, 313, 313, 313, 313, 313, 313, 
	313, 313, 313, 313, 313, 313, 313, 313, 
	313, 313, 313, 313, 313, 313, 313, 313, 
	313, 313, 313, 313, 313, 313, 313, 313, 
	313, 313, 313, 313, 313, 318, 313, 315, 
	319, 315, 315, 315, 319, 319, 319, 319, 
	319, 319, 319, 319, 319, 319, 319, 319, 
	319, 319, 319, 319, 319, 319, 315, 319, 
	321, 320, 320, 320, 320, 320, 320, 320, 
	320, 320, 320, 320, 320, 320, 320, 320, 
	320, 320, 320, 320, 320, 320, 320, 320, 
	320, 320, 320, 320, 322, 320, 320, 320, 
	320, 320, 320, 320, 320, 320, 320, 320, 
	320, 320, 320, 320, 320, 320, 320, 320, 
	320, 320, 320, 320, 320, 320, 320, 320, 
	320, 320, 320, 320, 320, 320, 320, 320, 
	320, 320, 320, 320, 320, 320, 320, 320, 
	320, 320, 320, 320, 320, 320, 320, 320, 
	320, 320, 320, 320, 320, 320, 320, 323, 
	320, 324, 324, 324, 324, 324, 324, 324, 
	324, 324, 324, 324, 324, 324, 324, 324, 
	324, 324, 324, 324, 324, 324, 324, 324, 
	324, 324, 324, 324, 324, 324, 324, 324, 
	324, 324, 324, 324, 324, 324, 324, 324, 
	324, 324, 324, 324, 324, 324, 324, 324, 
	324, 13, 13, 13, 13, 13, 13, 13, 
	13, 13, 13, 324, 324, 324, 324, 324, 
	324, 324, 13, 13, 13, 13, 13, 13, 
	13, 13, 13, 13, 13, 13, 13, 13, 
	13, 13, 13, 13, 13, 13, 13, 13, 
	13, 13, 13, 13, 324, 324, 324, 324, 
	13, 324, 13, 13, 13, 13, 13, 13, 
	13, 13, 13, 13, 13, 13, 13, 13, 
	13, 13, 13, 13, 13, 13, 13, 13, 
	13, 13, 13, 13, 324, 324, 324, 324, 
	324, 13, 16, 16, 16, 16, 16, 16, 
	16, 16, 16, 16, 324, 327, 326, 326, 
	326, 327, 326, 326, 326, 326, 326, 326, 
	326, 326, 326, 326, 326, 326, 326, 326, 
	326, 326, 326, 326, 326, 326, 326, 327, 
	326, 326, 326, 326, 326, 326, 326, 326, 
	326, 326, 326, 326, 326, 326, 326, 326, 
	326, 326, 326, 326, 326, 328, 328, 328, 
	328, 328, 328, 328, 328, 326, 326, 326, 
	326, 326, 326, 326, 326, 326, 326, 326, 
	329, 326, 326, 326, 326, 326, 326, 326, 
	326, 326, 330, 326, 326, 326, 326, 326, 
	326, 326, 326, 326, 326, 326, 326, 326, 
	326, 326, 326, 326, 326, 326, 326, 326, 
	331, 326, 326, 326, 326, 326, 326, 326, 
	326, 326, 326, 326, 326, 326, 326, 326, 
	326, 326, 332, 326, 326, 333, 326, 334, 
	335, 337, 337, 337, 337, 337, 337, 337, 
	337, 336, 338, 338, 338, 338, 338, 338, 
	338, 338, 336, 336, 339, 339, 38, 38, 
	38, 339, 38, 38, 38, 38, 38, 38, 
	38, 38, 38, 38, 38, 38, 38, 38, 
	38, 38, 38, 38, 38, 38, 38, 339, 
	38, 38, 38, 38, 38, 38, 38, 38, 
	38, 38, 38, 38, 38, 38, 38, 38, 
	38, 38, 38, 38, 38, 38, 38, 38, 
	38, 38, 38, 38, 38, 38, 38, 38, 
	38, 38, 38, 38, 39, 38, 38, 38, 
	38, 38, 38, 38, 38, 38, 38, 38, 
	38, 38, 38, 38, 38, 38, 38, 38, 
	38, 38, 38, 38, 38, 38, 38, 38, 
	38, 340, 38, 341, 342, 343, 343, 38, 
	38, 38, 343, 38, 38, 38, 38, 38, 
	38, 38, 38, 38, 38, 38, 38, 38, 
	38, 38, 38, 38, 38, 38, 38, 38, 
	343, 38, 38, 38, 38, 38, 38, 38, 
	38, 38, 38, 38, 38, 38, 38, 38, 
	38, 38, 38, 38, 38, 38, 38, 38, 
	38, 38, 38, 38, 38, 38, 38, 38, 
	38, 38, 38, 38, 38, 39, 38, 38, 
	38, 38, 38, 38, 38, 38, 38, 38, 
	38, 38, 38, 38, 38, 38, 38, 38, 
	38, 38, 38, 38, 38, 38, 38, 38, 
	38, 38, 344, 38, 37, 343, 345, 346, 
	347, 347, 25, 25, 25, 347, 25, 25, 
	25, 25, 25, 25, 25, 25, 25, 25, 
	25, 25, 25, 25, 25, 25, 25, 25, 
	25, 25, 25, 347, 25, 25, 25, 25, 
	25, 25, 25, 25, 25, 25, 25, 25, 
	25, 25, 25, 25, 25, 25, 25, 25, 
	25, 25, 25, 25, 25, 25, 25, 25, 
	25, 25, 25, 25, 25, 25, 25, 25, 
	26, 25, 25, 25, 25, 25, 25, 25, 
	25, 25, 25, 25, 25, 25, 25, 25, 
	25, 25, 25, 25, 25, 25, 25, 25, 
	25, 25, 25, 25, 25, 348, 25, 343, 
	19, 19, 19, 343, 19, 19, 19, 19, 
	19, 19, 19, 19, 19, 19, 19, 19, 
	19, 19, 19, 19, 19, 19, 19, 19, 
	19, 343, 19, 19, 19, 19, 19, 19, 
	19, 19, 19, 19, 19, 19, 19, 19, 
	19, 19, 19, 19, 349, 19, 339, 25, 
	25, 25, 339, 25, 25, 25, 25, 25, 
	25, 25, 25, 25, 25, 25, 25, 25, 
	25, 25, 25, 25, 25, 25, 25, 25, 
	339, 25, 25, 25, 25, 25, 25, 25, 
	25, 25, 25, 25, 25, 25, 25, 25, 
	25, 25, 25, 25, 25, 25, 25, 25, 
	25, 25, 25, 25, 25, 25, 25, 25, 
	25, 25, 25, 25, 25, 26, 25, 25, 
	25, 25, 25, 25, 25, 25, 25, 25, 
	25, 25, 25, 25, 25, 25, 25, 25, 
	25, 25, 25, 25, 25, 25, 25, 25, 
	25, 25, 27, 25, 339, 350, 350, 350, 
	339, 350, 350, 350, 350, 350, 350, 350, 
	350, 350, 350, 350, 350, 350, 350, 350, 
	350, 350, 350, 350, 350, 350, 339, 350, 
	350, 350, 350, 350, 350, 350, 350, 350, 
	350, 350, 350, 350, 350, 350, 350, 350, 
	350, 350, 350, 350, 350, 350, 350, 350, 
	350, 350, 350, 350, 350, 350, 350, 350, 
	350, 350, 350, 350, 350, 350, 350, 350, 
	350, 350, 350, 350, 350, 350, 350, 350, 
	350, 350, 350, 350, 350, 350, 350, 350, 
	350, 350, 350, 350, 350, 350, 350, 350, 
	351, 350, 352, 353, 353, 350, 350, 350, 
	353, 350, 350, 350, 350, 350, 350, 350, 
	350, 350, 350, 350, 350, 350, 350, 350, 
	350, 350, 350, 350, 350, 350, 353, 350, 
	350, 350, 350, 350, 350, 350, 350, 350, 
	350, 350, 350, 350, 350, 350, 350, 350, 
	350, 350, 350, 350, 350, 350, 350, 350, 
	350, 350, 350, 350, 350, 350, 350, 350, 
	350, 350, 350, 350, 350, 350, 350, 350, 
	350, 350, 350, 350, 350, 350, 350, 350, 
	350, 350, 350, 350, 350, 350, 350, 350, 
	350, 350, 350, 350, 350, 350, 350, 350, 
	354, 350, 353, 19, 19, 19, 353, 19, 
	19, 19, 19, 19, 19, 19, 19, 19, 
	19, 19, 19, 19, 19, 19, 19, 19, 
	19, 19, 19, 19, 353, 19, 19, 19, 
	19, 19, 19, 19, 19, 19, 19, 19, 
	19, 19, 19, 19, 19, 19, 19, 349, 
	19, 355, 353, 353, 25, 25, 25, 353, 
	25, 25, 25, 25, 25, 25, 25, 25, 
	25, 25, 25, 25, 25, 25, 25, 25, 
	25, 25, 25, 25, 25, 353, 25, 25, 
	25, 25, 25, 25, 25, 25, 25, 25, 
	25, 25, 25, 25, 25, 25, 25, 25, 
	25, 25, 25, 25, 25, 25, 25, 25, 
	25, 25, 25, 25, 25, 25, 25, 25, 
	25, 25, 26, 25, 25, 25, 25, 25, 
	25, 25, 25, 25, 25, 25, 25, 25, 
	25, 25, 25, 25, 25, 25, 25, 25, 
	25, 25, 25, 25, 25, 25, 25, 27, 
	25, 356, 357, 357, 357, 356, 357, 357, 
	357, 357, 357, 357, 357, 357, 357, 357, 
	357, 357, 357, 357, 357, 357, 357, 357, 
	357, 357, 357, 356, 357, 357, 357, 357, 
	357, 357, 357, 357, 357, 357, 357, 357, 
	357, 357, 357, 357, 357, 357, 357, 357, 
	357, 358, 358, 358, 358, 358, 358, 358, 
	358, 358, 358, 357, 357, 357, 357, 357, 
	357, 357, 358, 358, 358, 358, 358, 358, 
	357, 357, 357, 357, 357, 357, 357, 357, 
	357, 357, 357, 357, 357, 357, 357, 357, 
	357, 357, 357, 357, 357, 357, 357, 357, 
	357, 357, 358, 358, 358, 358, 358, 358, 
	357, 357, 357, 357, 357, 357, 357, 357, 
	357, 357, 357, 357, 357, 357, 357, 357, 
	357, 357, 357, 357, 359, 357, 356, 360, 
	360, 360, 356, 360, 360, 360, 360, 360, 
	360, 360, 360, 360, 360, 360, 360, 360, 
	360, 360, 360, 360, 360, 360, 360, 360, 
	356, 360, 356, 361, 361, 361, 356, 361, 
	361, 361, 361, 361, 361, 361, 361, 361, 
	361, 361, 361, 361, 361, 361, 361, 361, 
	361, 361, 361, 361, 356, 361, 356, 362, 
	362, 362, 356, 362, 362, 362, 362, 362, 
	362, 362, 362, 362, 362, 362, 362, 362, 
	362, 362, 362, 362, 362, 362, 362, 362, 
	356, 362, 356, 356, 360, 360, 360, 356, 
	360, 360, 360, 360, 360, 360, 360, 360, 
	360, 360, 360, 360, 360, 360, 360, 360, 
	360, 360, 360, 360, 360, 356, 360, 360, 
	360, 360, 360, 360, 360, 360, 360, 360, 
	360, 360, 360, 360, 360, 360, 360, 360, 
	360, 360, 360, 363, 363, 363, 363, 363, 
	363, 363, 363, 363, 363, 360, 360, 360, 
	360, 360, 360, 360, 363, 363, 363, 363, 
	363, 363, 360, 360, 360, 360, 360, 360, 
	360, 360, 360, 360, 360, 360, 360, 360, 
	360, 360, 360, 360, 360, 360, 360, 360, 
	360, 360, 360, 360, 363, 363, 363, 363, 
	363, 363, 360, 356, 361, 361, 361, 356, 
	361, 361, 361, 361, 361, 361, 361, 361, 
	361, 361, 361, 361, 361, 361, 361, 361, 
	361, 361, 361, 361, 361, 356, 361, 361, 
	361, 361, 361, 361, 361, 361, 361, 361, 
	361, 361, 361, 361, 361, 361, 361, 361, 
	361, 361, 361, 364, 364, 364, 364, 364, 
	364, 364, 364, 364, 364, 361, 361, 361, 
	361, 361, 361, 361, 364, 364, 364, 364, 
	364, 364, 361, 361, 361, 361, 361, 361, 
	361, 361, 361, 361, 361, 361, 361, 361, 
	361, 361, 361, 361, 361, 361, 361, 361, 
	361, 361, 361, 361, 364, 364, 364, 364, 
	364, 364, 361, 356, 362, 362, 362, 356, 
	362, 362, 362, 362, 362, 362, 362, 362, 
	362, 362, 362, 362, 362, 362, 362, 362, 
	362, 362, 362, 362, 362, 356, 362, 362, 
	362, 362, 362, 362, 362, 362, 362, 362, 
	362, 362, 362, 362, 362, 362, 362, 362, 
	362, 362, 362, 365, 365, 365, 365, 365, 
	365, 365, 365, 365, 365, 362, 362, 362, 
	362, 362, 362, 362, 365, 365, 365, 365, 
	365, 365, 362, 362, 362, 362, 362, 362, 
	362, 362, 362, 362, 362, 362, 362, 362, 
	362, 362, 362, 362, 362, 362, 362, 362, 
	362, 362, 362, 362, 365, 365, 365, 365, 
	365, 365, 362, 366, 369, 368, 368, 368, 
	369, 368, 368, 368, 368, 368, 368, 368, 
	368, 368, 368, 368, 368, 368, 368, 368, 
	368, 368, 368, 368, 368, 368, 369, 368, 
	368, 368, 368, 368, 368, 368, 368, 368, 
	368, 368, 368, 368, 368, 368, 368, 368, 
	368, 368, 368, 368, 370, 370, 370, 370, 
	370, 370, 370, 370, 370, 370, 368, 368, 
	368, 368, 368, 368, 368, 370, 370, 370, 
	370, 370, 370, 368, 368, 368, 368, 368, 
	368, 368, 368, 368, 368, 368, 368, 368, 
	368, 368, 368, 368, 368, 368, 368, 368, 
	368, 368, 368, 368, 368, 370, 370, 370, 
	370, 370, 370, 368, 368, 368, 368, 368, 
	368, 368, 368, 368, 368, 368, 368, 368, 
	368, 368, 368, 368, 368, 368, 368, 368, 
	368, 360, 368, 369, 371, 371, 371, 369, 
	371, 371, 371, 371, 371, 371, 371, 371, 
	371, 371, 371, 371, 371, 371, 371, 371, 
	371, 371, 371, 371, 371, 369, 371, 371, 
	371, 371, 371, 371, 371, 371, 371, 371, 
	371, 371, 371, 371, 371, 371, 371, 371, 
	371, 371, 371, 372, 372, 372, 372, 372, 
	372, 372, 372, 372, 372, 371, 371, 371, 
	371, 371, 371, 371, 372, 372, 372, 372, 
	372, 372, 371, 371, 371, 371, 371, 371, 
	371, 371, 371, 371, 371, 371, 371, 371, 
	371, 371, 371, 371, 371, 371, 371, 371, 
	371, 371, 371, 371, 372, 372, 372, 372, 
	372, 372, 371, 371, 371, 371, 371, 371, 
	371, 371, 371, 371, 371, 371, 371, 371, 
	371, 371, 371, 371, 371, 371, 371, 371, 
	373, 371, 369, 374, 374, 374, 369, 374, 
	374, 374, 374, 374, 374, 374, 374, 374, 
	374, 374, 374, 374, 374, 374, 374, 374, 
	374, 374, 374, 374, 369, 374, 374, 374, 
	374, 374, 374, 374, 374, 374, 374, 374, 
	374, 374, 374, 374, 374, 374, 374, 374, 
	374, 374, 375, 375, 375, 375, 375, 375, 
	375, 375, 375, 375, 374, 374, 374, 374, 
	374, 374, 374, 375, 375, 375, 375, 375, 
	375, 374, 374, 374, 374, 374, 374, 374, 
	374, 374, 374, 374, 374, 374, 374, 374, 
	374, 374, 374, 374, 374, 374, 374, 374, 
	374, 374, 374, 375, 375, 375, 375, 375, 
	375, 374, 374, 374, 374, 374, 374, 374, 
	374, 374, 374, 374, 374, 374, 374, 374, 
	374, 374, 374, 374, 374, 374, 374, 376, 
	374, 369, 377, 377, 377, 369, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 369, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 378, 378, 378, 378, 378, 378, 378, 
	378, 378, 378, 377, 377, 377, 377, 377, 
	377, 377, 378, 378, 378, 378, 378, 378, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 378, 378, 378, 378, 378, 378, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 369, 377, 
	369, 377, 377, 377, 369, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 369, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	378, 378, 378, 378, 378, 378, 378, 378, 
	378, 378, 377, 377, 377, 377, 377, 377, 
	377, 378, 378, 378, 378, 378, 378, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 378, 378, 378, 378, 378, 378, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 369, 377, 379, 
	369, 378, 378, 378, 369, 378, 378, 378, 
	378, 378, 378, 378, 378, 378, 378, 378, 
	378, 378, 378, 378, 378, 378, 378, 378, 
	378, 378, 369, 378, 378, 378, 378, 378, 
	378, 378, 378, 378, 378, 378, 378, 378, 
	378, 378, 378, 378, 378, 378, 378, 378, 
	378, 378, 378, 378, 378, 378, 378, 378, 
	378, 378, 378, 378, 378, 378, 378, 378, 
	378, 378, 378, 378, 378, 378, 378, 378, 
	378, 378, 378, 378, 378, 378, 378, 378, 
	378, 378, 378, 378, 378, 378, 378, 378, 
	378, 378, 378, 378, 378, 378, 378, 378, 
	378, 378, 378, 378, 378, 378, 378, 378, 
	378, 378, 378, 378, 378, 378, 378, 378, 
	378, 378, 378, 378, 378, 378, 378, 378, 
	378, 378, 378, 378, 378, 379, 378, 369, 
	378, 378, 378, 369, 378, 378, 378, 378, 
	378, 378, 378, 378, 378, 378, 378, 378, 
	378, 378, 378, 378, 378, 378, 378, 378, 
	378, 369, 378, 378, 378, 378, 378, 378, 
	378, 378, 378, 378, 378, 378, 378, 378, 
	378, 378, 378, 378, 378, 378, 378, 378, 
	378, 378, 378, 378, 378, 378, 378, 378, 
	378, 378, 378, 378, 378, 378, 378, 378, 
	378, 378, 378, 378, 378, 378, 378, 378, 
	378, 378, 378, 378, 378, 378, 378, 378, 
	378, 378, 378, 378, 378, 378, 378, 378, 
	378, 378, 378, 378, 378, 378, 378, 378, 
	378, 378, 378, 378, 378, 378, 378, 378, 
	378, 378, 378, 378, 378, 378, 378, 378, 
	378, 378, 378, 378, 378, 378, 378, 378, 
	378, 378, 378, 378, 367, 378, 367, 369, 
	375, 375, 375, 369, 375, 375, 375, 375, 
	375, 375, 375, 375, 375, 375, 375, 375, 
	375, 375, 375, 375, 375, 375, 375, 375, 
	375, 369, 375, 375, 375, 375, 375, 375, 
	375, 375, 375, 375, 375, 375, 375, 375, 
	375, 375, 375, 375, 375, 375, 375, 375, 
	375, 375, 375, 375, 375, 375, 375, 375, 
	375, 375, 375, 375, 375, 375, 375, 375, 
	375, 375, 375, 375, 375, 375, 375, 375, 
	375, 375, 375, 375, 375, 375, 375, 375, 
	375, 375, 375, 375, 375, 375, 375, 375, 
	375, 375, 375, 375, 375, 375, 375, 375, 
	375, 375, 375, 375, 375, 375, 375, 375, 
	375, 375, 375, 375, 375, 375, 375, 375, 
	375, 375, 375, 375, 375, 375, 375, 375, 
	375, 375, 375, 375, 362, 375, 367, 362, 
	362, 362, 367, 362, 362, 362, 362, 362, 
	362, 362, 362, 362, 362, 362, 362, 362, 
	362, 362, 362, 362, 362, 362, 362, 362, 
	367, 362, 369, 371, 371, 371, 369, 371, 
	371, 371, 371, 380, 371, 371, 371, 371, 
	371, 371, 371, 371, 371, 371, 371, 371, 
	371, 371, 371, 371, 369, 371, 371, 371, 
	371, 371, 380, 371, 371, 371, 371, 371, 
	371, 371, 371, 371, 371, 371, 371, 371, 
	371, 371, 381, 381, 381, 381, 381, 381, 
	381, 381, 381, 381, 371, 371, 371, 371, 
	371, 371, 371, 381, 381, 381, 381, 381, 
	381, 371, 371, 371, 371, 371, 371, 371, 
	371, 371, 371, 371, 371, 371, 371, 371, 
	371, 371, 371, 371, 371, 371, 371, 371, 
	371, 371, 371, 381, 381, 381, 381, 381, 
	381, 371, 371, 371, 371, 371, 371, 371, 
	371, 371, 371, 371, 371, 371, 371, 371, 
	371, 371, 371, 371, 371, 371, 371, 382, 
	371, 369, 377, 377, 377, 369, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 369, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 383, 383, 383, 383, 383, 383, 383, 
	383, 383, 383, 377, 377, 377, 377, 377, 
	377, 377, 383, 383, 383, 383, 383, 383, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 383, 383, 383, 383, 383, 383, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 369, 377, 
	369, 377, 377, 377, 369, 377, 377, 377, 
	377, 380, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 369, 377, 377, 377, 377, 377, 
	380, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	384, 384, 384, 384, 384, 384, 384, 384, 
	384, 384, 377, 377, 377, 377, 377, 377, 
	377, 384, 384, 384, 384, 384, 384, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 384, 384, 384, 384, 384, 384, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 382, 377, 369, 
	377, 377, 377, 369, 377, 377, 377, 377, 
	380, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 369, 377, 377, 377, 377, 377, 380, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 385, 
	385, 385, 385, 385, 385, 385, 385, 385, 
	385, 377, 377, 377, 377, 377, 377, 377, 
	385, 385, 385, 385, 385, 385, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	385, 385, 385, 385, 385, 385, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 382, 377, 369, 377, 
	377, 377, 369, 377, 377, 377, 377, 380, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	369, 377, 377, 377, 377, 377, 380, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 386, 386, 
	386, 386, 386, 386, 386, 386, 386, 386, 
	377, 377, 377, 377, 377, 377, 377, 386, 
	386, 386, 386, 386, 386, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 386, 
	386, 386, 386, 386, 386, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 382, 377, 369, 377, 377, 
	377, 369, 377, 377, 377, 377, 380, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 369, 
	377, 377, 377, 377, 377, 380, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 387, 387, 387, 
	387, 387, 387, 387, 387, 387, 387, 377, 
	377, 377, 377, 377, 377, 377, 387, 387, 
	387, 387, 387, 387, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 387, 387, 
	387, 387, 387, 387, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 382, 377, 369, 377, 377, 377, 
	369, 377, 377, 377, 377, 380, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 369, 377, 
	377, 377, 377, 377, 380, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 388, 388, 388, 388, 
	388, 388, 388, 388, 388, 388, 377, 377, 
	377, 377, 377, 377, 377, 388, 388, 388, 
	388, 388, 388, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 388, 388, 388, 
	388, 388, 388, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 382, 377, 369, 377, 377, 377, 369, 
	377, 377, 377, 377, 380, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 369, 377, 377, 
	377, 377, 377, 380, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 389, 389, 389, 389, 389, 
	389, 389, 389, 389, 389, 377, 377, 377, 
	377, 377, 377, 377, 389, 389, 389, 389, 
	389, 389, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 389, 389, 389, 389, 
	389, 389, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	382, 377, 369, 377, 377, 377, 369, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 369, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 389, 389, 389, 389, 389, 389, 
	389, 389, 389, 389, 377, 377, 377, 377, 
	377, 377, 377, 389, 389, 389, 389, 389, 
	389, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 389, 389, 389, 389, 389, 
	389, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 377, 
	377, 377, 377, 377, 377, 377, 377, 379, 
	377, 390, 369, 374, 374, 374, 369, 374, 
	374, 374, 374, 380, 374, 374, 374, 374, 
	374, 374, 374, 374, 374, 374, 374, 374, 
	374, 374, 374, 374, 369, 374, 374, 374, 
	374, 374, 380, 374, 374, 374, 374, 374, 
	374, 374, 374, 374, 374, 374, 374, 374, 
	374, 374, 385, 385, 385, 385, 385, 385, 
	385, 385, 385, 385, 374, 374, 374, 374, 
	374, 374, 374, 385, 385, 385, 385, 385, 
	385, 374, 374, 374, 374, 374, 374, 374, 
	374, 374, 374, 374, 374, 374, 374, 374, 
	374, 374, 374, 374, 374, 374, 374, 374, 
	374, 374, 374, 385, 385, 385, 385, 385, 
	385, 374, 374, 374, 374, 374, 374, 374, 
	374, 374, 374, 374, 374, 374, 374, 374, 
	374, 374, 374, 374, 374, 374, 374, 382, 
	374, 391, 393, 393, 393, 393, 393, 393, 
	393, 393, 393, 393, 392, 392, 392, 392, 
	392, 392, 392, 393, 393, 393, 393, 393, 
	393, 392, 392, 392, 392, 392, 392, 392, 
	392, 392, 392, 392, 392, 392, 392, 392, 
	392, 392, 392, 392, 392, 392, 392, 392, 
	392, 392, 392, 393, 393, 393, 393, 393, 
	393, 392, 392, 395, 394, 394, 394, 395, 
	394, 394, 394, 394, 394, 396, 394, 394, 
	394, 394, 394, 394, 394, 394, 394, 394, 
	394, 394, 394, 394, 394, 395, 394, 394, 
	394, 394, 394, 394, 394, 394, 397, 394, 
	394, 394, 394, 394, 394, 394, 394, 394, 
	394, 394, 394, 394, 394, 394, 394, 394, 
	394, 394, 394, 394, 394, 394, 394, 394, 
	394, 394, 394, 394, 394, 394, 394, 394, 
	394, 394, 394, 394, 394, 394, 394, 394, 
	394, 394, 394, 394, 394, 394, 394, 394, 
	394, 394, 394, 394, 394, 394, 394, 398, 
	394, 400, 399, 399, 399, 399, 399, 399, 
	399, 399, 399, 399, 399, 399, 399, 399, 
	399, 399, 399, 399, 399, 399, 399, 399, 
	399, 399, 399, 399, 399, 401, 399, 399, 
	399, 399, 399, 399, 399, 399, 399, 399, 
	399, 399, 399, 399, 399, 399, 399, 399, 
	399, 399, 399, 399, 399, 399, 399, 399, 
	399, 399, 399, 399, 399, 399, 399, 399, 
	399, 399, 399, 399, 399, 399, 399, 399, 
	399, 399, 399, 399, 399, 399, 399, 399, 
	399, 399, 399, 399, 399, 399, 399, 399, 
	402, 399, 403, 403, 403, 403, 403, 403, 
	403, 403, 403, 403, 403, 403, 403, 403, 
	403, 403, 403, 403, 403, 403, 403, 403, 
	403, 403, 403, 403, 403, 403, 403, 403, 
	403, 403, 403, 403, 403, 403, 403, 403, 
	403, 403, 403, 403, 403, 403, 403, 403, 
	403, 403, 44, 44, 44, 44, 44, 44, 
	44, 44, 44, 44, 403, 403, 403, 403, 
	403, 403, 403, 44, 44, 44, 44, 44, 
	44, 44, 44, 44, 44, 44, 44, 44, 
	44, 44, 44, 44, 44, 44, 44, 44, 
	44, 44, 44, 44, 44, 403, 403, 403, 
	403, 44, 403, 44, 44, 44, 44, 44, 
	44, 44, 44, 44, 44, 44, 44, 44, 
	44, 44, 44, 44, 44, 44, 44, 44, 
	44, 44, 44, 44, 44, 403, 403, 403, 
	403, 403, 44, 47, 47, 47, 47, 47, 
	47, 47, 47, 47, 47, 403, 406, 405, 
	405, 405, 406, 405, 405, 405, 405, 405, 
	405, 405, 405, 405, 405, 405, 405, 405, 
	405, 405, 405, 405, 405, 405, 405, 405, 
	406, 405, 405, 405, 405, 405, 405, 405, 
	405, 405, 405, 405, 405, 405, 405, 405, 
	405, 405, 405, 405, 405, 405, 407, 407, 
	407, 407, 407, 407, 407, 407, 405, 405, 
	405, 405, 405, 405, 405, 405, 405, 405, 
	405, 408, 405, 405, 405, 405, 405, 405, 
	405, 405, 405, 409, 405, 405, 405, 405, 
	405, 405, 405, 405, 405, 405, 405, 405, 
	405, 405, 405, 405, 405, 405, 405, 405, 
	405, 410, 405, 405, 405, 405, 405, 405, 
	405, 405, 405, 405, 405, 405, 405, 405, 
	405, 405, 405, 411, 405, 405, 412, 405, 
	413, 414, 416, 416, 416, 416, 416, 416, 
	416, 416, 415, 417, 417, 417, 417, 417, 
	417, 417, 417, 415, 415, 418, 418, 69, 
	69, 69, 418, 69, 69, 69, 69, 69, 
	69, 69, 69, 69, 69, 69, 69, 69, 
	69, 69, 69, 69, 69, 69, 69, 69, 
	418, 69, 69, 69, 69, 69, 69, 69, 
	69, 69, 69, 69, 69, 69, 69, 69, 
	69, 69, 69, 69, 69, 69, 69, 69, 
	69, 69, 69, 69, 69, 69, 69, 69, 
	69, 69, 69, 69, 69, 70, 69, 69, 
	69, 69, 69, 69, 69, 69, 69, 69, 
	69, 69, 69, 69, 69, 69, 69, 69, 
	69, 69, 69, 69, 69, 69, 69, 69, 
	69, 69, 419, 69, 420, 421, 422, 422, 
	69, 69, 69, 422, 69, 69, 69, 69, 
	69, 69, 69, 69, 69, 69, 69, 69, 
	69, 69, 69, 69, 69, 69, 69, 69, 
	69, 422, 69, 69, 69, 69, 69, 69, 
	69, 69, 69, 69, 69, 69, 69, 69, 
	69, 69, 69, 69, 69, 69, 69, 69, 
	69, 69, 69, 69, 69, 69, 69, 69, 
	69, 69, 69, 69, 69, 69, 70, 69, 
	69, 69, 69, 69, 69, 69, 69, 69, 
	69, 69, 69, 69, 69, 69, 69, 69, 
	69, 69, 69, 69, 69, 69, 69, 69, 
	69, 69, 69, 423, 69, 68, 422, 424, 
	425, 426, 426, 56, 56, 56, 426, 56, 
	56, 56, 56, 56, 56, 56, 56, 56, 
	56, 56, 56, 56, 56, 56, 56, 56, 
	56, 56, 56, 56, 426, 56, 56, 56, 
	56, 56, 56, 56, 56, 56, 56, 56, 
	56, 56, 56, 56, 56, 56, 56, 56, 
	56, 56, 56, 56, 56, 56, 56, 56, 
	56, 56, 56, 56, 56, 56, 56, 56, 
	56, 57, 56, 56, 56, 56, 56, 56, 
	56, 56, 56, 56, 56, 56, 56, 56, 
	56, 56, 56, 56, 56, 56, 56, 56, 
	56, 56, 56, 56, 56, 56, 427, 56, 
	422, 50, 50, 50, 422, 50, 50, 50, 
	50, 50, 50, 50, 50, 50, 50, 50, 
	50, 50, 50, 50, 50, 50, 50, 50, 
	50, 50, 422, 50, 50, 50, 50, 50, 
	50, 50, 50, 50, 50, 50, 50, 50, 
	50, 50, 50, 50, 50, 428, 50, 418, 
	56, 56, 56, 418, 56, 56, 56, 56, 
	56, 56, 56, 56, 56, 56, 56, 56, 
	56, 56, 56, 56, 56, 56, 56, 56, 
	56, 418, 56, 56, 56, 56, 56, 56, 
	56, 56, 56, 56, 56, 56, 56, 56, 
	56, 56, 56, 56, 56, 56, 56, 56, 
	56, 56, 56, 56, 56, 56, 56, 56, 
	56, 56, 56, 56, 56, 56, 57, 56, 
	56, 56, 56, 56, 56, 56, 56, 56, 
	56, 56, 56, 56, 56, 56, 56, 56, 
	56, 56, 56, 56, 56, 56, 56, 56, 
	56, 56, 56, 58, 56, 418, 429, 429, 
	429, 418, 429, 429, 429, 429, 429, 429, 
	429, 429, 429, 429, 429, 429, 429, 429, 
	429, 429, 429, 429, 429, 429, 429, 418, 
	429, 429, 429, 429, 429, 429, 429, 429, 
	429, 429, 429, 429, 429, 429, 429, 429, 
	429, 429, 429, 429, 429, 429, 429, 429, 
	429, 429, 429, 429, 429, 429, 429, 429, 
	429, 429, 429, 429, 429, 429, 429, 429, 
	429, 429, 429, 429, 429, 429, 429, 429, 
	429, 429, 429, 429, 429, 429, 429, 429, 
	429, 429, 429, 429, 429, 429, 429, 429, 
	429, 430, 429, 431, 432, 432, 429, 429, 
	429, 432, 429, 429, 429, 429, 429, 429, 
	429, 429, 429, 429, 429, 429, 429, 429, 
	429, 429, 429, 429, 429, 429, 429, 432, 
	429, 429, 429, 429, 429, 429, 429, 429, 
	429, 429, 429, 429, 429, 429, 429, 429, 
	429, 429, 429, 429, 429, 429, 429, 429, 
	429, 429, 429, 429, 429, 429, 429, 429, 
	429, 429, 429, 429, 429, 429, 429, 429, 
	429, 429, 429, 429, 429, 429, 429, 429, 
	429, 429, 429, 429, 429, 429, 429, 429, 
	429, 429, 429, 429, 429, 429, 429, 429, 
	429, 433, 429, 432, 50, 50, 50, 432, 
	50, 50, 50, 50, 50, 50, 50, 50, 
	50, 50, 50, 50, 50, 50, 50, 50, 
	50, 50, 50, 50, 50, 432, 50, 50, 
	50, 50, 50, 50, 50, 50, 50, 50, 
	50, 50, 50, 50, 50, 50, 50, 50, 
	428, 50, 434, 432, 432, 56, 56, 56, 
	432, 56, 56, 56, 56, 56, 56, 56, 
	56, 56, 56, 56, 56, 56, 56, 56, 
	56, 56, 56, 56, 56, 56, 432, 56, 
	56, 56, 56, 56, 56, 56, 56, 56, 
	56, 56, 56, 56, 56, 56, 56, 56, 
	56, 56, 56, 56, 56, 56, 56, 56, 
	56, 56, 56, 56, 56, 56, 56, 56, 
	56, 56, 56, 57, 56, 56, 56, 56, 
	56, 56, 56, 56, 56, 56, 56, 56, 
	56, 56, 56, 56, 56, 56, 56, 56, 
	56, 56, 56, 56, 56, 56, 56, 56, 
	58, 56, 435, 436, 436, 436, 435, 436, 
	436, 436, 436, 436, 436, 436, 436, 436, 
	436, 436, 436, 436, 436, 436, 436, 436, 
	436, 436, 436, 436, 435, 436, 436, 436, 
	436, 436, 436, 436, 436, 436, 436, 436, 
	436, 436, 436, 436, 436, 436, 436, 436, 
	436, 436, 437, 437, 437, 437, 437, 437, 
	437, 437, 437, 437, 436, 436, 436, 436, 
	436, 436, 436, 437, 437, 437, 437, 437, 
	437, 436, 436, 436, 436, 436, 436, 436, 
	436, 436, 436, 436, 436, 436, 436, 436, 
	436, 436, 436, 436, 436, 436, 436, 436, 
	436, 436, 436, 437, 437, 437, 437, 437, 
	437, 436, 436, 436, 436, 436, 436, 436, 
	436, 436, 436, 436, 436, 436, 436, 436, 
	436, 436, 436, 436, 436, 438, 436, 435, 
	439, 439, 439, 435, 439, 439, 439, 439, 
	439, 439, 439, 439, 439, 439, 439, 439, 
	439, 439, 439, 439, 439, 439, 439, 439, 
	439, 435, 439, 435, 440, 440, 440, 435, 
	440, 440, 440, 440, 440, 440, 440, 440, 
	440, 440, 440, 440, 440, 440, 440, 440, 
	440, 440, 440, 440, 440, 435, 440, 435, 
	441, 441, 441, 435, 441, 441, 441, 441, 
	441, 441, 441, 441, 441, 441, 441, 441, 
	441, 441, 441, 441, 441, 441, 441, 441, 
	441, 435, 441, 435, 435, 439, 439, 439, 
	435, 439, 439, 439, 439, 439, 439, 439, 
	439, 439, 439, 439, 439, 439, 439, 439, 
	439, 439, 439, 439, 439, 439, 435, 439, 
	439, 439, 439, 439, 439, 439, 439, 439, 
	439, 439, 439, 439, 439, 439, 439, 439, 
	439, 439, 439, 439, 442, 442, 442, 442, 
	442, 442, 442, 442, 442, 442, 439, 439, 
	439, 439, 439, 439, 439, 442, 442, 442, 
	442, 442, 442, 439, 439, 439, 439, 439, 
	439, 439, 439, 439, 439, 439, 439, 439, 
	439, 439, 439, 439, 439, 439, 439, 439, 
	439, 439, 439, 439, 439, 442, 442, 442, 
	442, 442, 442, 439, 435, 440, 440, 440, 
	435, 440, 440, 440, 440, 440, 440, 440, 
	440, 440, 440, 440, 440, 440, 440, 440, 
	440, 440, 440, 440, 440, 440, 435, 440, 
	440, 440, 440, 440, 440, 440, 440, 440, 
	440, 440, 440, 440, 440, 440, 440, 440, 
	440, 440, 440, 440, 443, 443, 443, 443, 
	443, 443, 443, 443, 443, 443, 440, 440, 
	440, 440, 440, 440, 440, 443, 443, 443, 
	443, 443, 443, 440, 440, 440, 440, 440, 
	440, 440, 440, 440, 440, 440, 440, 440, 
	440, 440, 440, 440, 440, 440, 440, 440, 
	440, 440, 440, 440, 440, 443, 443, 443, 
	443, 443, 443, 440, 435, 441, 441, 441, 
	435, 441, 441, 441, 441, 441, 441, 441, 
	441, 441, 441, 441, 441, 441, 441, 441, 
	441, 441, 441, 441, 441, 441, 435, 441, 
	441, 441, 441, 441, 441, 441, 441, 441, 
	441, 441, 441, 441, 441, 441, 441, 441, 
	441, 441, 441, 441, 444, 444, 444, 444, 
	444, 444, 444, 444, 444, 444, 441, 441, 
	441, 441, 441, 441, 441, 444, 444, 444, 
	444, 444, 444, 441, 441, 441, 441, 441, 
	441, 441, 441, 441, 441, 441, 441, 441, 
	441, 441, 441, 441, 441, 441, 441, 441, 
	441, 441, 441, 441, 441, 444, 444, 444, 
	444, 444, 444, 441, 445, 448, 447, 447, 
	447, 448, 447, 447, 447, 447, 447, 447, 
	447, 447, 447, 447, 447, 447, 447, 447, 
	447, 447, 447, 447, 447, 447, 447, 448, 
	447, 447, 447, 447, 447, 447, 447, 447, 
	447, 447, 447, 447, 447, 447, 447, 447, 
	447, 447, 447, 447, 447, 449, 449, 449, 
	449, 449, 449, 449, 449, 449, 449, 447, 
	447, 447, 447, 447, 447, 447, 449, 449, 
	449, 449, 449, 449, 447, 447, 447, 447, 
	447, 447, 447, 447, 447, 447, 447, 447, 
	447, 447, 447, 447, 447, 447, 447, 447, 
	447, 447, 447, 447, 447, 447, 449, 449, 
	449, 449, 449, 449, 447, 447, 447, 447, 
	447, 447, 447, 447, 447, 447, 447, 447, 
	447, 447, 447, 447, 447, 447, 447, 447, 
	447, 447, 439, 447, 448, 450, 450, 450, 
	448, 450, 450, 450, 450, 450, 450, 450, 
	450, 450, 450, 450, 450, 450, 450, 450, 
	450, 450, 450, 450, 450, 450, 448, 450, 
	450, 450, 450, 450, 450, 450, 450, 450, 
	450, 450, 450, 450, 450, 450, 450, 450, 
	450, 450, 450, 450, 451, 451, 451, 451, 
	451, 451, 451, 451, 451, 451, 450, 450, 
	450, 450, 450, 450, 450, 451, 451, 451, 
	451, 451, 451, 450, 450, 450, 450, 450, 
	450, 450, 450, 450, 450, 450, 450, 450, 
	450, 450, 450, 450, 450, 450, 450, 450, 
	450, 450, 450, 450, 450, 451, 451, 451, 
	451, 451, 451, 450, 450, 450, 450, 450, 
	450, 450, 450, 450, 450, 450, 450, 450, 
	450, 450, 450, 450, 450, 450, 450, 450, 
	450, 452, 450, 448, 453, 453, 453, 448, 
	453, 453, 453, 453, 453, 453, 453, 453, 
	453, 453, 453, 453, 453, 453, 453, 453, 
	453, 453, 453, 453, 453, 448, 453, 453, 
	453, 453, 453, 453, 453, 453, 453, 453, 
	453, 453, 453, 453, 453, 453, 453, 453, 
	453, 453, 453, 454, 454, 454, 454, 454, 
	454, 454, 454, 454, 454, 453, 453, 453, 
	453, 453, 453, 453, 454, 454, 454, 454, 
	454, 454, 453, 453, 453, 453, 453, 453, 
	453, 453, 453, 453, 453, 453, 453, 453, 
	453, 453, 453, 453, 453, 453, 453, 453, 
	453, 453, 453, 453, 454, 454, 454, 454, 
	454, 454, 453, 453, 453, 453, 453, 453, 
	453, 453, 453, 453, 453, 453, 453, 453, 
	453, 453, 453, 453, 453, 453, 453, 453, 
	455, 453, 448, 456, 456, 456, 448, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 448, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 457, 457, 457, 457, 457, 457, 
	457, 457, 457, 457, 456, 456, 456, 456, 
	456, 456, 456, 457, 457, 457, 457, 457, 
	457, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 457, 457, 457, 457, 457, 
	457, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 448, 
	456, 448, 456, 456, 456, 448, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 448, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 457, 457, 457, 457, 457, 457, 457, 
	457, 457, 457, 456, 456, 456, 456, 456, 
	456, 456, 457, 457, 457, 457, 457, 457, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 457, 457, 457, 457, 457, 457, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 448, 456, 
	458, 448, 457, 457, 457, 448, 457, 457, 
	457, 457, 457, 457, 457, 457, 457, 457, 
	457, 457, 457, 457, 457, 457, 457, 457, 
	457, 457, 457, 448, 457, 457, 457, 457, 
	457, 457, 457, 457, 457, 457, 457, 457, 
	457, 457, 457, 457, 457, 457, 457, 457, 
	457, 457, 457, 457, 457, 457, 457, 457, 
	457, 457, 457, 457, 457, 457, 457, 457, 
	457, 457, 457, 457, 457, 457, 457, 457, 
	457, 457, 457, 457, 457, 457, 457, 457, 
	457, 457, 457, 457, 457, 457, 457, 457, 
	457, 457, 457, 457, 457, 457, 457, 457, 
	457, 457, 457, 457, 457, 457, 457, 457, 
	457, 457, 457, 457, 457, 457, 457, 457, 
	457, 457, 457, 457, 457, 457, 457, 457, 
	457, 457, 457, 457, 457, 457, 458, 457, 
	448, 457, 457, 457, 448, 457, 457, 457, 
	457, 457, 457, 457, 457, 457, 457, 457, 
	457, 457, 457, 457, 457, 457, 457, 457, 
	457, 457, 448, 457, 457, 457, 457, 457, 
	457, 457, 457, 457, 457, 457, 457, 457, 
	457, 457, 457, 457, 457, 457, 457, 457, 
	457, 457, 457, 457, 457, 457, 457, 457, 
	457, 457, 457, 457, 457, 457, 457, 457, 
	457, 457, 457, 457, 457, 457, 457, 457, 
	457, 457, 457, 457, 457, 457, 457, 457, 
	457, 457, 457, 457, 457, 457, 457, 457, 
	457, 457, 457, 457, 457, 457, 457, 457, 
	457, 457, 457, 457, 457, 457, 457, 457, 
	457, 457, 457, 457, 457, 457, 457, 457, 
	457, 457, 457, 457, 457, 457, 457, 457, 
	457, 457, 457, 457, 457, 446, 457, 446, 
	448, 454, 454, 454, 448, 454, 454, 454, 
	454, 454, 454, 454, 454, 454, 454, 454, 
	454, 454, 454, 454, 454, 454, 454, 454, 
	454, 454, 448, 454, 454, 454, 454, 454, 
	454, 454, 454, 454, 454, 454, 454, 454, 
	454, 454, 454, 454, 454, 454, 454, 454, 
	454, 454, 454, 454, 454, 454, 454, 454, 
	454, 454, 454, 454, 454, 454, 454, 454, 
	454, 454, 454, 454, 454, 454, 454, 454, 
	454, 454, 454, 454, 454, 454, 454, 454, 
	454, 454, 454, 454, 454, 454, 454, 454, 
	454, 454, 454, 454, 454, 454, 454, 454, 
	454, 454, 454, 454, 454, 454, 454, 454, 
	454, 454, 454, 454, 454, 454, 454, 454, 
	454, 454, 454, 454, 454, 454, 454, 454, 
	454, 454, 454, 454, 454, 441, 454, 446, 
	441, 441, 441, 446, 441, 441, 441, 441, 
	441, 441, 441, 441, 441, 441, 441, 441, 
	441, 441, 441, 441, 441, 441, 441, 441, 
	441, 446, 441, 448, 450, 450, 450, 448, 
	450, 450, 450, 450, 459, 450, 450, 450, 
	450, 450, 450, 450, 450, 450, 450, 450, 
	450, 450, 450, 450, 450, 448, 450, 450, 
	450, 450, 450, 459, 450, 450, 450, 450, 
	450, 450, 450, 450, 450, 450, 450, 450, 
	450, 450, 450, 460, 460, 460, 460, 460, 
	460, 460, 460, 460, 460, 450, 450, 450, 
	450, 450, 450, 450, 460, 460, 460, 460, 
	460, 460, 450, 450, 450, 450, 450, 450, 
	450, 450, 450, 450, 450, 450, 450, 450, 
	450, 450, 450, 450, 450, 450, 450, 450, 
	450, 450, 450, 450, 460, 460, 460, 460, 
	460, 460, 450, 450, 450, 450, 450, 450, 
	450, 450, 450, 450, 450, 450, 450, 450, 
	450, 450, 450, 450, 450, 450, 450, 450, 
	461, 450, 448, 456, 456, 456, 448, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 448, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 462, 462, 462, 462, 462, 462, 
	462, 462, 462, 462, 456, 456, 456, 456, 
	456, 456, 456, 462, 462, 462, 462, 462, 
	462, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 462, 462, 462, 462, 462, 
	462, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 448, 
	456, 448, 456, 456, 456, 448, 456, 456, 
	456, 456, 459, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 448, 456, 456, 456, 456, 
	456, 459, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 463, 463, 463, 463, 463, 463, 463, 
	463, 463, 463, 456, 456, 456, 456, 456, 
	456, 456, 463, 463, 463, 463, 463, 463, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 463, 463, 463, 463, 463, 463, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 461, 456, 
	448, 456, 456, 456, 448, 456, 456, 456, 
	456, 459, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 448, 456, 456, 456, 456, 456, 
	459, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	464, 464, 464, 464, 464, 464, 464, 464, 
	464, 464, 456, 456, 456, 456, 456, 456, 
	456, 464, 464, 464, 464, 464, 464, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 464, 464, 464, 464, 464, 464, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 461, 456, 448, 
	456, 456, 456, 448, 456, 456, 456, 456, 
	459, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 448, 456, 456, 456, 456, 456, 459, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 465, 
	465, 465, 465, 465, 465, 465, 465, 465, 
	465, 456, 456, 456, 456, 456, 456, 456, 
	465, 465, 465, 465, 465, 465, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	465, 465, 465, 465, 465, 465, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 461, 456, 448, 456, 
	456, 456, 448, 456, 456, 456, 456, 459, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	448, 456, 456, 456, 456, 456, 459, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 466, 466, 
	466, 466, 466, 466, 466, 466, 466, 466, 
	456, 456, 456, 456, 456, 456, 456, 466, 
	466, 466, 466, 466, 466, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 466, 
	466, 466, 466, 466, 466, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 461, 456, 448, 456, 456, 
	456, 448, 456, 456, 456, 456, 459, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 448, 
	456, 456, 456, 456, 456, 459, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 467, 467, 467, 
	467, 467, 467, 467, 467, 467, 467, 456, 
	456, 456, 456, 456, 456, 456, 467, 467, 
	467, 467, 467, 467, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 467, 467, 
	467, 467, 467, 467, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 461, 456, 448, 456, 456, 456, 
	448, 456, 456, 456, 456, 459, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 448, 456, 
	456, 456, 456, 456, 459, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 468, 468, 468, 468, 
	468, 468, 468, 468, 468, 468, 456, 456, 
	456, 456, 456, 456, 456, 468, 468, 468, 
	468, 468, 468, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 468, 468, 468, 
	468, 468, 468, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 461, 456, 448, 456, 456, 456, 448, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 448, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 468, 468, 468, 468, 468, 
	468, 468, 468, 468, 468, 456, 456, 456, 
	456, 456, 456, 456, 468, 468, 468, 468, 
	468, 468, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 468, 468, 468, 468, 
	468, 468, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 456, 456, 
	458, 456, 469, 448, 453, 453, 453, 448, 
	453, 453, 453, 453, 459, 453, 453, 453, 
	453, 453, 453, 453, 453, 453, 453, 453, 
	453, 453, 453, 453, 453, 448, 453, 453, 
	453, 453, 453, 459, 453, 453, 453, 453, 
	453, 453, 453, 453, 453, 453, 453, 453, 
	453, 453, 453, 464, 464, 464, 464, 464, 
	464, 464, 464, 464, 464, 453, 453, 453, 
	453, 453, 453, 453, 464, 464, 464, 464, 
	464, 464, 453, 453, 453, 453, 453, 453, 
	453, 453, 453, 453, 453, 453, 453, 453, 
	453, 453, 453, 453, 453, 453, 453, 453, 
	453, 453, 453, 453, 464, 464, 464, 464, 
	464, 464, 453, 453, 453, 453, 453, 453, 
	453, 453, 453, 453, 453, 453, 453, 453, 
	453, 453, 453, 453, 453, 453, 453, 453, 
	461, 453, 470, 472, 472, 472, 472, 472, 
	472, 472, 472, 472, 472, 471, 471, 471, 
	471, 471, 471, 471, 472, 472, 472, 472, 
	472, 472, 471, 471, 471, 471, 471, 471, 
	471, 471, 471, 471, 471, 471, 471, 471, 
	471, 471, 471, 471, 471, 471, 471, 471, 
	471, 471, 471, 471, 472, 472, 472, 472, 
	472, 472, 471, 471, 474, 473, 473, 473, 
	474, 473, 473, 473, 473, 475, 476, 475, 
	475, 475, 473, 473, 473, 473, 473, 473, 
	473, 473, 473, 473, 473, 473, 474, 473, 
	473, 473, 473, 473, 475, 473, 473, 473, 
	473, 473, 473, 473, 473, 473, 473, 473, 
	473, 473, 473, 473, 473, 473, 473, 473, 
	473, 473, 473, 473, 473, 473, 473, 473, 
	473, 473, 473, 473, 473, 473, 473, 473, 
	473, 473, 473, 473, 473, 473, 473, 473, 
	473, 473, 473, 473, 473, 473, 473, 473, 
	473, 473, 473, 473, 473, 473, 473, 473, 
	477, 473, 475, 478, 475, 475, 475, 478, 
	478, 478, 478, 478, 478, 478, 478, 478, 
	478, 478, 478, 478, 478, 478, 478, 478, 
	478, 475, 478, 479, 480, 480, 480, 479, 
	480, 480, 480, 480, 480, 480, 480, 480, 
	480, 480, 480, 480, 480, 480, 480, 480, 
	480, 480, 480, 480, 480, 479, 480, 482, 
	481, 481, 481, 482, 481, 481, 481, 481, 
	481, 483, 481, 481, 481, 481, 481, 481, 
	481, 481, 481, 481, 481, 481, 481, 481, 
	481, 482, 481, 481, 481, 481, 481, 481, 
	481, 481, 481, 481, 481, 481, 481, 481, 
	481, 481, 481, 481, 481, 481, 481, 481, 
	481, 481, 481, 481, 481, 481, 481, 481, 
	481, 481, 481, 481, 481, 481, 481, 481, 
	481, 481, 481, 481, 481, 481, 481, 481, 
	481, 481, 481, 481, 481, 481, 481, 481, 
	481, 481, 481, 481, 481, 481, 481, 481, 
	481, 481, 481, 484, 481, 485, 486, 486, 
	486, 485, 486, 486, 486, 486, 486, 487, 
	486, 486, 486, 486, 486, 486, 486, 486, 
	486, 486, 486, 486, 486, 486, 486, 485, 
	486, 489, 488, 488, 488, 489, 488, 488, 
	488, 488, 488, 490, 488, 488, 488, 488, 
	488, 488, 488, 488, 488, 488, 488, 488, 
	488, 488, 488, 489, 488, 488, 488, 488, 
	488, 488, 488, 488, 491, 488, 493, 492, 
	492, 492, 492, 492, 492, 492, 492, 492, 
	492, 492, 492, 492, 492, 492, 492, 492, 
	492, 492, 492, 492, 492, 492, 492, 492, 
	492, 492, 494, 492, 492, 492, 492, 492, 
	492, 492, 492, 492, 492, 492, 492, 492, 
	492, 492, 492, 492, 492, 492, 492, 492, 
	492, 492, 492, 492, 492, 492, 492, 492, 
	492, 492, 492, 492, 492, 492, 492, 492, 
	492, 492, 492, 492, 492, 492, 492, 492, 
	492, 492, 492, 492, 492, 492, 492, 492, 
	492, 492, 492, 492, 492, 495, 492, 496, 
	496, 496, 496, 496, 496, 496, 496, 496, 
	496, 496, 496, 496, 496, 496, 496, 496, 
	496, 496, 496, 496, 496, 496, 496, 496, 
	496, 496, 496, 496, 496, 496, 496, 496, 
	496, 496, 496, 496, 496, 496, 496, 496, 
	496, 496, 496, 496, 496, 496, 496, 75, 
	75, 75, 75, 75, 75, 75, 75, 75, 
	75, 496, 496, 496, 496, 496, 496, 496, 
	75, 75, 75, 75, 75, 75, 75, 75, 
	75, 75, 75, 75, 75, 75, 75, 75, 
	75, 75, 75, 75, 75, 75, 75, 75, 
	75, 75, 496, 496, 496, 496, 75, 496, 
	75, 75, 75, 75, 75, 75, 75, 75, 
	75, 75, 75, 75, 75, 75, 75, 75, 
	75, 75, 75, 75, 75, 75, 75, 75, 
	75, 75, 496, 496, 496, 496, 496, 75, 
	78, 78, 78, 78, 78, 78, 78, 78, 
	78, 78, 496, 498, 497, 497, 497, 498, 
	497, 497, 497, 497, 497, 499, 497, 497, 
	497, 497, 497, 497, 497, 497, 497, 497, 
	497, 497, 497, 497, 497, 498, 497, 501, 
	501, 501, 501, 501, 501, 501, 501, 501, 
	501, 501, 501, 501, 501, 501, 501, 501, 
	501, 501, 501, 501, 501, 501, 501, 501, 
	501, 500, 500, 500, 500, 500, 500, 501, 
	501, 501, 501, 501, 501, 501, 501, 501, 
	501, 501, 501, 501, 501, 501, 501, 501, 
	501, 501, 501, 501, 501, 501, 501, 501, 
	501, 500, 501, 501, 501, 501, 501, 501, 
	501, 501, 501, 501, 501, 501, 501, 501, 
	501, 501, 501, 501, 501, 501, 501, 501, 
	501, 501, 501, 501, 502, 502, 502, 502, 
	502, 502, 501, 501, 501, 501, 501, 501, 
	501, 501, 501, 501, 501, 501, 501, 501, 
	501, 501, 501, 501, 501, 501, 501, 501, 
	501, 501, 501, 501, 502, 503, 82, 82, 
	82, 82, 82, 82, 82, 82, 82, 82, 
	82, 82, 82, 82, 82, 82, 82, 82, 
	82, 82, 82, 82, 82, 82, 82, 82, 
	82, 504, 82, 505, 505, 505, 505, 505, 
	505, 505, 505, 505, 505, 505, 505, 505, 
	505, 505, 505, 505, 505, 505, 505, 505, 
	505, 505, 505, 505, 505, 505, 505, 505, 
	505, 505, 505, 505, 505, 505, 505, 505, 
	505, 505, 505, 505, 505, 505, 505, 505, 
	505, 505, 505, 80, 80, 80, 80, 80, 
	80, 80, 80, 80, 80, 505, 505, 505, 
	505, 505, 505, 505, 80, 80, 80, 80, 
	80, 80, 80, 80, 80, 80, 80, 80, 
	80, 80, 80, 80, 80, 80, 80, 80, 
	80, 80, 80, 80, 80, 80, 505, 505, 
	505, 505, 80, 505, 80, 80, 80, 80, 
	80, 80, 80, 80, 80, 80, 80, 80, 
	80, 80, 80, 80, 80, 80, 80, 80, 
	80, 80, 80, 80, 80, 80, 505, 505, 
	505, 505, 505, 80, 84, 84, 84, 84, 
	84, 84, 84, 84, 84, 84, 505, 506, 
	506, 506, 506, 506, 506, 506, 506, 506, 
	506, 506, 506, 506, 506, 506, 506, 506, 
	506, 506, 506, 506, 506, 506, 506, 506, 
	506, 506, 506, 506, 506, 506, 506, 506, 
	506, 506, 506, 506, 506, 506, 506, 506, 
	506, 506, 506, 506, 506, 506, 506, 85, 
	85, 85, 85, 85, 85, 85, 85, 85, 
	85, 506, 506, 506, 506, 506, 506, 506, 
	85, 85, 85, 85, 85, 85, 85, 85, 
	85, 85, 85, 85, 85, 85, 85, 85, 
	85, 85, 85, 85, 85, 85, 85, 85, 
	85, 85, 506, 506, 506, 506, 85, 506, 
	85, 85, 85, 85, 85, 85, 85, 85, 
	85, 85, 85, 85, 85, 85, 85, 85, 
	85, 85, 85, 85, 85, 85, 85, 85, 
	85, 85, 506, 506, 506, 506, 506, 85, 
	507, 507, 507, 507, 507, 507, 507, 507, 
	507, 507, 507, 507, 507, 507, 507, 507, 
	507, 507, 507, 507, 507, 507, 507, 507, 
	507, 507, 507, 507, 507, 507, 507, 507, 
	507, 507, 507, 507, 507, 507, 507, 507, 
	507, 507, 507, 507, 507, 507, 507, 507, 
	87, 87, 87, 87, 87, 87, 87, 87, 
	87, 87, 507, 507, 507, 507, 507, 507, 
	507, 87, 87, 87, 87, 87, 87, 87, 
	87, 87, 87, 87, 87, 87, 87, 87, 
	87, 87, 87, 87, 87, 87, 87, 87, 
	87, 87, 87, 507, 507, 507, 507, 87, 
	507, 87, 87, 87, 87, 87, 87, 87, 
	87, 87, 87, 87, 87, 87, 87, 87, 
	87, 87, 87, 87, 87, 87, 87, 87, 
	87, 87, 87, 507, 507, 507, 507, 507, 
	87, 509, 510, 510, 510, 509, 510, 510, 
	510, 510, 511, 512, 511, 511, 511, 510, 
	510, 510, 510, 510, 510, 510, 510, 510, 
	510, 510, 510, 509, 510, 510, 510, 510, 
	510, 511, 513, 510, 514, 515, 516, 517, 
	510, 510, 510, 518, 519, 510, 519, 510, 
	520, 510, 510, 510, 510, 510, 510, 510, 
	510, 510, 510, 521, 510, 522, 523, 524, 
	510, 510, 525, 526, 525, 525, 527, 525, 
	525, 525, 525, 525, 525, 525, 525, 525, 
	525, 525, 525, 525, 525, 525, 525, 525, 
	525, 525, 525, 525, 528, 529, 510, 520, 
	530, 520, 531, 532, 533, 534, 535, 536, 
	508, 508, 537, 508, 508, 508, 538, 539, 
	540, 508, 508, 541, 542, 543, 544, 508, 
	545, 508, 546, 508, 510, 547, 510, 519, 
	510, 508, 548, 548, 548, 548, 548, 548, 
	548, 548, 548, 548, 548, 548, 548, 548, 
	548, 548, 548, 548, 548, 548, 548, 548, 
	548, 548, 548, 548, 548, 548, 548, 548, 
	548, 548, 548, 549, 548, 548, 548, 548, 
	548, 548, 548, 548, 548, 548, 548, 548, 
	548, 548, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 548, 548, 548, 549, 
	548, 549, 548, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 548, 548, 548, 
	548, 508, 548, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 548, 548, 548, 
	548, 548, 508, 511, 550, 511, 511, 511, 
	550, 550, 550, 550, 550, 550, 550, 550, 
	550, 550, 550, 550, 550, 550, 550, 550, 
	550, 550, 511, 550, 551, 520, 552, 552, 
	520, 552, 552, 552, 552, 552, 552, 552, 
	552, 552, 552, 552, 552, 552, 552, 552, 
	552, 552, 552, 552, 552, 552, 552, 552, 
	552, 552, 552, 552, 552, 552, 552, 552, 
	552, 552, 552, 552, 552, 552, 552, 552, 
	552, 552, 552, 552, 552, 552, 552, 552, 
	552, 552, 552, 552, 552, 552, 552, 552, 
	552, 552, 552, 552, 552, 552, 520, 552, 
	553, 554, 555, 556, 557, 557, 557, 557, 
	557, 557, 557, 557, 557, 557, 557, 557, 
	557, 557, 557, 557, 557, 557, 557, 557, 
	557, 557, 557, 557, 557, 557, 557, 557, 
	557, 557, 557, 557, 557, 89, 89, 557, 
	89, 557, 89, 89, 557, 557, 89, 89, 
	89, 559, 89, 89, 560, 560, 560, 560, 
	560, 560, 560, 560, 560, 560, 89, 89, 
	89, 89, 89, 89, 89, 558, 558, 558, 
	558, 558, 558, 558, 558, 558, 558, 558, 
	558, 558, 558, 558, 558, 558, 558, 558, 
	558, 558, 558, 558, 558, 558, 558, 557, 
	89, 557, 557, 558, 89, 558, 558, 558, 
	558, 558, 558, 558, 558, 558, 558, 558, 
	558, 558, 558, 558, 558, 558, 558, 558, 
	558, 558, 558, 558, 558, 558, 558, 557, 
	557, 557, 89, 557, 558, 561, 561, 561, 
	561, 561, 561, 561, 561, 561, 561, 561, 
	561, 561, 561, 561, 561, 561, 561, 561, 
	561, 561, 561, 561, 561, 561, 561, 561, 
	561, 561, 561, 561, 561, 561, 561, 561, 
	561, 561, 561, 561, 561, 561, 561, 561, 
	561, 561, 561, 561, 561, 558, 558, 558, 
	558, 558, 558, 558, 558, 558, 558, 561, 
	561, 561, 561, 561, 561, 561, 558, 558, 
	558, 558, 558, 558, 558, 558, 558, 558, 
	558, 558, 558, 558, 558, 558, 558, 558, 
	558, 558, 558, 558, 558, 558, 558, 558, 
	561, 561, 561, 561, 558, 561, 558, 558, 
	558, 558, 558, 558, 558, 558, 558, 558, 
	558, 558, 558, 558, 558, 558, 558, 558, 
	558, 558, 558, 558, 558, 558, 558, 558, 
	561, 561, 561, 561, 561, 558, 560, 560, 
	560, 560, 560, 560, 560, 560, 560, 560, 
	561, 562, 552, 520, 552, 520, 552, 520, 
	552, 564, 563, 520, 565, 552, 520, 552, 
	566, 520, 557, 557, 557, 557, 557, 557, 
	557, 557, 557, 557, 557, 557, 557, 557, 
	557, 557, 557, 557, 557, 557, 557, 557, 
	557, 557, 557, 557, 557, 557, 557, 557, 
	557, 557, 557, 557, 557, 557, 557, 557, 
	557, 557, 557, 557, 557, 557, 557, 557, 
	557, 557, 557, 557, 557, 557, 557, 557, 
	557, 557, 557, 557, 557, 557, 557, 557, 
	557, 520, 557, 520, 552, 520, 520, 552, 
	548, 548, 548, 548, 548, 548, 548, 548, 
	548, 548, 548, 548, 548, 548, 548, 548, 
	548, 548, 548, 548, 548, 548, 548, 548, 
	548, 548, 548, 548, 548, 548, 548, 548, 
	548, 549, 548, 548, 548, 548, 548, 548, 
	548, 548, 548, 548, 548, 548, 548, 548, 
	525, 525, 525, 525, 525, 525, 525, 525, 
	525, 525, 548, 548, 548, 549, 548, 549, 
	548, 525, 525, 525, 525, 525, 525, 525, 
	525, 525, 525, 525, 525, 525, 525, 525, 
	525, 525, 525, 525, 525, 525, 525, 525, 
	525, 525, 525, 548, 548, 548, 548, 525, 
	548, 525, 525, 525, 525, 525, 525, 525, 
	525, 525, 525, 525, 525, 525, 525, 525, 
	525, 525, 525, 525, 525, 525, 525, 525, 
	525, 525, 525, 548, 548, 548, 548, 548, 
	525, 567, 567, 567, 567, 567, 567, 567, 
	567, 567, 567, 567, 567, 567, 567, 567, 
	567, 567, 567, 567, 567, 567, 567, 567, 
	567, 567, 567, 567, 567, 567, 567, 567, 
	567, 567, 549, 567, 567, 567, 567, 567, 
	567, 567, 567, 567, 567, 567, 567, 567, 
	567, 525, 525, 525, 525, 525, 525, 525, 
	525, 525, 525, 567, 567, 567, 549, 567, 
	549, 567, 525, 525, 525, 525, 568, 525, 
	525, 525, 525, 525, 525, 525, 525, 525, 
	525, 525, 525, 525, 525, 525, 525, 525, 
	525, 525, 525, 525, 567, 567, 567, 567, 
	525, 567, 525, 525, 525, 525, 525, 525, 
	525, 525, 525, 525, 525, 525, 525, 525, 
	525, 525, 525, 525, 525, 525, 525, 525, 
	525, 525, 525, 525, 567, 567, 567, 567, 
	567, 525, 567, 567, 567, 567, 567, 567, 
	567, 567, 567, 567, 567, 567, 567, 567, 
	567, 567, 567, 567, 567, 567, 567, 567, 
	567, 567, 567, 567, 567, 567, 567, 567, 
	567, 567, 567, 549, 567, 567, 567, 567, 
	567, 567, 567, 567, 567, 567, 567, 567, 
	567, 567, 525, 525, 525, 525, 525, 525, 
	525, 525, 525, 525, 567, 567, 567, 549, 
	567, 549, 567, 525, 525, 525, 525, 525, 
	525, 569, 525, 525, 525, 525, 525, 525, 
	525, 525, 525, 525, 525, 525, 525, 525, 
	525, 525, 525, 525, 525, 567, 567, 567, 
	567, 525, 567, 525, 525, 525, 525, 525, 
	525, 525, 525, 525, 525, 525, 525, 525, 
	525, 525, 525, 525, 525, 525, 525, 525, 
	525, 525, 525, 525, 525, 567, 567, 567, 
	567, 567, 525, 567, 567, 567, 567, 567, 
	567, 567, 567, 567, 567, 567, 567, 567, 
	567, 567, 567, 567, 567, 567, 567, 567, 
	567, 567, 567, 567, 567, 567, 567, 567, 
	567, 567, 567, 567, 549, 567, 567, 567, 
	567, 567, 567, 567, 567, 567, 567, 567, 
	567, 567, 567, 525, 525, 525, 525, 525, 
	525, 525, 525, 525, 525, 567, 567, 567, 
	549, 567, 549, 567, 525, 525, 525, 525, 
	525, 525, 525, 525, 570, 525, 525, 525, 
	525, 525, 525, 525, 525, 525, 525, 525, 
	525, 525, 525, 525, 525, 525, 567, 567, 
	567, 567, 525, 567, 525, 525, 525, 525, 
	525, 525, 525, 525, 525, 525, 525, 525, 
	525, 525, 525, 525, 525, 525, 525, 525, 
	525, 525, 525, 525, 525, 525, 567, 567, 
	567, 567, 567, 525, 567, 567, 567, 567, 
	567, 567, 567, 567, 567, 567, 567, 567, 
	567, 567, 567, 567, 567, 567, 567, 567, 
	567, 567, 567, 567, 567, 567, 567, 567, 
	567, 567, 567, 567, 567, 549, 567, 567, 
	567, 567, 567, 567, 567, 567, 567, 567, 
	567, 567, 567, 567, 525, 525, 525, 525, 
	525, 525, 525, 525, 525, 525, 567, 567, 
	567, 549, 567, 549, 567, 525, 525, 525, 
	525, 525, 525, 525, 525, 525, 525, 525, 
	525, 525, 571, 525, 525, 525, 525, 525, 
	525, 525, 525, 525, 525, 525, 525, 567, 
	567, 567, 567, 525, 567, 525, 525, 525, 
	525, 525, 525, 525, 525, 525, 525, 525, 
	525, 525, 525, 525, 525, 525, 525, 525, 
	525, 525, 525, 525, 525, 525, 525, 567, 
	567, 567, 567, 567, 525, 567, 567, 567, 
	567, 567, 567, 567, 567, 567, 567, 567, 
	567, 567, 567, 567, 567, 567, 567, 567, 
	567, 567, 567, 567, 567, 567, 567, 567, 
	567, 567, 567, 567, 567, 567, 549, 567, 
	567, 567, 567, 567, 567, 567, 567, 567, 
	567, 567, 567, 567, 567, 525, 525, 525, 
	525, 525, 525, 525, 525, 525, 525, 567, 
	567, 567, 549, 567, 549, 567, 525, 525, 
	525, 525, 525, 525, 525, 525, 525, 525, 
	525, 525, 525, 572, 525, 525, 525, 525, 
	525, 525, 525, 525, 525, 525, 525, 525, 
	567, 567, 567, 567, 525, 567, 525, 525, 
	525, 525, 525, 525, 525, 525, 525, 525, 
	525, 525, 525, 525, 525, 525, 525, 525, 
	525, 525, 525, 525, 525, 525, 525, 525, 
	567, 567, 567, 567, 567, 525, 567, 567, 
	567, 567, 567, 567, 567, 567, 567, 567, 
	567, 567, 567, 567, 567, 567, 567, 567, 
	567, 567, 567, 567, 567, 567, 567, 567, 
	567, 567, 567, 567, 567, 567, 567, 549, 
	567, 567, 567, 567, 567, 567, 567, 567, 
	567, 567, 567, 567, 567, 567, 525, 525, 
	525, 525, 525, 525, 525, 525, 525, 525, 
	567, 567, 567, 549, 567, 549, 567, 525, 
	525, 525, 571, 525, 525, 525, 525, 525, 
	525, 525, 525, 525, 525, 525, 525, 525, 
	525, 525, 525, 525, 525, 525, 525, 525, 
	525, 567, 567, 567, 567, 525, 567, 525, 
	525, 525, 525, 525, 525, 525, 525, 525, 
	525, 525, 525, 525, 525, 525, 525, 525, 
	525, 525, 525, 525, 525, 525, 525, 525, 
	525, 567, 567, 567, 567, 567, 525, 566, 
	557, 512, 557, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 549, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 573, 573, 573, 
	549, 573, 549, 573, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 573, 573, 
	573, 573, 574, 573, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 573, 573, 
	573, 573, 573, 508, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 549, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 573, 573, 
	573, 549, 573, 549, 573, 508, 508, 508, 
	508, 575, 576, 508, 508, 508, 508, 508, 
	577, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 573, 
	573, 573, 573, 508, 573, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 573, 
	573, 573, 573, 573, 508, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 549, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 573, 
	573, 573, 549, 573, 549, 573, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 578, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	573, 573, 573, 573, 508, 573, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	573, 573, 573, 573, 573, 508, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 549, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	573, 573, 573, 549, 573, 549, 573, 508, 
	508, 579, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 573, 573, 573, 573, 508, 573, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 573, 573, 573, 573, 573, 508, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	549, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 573, 573, 573, 549, 573, 549, 573, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 580, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 573, 573, 573, 573, 508, 573, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 573, 573, 573, 573, 573, 508, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 549, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 573, 573, 573, 549, 573, 549, 
	573, 508, 508, 508, 581, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 573, 573, 573, 573, 508, 
	573, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 573, 573, 573, 573, 573, 
	508, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 549, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 573, 573, 573, 549, 573, 
	549, 573, 508, 508, 508, 508, 508, 508, 
	508, 508, 582, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 573, 573, 573, 573, 
	508, 573, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 573, 573, 573, 573, 
	573, 508, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 549, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 573, 573, 573, 549, 
	573, 549, 573, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	583, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 573, 573, 573, 
	573, 508, 573, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 573, 573, 573, 
	573, 573, 508, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 549, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 573, 573, 573, 
	549, 573, 549, 573, 508, 508, 508, 508, 
	508, 508, 584, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 573, 573, 
	573, 573, 508, 573, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 573, 573, 
	573, 573, 573, 508, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 549, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 573, 573, 
	573, 549, 573, 549, 573, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 573, 
	573, 573, 573, 585, 573, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 573, 
	573, 573, 573, 573, 508, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 549, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 573, 
	573, 573, 549, 573, 549, 573, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	573, 573, 573, 573, 586, 573, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	573, 573, 573, 573, 573, 508, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 549, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	573, 573, 573, 549, 573, 549, 573, 508, 
	508, 508, 508, 508, 508, 508, 508, 587, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 573, 573, 573, 573, 508, 573, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 573, 573, 573, 573, 573, 508, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	549, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 573, 573, 573, 549, 573, 549, 573, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 588, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 573, 573, 573, 573, 508, 573, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 573, 573, 573, 573, 573, 508, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 549, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 573, 573, 573, 549, 573, 549, 
	573, 508, 508, 508, 508, 584, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 573, 573, 573, 573, 508, 
	573, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 573, 573, 573, 573, 573, 
	508, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 549, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 573, 573, 573, 549, 573, 
	549, 573, 508, 508, 508, 508, 508, 508, 
	508, 508, 589, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 573, 573, 573, 573, 
	508, 573, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 573, 573, 573, 573, 
	573, 508, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 549, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 573, 573, 573, 549, 
	573, 549, 573, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	588, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 573, 573, 573, 
	573, 508, 573, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 573, 573, 573, 
	573, 573, 508, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 549, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 573, 573, 573, 
	549, 573, 549, 573, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 573, 573, 
	573, 573, 508, 573, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 590, 
	508, 591, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 573, 573, 
	573, 573, 573, 508, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 549, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 573, 573, 
	573, 549, 573, 549, 573, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 573, 
	573, 573, 573, 508, 573, 508, 508, 508, 
	508, 508, 508, 508, 508, 592, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 573, 
	573, 573, 573, 573, 508, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 549, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 573, 
	573, 573, 549, 573, 549, 573, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	573, 573, 573, 573, 508, 573, 593, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	573, 573, 573, 573, 573, 508, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 549, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	573, 573, 573, 549, 573, 549, 573, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 573, 573, 573, 573, 508, 573, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 586, 508, 508, 508, 508, 508, 508, 
	508, 573, 573, 573, 573, 573, 508, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	549, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 573, 573, 573, 549, 573, 549, 573, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 573, 573, 573, 573, 508, 573, 
	508, 508, 508, 586, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 573, 573, 573, 573, 573, 508, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 549, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 573, 573, 573, 549, 573, 549, 
	573, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 573, 573, 573, 573, 508, 
	573, 508, 508, 508, 508, 594, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 595, 508, 508, 508, 508, 508, 
	508, 508, 508, 573, 573, 573, 573, 573, 
	508, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 549, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 573, 573, 573, 549, 573, 
	549, 573, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 573, 573, 573, 573, 
	508, 573, 508, 508, 508, 508, 508, 508, 
	596, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 573, 573, 573, 573, 
	573, 508, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 549, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 573, 573, 573, 549, 
	573, 549, 573, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 573, 573, 573, 
	573, 508, 573, 508, 508, 508, 508, 508, 
	508, 508, 508, 597, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 573, 573, 573, 
	573, 573, 508, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 549, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 573, 573, 573, 
	549, 573, 549, 573, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 573, 573, 
	573, 573, 508, 573, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 586, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 573, 573, 
	573, 573, 573, 508, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 549, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 573, 573, 
	573, 549, 573, 549, 573, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 573, 
	573, 573, 573, 508, 573, 508, 508, 508, 
	508, 598, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 573, 
	573, 573, 573, 573, 508, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 549, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 573, 
	573, 573, 549, 573, 549, 573, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	573, 573, 573, 573, 508, 573, 599, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	573, 573, 573, 573, 573, 508, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 549, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	573, 573, 573, 549, 573, 549, 573, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 573, 573, 573, 573, 508, 573, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 586, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 573, 573, 573, 573, 573, 508, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	549, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 573, 573, 573, 549, 573, 549, 573, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 573, 573, 573, 573, 508, 573, 
	600, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 601, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 573, 573, 573, 573, 573, 508, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 549, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 573, 573, 573, 549, 573, 549, 
	573, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 573, 573, 573, 573, 508, 
	573, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 602, 508, 508, 508, 508, 
	508, 508, 508, 573, 573, 573, 573, 573, 
	508, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 549, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 573, 573, 573, 549, 573, 
	549, 573, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 573, 573, 573, 573, 
	508, 573, 508, 508, 508, 508, 586, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 573, 573, 573, 573, 
	573, 508, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 549, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 573, 573, 573, 549, 
	573, 549, 573, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 573, 573, 573, 
	573, 508, 573, 603, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 573, 573, 573, 
	573, 573, 508, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 549, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 573, 573, 573, 
	549, 573, 549, 573, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 573, 573, 
	573, 573, 508, 573, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 593, 508, 
	508, 508, 508, 508, 508, 508, 573, 573, 
	573, 573, 573, 508, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 549, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 573, 573, 
	573, 549, 573, 549, 573, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 573, 
	573, 573, 573, 508, 573, 508, 508, 508, 
	508, 604, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 586, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 573, 
	573, 573, 573, 573, 508, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 549, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 573, 
	573, 573, 549, 573, 549, 573, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	573, 573, 573, 573, 508, 573, 508, 508, 
	508, 508, 508, 605, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	573, 573, 573, 573, 573, 508, 606, 606, 
	606, 606, 606, 606, 606, 606, 606, 606, 
	606, 606, 606, 606, 606, 606, 606, 606, 
	606, 606, 606, 606, 606, 606, 606, 606, 
	606, 606, 606, 606, 606, 606, 606, 549, 
	606, 606, 606, 606, 606, 606, 606, 606, 
	606, 606, 606, 606, 606, 606, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	606, 606, 606, 549, 606, 549, 606, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 606, 606, 606, 606, 508, 606, 508, 
	508, 508, 508, 508, 508, 508, 508, 607, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 606, 606, 606, 606, 606, 508, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	549, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 573, 573, 573, 549, 573, 549, 573, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 573, 573, 573, 573, 508, 573, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 608, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 573, 573, 573, 573, 573, 508, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 549, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 573, 573, 573, 549, 573, 549, 
	573, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 573, 573, 573, 573, 508, 
	573, 508, 508, 508, 508, 609, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 573, 573, 573, 573, 573, 
	508, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 549, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 573, 573, 573, 549, 573, 
	549, 573, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 573, 573, 573, 573, 
	508, 573, 508, 508, 508, 610, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 573, 573, 573, 573, 
	573, 508, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 549, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 573, 573, 573, 549, 
	573, 611, 573, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 573, 573, 573, 
	573, 508, 573, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 573, 573, 573, 
	573, 573, 508, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 549, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 573, 573, 573, 
	549, 573, 549, 573, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 573, 573, 
	573, 573, 508, 573, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 612, 
	508, 613, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 573, 573, 
	573, 573, 573, 508, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 549, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 573, 573, 
	573, 549, 573, 549, 573, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 573, 
	573, 573, 573, 508, 573, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 614, 
	508, 508, 508, 508, 508, 508, 508, 573, 
	573, 573, 573, 573, 508, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 549, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 573, 
	573, 573, 549, 573, 549, 573, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	573, 573, 573, 573, 508, 573, 508, 508, 
	508, 508, 586, 508, 508, 508, 615, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	573, 573, 573, 573, 573, 508, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 549, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	573, 573, 573, 549, 573, 549, 573, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 573, 573, 573, 573, 508, 573, 508, 
	508, 508, 508, 508, 586, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 573, 573, 573, 573, 573, 508, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	549, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 573, 573, 573, 549, 573, 549, 573, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 573, 573, 573, 573, 508, 573, 
	508, 508, 508, 586, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 616, 508, 508, 508, 508, 508, 
	508, 508, 573, 573, 573, 573, 573, 508, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 549, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 573, 573, 573, 549, 573, 549, 
	573, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 573, 573, 573, 573, 508, 
	573, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 617, 508, 508, 
	508, 508, 508, 573, 573, 573, 573, 573, 
	508, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 549, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 573, 573, 573, 549, 573, 
	549, 573, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 573, 573, 573, 573, 
	508, 573, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 602, 508, 508, 508, 508, 
	508, 508, 508, 508, 573, 573, 573, 573, 
	573, 508, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 549, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 573, 573, 573, 549, 
	573, 549, 573, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 573, 573, 573, 
	573, 508, 573, 618, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 540, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 573, 573, 573, 
	573, 573, 508, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 549, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 573, 573, 573, 
	549, 573, 549, 573, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 573, 573, 
	573, 573, 508, 573, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 600, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 573, 573, 
	573, 573, 573, 508, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 549, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 573, 573, 
	573, 549, 573, 549, 573, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 573, 
	573, 573, 573, 508, 573, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 586, 508, 
	508, 508, 508, 508, 508, 508, 508, 573, 
	573, 573, 573, 573, 508, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 549, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 573, 
	573, 573, 549, 573, 549, 573, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	573, 573, 573, 573, 508, 573, 508, 508, 
	508, 508, 508, 586, 508, 508, 508, 508, 
	508, 508, 508, 586, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	573, 573, 573, 573, 573, 508, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 549, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	573, 573, 573, 549, 573, 549, 573, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 573, 573, 573, 573, 508, 573, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 619, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 573, 573, 573, 573, 573, 508, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	549, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 573, 573, 573, 549, 573, 549, 573, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 573, 573, 573, 573, 508, 573, 
	508, 508, 508, 620, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 573, 573, 573, 573, 573, 508, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 549, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 573, 573, 573, 549, 573, 549, 
	573, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 573, 573, 573, 573, 508, 
	573, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 621, 508, 508, 
	508, 508, 508, 573, 573, 573, 573, 573, 
	508, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 549, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 573, 573, 573, 549, 573, 
	549, 573, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 573, 573, 573, 573, 
	508, 573, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 602, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 573, 573, 573, 573, 
	573, 508, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 549, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 573, 573, 573, 549, 
	573, 549, 573, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 573, 573, 573, 
	573, 508, 573, 508, 508, 508, 508, 622, 
	508, 508, 508, 623, 508, 508, 508, 508, 
	508, 624, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 573, 573, 573, 
	573, 573, 508, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 549, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 573, 573, 573, 
	549, 573, 549, 573, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 573, 573, 
	573, 573, 508, 573, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 624, 508, 508, 573, 573, 
	573, 573, 573, 508, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 549, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 573, 573, 
	573, 549, 573, 549, 573, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 573, 
	573, 573, 573, 508, 573, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	586, 508, 508, 508, 508, 508, 508, 573, 
	573, 573, 573, 573, 508, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 549, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 573, 
	573, 573, 549, 573, 549, 573, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	573, 573, 573, 573, 508, 573, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 586, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	573, 573, 573, 573, 573, 508, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 549, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	573, 573, 573, 549, 573, 549, 573, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 573, 573, 573, 573, 508, 573, 508, 
	508, 508, 508, 625, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 573, 573, 573, 573, 573, 508, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	549, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 573, 573, 573, 549, 573, 549, 573, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 573, 573, 573, 573, 508, 573, 
	508, 508, 508, 626, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 627, 628, 508, 508, 508, 508, 
	508, 508, 573, 573, 573, 573, 573, 508, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 549, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 573, 573, 573, 549, 573, 549, 
	573, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 573, 573, 573, 573, 508, 
	573, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 586, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 573, 573, 573, 573, 573, 
	508, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 549, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 573, 573, 573, 549, 573, 
	549, 573, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 573, 573, 573, 573, 
	508, 573, 508, 508, 629, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 573, 573, 573, 573, 
	573, 508, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 549, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 573, 573, 573, 549, 
	573, 549, 573, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 573, 573, 573, 
	573, 508, 573, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 602, 
	508, 508, 508, 508, 508, 573, 573, 573, 
	573, 573, 508, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 549, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 573, 573, 573, 
	549, 573, 549, 573, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 573, 573, 
	573, 573, 508, 573, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 630, 508, 508, 
	631, 508, 508, 508, 508, 508, 573, 573, 
	573, 573, 573, 508, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 549, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 573, 573, 
	573, 549, 573, 549, 573, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 573, 
	573, 573, 573, 508, 573, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 586, 508, 573, 
	573, 573, 573, 573, 508, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 549, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 573, 
	573, 573, 549, 573, 549, 573, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	573, 573, 573, 573, 508, 573, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 597, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	573, 573, 573, 573, 573, 508, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 549, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	573, 573, 573, 549, 573, 549, 573, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 573, 573, 573, 573, 508, 573, 508, 
	508, 508, 508, 632, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 633, 508, 508, 508, 508, 
	508, 573, 573, 573, 573, 573, 508, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	549, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 573, 573, 573, 549, 573, 549, 573, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 573, 573, 573, 573, 508, 573, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 615, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 573, 573, 573, 573, 573, 508, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 549, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 573, 573, 573, 549, 573, 549, 
	573, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 573, 573, 573, 573, 508, 
	573, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	634, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 573, 573, 573, 573, 573, 
	508, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 549, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 573, 573, 573, 549, 573, 
	549, 573, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 573, 573, 573, 573, 
	508, 573, 508, 508, 508, 508, 540, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 573, 573, 573, 573, 
	573, 508, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 549, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 573, 573, 573, 549, 
	573, 549, 573, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 573, 573, 573, 
	573, 508, 573, 508, 508, 508, 508, 508, 
	508, 508, 635, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 629, 508, 508, 508, 
	508, 508, 508, 508, 508, 573, 573, 573, 
	573, 573, 508, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 549, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 573, 573, 573, 
	549, 573, 549, 573, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 573, 573, 
	573, 573, 508, 573, 508, 508, 508, 508, 
	597, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 573, 573, 
	573, 573, 573, 508, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 549, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 573, 573, 
	573, 549, 573, 549, 573, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 573, 
	573, 573, 573, 508, 573, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 636, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 573, 
	573, 573, 573, 573, 508, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 549, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 573, 
	573, 573, 549, 573, 549, 573, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	573, 573, 573, 573, 508, 573, 508, 508, 
	508, 637, 508, 508, 508, 508, 508, 508, 
	508, 638, 508, 508, 508, 508, 508, 508, 
	508, 639, 508, 508, 508, 508, 508, 508, 
	573, 573, 573, 573, 573, 508, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 549, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	573, 573, 573, 549, 573, 549, 573, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 573, 573, 573, 573, 508, 573, 508, 
	508, 508, 508, 615, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 573, 573, 573, 573, 573, 508, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	549, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 573, 573, 573, 549, 573, 549, 573, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 573, 573, 573, 573, 508, 573, 
	508, 508, 508, 508, 603, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 573, 573, 573, 573, 573, 508, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 549, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 573, 573, 573, 549, 573, 549, 
	573, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 573, 573, 573, 573, 508, 
	573, 508, 508, 508, 508, 508, 508, 508, 
	508, 623, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 573, 573, 573, 573, 573, 
	508, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 549, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 573, 573, 573, 549, 573, 
	549, 573, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 573, 573, 573, 573, 
	508, 573, 508, 508, 508, 508, 508, 508, 
	508, 640, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 573, 573, 573, 573, 
	573, 508, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 549, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 573, 573, 573, 549, 
	573, 549, 573, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 573, 573, 573, 
	573, 508, 573, 508, 508, 508, 508, 597, 
	508, 508, 508, 621, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 573, 573, 573, 
	573, 573, 508, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 549, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 573, 573, 573, 
	549, 573, 549, 573, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 573, 573, 
	573, 573, 508, 573, 508, 508, 508, 508, 
	508, 508, 508, 508, 641, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 573, 573, 
	573, 573, 573, 508, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 549, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 573, 573, 
	573, 549, 573, 549, 573, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 573, 
	573, 573, 573, 508, 573, 508, 508, 508, 
	508, 642, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 573, 
	573, 573, 573, 573, 508, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 573, 549, 573, 
	573, 573, 573, 573, 573, 573, 573, 573, 
	573, 573, 573, 573, 573, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 573, 
	573, 573, 549, 573, 549, 573, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	573, 573, 573, 573, 508, 573, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	508, 591, 508, 508, 508, 508, 508, 508, 
	508, 508, 508, 508, 508, 508, 508, 508, 
	573, 573, 573, 573, 573, 508, 520, 552, 
	644, 645, 645, 645, 644, 645, 645, 645, 
	645, 646, 645, 646, 646, 646, 645, 645, 
	645, 645, 645, 645, 645, 645, 645, 645, 
	645, 645, 644, 645, 645, 645, 645, 645, 
	646, 645, 645, 647, 645, 645, 645, 645, 
	645, 645, 645, 645, 645, 645, 645, 645, 
	645, 645, 645, 645, 645, 645, 645, 645, 
	645, 645, 645, 645, 645, 645, 645, 645, 
	645, 643, 643, 643, 643, 643, 643, 643, 
	643, 643, 643, 643, 643, 643, 643, 643, 
	643, 643, 643, 643, 643, 643, 643, 643, 
	643, 643, 643, 645, 648, 645, 645, 643, 
	645, 643, 643, 643, 643, 643, 643, 643, 
	643, 643, 643, 643, 643, 643, 643, 643, 
	643, 643, 643, 643, 643, 643, 643, 643, 
	643, 643, 643, 645, 645, 645, 645, 645, 
	643, 649, 649, 649, 649, 649, 649, 649, 
	649, 649, 649, 649, 649, 649, 649, 649, 
	649, 649, 649, 649, 649, 649, 649, 649, 
	649, 649, 649, 649, 649, 649, 649, 649, 
	649, 649, 94, 649, 649, 649, 649, 649, 
	649, 649, 649, 649, 649, 649, 649, 649, 
	649, 93, 93, 93, 93, 93, 93, 93, 
	93, 93, 93, 95, 649, 649, 649, 649, 
	94, 649, 93, 93, 93, 93, 93, 93, 
	93, 93, 93, 93, 93, 93, 93, 93, 
	93, 93, 93, 93, 93, 93, 93, 93, 
	93, 93, 93, 93, 649, 649, 649, 649, 
	93, 649, 93, 93, 93, 93, 93, 93, 
	93, 93, 93, 93, 93, 93, 93, 93, 
	93, 93, 93, 93, 93, 93, 93, 93, 
	93, 93, 93, 93, 649, 649, 649, 649, 
	649, 93, 646, 650, 646, 646, 646, 650, 
	650, 650, 650, 650, 650, 650, 650, 650, 
	650, 650, 650, 650, 650, 650, 650, 650, 
	650, 646, 650, 651, 652, 653, 654, 655, 
	649, 656, 658, 659, 659, 659, 658, 659, 
	659, 659, 659, 660, 661, 660, 660, 660, 
	659, 659, 659, 659, 659, 659, 659, 659, 
	659, 659, 659, 659, 658, 659, 659, 659, 
	659, 659, 660, 662, 659, 663, 659, 664, 
	665, 659, 659, 659, 666, 667, 659, 667, 
	659, 664, 659, 659, 659, 659, 659, 659, 
	659, 659, 659, 659, 659, 659, 668, 669, 
	670, 659, 659, 671, 671, 671, 671, 671, 
	671, 671, 671, 671, 671, 671, 671, 671, 
	671, 671, 671, 671, 671, 671, 671, 671, 
	671, 671, 671, 671, 671, 672, 673, 659, 
	664, 657, 664, 657, 657, 657, 657, 657, 
	657, 657, 657, 657, 657, 657, 657, 657, 
	657, 657, 657, 657, 657, 657, 657, 657, 
	657, 657, 657, 657, 657, 659, 674, 659, 
	667, 659, 657, 675, 675, 675, 675, 675, 
	675, 675, 675, 675, 675, 675, 675, 675, 
	675, 675, 675, 675, 675, 675, 675, 675, 
	675, 675, 675, 675, 675, 675, 675, 675, 
	675, 675, 675, 675, 676, 675, 675, 675, 
	675, 675, 675, 675, 675, 675, 675, 675, 
	675, 675, 675, 657, 657, 657, 657, 657, 
	657, 657, 657, 657, 657, 675, 675, 675, 
	675, 675, 676, 675, 657, 657, 657, 657, 
	657, 657, 657, 657, 657, 657, 657, 657, 
	657, 657, 657, 657, 657, 657, 657, 657, 
	657, 657, 657, 657, 657, 657, 675, 675, 
	675, 675, 657, 675, 657, 657, 657, 657, 
	657, 657, 657, 657, 657, 657, 657, 657, 
	657, 657, 657, 657, 657, 657, 657, 657, 
	657, 657, 657, 657, 657, 657, 675, 675, 
	675, 675, 675, 657, 678, 677, 679, 660, 
	680, 660, 660, 660, 680, 680, 680, 680, 
	680, 680, 680, 680, 680, 680, 680, 680, 
	680, 680, 680, 680, 680, 680, 660, 680, 
	681, 664, 682, 682, 664, 682, 682, 682, 
	682, 682, 682, 682, 682, 682, 682, 682, 
	682, 682, 682, 682, 682, 682, 682, 682, 
	682, 682, 682, 682, 682, 682, 682, 682, 
	682, 682, 682, 682, 682, 682, 682, 682, 
	682, 682, 682, 682, 682, 682, 682, 682, 
	682, 682, 682, 682, 682, 682, 682, 682, 
	682, 682, 682, 682, 682, 682, 682, 682, 
	682, 682, 664, 682, 683, 684, 685, 686, 
	664, 682, 664, 682, 664, 682, 664, 687, 
	682, 664, 682, 689, 664, 688, 688, 688, 
	688, 688, 688, 688, 688, 688, 688, 688, 
	688, 688, 688, 688, 688, 688, 688, 688, 
	688, 688, 688, 688, 688, 688, 688, 688, 
	688, 688, 688, 688, 688, 688, 688, 688, 
	688, 688, 688, 688, 688, 688, 688, 688, 
	688, 688, 688, 688, 688, 688, 688, 688, 
	688, 688, 688, 688, 688, 688, 688, 688, 
	688, 688, 688, 688, 664, 688, 664, 682, 
	664, 664, 682, 690, 690, 690, 690, 690, 
	690, 690, 690, 690, 690, 690, 690, 690, 
	690, 690, 690, 690, 690, 690, 690, 690, 
	690, 690, 690, 690, 690, 690, 690, 690, 
	690, 690, 690, 690, 676, 690, 690, 690, 
	690, 690, 690, 690, 690, 690, 690, 690, 
	690, 690, 690, 671, 671, 671, 671, 671, 
	671, 671, 671, 671, 671, 690, 690, 690, 
	690, 690, 676, 690, 671, 671, 671, 671, 
	671, 671, 671, 671, 671, 671, 671, 671, 
	671, 671, 671, 671, 671, 671, 671, 671, 
	671, 671, 671, 671, 671, 671, 690, 690, 
	690, 690, 671, 690, 671, 671, 671, 671, 
	671, 671, 671, 671, 671, 671, 671, 671, 
	671, 671, 671, 671, 671, 671, 671, 671, 
	671, 671, 671, 671, 671, 671, 690, 690, 
	690, 690, 690, 671, 689, 688, 661, 688, 
	664, 682, 692, 691, 691, 691, 692, 691, 
	691, 691, 691, 693, 694, 693, 693, 693, 
	691, 691, 691, 691, 691, 691, 691, 691, 
	691, 691, 691, 691, 692, 691, 691, 691, 
	691, 691, 693, 691, 691, 695, 691, 102, 
	696, 691, 697, 691, 698, 102, 133, 699, 
	135, 102, 691, 691, 691, 691, 691, 691, 
	691, 691, 691, 691, 700, 691, 701, 133, 
	702, 703, 691, 691, 691, 691, 691, 691, 
	691, 691, 691, 691, 691, 691, 691, 691, 
	691, 691, 691, 691, 691, 691, 691, 691, 
	691, 691, 691, 691, 691, 133, 704, 133, 
	102, 691, 691, 691, 691, 691, 691, 691, 
	691, 691, 691, 705, 691, 691, 691, 691, 
	691, 691, 691, 691, 706, 691, 691, 707, 
	691, 708, 691, 691, 691, 146, 147, 691, 
	102, 691, 709, 709, 709, 709, 709, 709, 
	709, 709, 709, 693, 709, 693, 693, 693, 
	709, 709, 709, 709, 709, 709, 709, 709, 
	709, 709, 709, 709, 709, 709, 709, 709, 
	709, 709, 693, 709, 709, 709, 709, 128, 
	129, 709, 130, 709, 131, 132, 133, 134, 
	135, 128, 709, 709, 709, 709, 709, 709, 
	709, 709, 709, 709, 136, 709, 137, 133, 
	138, 139, 709, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 140, 141, 133, 
	102, 97, 709, 97, 97, 97, 97, 97, 
	97, 97, 97, 142, 97, 97, 97, 97, 
	97, 97, 97, 97, 143, 97, 97, 144, 
	97, 145, 97, 97, 97, 146, 147, 709, 
	102, 709, 97, 710, 711, 711, 711, 710, 
	711, 711, 711, 711, 133, 712, 133, 133, 
	133, 711, 711, 711, 711, 711, 711, 711, 
	711, 711, 711, 711, 711, 710, 711, 711, 
	711, 711, 711, 133, 711, 711, 711, 711, 
	711, 711, 711, 711, 711, 711, 711, 711, 
	711, 711, 711, 711, 711, 711, 711, 711, 
	711, 711, 711, 711, 711, 711, 711, 711, 
	133, 711, 133, 712, 133, 133, 133, 96, 
	96, 96, 96, 96, 96, 96, 96, 96, 
	96, 96, 96, 96, 96, 96, 96, 96, 
	96, 133, 96, 96, 96, 96, 96, 102, 
	96, 96, 96, 96, 96, 96, 96, 96, 
	96, 96, 96, 96, 96, 96, 96, 96, 
	96, 96, 96, 96, 96, 96, 133, 96, 
	133, 712, 133, 133, 133, 96, 96, 96, 
	96, 96, 96, 96, 96, 96, 96, 96, 
	96, 96, 96, 96, 96, 96, 96, 133, 
	96, 96, 96, 96, 96, 96, 96, 96, 
	96, 96, 96, 96, 96, 96, 96, 96, 
	96, 96, 96, 96, 96, 96, 96, 96, 
	96, 96, 96, 96, 133, 96, 713, 133, 
	712, 133, 133, 133, 714, 714, 714, 714, 
	714, 714, 714, 714, 714, 714, 714, 714, 
	714, 714, 714, 714, 714, 714, 133, 714, 
	714, 714, 714, 714, 714, 714, 714, 714, 
	715, 714, 714, 714, 714, 714, 714, 714, 
	714, 714, 714, 714, 714, 714, 714, 714, 
	714, 714, 714, 133, 714, 133, 712, 133, 
	133, 133, 96, 96, 96, 96, 96, 96, 
	96, 96, 96, 96, 96, 96, 96, 96, 
	96, 96, 96, 96, 133, 96, 96, 96, 
	96, 96, 96, 96, 96, 96, 96, 96, 
	96, 96, 96, 96, 96, 96, 96, 96, 
	96, 96, 96, 96, 96, 96, 96, 96, 
	96, 133, 133, 96, 716, 710, 133, 710, 
	718, 717, 720, 721, 720, 720, 720, 719, 
	719, 719, 719, 719, 719, 719, 719, 719, 
	719, 719, 719, 719, 719, 719, 719, 719, 
	719, 720, 719, 722, 710, 710, 710, 710, 
	710, 710, 710, 710, 710, 710, 710, 710, 
	710, 710, 710, 710, 710, 710, 710, 710, 
	710, 710, 710, 710, 710, 710, 710, 710, 
	710, 710, 710, 710, 710, 98, 710, 710, 
	710, 710, 710, 710, 710, 710, 710, 710, 
	710, 710, 710, 710, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 99, 710, 
	710, 710, 710, 98, 710, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 710, 
	710, 710, 710, 97, 710, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 97, 
	97, 97, 97, 97, 97, 97, 97, 710, 
	710, 710, 710, 710, 97, 723, 133, 712, 
	133, 133, 133, 710, 710, 710, 710, 710, 
	710, 710, 710, 710, 710, 710, 710, 710, 
	710, 710, 710, 710, 710, 133, 710, 710, 
	710, 710, 710, 710, 710, 710, 710, 710, 
	710, 710, 710, 710, 710, 710, 710, 710, 
	710, 710, 710, 710, 710, 710, 710, 710, 
	710, 710, 133, 710, 710, 710, 710, 710, 
	710, 710, 710, 710, 710, 710, 710, 710, 
	710, 710, 710, 710, 710, 710, 710, 710, 
	710, 710, 710, 710, 710, 710, 710, 710, 
	710, 710, 710, 710, 710, 710, 710, 710, 
	710, 710, 710, 710, 710, 710, 710, 710, 
	710, 710, 710, 710, 710, 710, 710, 710, 
	710, 710, 710, 710, 710, 710, 710, 710, 
	710, 102, 710, 724, 725, 726, 727, 728, 
	729, 133, 712, 133, 133, 133, 710, 710, 
	710, 710, 710, 710, 710, 710, 710, 710, 
	710, 710, 710, 710, 710, 710, 710, 710, 
	133, 710, 710, 710, 710, 710, 710, 710, 
	710, 710, 102, 710, 710, 710, 710, 710, 
	710, 710, 710, 710, 710, 710, 710, 710, 
	710, 710, 710, 710, 710, 133, 710, 133, 
	717, 102, 730, 102, 730, 731, 732, 731, 
	731, 731, 719, 719, 719, 719, 719, 719, 
	719, 719, 719, 719, 719, 719, 719, 719, 
	719, 719, 719, 719, 731, 719, 733, 730, 
	734, 734, 734, 734, 734, 734, 734, 734, 
	734, 105, 734, 105, 105, 105, 734, 734, 
	734, 734, 734, 734, 734, 734, 734, 734, 
	734, 734, 734, 734, 734, 734, 734, 734, 
	105, 734, 734, 734, 734, 106, 107, 734, 
	108, 734, 109, 110, 111, 112, 113, 106, 
	734, 734, 734, 734, 734, 734, 734, 734, 
	734, 734, 114, 734, 115, 111, 116, 117, 
	734, 104, 104, 104, 104, 104, 104, 104, 
	104, 104, 104, 104, 104, 104, 104, 104, 
	104, 104, 104, 104, 104, 104, 104, 104, 
	104, 104, 104, 118, 119, 111, 120, 104, 
	734, 104, 104, 104, 104, 104, 104, 104, 
	104, 121, 104, 104, 104, 104, 104, 104, 
	104, 104, 122, 104, 104, 123, 104, 124, 
	104, 104, 104, 125, 126, 734, 120, 734, 
	104, 133, 730, 735, 730, 736, 730, 737, 
	730, 738, 172, 172, 172, 738, 172, 172, 
	172, 172, 739, 172, 739, 739, 739, 172, 
	172, 172, 172, 172, 172, 172, 172, 172, 
	172, 172, 172, 738, 172, 172, 172, 172, 
	172, 739, 172, 172, 172, 172, 172, 172, 
	172, 172, 172, 172, 172, 172, 172, 172, 
	172, 172, 172, 172, 172, 172, 172, 172, 
	172, 172, 172, 172, 172, 172, 172, 172, 
	172, 172, 174, 174, 174, 174, 174, 174, 
	174, 174, 174, 174, 174, 174, 174, 174, 
	174, 174, 174, 174, 174, 174, 174, 174, 
	174, 174, 174, 174, 172, 740, 172, 172, 
	174, 172, 174, 174, 174, 178, 174, 174, 
	174, 174, 174, 174, 174, 174, 174, 174, 
	174, 174, 174, 174, 174, 174, 174, 174, 
	174, 174, 174, 174, 172, 172, 172, 172, 
	172, 174, 171, 171, 171, 171, 171, 171, 
	171, 171, 171, 171, 171, 171, 171, 171, 
	171, 171, 171, 171, 171, 171, 171, 171, 
	171, 171, 171, 171, 171, 171, 171, 171, 
	171, 171, 171, 741, 171, 171, 171, 171, 
	171, 171, 171, 171, 171, 171, 171, 171, 
	171, 171, 174, 174, 174, 174, 174, 174, 
	174, 174, 174, 174, 172, 171, 171, 171, 
	171, 741, 171, 174, 174, 174, 174, 174, 
	174, 174, 174, 174, 174, 174, 174, 174, 
	174, 174, 174, 174, 174, 174, 174, 174, 
	174, 174, 174, 174, 174, 171, 171, 171, 
	171, 174, 171, 174, 174, 174, 174, 174, 
	174, 174, 174, 174, 174, 174, 174, 174, 
	174, 174, 174, 174, 174, 174, 174, 174, 
	174, 174, 174, 174, 174, 171, 171, 171, 
	171, 171, 174, 742, 742, 742, 742, 742, 
	742, 742, 742, 742, 175, 742, 175, 175, 
	175, 742, 742, 742, 742, 742, 742, 742, 
	742, 742, 742, 742, 742, 742, 742, 742, 
	742, 742, 742, 175, 742, 742, 742, 742, 
	742, 742, 742, 176, 742, 742, 742, 742, 
	742, 742, 742, 742, 742, 742, 742, 742, 
	742, 742, 742, 742, 742, 742, 742, 742, 
	742, 742, 742, 742, 174, 174, 174, 174, 
	174, 174, 174, 174, 174, 174, 174, 174, 
	174, 174, 174, 174, 174, 174, 174, 174, 
	174, 174, 174, 174, 174, 174, 742, 177, 
	742, 742, 174, 742, 174, 174, 174, 178, 
	174, 174, 174, 174, 174, 174, 174, 174, 
	174, 174, 174, 174, 174, 174, 174, 174, 
	174, 174, 174, 174, 174, 174, 742, 742, 
	742, 742, 742, 174, 743, 742, 742, 742, 
	742, 742, 742, 742, 742, 742, 742, 742, 
	742, 742, 742, 742, 742, 742, 742, 742, 
	742, 742, 742, 742, 742, 742, 742, 742, 
	742, 742, 742, 742, 742, 742, 741, 742, 
	742, 742, 742, 742, 742, 742, 742, 742, 
	742, 742, 742, 742, 742, 174, 174, 174, 
	174, 174, 174, 174, 174, 174, 174, 172, 
	742, 742, 742, 742, 741, 742, 174, 174, 
	174, 174, 174, 174, 174, 174, 174, 174, 
	174, 174, 174, 174, 174, 174, 174, 174, 
	174, 174, 174, 174, 174, 174, 174, 174, 
	742, 742, 742, 742, 174, 742, 174, 174, 
	174, 174, 174, 174, 174, 174, 174, 174, 
	174, 174, 174, 174, 744, 174, 174, 174, 
	174, 174, 174, 174, 174, 174, 174, 174, 
	742, 742, 742, 742, 742, 174, 179, 742, 
	746, 745, 745, 745, 746, 745, 745, 745, 
	745, 747, 745, 747, 747, 747, 745, 745, 
	745, 745, 745, 745, 745, 745, 745, 745, 
	745, 745, 746, 745, 745, 745, 745, 745, 
	747, 745, 745, 748, 745, 745, 745, 745, 
	745, 745, 745, 745, 745, 745, 745, 745, 
	745, 745, 745, 745, 745, 745, 745, 745, 
	745, 745, 745, 745, 745, 745, 745, 745, 
	745, 745, 745, 745, 745, 745, 745, 745, 
	745, 745, 745, 745, 745, 745, 745, 745, 
	745, 745, 745, 745, 745, 745, 745, 745, 
	745, 745, 745, 745, 749, 745, 745, 745, 
	745, 745, 745, 745, 750, 745, 745, 745, 
	745, 745, 745, 745, 745, 745, 745, 745, 
	745, 745, 745, 745, 745, 745, 745, 745, 
	745, 745, 745, 751, 745, 747, 752, 747, 
	747, 747, 752, 752, 752, 752, 752, 752, 
	752, 752, 752, 752, 752, 752, 752, 752, 
	752, 752, 752, 752, 747, 752, 753, 754, 
	755, 756, 758, 757, 759, 760, 757, 761, 
	763, 764, 764, 764, 763, 764, 764, 764, 
	764, 765, 766, 765, 765, 765, 764, 764, 
	764, 764, 764, 764, 764, 764, 764, 764, 
	764, 764, 763, 764, 764, 764, 764, 764, 
	765, 764, 764, 767, 764, 764, 764, 764, 
	764, 764, 764, 764, 764, 764, 764, 764, 
	764, 764, 764, 764, 764, 764, 764, 764, 
	764, 764, 764, 764, 764, 764, 764, 764, 
	764, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 764, 768, 764, 764, 762, 
	764, 762, 762, 762, 762, 762, 762, 762, 
	762, 769, 762, 762, 762, 762, 762, 762, 
	762, 762, 770, 762, 762, 771, 762, 772, 
	762, 762, 762, 764, 764, 764, 764, 764, 
	762, 773, 773, 773, 773, 773, 773, 773, 
	773, 773, 773, 773, 773, 773, 773, 773, 
	773, 773, 773, 773, 773, 773, 773, 773, 
	773, 773, 773, 773, 773, 773, 773, 773, 
	773, 773, 773, 773, 773, 773, 773, 773, 
	773, 773, 773, 773, 773, 773, 773, 773, 
	773, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 773, 773, 773, 773, 773, 
	773, 773, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 773, 773, 773, 773, 
	762, 773, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 773, 773, 773, 773, 
	773, 762, 765, 774, 765, 765, 765, 774, 
	774, 774, 774, 774, 774, 774, 774, 774, 
	774, 774, 774, 774, 774, 774, 774, 774, 
	774, 765, 774, 775, 776, 777, 778, 779, 
	781, 780, 782, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 783, 783, 783, 
	783, 783, 783, 783, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 783, 783, 
	783, 783, 762, 783, 762, 762, 762, 762, 
	762, 784, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 783, 783, 
	783, 783, 783, 762, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 783, 783, 
	783, 783, 783, 783, 783, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 783, 
	783, 783, 783, 762, 783, 762, 762, 762, 
	762, 785, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 783, 
	783, 783, 783, 783, 762, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 783, 
	783, 783, 783, 783, 783, 783, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	783, 783, 783, 783, 762, 783, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	786, 762, 762, 762, 762, 762, 762, 762, 
	783, 783, 783, 783, 783, 762, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	783, 783, 783, 783, 783, 783, 783, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 783, 783, 783, 783, 762, 783, 762, 
	762, 787, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 783, 783, 783, 783, 783, 762, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 783, 783, 783, 783, 783, 783, 783, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 783, 783, 783, 783, 762, 783, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 788, 762, 762, 762, 
	762, 762, 783, 783, 783, 783, 783, 762, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 783, 783, 783, 783, 783, 783, 
	783, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 783, 783, 783, 783, 762, 
	783, 762, 762, 762, 762, 784, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 783, 783, 783, 783, 783, 
	762, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 783, 783, 783, 783, 783, 
	783, 783, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 783, 783, 783, 783, 
	762, 783, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 789, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 783, 783, 783, 783, 
	783, 762, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 783, 783, 783, 783, 
	783, 783, 783, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 783, 783, 783, 
	783, 762, 783, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 790, 762, 
	762, 762, 762, 762, 762, 762, 791, 762, 
	762, 762, 762, 762, 762, 783, 783, 783, 
	783, 783, 762, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 783, 783, 783, 
	783, 783, 783, 783, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 783, 783, 
	783, 783, 762, 783, 762, 762, 762, 762, 
	792, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 783, 783, 
	783, 783, 783, 762, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 783, 783, 
	783, 783, 783, 783, 783, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 783, 
	783, 783, 783, 762, 783, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 793, 
	762, 762, 762, 762, 762, 762, 762, 783, 
	783, 783, 783, 783, 762, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 783, 
	783, 783, 783, 783, 783, 783, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	783, 783, 783, 783, 762, 783, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	784, 762, 762, 762, 762, 762, 762, 762, 
	783, 783, 783, 783, 783, 762, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	783, 783, 783, 783, 783, 783, 783, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 783, 783, 783, 783, 762, 783, 762, 
	762, 762, 762, 762, 762, 762, 762, 794, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 783, 783, 783, 783, 783, 762, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 783, 783, 783, 783, 783, 783, 783, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 783, 783, 783, 783, 762, 783, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 784, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 783, 783, 783, 783, 783, 762, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 783, 783, 783, 783, 783, 783, 
	783, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 783, 783, 783, 783, 762, 
	783, 762, 762, 762, 762, 762, 762, 762, 
	795, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 783, 783, 783, 783, 783, 
	762, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 783, 783, 783, 783, 783, 
	783, 783, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 783, 783, 783, 783, 
	762, 783, 762, 762, 762, 762, 762, 762, 
	762, 762, 796, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 783, 783, 783, 783, 
	783, 762, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 783, 783, 783, 783, 783, 783, 
	783, 783, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 783, 783, 783, 783, 
	783, 783, 783, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 783, 783, 783, 
	783, 762, 783, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 762, 788, 762, 
	762, 762, 762, 762, 762, 762, 762, 762, 
	762, 762, 762, 762, 762, 783, 783, 783, 
	783, 783, 762, 798, 196, 196, 196, 798, 
	196, 196, 196, 196, 799, 800, 799, 799, 
	799, 196, 196, 196, 196, 196, 196, 196, 
	196, 196, 196, 196, 196, 798, 196, 196, 
	196, 196, 196, 799, 801, 196, 802, 196, 
	803, 804, 196, 805, 196, 806, 807, 196, 
	808, 809, 810, 196, 196, 196, 196, 196, 
	196, 196, 196, 196, 196, 811, 196, 812, 
	813, 814, 815, 196, 816, 817, 816, 816, 
	818, 816, 816, 816, 816, 816, 816, 816, 
	816, 816, 816, 816, 816, 816, 816, 816, 
	816, 816, 816, 816, 816, 816, 819, 820, 
	196, 821, 822, 196, 823, 824, 825, 826, 
	827, 828, 797, 797, 829, 797, 797, 797, 
	830, 831, 832, 797, 797, 833, 834, 835, 
	836, 797, 837, 797, 838, 797, 839, 840, 
	196, 821, 196, 797, 185, 185, 185, 185, 
	185, 185, 185, 185, 185, 185, 185, 185, 
	185, 185, 185, 185, 185, 185, 185, 185, 
	185, 185, 185, 185, 185, 185, 185, 185, 
	185, 185, 185, 185, 185, 241, 185, 185, 
	185, 185, 185, 185, 185, 185, 185, 185, 
	185, 185, 185, 185, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 242, 185, 
	185, 243, 185, 241, 185, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 185, 
	185, 185, 185, 797, 185, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 185, 
	185, 185, 185, 185, 797, 842, 841, 841, 
	843, 841, 844, 846, 847, 845, 845, 845, 
	845, 845, 845, 845, 845, 845, 845, 845, 
	845, 845, 845, 845, 845, 845, 845, 845, 
	845, 845, 845, 845, 845, 845, 845, 845, 
	845, 845, 845, 845, 845, 845, 845, 845, 
	845, 845, 845, 845, 845, 845, 845, 845, 
	845, 845, 845, 845, 845, 845, 845, 845, 
	845, 845, 845, 845, 845, 845, 845, 845, 
	845, 845, 845, 845, 848, 845, 850, 849, 
	851, 852, 853, 799, 854, 799, 799, 799, 
	854, 854, 854, 854, 854, 854, 854, 854, 
	854, 854, 854, 854, 854, 854, 854, 854, 
	854, 854, 799, 854, 856, 855, 858, 859, 
	858, 858, 858, 857, 857, 857, 857, 857, 
	857, 857, 857, 857, 857, 857, 857, 857, 
	857, 857, 857, 857, 857, 858, 857, 196, 
	860, 860, 860, 860, 860, 860, 860, 860, 
	860, 860, 860, 860, 860, 860, 860, 860, 
	860, 860, 860, 860, 860, 860, 860, 860, 
	860, 860, 860, 860, 860, 860, 860, 860, 
	860, 860, 860, 860, 860, 860, 860, 860, 
	860, 860, 860, 860, 860, 860, 860, 860, 
	860, 860, 860, 860, 860, 860, 860, 860, 
	860, 860, 860, 860, 860, 860, 860, 860, 
	196, 860, 861, 862, 863, 864, 867, 867, 
	867, 867, 867, 867, 867, 867, 867, 867, 
	867, 867, 867, 867, 867, 867, 867, 867, 
	867, 867, 867, 867, 867, 867, 867, 867, 
	866, 866, 866, 866, 866, 866, 867, 867, 
	867, 867, 867, 867, 867, 867, 867, 867, 
	867, 867, 867, 867, 867, 867, 867, 867, 
	867, 867, 867, 867, 867, 867, 867, 867, 
	866, 868, 194, 194, 194, 868, 194, 194, 
	194, 194, 194, 194, 194, 194, 194, 194, 
	194, 194, 194, 194, 194, 194, 194, 194, 
	194, 194, 194, 868, 194, 194, 194, 194, 
	194, 194, 194, 194, 194, 194, 194, 194, 
	194, 194, 194, 194, 194, 194, 194, 194, 
	194, 194, 194, 194, 194, 194, 194, 194, 
	194, 194, 194, 194, 194, 194, 194, 194, 
	194, 194, 195, 195, 195, 195, 195, 195, 
	195, 195, 195, 195, 195, 195, 195, 195, 
	195, 195, 195, 195, 195, 195, 195, 195, 
	195, 195, 195, 195, 194, 194, 194, 194, 
	194, 194, 195, 195, 195, 195, 195, 195, 
	195, 195, 195, 195, 195, 195, 195, 195, 
	195, 195, 195, 195, 195, 195, 195, 195, 
	195, 195, 195, 195, 194, 870, 869, 869, 
	869, 869, 869, 869, 869, 869, 869, 869, 
	869, 869, 869, 869, 869, 869, 869, 869, 
	869, 869, 869, 869, 196, 869, 871, 873, 
	872, 872, 872, 872, 872, 872, 872, 872, 
	872, 872, 872, 872, 872, 872, 872, 872, 
	872, 872, 196, 872, 196, 185, 874, 874, 
	874, 874, 874, 874, 874, 874, 874, 874, 
	869, 869, 869, 196, 869, 874, 874, 874, 
	874, 874, 874, 874, 874, 874, 874, 869, 
	869, 869, 196, 196, 869, 875, 860, 196, 
	860, 860, 876, 876, 876, 860, 876, 876, 
	876, 876, 876, 876, 876, 876, 876, 876, 
	876, 876, 876, 876, 876, 876, 876, 876, 
	876, 876, 876, 860, 876, 860, 860, 860, 
	860, 860, 860, 860, 860, 860, 860, 860, 
	860, 860, 860, 860, 860, 860, 860, 860, 
	860, 860, 860, 860, 860, 860, 860, 860, 
	860, 860, 860, 860, 860, 860, 878, 879, 
	860, 880, 199, 881, 879, 860, 860, 882, 
	883, 860, 883, 860, 199, 860, 860, 860, 
	860, 860, 860, 860, 860, 860, 860, 884, 
	860, 885, 886, 887, 860, 888, 877, 877, 
	877, 877, 877, 877, 877, 877, 877, 877, 
	877, 877, 877, 877, 877, 877, 877, 877, 
	877, 877, 877, 877, 877, 877, 877, 877, 
	889, 860, 860, 199, 877, 199, 877, 877, 
	877, 877, 877, 877, 877, 877, 877, 877, 
	877, 877, 877, 877, 877, 877, 877, 877, 
	877, 877, 877, 877, 877, 877, 877, 877, 
	860, 890, 860, 883, 860, 877, 891, 891, 
	891, 891, 891, 891, 891, 891, 891, 891, 
	891, 891, 891, 891, 891, 891, 891, 891, 
	891, 891, 891, 891, 891, 891, 891, 891, 
	891, 891, 891, 891, 891, 891, 891, 892, 
	891, 891, 891, 891, 891, 891, 891, 891, 
	891, 891, 891, 891, 891, 891, 877, 877, 
	877, 877, 877, 877, 877, 877, 877, 877, 
	891, 891, 891, 893, 891, 892, 891, 877, 
	877, 877, 877, 877, 877, 877, 877, 877, 
	877, 877, 877, 877, 877, 877, 877, 877, 
	877, 877, 877, 877, 877, 877, 877, 877, 
	877, 891, 891, 891, 891, 877, 891, 877, 
	877, 877, 877, 877, 877, 877, 877, 877, 
	877, 877, 877, 877, 877, 877, 877, 877, 
	877, 877, 877, 877, 877, 877, 877, 877, 
	877, 891, 891, 891, 891, 891, 877, 895, 
	894, 896, 898, 899, 897, 897, 897, 897, 
	897, 897, 897, 897, 897, 897, 897, 897, 
	897, 897, 897, 897, 897, 897, 897, 897, 
	897, 897, 897, 897, 897, 897, 897, 897, 
	897, 897, 897, 897, 897, 897, 897, 897, 
	897, 897, 897, 897, 897, 897, 897, 897, 
	897, 897, 897, 897, 897, 897, 897, 897, 
	897, 897, 897, 897, 897, 897, 897, 897, 
	897, 897, 897, 900, 897, 902, 903, 901, 
	904, 905, 906, 907, 199, 891, 891, 199, 
	891, 891, 891, 891, 891, 891, 891, 891, 
	891, 891, 891, 891, 891, 891, 891, 891, 
	891, 891, 891, 891, 891, 891, 891, 891, 
	891, 891, 891, 891, 891, 891, 891, 891, 
	891, 891, 891, 891, 891, 891, 891, 891, 
	891, 891, 891, 891, 891, 891, 891, 891, 
	891, 891, 891, 891, 891, 891, 891, 891, 
	891, 891, 891, 891, 891, 199, 891, 891, 
	891, 891, 891, 891, 891, 891, 891, 891, 
	891, 891, 891, 891, 891, 891, 891, 891, 
	891, 891, 891, 891, 891, 891, 891, 891, 
	891, 891, 891, 891, 891, 891, 891, 891, 
	891, 891, 891, 891, 891, 891, 891, 891, 
	891, 891, 891, 891, 891, 891, 891, 198, 
	198, 198, 198, 198, 198, 198, 198, 198, 
	198, 891, 891, 891, 891, 891, 891, 891, 
	198, 198, 198, 198, 198, 198, 198, 198, 
	198, 198, 198, 198, 198, 198, 198, 198, 
	198, 198, 198, 198, 198, 198, 198, 198, 
	198, 198, 891, 891, 891, 891, 198, 891, 
	198, 198, 198, 198, 198, 198, 198, 198, 
	198, 198, 198, 198, 198, 198, 198, 198, 
	198, 198, 198, 198, 198, 198, 198, 198, 
	198, 198, 891, 891, 891, 891, 891, 198, 
	201, 201, 201, 201, 201, 201, 201, 201, 
	201, 201, 891, 199, 891, 199, 891, 199, 
	891, 199, 908, 891, 199, 891, 199, 891, 
	199, 199, 891, 199, 891, 909, 910, 860, 
	911, 196, 860, 821, 196, 860, 860, 860, 
	860, 860, 860, 860, 860, 860, 860, 860, 
	860, 860, 860, 860, 860, 860, 860, 860, 
	860, 860, 860, 860, 860, 860, 860, 860, 
	860, 860, 860, 860, 860, 860, 860, 860, 
	860, 860, 860, 860, 860, 860, 860, 860, 
	860, 860, 860, 860, 860, 860, 860, 860, 
	860, 860, 860, 860, 860, 860, 860, 860, 
	860, 860, 860, 860, 196, 860, 196, 870, 
	860, 914, 913, 913, 913, 914, 913, 913, 
	913, 913, 915, 916, 915, 915, 915, 913, 
	913, 913, 913, 913, 913, 913, 913, 913, 
	913, 913, 913, 914, 913, 913, 913, 913, 
	913, 915, 913, 913, 913, 913, 913, 913, 
	913, 913, 913, 913, 913, 913, 913, 913, 
	913, 913, 913, 913, 913, 913, 913, 913, 
	913, 913, 913, 913, 913, 913, 913, 913, 
	913, 913, 917, 917, 917, 917, 917, 917, 
	917, 917, 917, 917, 917, 917, 917, 917, 
	917, 917, 917, 917, 917, 917, 917, 917, 
	917, 917, 917, 917, 913, 918, 913, 913, 
	917, 913, 917, 917, 917, 917, 917, 917, 
	917, 917, 917, 917, 917, 917, 917, 917, 
	917, 917, 917, 917, 917, 917, 917, 917, 
	917, 917, 917, 917, 913, 919, 919, 919, 
	919, 919, 919, 919, 919, 919, 919, 919, 
	919, 919, 919, 919, 919, 919, 919, 919, 
	919, 919, 919, 919, 919, 919, 919, 919, 
	919, 919, 919, 919, 919, 919, 919, 919, 
	919, 919, 919, 919, 919, 919, 919, 919, 
	919, 919, 919, 919, 919, 919, 919, 919, 
	919, 919, 919, 919, 919, 919, 919, 919, 
	919, 919, 919, 919, 919, 919, 919, 920, 
	920, 920, 920, 920, 920, 920, 920, 920, 
	920, 920, 920, 920, 920, 920, 920, 920, 
	920, 920, 920, 920, 920, 920, 920, 920, 
	920, 919, 919, 919, 919, 920, 919, 920, 
	920, 920, 920, 920, 920, 920, 920, 920, 
	920, 920, 920, 920, 920, 920, 920, 920, 
	920, 920, 920, 920, 920, 920, 920, 920, 
	920, 919, 919, 919, 919, 919, 920, 921, 
	921, 921, 921, 921, 921, 921, 921, 921, 
	921, 921, 921, 921, 921, 921, 921, 921, 
	921, 921, 921, 921, 921, 921, 921, 921, 
	921, 921, 921, 921, 921, 921, 921, 921, 
	921, 921, 921, 921, 921, 921, 921, 921, 
	921, 921, 921, 921, 921, 921, 921, 922, 
	922, 922, 922, 922, 922, 922, 922, 922, 
	922, 921, 921, 921, 921, 921, 921, 921, 
	922, 922, 922, 922, 922, 922, 922, 922, 
	922, 922, 922, 922, 922, 922, 922, 922, 
	922, 922, 922, 922, 922, 922, 922, 922, 
	922, 922, 921, 921, 921, 921, 922, 921, 
	922, 922, 922, 922, 922, 922, 922, 922, 
	922, 922, 922, 922, 922, 922, 922, 922, 
	922, 922, 922, 922, 922, 922, 922, 922, 
	922, 922, 921, 921, 921, 921, 921, 922, 
	925, 924, 924, 924, 925, 924, 924, 924, 
	924, 924, 924, 924, 924, 924, 924, 924, 
	924, 924, 924, 924, 924, 924, 924, 924, 
	924, 924, 925, 924, 924, 924, 924, 924, 
	924, 924, 924, 924, 924, 924, 924, 924, 
	924, 924, 924, 924, 924, 924, 924, 924, 
	926, 926, 926, 926, 926, 926, 926, 926, 
	924, 924, 924, 924, 924, 924, 924, 924, 
	924, 924, 924, 927, 924, 924, 924, 924, 
	924, 924, 924, 924, 924, 928, 924, 924, 
	924, 924, 924, 924, 924, 924, 924, 924, 
	924, 924, 924, 924, 924, 924, 924, 924, 
	924, 924, 924, 929, 924, 924, 924, 924, 
	924, 924, 924, 924, 924, 924, 924, 924, 
	924, 924, 924, 924, 924, 930, 924, 924, 
	931, 924, 932, 933, 935, 935, 935, 935, 
	935, 935, 935, 935, 934, 936, 936, 936, 
	936, 936, 936, 936, 936, 934, 934, 937, 
	937, 235, 235, 235, 937, 235, 235, 235, 
	235, 235, 235, 235, 235, 235, 235, 235, 
	235, 235, 235, 235, 235, 235, 235, 235, 
	235, 235, 937, 235, 235, 235, 235, 235, 
	235, 235, 235, 235, 235, 235, 235, 235, 
	235, 235, 235, 235, 235, 235, 235, 235, 
	235, 235, 235, 235, 235, 235, 235, 235, 
	235, 235, 235, 235, 235, 235, 235, 236, 
	235, 235, 235, 235, 235, 235, 235, 235, 
	235, 235, 235, 235, 235, 235, 235, 235, 
	235, 235, 235, 235, 235, 235, 235, 235, 
	235, 235, 235, 235, 938, 235, 939, 940, 
	941, 941, 235, 235, 235, 941, 235, 235, 
	235, 235, 235, 235, 235, 235, 235, 235, 
	235, 235, 235, 235, 235, 235, 235, 235, 
	235, 235, 235, 941, 235, 235, 235, 235, 
	235, 235, 235, 235, 235, 235, 235, 235, 
	235, 235, 235, 235, 235, 235, 235, 235, 
	235, 235, 235, 235, 235, 235, 235, 235, 
	235, 235, 235, 235, 235, 235, 235, 235, 
	236, 235, 235, 235, 235, 235, 235, 235, 
	235, 235, 235, 235, 235, 235, 235, 235, 
	235, 235, 235, 235, 235, 235, 235, 235, 
	235, 235, 235, 235, 235, 942, 235, 234, 
	941, 943, 944, 945, 945, 222, 222, 222, 
	945, 222, 222, 222, 222, 222, 222, 222, 
	222, 222, 222, 222, 222, 222, 222, 222, 
	222, 222, 222, 222, 222, 222, 945, 222, 
	222, 222, 222, 222, 222, 222, 222, 222, 
	222, 222, 222, 222, 222, 222, 222, 222, 
	222, 222, 222, 222, 222, 222, 222, 222, 
	222, 222, 222, 222, 222, 222, 222, 222, 
	222, 222, 222, 223, 222, 222, 222, 222, 
	222, 222, 222, 222, 222, 222, 222, 222, 
	222, 222, 222, 222, 222, 222, 222, 222, 
	222, 222, 222, 222, 222, 222, 222, 222, 
	946, 222, 941, 216, 216, 216, 941, 216, 
	216, 216, 216, 216, 216, 216, 216, 216, 
	216, 216, 216, 216, 216, 216, 216, 216, 
	216, 216, 216, 216, 941, 216, 216, 216, 
	216, 216, 216, 216, 216, 216, 216, 216, 
	216, 216, 216, 216, 216, 216, 216, 947, 
	216, 937, 222, 222, 222, 937, 222, 222, 
	222, 222, 222, 222, 222, 222, 222, 222, 
	222, 222, 222, 222, 222, 222, 222, 222, 
	222, 222, 222, 937, 222, 222, 222, 222, 
	222, 222, 222, 222, 222, 222, 222, 222, 
	222, 222, 222, 222, 222, 222, 222, 222, 
	222, 222, 222, 222, 222, 222, 222, 222, 
	222, 222, 222, 222, 222, 222, 222, 222, 
	223, 222, 222, 222, 222, 222, 222, 222, 
	222, 222, 222, 222, 222, 222, 222, 222, 
	222, 222, 222, 222, 222, 222, 222, 222, 
	222, 222, 222, 222, 222, 224, 222, 937, 
	948, 948, 948, 937, 948, 948, 948, 948, 
	948, 948, 948, 948, 948, 948, 948, 948, 
	948, 948, 948, 948, 948, 948, 948, 948, 
	948, 937, 948, 948, 948, 948, 948, 948, 
	948, 948, 948, 948, 948, 948, 948, 948, 
	948, 948, 948, 948, 948, 948, 948, 948, 
	948, 948, 948, 948, 948, 948, 948, 948, 
	948, 948, 948, 948, 948, 948, 948, 948, 
	948, 948, 948, 948, 948, 948, 948, 948, 
	948, 948, 948, 948, 948, 948, 948, 948, 
	948, 948, 948, 948, 948, 948, 948, 948, 
	948, 948, 948, 949, 948, 950, 951, 951, 
	948, 948, 948, 951, 948, 948, 948, 948, 
	948, 948, 948, 948, 948, 948, 948, 948, 
	948, 948, 948, 948, 948, 948, 948, 948, 
	948, 951, 948, 948, 948, 948, 948, 948, 
	948, 948, 948, 948, 948, 948, 948, 948, 
	948, 948, 948, 948, 948, 948, 948, 948, 
	948, 948, 948, 948, 948, 948, 948, 948, 
	948, 948, 948, 948, 948, 948, 948, 948, 
	948, 948, 948, 948, 948, 948, 948, 948, 
	948, 948, 948, 948, 948, 948, 948, 948, 
	948, 948, 948, 948, 948, 948, 948, 948, 
	948, 948, 948, 952, 948, 951, 216, 216, 
	216, 951, 216, 216, 216, 216, 216, 216, 
	216, 216, 216, 216, 216, 216, 216, 216, 
	216, 216, 216, 216, 216, 216, 216, 951, 
	216, 216, 216, 216, 216, 216, 216, 216, 
	216, 216, 216, 216, 216, 216, 216, 216, 
	216, 216, 947, 216, 953, 951, 951, 222, 
	222, 222, 951, 222, 222, 222, 222, 222, 
	222, 222, 222, 222, 222, 222, 222, 222, 
	222, 222, 222, 222, 222, 222, 222, 222, 
	951, 222, 222, 222, 222, 222, 222, 222, 
	222, 222, 222, 222, 222, 222, 222, 222, 
	222, 222, 222, 222, 222, 222, 222, 222, 
	222, 222, 222, 222, 222, 222, 222, 222, 
	222, 222, 222, 222, 222, 223, 222, 222, 
	222, 222, 222, 222, 222, 222, 222, 222, 
	222, 222, 222, 222, 222, 222, 222, 222, 
	222, 222, 222, 222, 222, 222, 222, 222, 
	222, 222, 224, 222, 954, 955, 955, 955, 
	954, 955, 955, 955, 955, 955, 955, 955, 
	955, 955, 955, 955, 955, 955, 955, 955, 
	955, 955, 955, 955, 955, 955, 954, 955, 
	955, 955, 955, 955, 955, 955, 955, 955, 
	955, 955, 955, 955, 955, 955, 955, 955, 
	955, 955, 955, 955, 956, 956, 956, 956, 
	956, 956, 956, 956, 956, 956, 955, 955, 
	955, 955, 955, 955, 955, 956, 956, 956, 
	956, 956, 956, 955, 955, 955, 955, 955, 
	955, 955, 955, 955, 955, 955, 955, 955, 
	955, 955, 955, 955, 955, 955, 955, 955, 
	955, 955, 955, 955, 955, 956, 956, 956, 
	956, 956, 956, 955, 955, 955, 955, 955, 
	955, 955, 955, 955, 955, 955, 955, 955, 
	955, 955, 955, 955, 955, 955, 955, 957, 
	955, 954, 958, 958, 958, 954, 958, 958, 
	958, 958, 958, 958, 958, 958, 958, 958, 
	958, 958, 958, 958, 958, 958, 958, 958, 
	958, 958, 958, 954, 958, 954, 959, 959, 
	959, 954, 959, 959, 959, 959, 959, 959, 
	959, 959, 959, 959, 959, 959, 959, 959, 
	959, 959, 959, 959, 959, 959, 959, 954, 
	959, 954, 960, 960, 960, 954, 960, 960, 
	960, 960, 960, 960, 960, 960, 960, 960, 
	960, 960, 960, 960, 960, 960, 960, 960, 
	960, 960, 960, 954, 960, 954, 954, 958, 
	958, 958, 954, 958, 958, 958, 958, 958, 
	958, 958, 958, 958, 958, 958, 958, 958, 
	958, 958, 958, 958, 958, 958, 958, 958, 
	954, 958, 958, 958, 958, 958, 958, 958, 
	958, 958, 958, 958, 958, 958, 958, 958, 
	958, 958, 958, 958, 958, 958, 961, 961, 
	961, 961, 961, 961, 961, 961, 961, 961, 
	958, 958, 958, 958, 958, 958, 958, 961, 
	961, 961, 961, 961, 961, 958, 958, 958, 
	958, 958, 958, 958, 958, 958, 958, 958, 
	958, 958, 958, 958, 958, 958, 958, 958, 
	958, 958, 958, 958, 958, 958, 958, 961, 
	961, 961, 961, 961, 961, 958, 954, 959, 
	959, 959, 954, 959, 959, 959, 959, 959, 
	959, 959, 959, 959, 959, 959, 959, 959, 
	959, 959, 959, 959, 959, 959, 959, 959, 
	954, 959, 959, 959, 959, 959, 959, 959, 
	959, 959, 959, 959, 959, 959, 959, 959, 
	959, 959, 959, 959, 959, 959, 962, 962, 
	962, 962, 962, 962, 962, 962, 962, 962, 
	959, 959, 959, 959, 959, 959, 959, 962, 
	962, 962, 962, 962, 962, 959, 959, 959, 
	959, 959, 959, 959, 959, 959, 959, 959, 
	959, 959, 959, 959, 959, 959, 959, 959, 
	959, 959, 959, 959, 959, 959, 959, 962, 
	962, 962, 962, 962, 962, 959, 954, 960, 
	960, 960, 954, 960, 960, 960, 960, 960, 
	960, 960, 960, 960, 960, 960, 960, 960, 
	960, 960, 960, 960, 960, 960, 960, 960, 
	954, 960, 960, 960, 960, 960, 960, 960, 
	960, 960, 960, 960, 960, 960, 960, 960, 
	960, 960, 960, 960, 960, 960, 963, 963, 
	963, 963, 963, 963, 963, 963, 963, 963, 
	960, 960, 960, 960, 960, 960, 960, 963, 
	963, 963, 963, 963, 963, 960, 960, 960, 
	960, 960, 960, 960, 960, 960, 960, 960, 
	960, 960, 960, 960, 960, 960, 960, 960, 
	960, 960, 960, 960, 960, 960, 960, 963, 
	963, 963, 963, 963, 963, 960, 964, 967, 
	966, 966, 966, 967, 966, 966, 966, 966, 
	966, 966, 966, 966, 966, 966, 966, 966, 
	966, 966, 966, 966, 966, 966, 966, 966, 
	966, 967, 966, 966, 966, 966, 966, 966, 
	966, 966, 966, 966, 966, 966, 966, 966, 
	966, 966, 966, 966, 966, 966, 966, 968, 
	968, 968, 968, 968, 968, 968, 968, 968, 
	968, 966, 966, 966, 966, 966, 966, 966, 
	968, 968, 968, 968, 968, 968, 966, 966, 
	966, 966, 966, 966, 966, 966, 966, 966, 
	966, 966, 966, 966, 966, 966, 966, 966, 
	966, 966, 966, 966, 966, 966, 966, 966, 
	968, 968, 968, 968, 968, 968, 966, 966, 
	966, 966, 966, 966, 966, 966, 966, 966, 
	966, 966, 966, 966, 966, 966, 966, 966, 
	966, 966, 966, 966, 958, 966, 967, 969, 
	969, 969, 967, 969, 969, 969, 969, 969, 
	969, 969, 969, 969, 969, 969, 969, 969, 
	969, 969, 969, 969, 969, 969, 969, 969, 
	967, 969, 969, 969, 969, 969, 969, 969, 
	969, 969, 969, 969, 969, 969, 969, 969, 
	969, 969, 969, 969, 969, 969, 970, 970, 
	970, 970, 970, 970, 970, 970, 970, 970, 
	969, 969, 969, 969, 969, 969, 969, 970, 
	970, 970, 970, 970, 970, 969, 969, 969, 
	969, 969, 969, 969, 969, 969, 969, 969, 
	969, 969, 969, 969, 969, 969, 969, 969, 
	969, 969, 969, 969, 969, 969, 969, 970, 
	970, 970, 970, 970, 970, 969, 969, 969, 
	969, 969, 969, 969, 969, 969, 969, 969, 
	969, 969, 969, 969, 969, 969, 969, 969, 
	969, 969, 969, 971, 969, 967, 972, 972, 
	972, 967, 972, 972, 972, 972, 972, 972, 
	972, 972, 972, 972, 972, 972, 972, 972, 
	972, 972, 972, 972, 972, 972, 972, 967, 
	972, 972, 972, 972, 972, 972, 972, 972, 
	972, 972, 972, 972, 972, 972, 972, 972, 
	972, 972, 972, 972, 972, 973, 973, 973, 
	973, 973, 973, 973, 973, 973, 973, 972, 
	972, 972, 972, 972, 972, 972, 973, 973, 
	973, 973, 973, 973, 972, 972, 972, 972, 
	972, 972, 972, 972, 972, 972, 972, 972, 
	972, 972, 972, 972, 972, 972, 972, 972, 
	972, 972, 972, 972, 972, 972, 973, 973, 
	973, 973, 973, 973, 972, 972, 972, 972, 
	972, 972, 972, 972, 972, 972, 972, 972, 
	972, 972, 972, 972, 972, 972, 972, 972, 
	972, 972, 974, 972, 967, 975, 975, 975, 
	967, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 967, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 976, 976, 976, 976, 
	976, 976, 976, 976, 976, 976, 975, 975, 
	975, 975, 975, 975, 975, 976, 976, 976, 
	976, 976, 976, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 976, 976, 976, 
	976, 976, 976, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 967, 975, 967, 975, 975, 975, 967, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 967, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 976, 976, 976, 976, 976, 
	976, 976, 976, 976, 976, 975, 975, 975, 
	975, 975, 975, 975, 976, 976, 976, 976, 
	976, 976, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 976, 976, 976, 976, 
	976, 976, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	967, 975, 977, 967, 976, 976, 976, 967, 
	976, 976, 976, 976, 976, 976, 976, 976, 
	976, 976, 976, 976, 976, 976, 976, 976, 
	976, 976, 976, 976, 976, 967, 976, 976, 
	976, 976, 976, 976, 976, 976, 976, 976, 
	976, 976, 976, 976, 976, 976, 976, 976, 
	976, 976, 976, 976, 976, 976, 976, 976, 
	976, 976, 976, 976, 976, 976, 976, 976, 
	976, 976, 976, 976, 976, 976, 976, 976, 
	976, 976, 976, 976, 976, 976, 976, 976, 
	976, 976, 976, 976, 976, 976, 976, 976, 
	976, 976, 976, 976, 976, 976, 976, 976, 
	976, 976, 976, 976, 976, 976, 976, 976, 
	976, 976, 976, 976, 976, 976, 976, 976, 
	976, 976, 976, 976, 976, 976, 976, 976, 
	976, 976, 976, 976, 976, 976, 976, 976, 
	977, 976, 967, 976, 976, 976, 967, 976, 
	976, 976, 976, 976, 976, 976, 976, 976, 
	976, 976, 976, 976, 976, 976, 976, 976, 
	976, 976, 976, 976, 967, 976, 976, 976, 
	976, 976, 976, 976, 976, 976, 976, 976, 
	976, 976, 976, 976, 976, 976, 976, 976, 
	976, 976, 976, 976, 976, 976, 976, 976, 
	976, 976, 976, 976, 976, 976, 976, 976, 
	976, 976, 976, 976, 976, 976, 976, 976, 
	976, 976, 976, 976, 976, 976, 976, 976, 
	976, 976, 976, 976, 976, 976, 976, 976, 
	976, 976, 976, 976, 976, 976, 976, 976, 
	976, 976, 976, 976, 976, 976, 976, 976, 
	976, 976, 976, 976, 976, 976, 976, 976, 
	976, 976, 976, 976, 976, 976, 976, 976, 
	976, 976, 976, 976, 976, 976, 976, 965, 
	976, 965, 967, 973, 973, 973, 967, 973, 
	973, 973, 973, 973, 973, 973, 973, 973, 
	973, 973, 973, 973, 973, 973, 973, 973, 
	973, 973, 973, 973, 967, 973, 973, 973, 
	973, 973, 973, 973, 973, 973, 973, 973, 
	973, 973, 973, 973, 973, 973, 973, 973, 
	973, 973, 973, 973, 973, 973, 973, 973, 
	973, 973, 973, 973, 973, 973, 973, 973, 
	973, 973, 973, 973, 973, 973, 973, 973, 
	973, 973, 973, 973, 973, 973, 973, 973, 
	973, 973, 973, 973, 973, 973, 973, 973, 
	973, 973, 973, 973, 973, 973, 973, 973, 
	973, 973, 973, 973, 973, 973, 973, 973, 
	973, 973, 973, 973, 973, 973, 973, 973, 
	973, 973, 973, 973, 973, 973, 973, 973, 
	973, 973, 973, 973, 973, 973, 973, 960, 
	973, 965, 960, 960, 960, 965, 960, 960, 
	960, 960, 960, 960, 960, 960, 960, 960, 
	960, 960, 960, 960, 960, 960, 960, 960, 
	960, 960, 960, 965, 960, 967, 969, 969, 
	969, 967, 969, 969, 969, 969, 978, 969, 
	969, 969, 969, 969, 969, 969, 969, 969, 
	969, 969, 969, 969, 969, 969, 969, 967, 
	969, 969, 969, 969, 969, 978, 969, 969, 
	969, 969, 969, 969, 969, 969, 969, 969, 
	969, 969, 969, 969, 969, 979, 979, 979, 
	979, 979, 979, 979, 979, 979, 979, 969, 
	969, 969, 969, 969, 969, 969, 979, 979, 
	979, 979, 979, 979, 969, 969, 969, 969, 
	969, 969, 969, 969, 969, 969, 969, 969, 
	969, 969, 969, 969, 969, 969, 969, 969, 
	969, 969, 969, 969, 969, 969, 979, 979, 
	979, 979, 979, 979, 969, 969, 969, 969, 
	969, 969, 969, 969, 969, 969, 969, 969, 
	969, 969, 969, 969, 969, 969, 969, 969, 
	969, 969, 980, 969, 967, 975, 975, 975, 
	967, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 967, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 981, 981, 981, 981, 
	981, 981, 981, 981, 981, 981, 975, 975, 
	975, 975, 975, 975, 975, 981, 981, 981, 
	981, 981, 981, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 981, 981, 981, 
	981, 981, 981, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 967, 975, 967, 975, 975, 975, 967, 
	975, 975, 975, 975, 978, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 967, 975, 975, 
	975, 975, 975, 978, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 982, 982, 982, 982, 982, 
	982, 982, 982, 982, 982, 975, 975, 975, 
	975, 975, 975, 975, 982, 982, 982, 982, 
	982, 982, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 982, 982, 982, 982, 
	982, 982, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	977, 975, 967, 975, 975, 975, 967, 975, 
	975, 975, 975, 978, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 967, 975, 975, 975, 
	975, 975, 978, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 983, 983, 983, 983, 983, 983, 
	983, 983, 983, 983, 975, 975, 975, 975, 
	975, 975, 975, 983, 983, 983, 983, 983, 
	983, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 983, 983, 983, 983, 983, 
	983, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 977, 
	975, 967, 975, 975, 975, 967, 975, 975, 
	975, 975, 978, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 967, 975, 975, 975, 975, 
	975, 978, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 984, 984, 984, 984, 984, 984, 984, 
	984, 984, 984, 975, 975, 975, 975, 975, 
	975, 975, 984, 984, 984, 984, 984, 984, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 984, 984, 984, 984, 984, 984, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 977, 975, 
	967, 975, 975, 975, 967, 975, 975, 975, 
	975, 978, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 967, 975, 975, 975, 975, 975, 
	978, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	985, 985, 985, 985, 985, 985, 985, 985, 
	985, 985, 975, 975, 975, 975, 975, 975, 
	975, 985, 985, 985, 985, 985, 985, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 985, 985, 985, 985, 985, 985, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 977, 975, 967, 
	975, 975, 975, 967, 975, 975, 975, 975, 
	978, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 967, 975, 975, 975, 975, 975, 978, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 986, 
	986, 986, 986, 986, 986, 986, 986, 986, 
	986, 975, 975, 975, 975, 975, 975, 975, 
	986, 986, 986, 986, 986, 986, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	986, 986, 986, 986, 986, 986, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 977, 975, 967, 975, 
	975, 975, 967, 975, 975, 975, 975, 978, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	967, 975, 975, 975, 975, 975, 978, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 987, 987, 
	987, 987, 987, 987, 987, 987, 987, 987, 
	975, 975, 975, 975, 975, 975, 975, 987, 
	987, 987, 987, 987, 987, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 987, 
	987, 987, 987, 987, 987, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 977, 975, 967, 975, 975, 
	975, 967, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 967, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 987, 987, 987, 
	987, 987, 987, 987, 987, 987, 987, 975, 
	975, 975, 975, 975, 975, 975, 987, 987, 
	987, 987, 987, 987, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 987, 987, 
	987, 987, 987, 987, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 977, 975, 967, 972, 972, 972, 
	967, 972, 972, 972, 972, 978, 972, 972, 
	972, 972, 972, 972, 972, 972, 972, 972, 
	972, 972, 972, 972, 972, 972, 967, 972, 
	972, 972, 972, 972, 978, 972, 972, 972, 
	972, 972, 972, 972, 972, 972, 972, 972, 
	972, 972, 972, 972, 988, 988, 988, 988, 
	988, 988, 988, 988, 988, 988, 972, 972, 
	972, 972, 972, 972, 972, 988, 988, 988, 
	988, 988, 988, 972, 972, 972, 972, 972, 
	972, 972, 972, 972, 972, 972, 972, 972, 
	972, 972, 972, 972, 972, 972, 972, 972, 
	972, 972, 972, 972, 972, 988, 988, 988, 
	988, 988, 988, 972, 972, 972, 972, 972, 
	972, 972, 972, 972, 972, 972, 972, 972, 
	972, 972, 972, 972, 972, 972, 972, 972, 
	972, 980, 972, 967, 975, 975, 975, 967, 
	975, 975, 975, 975, 978, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 967, 975, 975, 
	975, 975, 975, 978, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 989, 989, 989, 989, 989, 
	989, 989, 989, 989, 989, 975, 975, 975, 
	975, 975, 975, 975, 989, 989, 989, 989, 
	989, 989, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 989, 989, 989, 989, 
	989, 989, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	980, 975, 967, 975, 975, 975, 967, 975, 
	975, 975, 975, 978, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 967, 975, 975, 975, 
	975, 975, 978, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 990, 990, 990, 990, 990, 990, 
	990, 990, 990, 990, 975, 975, 975, 975, 
	975, 975, 975, 990, 990, 990, 990, 990, 
	990, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 990, 990, 990, 990, 990, 
	990, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 980, 
	975, 967, 975, 975, 975, 967, 975, 975, 
	975, 975, 978, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 967, 975, 975, 975, 975, 
	975, 978, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 991, 991, 991, 991, 991, 991, 991, 
	991, 991, 991, 975, 975, 975, 975, 975, 
	975, 975, 991, 991, 991, 991, 991, 991, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 991, 991, 991, 991, 991, 991, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 980, 975, 
	967, 975, 975, 975, 967, 975, 975, 975, 
	975, 978, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 967, 975, 975, 975, 975, 975, 
	978, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	987, 987, 987, 987, 987, 987, 987, 987, 
	987, 987, 975, 975, 975, 975, 975, 975, 
	975, 987, 987, 987, 987, 987, 987, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 987, 987, 987, 987, 987, 987, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 975, 975, 975, 
	975, 975, 975, 975, 975, 980, 975, 992, 
	993, 995, 995, 995, 995, 995, 995, 995, 
	995, 995, 995, 994, 994, 994, 994, 994, 
	994, 994, 995, 995, 995, 995, 995, 995, 
	994, 994, 994, 994, 994, 994, 994, 994, 
	994, 994, 994, 994, 994, 994, 994, 994, 
	994, 994, 994, 994, 994, 994, 994, 994, 
	994, 994, 995, 995, 995, 995, 995, 995, 
	994, 994, 185, 185, 185, 185, 185, 185, 
	185, 185, 185, 185, 185, 185, 185, 185, 
	185, 185, 185, 185, 185, 185, 185, 185, 
	185, 185, 185, 185, 185, 185, 185, 185, 
	185, 185, 185, 241, 185, 185, 185, 185, 
	185, 185, 185, 185, 185, 185, 185, 185, 
	185, 185, 240, 240, 240, 240, 240, 240, 
	240, 240, 240, 240, 242, 185, 185, 243, 
	185, 241, 185, 240, 240, 240, 240, 240, 
	240, 240, 240, 240, 240, 240, 240, 240, 
	240, 240, 240, 240, 240, 240, 240, 240, 
	240, 240, 240, 240, 240, 185, 185, 185, 
	185, 240, 185, 240, 240, 240, 240, 240, 
	240, 240, 240, 240, 240, 240, 240, 240, 
	240, 240, 240, 240, 240, 240, 240, 240, 
	240, 240, 240, 240, 240, 185, 185, 185, 
	185, 185, 240, 860, 860, 860, 860, 860, 
	860, 860, 860, 860, 860, 860, 860, 860, 
	860, 860, 860, 860, 860, 860, 860, 860, 
	860, 860, 860, 860, 860, 860, 860, 860, 
	860, 860, 860, 860, 241, 860, 860, 860, 
	860, 860, 860, 860, 860, 860, 860, 860, 
	860, 860, 860, 240, 240, 240, 240, 240, 
	240, 240, 240, 240, 240, 242, 860, 860, 
	243, 860, 241, 860, 240, 240, 240, 240, 
	996, 240, 240, 240, 240, 240, 240, 240, 
	240, 240, 240, 240, 240, 240, 240, 240, 
	240, 240, 240, 240, 240, 240, 860, 860, 
	860, 860, 240, 860, 240, 240, 240, 240, 
	240, 240, 240, 240, 240, 240, 240, 240, 
	240, 240, 240, 240, 240, 240, 240, 240, 
	240, 240, 240, 240, 240, 240, 860, 860, 
	860, 860, 860, 240, 860, 860, 860, 860, 
	860, 860, 860, 860, 860, 860, 860, 860, 
	860, 860, 860, 860, 860, 860, 860, 860, 
	860, 860, 860, 860, 860, 860, 860, 860, 
	860, 860, 860, 860, 860, 241, 860, 860, 
	860, 860, 860, 860, 860, 860, 860, 860, 
	860, 860, 860, 860, 240, 240, 240, 240, 
	240, 240, 240, 240, 240, 240, 242, 860, 
	860, 243, 860, 241, 860, 240, 240, 240, 
	240, 240, 240, 240, 240, 240, 240, 240, 
	240, 240, 997, 240, 240, 240, 240, 240, 
	240, 240, 240, 240, 240, 240, 240, 860, 
	860, 860, 860, 240, 860, 240, 240, 240, 
	240, 240, 240, 240, 240, 240, 240, 240, 
	240, 240, 240, 240, 240, 240, 240, 240, 
	240, 240, 240, 240, 240, 240, 240, 860, 
	860, 860, 860, 860, 240, 998, 999, 860, 
	855, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 241, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 242, 1000, 1000, 243, 1000, 
	241, 1000, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 1000, 1000, 1000, 1000, 
	1001, 1000, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 1000, 1000, 1000, 1000, 
	1000, 797, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 241, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 242, 1000, 1000, 243, 
	1000, 241, 1000, 797, 797, 797, 797, 1002, 
	1003, 797, 797, 797, 797, 797, 1004, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 1000, 1000, 1000, 
	1000, 797, 1000, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 1000, 1000, 1000, 
	1000, 1000, 797, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 241, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 242, 1000, 1000, 
	243, 1000, 241, 1000, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 1005, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 1000, 1000, 
	1000, 1000, 797, 1000, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 1000, 1000, 
	1000, 1000, 1000, 797, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 241, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 242, 1000, 
	1000, 243, 1000, 241, 1000, 797, 797, 1006, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 1000, 
	1000, 1000, 1000, 797, 1000, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 1000, 
	1000, 1000, 1000, 1000, 797, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 241, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 242, 
	1000, 1000, 243, 1000, 241, 1000, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 1007, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	1000, 1000, 1000, 1000, 797, 1000, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	1000, 1000, 1000, 1000, 1000, 797, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 241, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	242, 1000, 1000, 243, 1000, 241, 1000, 797, 
	797, 797, 1008, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 1000, 1000, 1000, 1000, 797, 1000, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 1000, 1000, 1000, 1000, 1000, 797, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	241, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 242, 1000, 1000, 243, 1000, 241, 1000, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	1009, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 1000, 1000, 1000, 1000, 797, 1000, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 1000, 1000, 1000, 1000, 1000, 797, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 241, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 242, 1000, 1000, 243, 1000, 241, 
	1000, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 1010, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 1000, 1000, 1000, 1000, 797, 
	1000, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 1000, 1000, 1000, 1000, 1000, 
	797, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 241, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 242, 1000, 1000, 243, 1000, 
	241, 1000, 797, 797, 797, 797, 797, 797, 
	1011, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 1000, 1000, 1000, 1000, 
	797, 1000, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 1000, 1000, 1000, 1000, 
	1000, 797, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 241, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 242, 1000, 1000, 243, 
	1000, 241, 1000, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 1000, 1000, 1000, 
	1000, 1012, 1000, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 1000, 1000, 1000, 
	1000, 1000, 797, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 241, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 242, 1000, 1000, 
	243, 1000, 241, 1000, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 1000, 1000, 
	1000, 1000, 1013, 1000, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 1000, 1000, 
	1000, 1000, 1000, 797, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 241, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 242, 1000, 
	1000, 243, 1000, 241, 1000, 797, 797, 797, 
	797, 797, 797, 797, 797, 1014, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 1000, 
	1000, 1000, 1000, 797, 1000, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 1000, 
	1000, 1000, 1000, 1000, 797, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 241, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 242, 
	1000, 1000, 243, 1000, 241, 1000, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 1015, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	1000, 1000, 1000, 1000, 797, 1000, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	1000, 1000, 1000, 1000, 1000, 797, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 241, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	242, 1000, 1000, 243, 1000, 241, 1000, 797, 
	797, 797, 797, 1011, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 1000, 1000, 1000, 1000, 797, 1000, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 1000, 1000, 1000, 1000, 1000, 797, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	241, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 242, 1000, 1000, 243, 1000, 241, 1000, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	1016, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 1000, 1000, 1000, 1000, 797, 1000, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 1000, 1000, 1000, 1000, 1000, 797, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 241, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 242, 1000, 1000, 243, 1000, 241, 
	1000, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 1015, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 1000, 1000, 1000, 1000, 797, 
	1000, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 1000, 1000, 1000, 1000, 1000, 
	797, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 241, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 242, 1000, 1000, 243, 1000, 
	241, 1000, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 1000, 1000, 1000, 1000, 
	797, 1000, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 1017, 797, 1018, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 1000, 1000, 1000, 1000, 
	1000, 797, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 241, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 242, 1000, 1000, 243, 
	1000, 241, 1000, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 1000, 1000, 1000, 
	1000, 797, 1000, 797, 797, 797, 797, 797, 
	797, 797, 797, 1019, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 1000, 1000, 1000, 
	1000, 1000, 797, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 241, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 242, 1000, 1000, 
	243, 1000, 241, 1000, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 1000, 1000, 
	1000, 1000, 797, 1000, 1020, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 1000, 1000, 
	1000, 1000, 1000, 797, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 241, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 242, 1000, 
	1000, 243, 1000, 241, 1000, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 1000, 
	1000, 1000, 1000, 797, 1000, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 1013, 
	797, 797, 797, 797, 797, 797, 797, 1000, 
	1000, 1000, 1000, 1000, 797, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 241, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 242, 
	1000, 1000, 243, 1000, 241, 1000, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	1000, 1000, 1000, 1000, 797, 1000, 797, 797, 
	797, 1013, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	1000, 1000, 1000, 1000, 1000, 797, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 241, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	242, 1000, 1000, 243, 1000, 241, 1000, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 1000, 1000, 1000, 1000, 797, 1000, 797, 
	797, 797, 797, 1021, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	1022, 797, 797, 797, 797, 797, 797, 797, 
	797, 1000, 1000, 1000, 1000, 1000, 797, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	241, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 242, 1000, 1000, 243, 1000, 241, 1000, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 1000, 1000, 1000, 1000, 797, 1000, 
	797, 797, 797, 797, 797, 797, 1023, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 1000, 1000, 1000, 1000, 1000, 797, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 241, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 242, 1000, 1000, 243, 1000, 241, 
	1000, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 1000, 1000, 1000, 1000, 797, 
	1000, 797, 797, 797, 797, 797, 797, 797, 
	797, 1024, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 1000, 1000, 1000, 1000, 1000, 
	797, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 241, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 242, 1000, 1000, 243, 1000, 
	241, 1000, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 1000, 1000, 1000, 1000, 
	797, 1000, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 1013, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 1000, 1000, 1000, 1000, 
	1000, 797, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 241, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 242, 1000, 1000, 243, 
	1000, 241, 1000, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 1000, 1000, 1000, 
	1000, 797, 1000, 797, 797, 797, 797, 1025, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 1000, 1000, 1000, 
	1000, 1000, 797, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 241, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 242, 1000, 1000, 
	243, 1000, 241, 1000, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 1000, 1000, 
	1000, 1000, 797, 1000, 1026, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 1000, 1000, 
	1000, 1000, 1000, 797, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 241, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 242, 1000, 
	1000, 243, 1000, 241, 1000, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 1000, 
	1000, 1000, 1000, 797, 1000, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 1013, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 1000, 
	1000, 1000, 1000, 1000, 797, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 241, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 242, 
	1000, 1000, 243, 1000, 241, 1000, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	1000, 1000, 1000, 1000, 797, 1000, 1027, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 1028, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	1000, 1000, 1000, 1000, 1000, 797, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 241, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	242, 1000, 1000, 243, 1000, 241, 1000, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 1000, 1000, 1000, 1000, 797, 1000, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 1029, 797, 797, 797, 797, 797, 797, 
	797, 1000, 1000, 1000, 1000, 1000, 797, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	241, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 242, 1000, 1000, 243, 1000, 241, 1000, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 1000, 1000, 1000, 1000, 797, 1000, 
	797, 797, 797, 797, 1013, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 1000, 1000, 1000, 1000, 1000, 797, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 241, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 242, 1000, 1000, 243, 1000, 241, 
	1000, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 1000, 1000, 1000, 1000, 797, 
	1000, 1030, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 1000, 1000, 1000, 1000, 1000, 
	797, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 241, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 242, 1000, 1000, 243, 1000, 
	241, 1000, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 1000, 1000, 1000, 1000, 
	797, 1000, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 1020, 797, 797, 797, 
	797, 797, 797, 797, 1000, 1000, 1000, 1000, 
	1000, 797, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 241, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 242, 1000, 1000, 243, 
	1000, 241, 1000, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 1000, 1000, 1000, 
	1000, 797, 1000, 797, 797, 797, 797, 1031, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 1013, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 1000, 1000, 1000, 
	1000, 1000, 797, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 241, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 242, 1000, 1000, 
	243, 1000, 241, 1000, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 1000, 1000, 
	1000, 1000, 797, 1000, 797, 797, 797, 797, 
	797, 1013, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 1000, 1000, 
	1000, 1000, 1000, 797, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 241, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 242, 1000, 
	1000, 243, 1000, 241, 1000, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 1000, 
	1000, 1000, 1000, 797, 1000, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	1032, 797, 1033, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 1000, 
	1000, 1000, 1000, 1000, 797, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 241, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 242, 
	1000, 1000, 243, 1000, 241, 1000, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	1000, 1000, 1000, 1000, 797, 1000, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	1034, 797, 797, 797, 797, 797, 797, 797, 
	1000, 1000, 1000, 1000, 1000, 797, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 241, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	242, 1000, 1000, 243, 1000, 241, 1000, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 1000, 1000, 1000, 1000, 797, 1000, 797, 
	797, 797, 797, 1013, 797, 797, 797, 1031, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 1000, 1000, 1000, 1000, 1000, 797, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	241, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 242, 1000, 1000, 243, 1000, 241, 1000, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 1000, 1000, 1000, 1000, 797, 1000, 
	797, 797, 797, 1013, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 1035, 797, 797, 797, 797, 797, 
	797, 797, 1000, 1000, 1000, 1000, 1000, 797, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 241, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 242, 1000, 1000, 243, 1000, 241, 
	1000, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 1000, 1000, 1000, 1000, 797, 
	1000, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 1036, 797, 797, 
	797, 797, 797, 1000, 1000, 1000, 1000, 1000, 
	797, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 241, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 242, 1000, 1000, 243, 1000, 
	241, 1000, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 1000, 1000, 1000, 1000, 
	797, 1000, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 1029, 797, 797, 797, 797, 
	797, 797, 797, 797, 1000, 1000, 1000, 1000, 
	1000, 797, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 241, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 242, 1000, 1000, 243, 
	1000, 241, 1000, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 1000, 1000, 1000, 
	1000, 797, 1000, 1037, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 832, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 1000, 1000, 1000, 
	1000, 1000, 797, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 241, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 242, 1000, 1000, 
	243, 1000, 241, 1000, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 1000, 1000, 
	1000, 1000, 797, 1000, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 1027, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 1000, 1000, 
	1000, 1000, 1000, 797, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 241, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 242, 1000, 
	1000, 243, 1000, 241, 1000, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 1000, 
	1000, 1000, 1000, 797, 1000, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 1013, 797, 
	797, 797, 797, 797, 797, 797, 797, 1000, 
	1000, 1000, 1000, 1000, 797, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 241, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 242, 
	1000, 1000, 243, 1000, 241, 1000, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	1000, 1000, 1000, 1000, 797, 1000, 797, 797, 
	797, 797, 797, 1038, 797, 797, 797, 797, 
	797, 797, 797, 1013, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	1000, 1000, 1000, 1000, 1000, 797, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 241, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	242, 1000, 1000, 243, 1000, 241, 1000, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 1000, 1000, 1000, 1000, 797, 1000, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 1039, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 1000, 1000, 1000, 1000, 1000, 797, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	241, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 242, 1000, 1000, 243, 1000, 241, 1000, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 1000, 1000, 1000, 1000, 797, 1000, 
	797, 797, 797, 1040, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 1000, 1000, 1000, 1000, 1000, 797, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 241, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 242, 1000, 1000, 243, 1000, 241, 
	1000, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 1000, 1000, 1000, 1000, 797, 
	1000, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 1041, 797, 797, 
	797, 797, 797, 1000, 1000, 1000, 1000, 1000, 
	797, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 241, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 242, 1000, 1000, 243, 1000, 
	241, 1000, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 1000, 1000, 1000, 1000, 
	797, 1000, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 1029, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 1000, 1000, 1000, 1000, 
	1000, 797, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 241, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 242, 1000, 1000, 243, 
	1000, 241, 1000, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 1000, 1000, 1000, 
	1000, 797, 1000, 797, 797, 797, 797, 1042, 
	797, 797, 797, 1043, 797, 797, 797, 797, 
	797, 1044, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 1000, 1000, 1000, 
	1000, 1000, 797, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 241, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 242, 1000, 1000, 
	243, 1000, 241, 1000, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 1000, 1000, 
	1000, 1000, 797, 1000, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 1044, 797, 797, 1000, 1000, 
	1000, 1000, 1000, 797, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 241, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 242, 1000, 
	1000, 243, 1000, 241, 1000, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 1000, 
	1000, 1000, 1000, 797, 1000, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	1013, 797, 797, 797, 797, 797, 797, 1000, 
	1000, 1000, 1000, 1000, 797, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 241, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 242, 
	1000, 1000, 243, 1000, 241, 1000, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	1000, 1000, 1000, 1000, 797, 1000, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 1013, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	1000, 1000, 1000, 1000, 1000, 797, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 241, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	242, 1000, 1000, 243, 1000, 241, 1000, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 1000, 1000, 1000, 1000, 797, 1000, 797, 
	797, 797, 797, 1045, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 1000, 1000, 1000, 1000, 1000, 797, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	241, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 242, 1000, 1000, 243, 1000, 241, 1000, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 1000, 1000, 1000, 1000, 797, 1000, 
	797, 797, 797, 1046, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 1047, 1048, 797, 797, 797, 797, 
	797, 797, 1000, 1000, 1000, 1000, 1000, 797, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 241, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 242, 1000, 1000, 243, 1000, 241, 
	1000, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 1000, 1000, 1000, 1000, 797, 
	1000, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 1013, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 1000, 1000, 1000, 1000, 1000, 
	797, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 241, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 242, 1000, 1000, 243, 1000, 
	241, 1000, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 1000, 1000, 1000, 1000, 
	797, 1000, 797, 797, 1049, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 1000, 1000, 1000, 1000, 
	1000, 797, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 241, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 242, 1000, 1000, 243, 
	1000, 241, 1000, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 1000, 1000, 1000, 
	1000, 797, 1000, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 1050, 
	797, 797, 797, 797, 797, 1000, 1000, 1000, 
	1000, 1000, 797, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 241, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 242, 1000, 1000, 
	243, 1000, 241, 1000, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 1000, 1000, 
	1000, 1000, 797, 1000, 797, 797, 797, 797, 
	1051, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 1000, 1000, 
	1000, 1000, 1000, 797, 1052, 1052, 1052, 1052, 
	1052, 1052, 1052, 1052, 1052, 1052, 1052, 1052, 
	1052, 1052, 1052, 1052, 1052, 1052, 1052, 1052, 
	1052, 1052, 1052, 1052, 1052, 1052, 1052, 1052, 
	1052, 1052, 1052, 1052, 1052, 1054, 1052, 1052, 
	1052, 1052, 1052, 1052, 1052, 1052, 1052, 1052, 
	1052, 1052, 1052, 1052, 1053, 1053, 1053, 1053, 
	1053, 1053, 1053, 1053, 1053, 1053, 1055, 1052, 
	1052, 1056, 1052, 1054, 1052, 1053, 1053, 1053, 
	1053, 1053, 1053, 1053, 1053, 1053, 1053, 1053, 
	1053, 1053, 1053, 1053, 1053, 1053, 1053, 1053, 
	1053, 1053, 1053, 1053, 1053, 1053, 1053, 1052, 
	1052, 1052, 1052, 1053, 1052, 1053, 1053, 1053, 
	1053, 1053, 1053, 1053, 1053, 1053, 1053, 1053, 
	1053, 1053, 1053, 1053, 1053, 1053, 1053, 1053, 
	1053, 1053, 1053, 1053, 1053, 1053, 1053, 1052, 
	1052, 1052, 1052, 1052, 1053, 846, 1057, 845, 
	845, 845, 845, 845, 845, 845, 845, 845, 
	845, 845, 845, 845, 845, 845, 845, 845, 
	845, 845, 845, 845, 845, 845, 845, 845, 
	845, 845, 845, 845, 845, 845, 845, 845, 
	845, 845, 845, 845, 845, 845, 845, 845, 
	845, 845, 845, 845, 845, 845, 845, 845, 
	845, 845, 845, 845, 845, 845, 845, 845, 
	845, 845, 845, 845, 845, 845, 848, 845, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 241, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 242, 1000, 1000, 243, 1000, 241, 
	1000, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 1000, 1000, 1000, 1000, 797, 
	1000, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 1058, 797, 797, 1059, 797, 797, 
	797, 797, 797, 1000, 1000, 1000, 1000, 1000, 
	797, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 241, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 242, 1000, 1000, 243, 1000, 
	241, 1000, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 1000, 1000, 1000, 1000, 
	797, 1000, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 1013, 797, 1000, 1000, 1000, 1000, 
	1000, 797, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 241, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 242, 1000, 1000, 243, 
	1000, 241, 1000, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 1000, 1000, 1000, 
	1000, 797, 1000, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 1024, 797, 797, 797, 
	797, 797, 797, 797, 797, 1000, 1000, 1000, 
	1000, 1000, 797, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 241, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 242, 1000, 1000, 
	243, 1000, 241, 1000, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 1000, 1000, 
	1000, 1000, 797, 1000, 797, 797, 797, 797, 
	1060, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	1061, 797, 797, 797, 797, 797, 1000, 1000, 
	1000, 1000, 1000, 797, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 241, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 242, 1000, 
	1000, 243, 1000, 241, 1000, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 1000, 
	1000, 1000, 1000, 797, 1000, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	1031, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 1000, 
	1000, 1000, 1000, 1000, 797, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 241, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 242, 
	1000, 1000, 243, 1000, 241, 1000, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	1000, 1000, 1000, 1000, 797, 1000, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 1062, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	1000, 1000, 1000, 1000, 1000, 797, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 241, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	242, 1000, 1000, 243, 1000, 241, 1000, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 1000, 1000, 1000, 1000, 797, 1000, 797, 
	797, 797, 797, 832, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 1000, 1000, 1000, 1000, 1000, 797, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	241, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 242, 1000, 1000, 243, 1000, 241, 1000, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 1000, 1000, 1000, 1000, 797, 1000, 
	797, 797, 797, 797, 797, 797, 797, 1063, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 1064, 797, 797, 797, 797, 797, 797, 
	797, 797, 1000, 1000, 1000, 1000, 1000, 797, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 241, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 242, 1000, 1000, 243, 1000, 241, 
	1000, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 1000, 1000, 1000, 1000, 797, 
	1000, 797, 797, 797, 797, 1024, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 1000, 1000, 1000, 1000, 1000, 
	797, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 241, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 242, 1000, 1000, 243, 1000, 
	241, 1000, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 1000, 1000, 1000, 1000, 
	797, 1000, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 1029, 797, 
	797, 797, 797, 797, 1000, 1000, 1000, 1000, 
	1000, 797, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 241, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 242, 1000, 1000, 243, 
	1000, 241, 1000, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 1000, 1000, 1000, 
	1000, 797, 1000, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	1065, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 1000, 1000, 1000, 
	1000, 1000, 797, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 241, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 242, 1000, 1000, 
	243, 1000, 241, 1000, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 1000, 1000, 
	1000, 1000, 797, 1000, 797, 797, 797, 1066, 
	797, 797, 797, 797, 797, 797, 797, 1067, 
	797, 797, 797, 797, 797, 797, 797, 1068, 
	797, 797, 797, 797, 797, 797, 1000, 1000, 
	1000, 1000, 1000, 797, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 241, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 242, 1000, 
	1000, 243, 1000, 241, 1000, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 1000, 
	1000, 1000, 1000, 797, 1000, 797, 797, 797, 
	797, 1031, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 1000, 
	1000, 1000, 1000, 1000, 797, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 241, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 242, 
	1000, 1000, 243, 1000, 241, 1000, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	1000, 1000, 1000, 1000, 797, 1000, 797, 797, 
	797, 797, 1069, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	1000, 1000, 1000, 1000, 1000, 797, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 241, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	242, 1000, 1000, 243, 1000, 241, 1000, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 1000, 1000, 1000, 1000, 797, 1000, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 1070, 797, 797, 797, 797, 797, 797, 
	797, 1000, 1000, 1000, 1000, 1000, 797, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	241, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 242, 1000, 1000, 243, 1000, 241, 1000, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 1000, 1000, 1000, 1000, 797, 1000, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 1038, 797, 797, 797, 797, 797, 
	797, 797, 1000, 1000, 1000, 1000, 1000, 797, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 241, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 242, 1000, 1000, 243, 1000, 241, 
	1000, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 1000, 1000, 1000, 1000, 797, 
	1000, 797, 797, 797, 797, 797, 797, 797, 
	797, 1071, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 1000, 1000, 1000, 1000, 1000, 
	797, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 241, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 242, 1000, 1000, 243, 1000, 
	241, 1000, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 1000, 1000, 1000, 1000, 
	797, 1000, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 1038, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 1000, 1000, 1000, 1000, 
	1000, 797, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 241, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 242, 1000, 1000, 243, 
	1000, 241, 1000, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 1000, 1000, 1000, 
	1000, 797, 1000, 797, 797, 797, 797, 797, 
	797, 797, 1072, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 1000, 1000, 1000, 
	1000, 1000, 797, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 241, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 242, 1000, 1000, 
	243, 1000, 241, 1000, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 1000, 1000, 
	1000, 1000, 797, 1000, 797, 797, 797, 797, 
	1024, 797, 797, 797, 1073, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 1000, 1000, 
	1000, 1000, 1000, 797, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 241, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 242, 1000, 
	1000, 243, 1000, 241, 1000, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 1000, 
	1000, 1000, 1000, 797, 1000, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	1074, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 1000, 
	1000, 1000, 1000, 1000, 797, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 241, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 242, 
	1000, 1000, 243, 1000, 241, 1000, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	1000, 1000, 1000, 1000, 797, 1000, 797, 797, 
	797, 797, 1038, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	1000, 1000, 1000, 1000, 1000, 797, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 241, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	242, 1000, 1000, 243, 1000, 241, 1000, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 1000, 1000, 1000, 1000, 797, 1000, 797, 
	797, 797, 797, 797, 797, 797, 797, 1075, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 1000, 1000, 1000, 1000, 1000, 797, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	241, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 242, 1000, 1000, 243, 1000, 241, 1000, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 1000, 1000, 1000, 1000, 797, 1000, 
	797, 797, 797, 797, 1076, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 1000, 1000, 1000, 1000, 1000, 797, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 241, 1000, 1000, 1000, 1000, 1000, 1000, 
	1000, 1000, 1000, 1000, 1000, 1000, 1000, 1000, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 242, 1000, 1000, 243, 1000, 241, 
	1000, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 1000, 1000, 1000, 1000, 797, 
	1000, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 797, 1018, 797, 797, 797, 
	797, 797, 797, 797, 797, 797, 797, 797, 
	797, 797, 797, 1000, 1000, 1000, 1000, 1000, 
	797, 1077, 196, 860, 860, 860, 860, 860, 
	860, 860, 860, 860, 860, 860, 860, 860, 
	860, 860, 860, 860, 860, 860, 860, 860, 
	860, 860, 860, 860, 860, 860, 860, 860, 
	860, 860, 860, 860, 860, 860, 860, 860, 
	860, 860, 860, 860, 860, 860, 860, 860, 
	860, 860, 860, 860, 860, 860, 860, 860, 
	860, 860, 860, 860, 860, 860, 860, 860, 
	860, 870, 860, 1079, 1078, 1078, 1078, 1079, 
	1078, 1078, 1078, 1078, 1080, 1081, 1080, 1080, 
	1080, 1078, 1078, 1078, 1078, 1078, 1078, 1078, 
	1078, 1078, 1078, 1078, 1078, 1079, 1078, 1078, 
	1078, 1078, 1078, 1080, 1078, 1078, 1082, 1078, 
	1078, 1078, 1078, 1078, 1078, 1078, 1078, 1078, 
	1078, 1078, 1078, 1078, 1078, 1078, 1078, 1078, 
	1078, 1078, 1078, 1078, 1078, 1078, 1078, 1078, 
	1078, 1078, 1078, 1078, 1078, 1078, 1078, 1078, 
	1078, 1078, 1078, 1078, 1078, 1078, 1078, 1078, 
	1078, 1078, 1078, 1078, 1078, 1078, 1078, 1078, 
	1078, 1078, 1078, 1078, 1078, 1078, 1078, 1083, 
	1078, 1080, 1084, 1080, 1080, 1080, 1084, 1084, 
	1084, 1084, 1084, 1084, 1084, 1084, 1084, 1084, 
	1084, 1084, 1084, 1084, 1084, 1084, 1084, 1084, 
	1080, 1084, 1085, 1086, 1087, 1088, 1089, 1091, 
	1090, 1092, 1094, 1095, 1095, 1095, 1094, 1095, 
	1095, 1095, 1095, 1096, 1097, 1096, 1096, 1096, 
	1095, 1095, 1095, 1095, 1095, 1095, 1095, 1095, 
	1095, 1095, 1095, 1095, 1094, 1095, 1095, 1095, 
	1095, 1095, 1096, 1095, 1098, 1099, 1095, 1095, 
	1095, 1098, 1095, 1095, 1095, 1095, 1095, 1095, 
	1095, 1095, 1095, 1095, 1095, 1095, 1095, 1095, 
	1095, 1095, 1095, 1095, 1095, 1095, 1095, 1095, 
	1095, 1095, 1095, 1093, 1093, 1093, 1093, 1093, 
	1093, 1093, 1093, 1093, 1093, 1093, 1093, 1093, 
	1093, 1093, 1093, 1093, 1093, 1093, 1093, 1093, 
	1093, 1093, 1093, 1093, 1093, 1095, 1100, 1095, 
	1095, 1093, 1095, 1093, 1093, 1093, 1093, 1093, 
	1093, 1093, 1093, 1093, 1093, 1093, 1093, 1093, 
	1093, 1093, 1093, 1093, 1093, 1093, 1093, 1093, 
	1093, 1093, 1093, 1093, 1093, 1095, 1095, 1095, 
	1095, 1095, 1093, 1101, 1101, 1101, 1101, 1101, 
	1101, 1101, 1101, 1101, 1101, 1101, 1101, 1101, 
	1101, 1101, 1101, 1101, 1101, 1101, 1101, 1101, 
	1101, 1101, 1101, 1101, 1101, 1101, 1101, 1101, 
	1101, 1101, 1101, 1101, 249, 1101, 1101, 1101, 
	1101, 1101, 1101, 1101, 1101, 1101, 1101, 1101, 
	1101, 1101, 1101, 248, 248, 248, 248, 248, 
	248, 248, 248, 248, 248, 250, 1101, 1101, 
	1101, 1101, 249, 1101, 248, 248, 248, 248, 
	248, 248, 248, 248, 248, 248, 248, 248, 
	248, 248, 248, 248, 248, 248, 248, 248, 
	248, 248, 248, 248, 248, 248, 1101, 1101, 
	1101, 1101, 248, 1101, 248, 248, 248, 248, 
	248, 248, 248, 248, 248, 248, 248, 248, 
	248, 248, 248, 248, 248, 248, 248, 248, 
	248, 248, 248, 248, 248, 248, 1101, 1101, 
	1101, 1101, 1101, 248, 1096, 1102, 1096, 1096, 
	1096, 1102, 1102, 1102, 1102, 1102, 1102, 1102, 
	1102, 1102, 1102, 1102, 1102, 1102, 1102, 1102, 
	1102, 1102, 1102, 1096, 1102, 1103, 1104, 1105, 
	1106, 1107, 1108, 1101, 1109, 1111, 1112, 1112, 
	1112, 1111, 1112, 1112, 1112, 1112, 1113, 1114, 
	1113, 1113, 1113, 1112, 1112, 1112, 1112, 1112, 
	1112, 1112, 1112, 1112, 1112, 1112, 1112, 1111, 
	1112, 1112, 1112, 1112, 1112, 1113, 1115, 1116, 
	1117, 1118, 1119, 1120, 1116, 1121, 1122, 1123, 
	1119, 1124, 1125, 1126, 1119, 1127, 1128, 1128, 
	1128, 1128, 1128, 1128, 1128, 1128, 1128, 1129, 
	1130, 1131, 1132, 1133, 1134, 1135, 1136, 1137, 
	1136, 1136, 1138, 1136, 1136, 1136, 1136, 1136, 
	1136, 1136, 1136, 1136, 1136, 1136, 1136, 1136, 
	1136, 1136, 1136, 1136, 1136, 1136, 1136, 1136, 
	1139, 1140, 1141, 1119, 1142, 1116, 1143, 1144, 
	1145, 1146, 1147, 1148, 1110, 1110, 1149, 1110, 
	1110, 1110, 1150, 1151, 1152, 1110, 1110, 1153, 
	1154, 1155, 1156, 1110, 1157, 1110, 1158, 1110, 
	1159, 1160, 1161, 1119, 1112, 1110, 258, 258, 
	258, 258, 258, 258, 258, 258, 258, 258, 
	258, 258, 258, 258, 258, 258, 258, 258, 
	258, 258, 258, 258, 258, 258, 258, 258, 
	258, 258, 258, 258, 258, 258, 258, 1162, 
	258, 258, 258, 258, 258, 258, 258, 258, 
	258, 258, 258, 258, 258, 258, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	258, 258, 258, 258, 258, 1162, 258, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 258, 258, 258, 258, 1110, 258, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 258, 258, 258, 258, 258, 1110, 1164, 
	1163, 1165, 1113, 1166, 1113, 1113, 1113, 1166, 
	1166, 1166, 1166, 1166, 1166, 1166, 1166, 1166, 
	1166, 1166, 1166, 1166, 1166, 1166, 1166, 1166, 
	1166, 1113, 1166, 1167, 1169, 1168, 1168, 1168, 
	1168, 1168, 1168, 1168, 1168, 1168, 1168, 1168, 
	1168, 1168, 1168, 1168, 1168, 1168, 1168, 1168, 
	1168, 1168, 1168, 1168, 1168, 1168, 1168, 1168, 
	1168, 1168, 1168, 1168, 1168, 1168, 1168, 1168, 
	1168, 1168, 1168, 1168, 1168, 1168, 1168, 1168, 
	1168, 1168, 1168, 1168, 1168, 1168, 1168, 1168, 
	1168, 1168, 1168, 1168, 1168, 1168, 1168, 1168, 
	1168, 1168, 1168, 1168, 1168, 1169, 1168, 1170, 
	1171, 1172, 1173, 1174, 1174, 1174, 1174, 1174, 
	1174, 1174, 1174, 1174, 1174, 1174, 1174, 1174, 
	1174, 1174, 1174, 1174, 1174, 1174, 1174, 1174, 
	1174, 1174, 1174, 1174, 1174, 1174, 1174, 1174, 
	1174, 1174, 1174, 1174, 253, 253, 1174, 253, 
	1174, 253, 253, 1174, 1174, 253, 253, 253, 
	1175, 253, 253, 1176, 1176, 1176, 1176, 1176, 
	1176, 1176, 1176, 1176, 1176, 253, 253, 253, 
	253, 253, 253, 253, 261, 261, 261, 261, 
	261, 261, 261, 261, 261, 261, 261, 261, 
	261, 261, 261, 261, 261, 261, 261, 261, 
	261, 261, 261, 261, 261, 261, 1174, 253, 
	1174, 1174, 261, 253, 261, 261, 261, 261, 
	261, 261, 261, 261, 261, 261, 261, 261, 
	261, 261, 261, 261, 261, 261, 261, 261, 
	261, 261, 261, 261, 261, 261, 1174, 1174, 
	1174, 253, 1174, 261, 1177, 1177, 1177, 1177, 
	1177, 1177, 1177, 1177, 1177, 1177, 1177, 1177, 
	1177, 1177, 1177, 1177, 1177, 1177, 1177, 1177, 
	1177, 1177, 1177, 1177, 1177, 1177, 1177, 1177, 
	1177, 1177, 1177, 1177, 1177, 1177, 1177, 1177, 
	1177, 1177, 1177, 1177, 1177, 1177, 1177, 1177, 
	1177, 1177, 1177, 1177, 261, 261, 261, 261, 
	261, 261, 261, 261, 261, 261, 1177, 1177, 
	1177, 1177, 1177, 1177, 1177, 261, 261, 261, 
	261, 261, 261, 261, 261, 261, 261, 261, 
	261, 261, 261, 261, 261, 261, 261, 261, 
	261, 261, 261, 261, 261, 261, 261, 1177, 
	1177, 1177, 1177, 261, 1177, 261, 261, 261, 
	261, 261, 261, 261, 261, 261, 261, 261, 
	261, 261, 261, 261, 261, 261, 261, 261, 
	261, 261, 261, 261, 261, 261, 261, 1177, 
	1177, 1177, 1177, 1177, 261, 1176, 1176, 1176, 
	1176, 1176, 1176, 1176, 1176, 1176, 1176, 1177, 
	1178, 1168, 1119, 1168, 1168, 1168, 1168, 1168, 
	1168, 1168, 1179, 1168, 1168, 1168, 1168, 1168, 
	1168, 1168, 1168, 1168, 1168, 1168, 1168, 1168, 
	1168, 1178, 1168, 1180, 1181, 1119, 1168, 1168, 
	1168, 1168, 1168, 1168, 1168, 1168, 1168, 1168, 
	1168, 1168, 1168, 1168, 1168, 1168, 1168, 1168, 
	1178, 1168, 1178, 1182, 1168, 1184, 1183, 257, 
	257, 257, 257, 257, 257, 257, 257, 257, 
	257, 1183, 1124, 1185, 257, 257, 257, 257, 
	257, 257, 257, 257, 257, 257, 1186, 1186, 
	1186, 1186, 1186, 1186, 1186, 1186, 1186, 1186, 
	1186, 1187, 1186, 1186, 1186, 1186, 1186, 1186, 
	1186, 1186, 1186, 1186, 1186, 1186, 1186, 1186, 
	1186, 1186, 1186, 1186, 1186, 1186, 1186, 1186, 
	1186, 1186, 1186, 1188, 1186, 1186, 1186, 1186, 
	1186, 1187, 1186, 256, 256, 256, 256, 256, 
	256, 256, 256, 256, 256, 1186, 1186, 1186, 
	1186, 1186, 1186, 1186, 1186, 1186, 1186, 1186, 
	1186, 1186, 1186, 1186, 1186, 1186, 1186, 1186, 
	1186, 1186, 1186, 1186, 1186, 1186, 1186, 1186, 
	1186, 1186, 1186, 1186, 1186, 1186, 1186, 1186, 
	1186, 1186, 255, 1186, 1190, 1189, 1191, 1191, 
	1191, 1191, 1191, 1191, 1191, 1191, 1191, 1191, 
	1189, 1189, 1189, 1189, 1189, 1189, 1189, 1189, 
	1192, 1189, 1193, 1194, 1189, 1189, 1189, 1189, 
	1189, 1189, 1189, 1189, 1189, 1195, 1189, 1189, 
	1189, 1189, 1189, 1189, 1189, 1189, 1196, 1189, 
	1189, 1189, 1189, 1189, 1189, 1197, 1189, 1189, 
	1192, 1189, 1193, 1194, 1189, 1189, 1189, 1198, 
	1189, 1189, 1189, 1189, 1189, 1195, 1189, 1189, 
	1199, 1189, 1189, 1189, 1189, 1189, 1196, 1189, 
	259, 259, 259, 259, 259, 259, 259, 259, 
	259, 259, 1200, 1200, 1200, 1200, 1200, 1200, 
	1200, 1200, 1200, 1200, 1200, 1201, 1200, 1200, 
	1200, 1200, 1200, 1200, 1200, 1200, 1200, 1200, 
	1200, 1200, 1200, 1200, 1200, 1200, 1200, 1200, 
	1200, 1200, 1200, 1200, 1200, 1200, 1200, 1202, 
	1200, 1200, 1200, 1200, 1200, 1201, 1200, 1200, 
	1200, 1203, 1200, 1200, 1200, 1200, 1200, 1200, 
	1200, 1200, 1204, 1200, 1205, 258, 1205, 258, 
	258, 260, 260, 260, 260, 260, 260, 260, 
	260, 260, 260, 258, 260, 260, 260, 260, 
	260, 260, 260, 260, 260, 260, 1206, 1206, 
	1206, 1206, 1206, 1206, 1206, 1206, 1206, 1206, 
	1206, 1206, 1206, 1206, 1206, 1206, 1206, 1206, 
	1206, 1206, 1206, 1206, 1206, 1206, 1206, 1206, 
	1206, 1206, 1206, 1206, 1206, 1206, 1206, 1206, 
	1206, 1206, 1206, 1205, 1206, 1206, 1206, 1206, 
	1206, 1206, 1206, 1206, 1206, 1207, 1206, 1208, 
	1210, 1209, 1211, 1191, 1191, 1191, 1191, 1191, 
	1191, 1191, 1191, 1191, 1191, 1189, 1189, 1189, 
	1189, 1189, 1189, 1189, 1189, 1189, 1189, 1189, 
	1189, 1189, 1189, 1189, 1189, 1189, 1189, 1189, 
	1189, 1189, 1189, 1189, 1189, 1189, 1189, 1189, 
	1189, 1189, 1189, 1189, 1189, 1189, 1189, 1189, 
	1189, 1189, 1197, 1189, 1189, 1189, 1189, 1189, 
	1189, 1189, 1189, 1189, 1198, 1189, 1189, 1189, 
	1189, 1189, 1189, 1189, 1189, 1199, 1189, 1212, 
	1212, 1212, 1212, 1212, 1212, 1212, 1212, 1212, 
	1212, 1189, 1189, 1189, 1189, 1189, 1189, 1189, 
	1189, 1189, 1189, 1189, 1189, 1189, 1189, 1189, 
	1189, 1189, 1189, 1189, 1189, 1189, 1189, 1189, 
	1189, 1189, 1189, 1189, 1189, 1189, 1189, 1189, 
	1189, 1189, 1189, 1189, 1189, 1189, 1213, 1189, 
	1189, 1189, 1189, 1189, 1189, 1189, 1189, 1189, 
	1198, 1189, 1189, 1189, 1189, 1189, 1189, 1189, 
	1189, 1199, 1189, 1215, 1215, 1215, 1215, 1215, 
	1215, 1215, 1215, 1215, 1215, 1214, 1214, 1214, 
	1214, 1214, 1214, 1214, 1214, 1214, 1214, 1214, 
	1214, 1214, 1214, 1214, 1214, 1214, 1214, 1214, 
	1214, 1214, 1214, 1214, 1214, 1214, 1214, 1214, 
	1214, 1214, 1214, 1214, 1214, 1214, 1214, 1214, 
	1214, 1214, 1216, 1214, 1214, 1214, 1214, 1214, 
	1214, 1214, 1214, 1214, 1217, 1214, 1214, 1214, 
	1214, 1214, 1214, 1214, 1214, 1218, 1214, 1215, 
	1215, 1215, 1215, 1215, 1215, 1215, 1215, 1215, 
	1215, 1214, 1214, 1214, 1214, 1214, 1214, 1214, 
	1214, 1214, 1214, 1214, 1214, 1214, 1214, 1214, 
	1214, 1214, 1214, 1214, 1214, 1214, 1214, 1214, 
	1214, 1214, 1214, 1214, 1214, 1214, 1214, 1214, 
	1214, 1214, 1214, 1214, 1214, 1214, 1219, 1214, 
	1214, 1214, 1214, 1214, 1214, 1214, 1214, 1214, 
	1217, 1214, 1214, 1214, 1214, 1214, 1214, 1214, 
	1214, 1218, 1214, 1217, 1214, 1214, 1214, 1214, 
	1214, 1214, 1214, 1214, 1218, 1214, 1220, 1222, 
	1221, 1223, 1225, 1225, 1224, 1224, 1224, 1224, 
	1224, 1224, 1224, 1224, 1224, 1224, 1224, 1224, 
	1224, 1224, 1224, 1224, 1224, 1224, 1224, 1224, 
	1224, 1224, 1224, 1224, 1224, 1224, 1224, 1224, 
	1224, 1224, 1224, 1224, 1224, 1224, 1224, 1224, 
	1224, 1224, 1224, 1224, 1224, 1224, 1224, 1224, 
	1224, 1226, 1224, 1224, 1224, 1224, 1224, 1224, 
	1224, 1224, 1224, 1227, 1224, 1224, 1224, 1224, 
	1224, 1224, 1224, 1224, 1228, 1224, 1229, 1229, 
	1214, 1214, 1214, 1214, 1214, 1214, 1214, 1214, 
	1214, 1214, 1214, 1214, 1214, 1214, 1214, 1214, 
	1214, 1214, 1214, 1214, 1214, 1214, 1214, 1214, 
	1214, 1214, 1214, 1214, 1214, 1214, 1214, 1214, 
	1214, 1214, 1214, 1214, 1214, 1214, 1214, 1214, 
	1214, 1214, 1214, 1214, 1214, 1230, 1214, 1214, 
	1214, 1214, 1214, 1214, 1214, 1214, 1214, 1217, 
	1214, 1214, 1214, 1214, 1214, 1214, 1214, 1214, 
	1218, 1214, 1229, 1229, 1214, 1214, 1214, 1214, 
	1214, 1214, 1214, 1214, 1214, 1214, 1214, 1214, 
	1214, 1214, 1214, 1214, 1214, 1214, 1214, 1214, 
	1214, 1214, 1214, 1214, 1214, 1214, 1214, 1214, 
	1214, 1214, 1214, 1214, 1214, 1214, 1214, 1214, 
	1214, 1214, 1214, 1214, 1214, 1214, 1214, 1214, 
	1214, 1219, 1214, 1214, 1214, 1214, 1214, 1214, 
	1214, 1214, 1214, 1217, 1214, 1214, 1214, 1214, 
	1214, 1214, 1214, 1214, 1218, 1214, 1232, 1232, 
	1232, 1232, 1232, 1232, 1232, 1232, 1232, 1232, 
	1231, 1231, 1231, 1231, 1231, 1231, 1231, 1231, 
	1231, 1231, 1231, 1231, 1231, 1231, 1231, 1231, 
	1231, 1231, 1231, 1231, 1231, 1231, 1231, 1231, 
	1231, 1231, 1231, 1231, 1231, 1231, 1231, 1231, 
	1231, 1231, 1231, 1231, 1231, 1233, 1231, 1231, 
	1231, 1231, 1231, 1231, 1231, 1231, 1231, 1234, 
	1231, 1231, 1231, 1231, 1231, 1231, 1231, 1231, 
	1235, 1231, 1237, 1237, 1237, 1237, 1237, 1237, 
	1237, 1237, 1237, 1237, 1236, 1236, 1236, 1236, 
	1236, 1236, 1236, 1236, 1236, 1236, 1236, 1236, 
	1236, 1236, 1236, 1236, 1236, 1236, 1236, 1236, 
	1236, 1236, 1236, 1236, 1236, 1236, 1236, 1236, 
	1236, 1236, 1236, 1236, 1236, 1236, 1236, 1236, 
	1236, 1238, 1236, 1236, 1236, 1236, 1236, 1236, 
	1236, 1236, 1236, 1239, 1236, 1236, 1236, 1236, 
	1236, 1236, 1236, 1236, 1240, 1236, 1242, 1242, 
	1242, 1242, 1242, 1242, 1242, 1242, 1242, 1242, 
	1241, 1241, 1241, 1241, 1241, 1241, 1241, 1242, 
	1242, 1242, 1242, 1242, 1242, 1241, 1241, 1241, 
	1241, 1241, 1241, 1241, 1241, 1241, 1241, 1241, 
	1241, 1241, 1241, 1241, 1241, 1241, 1241, 1241, 
	1241, 1241, 1241, 1241, 1241, 1243, 1241, 1242, 
	1242, 1242, 1242, 1242, 1242, 1241, 1241, 1244, 
	1241, 1241, 1241, 1241, 1241, 1241, 1241, 1241, 
	1245, 1241, 1246, 1246, 1246, 1246, 1246, 1246, 
	1246, 1246, 1246, 1246, 1214, 1214, 1214, 1214, 
	1214, 1214, 1214, 1246, 1246, 1246, 1246, 1246, 
	1246, 1214, 1214, 1214, 1214, 1214, 1214, 1214, 
	1214, 1214, 1214, 1214, 1214, 1214, 1214, 1214, 
	1214, 1214, 1214, 1214, 1214, 1214, 1214, 1214, 
	1214, 1247, 1214, 1246, 1246, 1246, 1246, 1246, 
	1246, 1214, 1214, 1217, 1214, 1214, 1214, 1214, 
	1214, 1214, 1214, 1214, 1218, 1214, 1246, 1246, 
	1246, 1246, 1246, 1246, 1246, 1246, 1246, 1246, 
	1214, 1214, 1214, 1214, 1214, 1214, 1214, 1246, 
	1246, 1246, 1246, 1246, 1246, 1214, 1214, 1214, 
	1214, 1214, 1214, 1214, 1214, 1214, 1214, 1214, 
	1214, 1214, 1214, 1214, 1214, 1214, 1214, 1214, 
	1214, 1214, 1214, 1214, 1214, 1219, 1214, 1246, 
	1246, 1246, 1246, 1246, 1246, 1214, 1214, 1217, 
	1214, 1214, 1214, 1214, 1214, 1214, 1214, 1214, 
	1218, 1214, 1249, 1248, 1250, 1250, 1250, 1250, 
	1250, 1250, 1250, 1250, 1250, 1250, 1248, 1248, 
	1248, 1248, 1248, 1248, 1248, 1248, 1248, 1248, 
	1248, 1251, 1248, 1248, 1248, 1248, 1248, 1248, 
	1248, 1248, 1248, 1248, 1248, 1248, 1248, 1248, 
	1248, 1248, 1248, 1248, 1248, 1248, 1248, 1248, 
	1248, 1248, 1248, 1252, 1248, 1248, 1248, 1248, 
	1248, 1251, 1248, 1248, 1248, 1253, 1248, 1248, 
	1248, 1248, 1248, 1248, 1248, 1248, 1254, 1248, 
	1255, 1255, 1255, 1255, 1255, 1255, 1255, 1255, 
	1255, 1255, 1248, 1248, 1248, 1248, 1248, 1248, 
	1248, 1248, 1248, 1248, 1248, 1248, 1248, 1248, 
	1248, 1248, 1248, 1248, 1248, 1248, 1248, 1248, 
	1248, 1248, 1248, 1248, 1248, 1248, 1248, 1248, 
	1248, 1248, 1248, 1248, 1248, 1248, 1248, 1256, 
	1248, 1248, 1248, 1248, 1248, 1248, 1248, 1248, 
	1248, 1253, 1248, 1248, 1248, 1248, 1248, 1248, 
	1248, 1248, 1254, 1248, 1257, 1214, 1258, 1258, 
	1258, 1258, 1258, 1258, 1258, 1258, 1258, 1258, 
	1214, 1214, 1214, 1214, 1214, 1214, 1214, 1214, 
	1214, 1214, 1214, 1259, 1214, 1214, 1214, 1214, 
	1214, 1214, 1214, 1214, 1214, 1214, 1214, 1214, 
	1214, 1214, 1214, 1214, 1214, 1214, 1214, 1214, 
	1214, 1214, 1214, 1214, 1214, 1260, 1214, 1214, 
	1214, 1214, 1214, 1259, 1214, 1214, 1214, 1217, 
	1214, 1214, 1214, 1214, 1214, 1214, 1214, 1214, 
	1218, 1214, 1258, 1258, 1258, 1258, 1258, 1258, 
	1258, 1258, 1258, 1258, 1214, 1214, 1214, 1214, 
	1214, 1214, 1214, 1214, 1214, 1214, 1214, 1214, 
	1214, 1214, 1214, 1214, 1214, 1214, 1214, 1214, 
	1214, 1214, 1214, 1214, 1214, 1214, 1214, 1214, 
	1214, 1214, 1214, 1214, 1214, 1214, 1214, 1214, 
	1214, 1219, 1214, 1214, 1214, 1214, 1214, 1214, 
	1214, 1214, 1214, 1217, 1214, 1214, 1214, 1214, 
	1214, 1214, 1214, 1214, 1218, 1214, 1179, 1185, 
	1119, 1261, 1168, 1169, 1168, 1262, 1169, 1185, 
	1185, 1185, 1185, 1185, 1185, 1185, 1185, 1185, 
	1185, 1185, 1185, 1185, 1185, 1185, 1185, 1185, 
	1185, 1185, 1185, 1185, 1185, 1185, 1185, 1185, 
	1185, 1185, 1185, 1185, 1185, 1185, 1185, 1185, 
	1185, 1185, 1185, 1185, 1185, 1185, 1185, 1185, 
	1185, 1185, 1185, 1185, 1185, 1185, 1185, 1185, 
	1185, 1185, 1185, 1185, 1185, 1185, 1185, 1185, 
	1185, 1185, 1185, 1185, 1185, 1185, 1169, 1185, 
	1169, 1168, 1169, 1119, 1168, 1174, 1174, 1174, 
	1174, 1174, 1174, 1174, 1174, 1174, 1174, 1174, 
	1174, 1174, 1174, 1174, 1174, 1174, 1174, 1174, 
	1174, 1174, 1174, 1174, 1174, 1174, 1174, 1174, 
	1174, 1174, 1174, 1174, 1174, 1174, 1174, 1174, 
	1174, 1174, 1174, 1174, 1174, 1174, 1174, 1174, 
	1174, 1174, 1174, 1174, 1174, 261, 261, 261, 
	261, 261, 261, 261, 261, 261, 261, 1174, 
	1174, 1174, 1174, 1174, 1174, 1263, 261, 261, 
	261, 261, 261, 261, 261, 261, 261, 261, 
	261, 261, 261, 261, 261, 261, 261, 261, 
	261, 261, 261, 261, 261, 261, 261, 261, 
	1174, 1174, 1174, 1174, 261, 1174, 261, 261, 
	261, 261, 261, 261, 261, 261, 261, 261, 
	261, 261, 261, 261, 261, 261, 261, 261, 
	261, 261, 261, 261, 261, 261, 261, 261, 
	1174, 1174, 1174, 1174, 1174, 261, 258, 258, 
	258, 258, 258, 258, 258, 258, 258, 258, 
	258, 258, 258, 258, 258, 258, 258, 258, 
	258, 258, 258, 258, 258, 258, 258, 258, 
	258, 258, 258, 258, 258, 258, 258, 1162, 
	258, 258, 258, 258, 258, 258, 258, 258, 
	258, 258, 258, 258, 258, 258, 1136, 1136, 
	1136, 1136, 1136, 1136, 1136, 1136, 1136, 1136, 
	1264, 258, 258, 258, 258, 1162, 258, 1136, 
	1136, 1136, 1136, 1136, 1136, 1136, 1136, 1136, 
	1136, 1136, 1136, 1136, 1136, 1136, 1136, 1136, 
	1136, 1136, 1136, 1136, 1136, 1136, 1136, 1136, 
	1136, 258, 258, 258, 258, 1136, 258, 1136, 
	1136, 1136, 1136, 1136, 1136, 1136, 1136, 1136, 
	1136, 1136, 1136, 1136, 1136, 1136, 1136, 1136, 
	1136, 1136, 1136, 1136, 1136, 1136, 1136, 1136, 
	1136, 258, 258, 258, 258, 258, 1136, 1265, 
	1266, 1266, 1266, 1266, 1266, 1266, 1266, 1266, 
	1266, 1266, 1266, 1266, 1266, 1266, 1266, 1266, 
	1266, 1266, 1266, 1266, 1266, 1266, 1266, 1266, 
	1266, 1266, 1266, 1266, 1266, 1266, 1266, 1266, 
	1266, 1162, 1266, 1266, 1266, 1266, 1266, 1266, 
	1266, 1266, 1266, 1266, 1266, 1266, 1266, 1266, 
	1136, 1136, 1136, 1136, 1136, 1136, 1136, 1136, 
	1136, 1136, 1264, 1266, 1266, 1266, 1266, 1162, 
	1266, 1136, 1136, 1136, 1136, 1267, 1136, 1136, 
	1136, 1136, 1136, 1136, 1136, 1136, 1136, 1136, 
	1136, 1136, 1136, 1136, 1136, 1136, 1136, 1136, 
	1136, 1136, 1136, 1266, 1266, 1266, 1266, 1136, 
	1266, 1136, 1136, 1136, 1136, 1136, 1136, 1136, 
	1136, 1136, 1136, 1136, 1136, 1136, 1136, 1136, 
	1136, 1136, 1136, 1136, 1136, 1136, 1136, 1136, 
	1136, 1136, 1136, 1266, 1266, 1266, 1266, 1266, 
	1136, 1266, 1266, 1266, 1266, 1266, 1266, 1266, 
	1266, 1266, 1266, 1266, 1266, 1266, 1266, 1266, 
	1266, 1266, 1266, 1266, 1266, 1266, 1266, 1266, 
	1266, 1266, 1266, 1266, 1266, 1266, 1266, 1266, 
	1266, 1266, 1162, 1266, 1266, 1266, 1266, 1266, 
	1266, 1266, 1266, 1266, 1266, 1266, 1266, 1266, 
	1266, 1136, 1136, 1136, 1136, 1136, 1136, 1136, 
	1136, 1136, 1136, 1264, 1266, 1266, 1266, 1266, 
	1162, 1266, 1136, 1136, 1136, 1136, 1136, 1136, 
	1268, 1136, 1136, 1136, 1136, 1136, 1136, 1136, 
	1136, 1136, 1136, 1136, 1136, 1136, 1136, 1136, 
	1136, 1136, 1136, 1136, 1266, 1266, 1266, 1266, 
	1136, 1266, 1136, 1136, 1136, 1136, 1136, 1136, 
	1136, 1136, 1136, 1136, 1136, 1136, 1136, 1136, 
	1136, 1136, 1136, 1136, 1136, 1136, 1136, 1136, 
	1136, 1136, 1136, 1136, 1266, 1266, 1266, 1266, 
	1266, 1136, 1266, 1266, 1266, 1266, 1266, 1266, 
	1266, 1266, 1266, 1266, 1266, 1266, 1266, 1266, 
	1266, 1266, 1266, 1266, 1266, 1266, 1266, 1266, 
	1266, 1266, 1266, 1266, 1266, 1266, 1266, 1266, 
	1266, 1266, 1266, 1162, 1266, 1266, 1266, 1266, 
	1266, 1266, 1266, 1266, 1266, 1266, 1266, 1266, 
	1266, 1266, 1136, 1136, 1136, 1136, 1136, 1136, 
	1136, 1136, 1136, 1136, 1264, 1266, 1266, 1266, 
	1266, 1162, 1266, 1136, 1136, 1136, 1136, 1136, 
	1136, 1136, 1136, 1269, 1136, 1136, 1136, 1136, 
	1136, 1136, 1136, 1136, 1136, 1136, 1136, 1136, 
	1136, 1136, 1136, 1136, 1136, 1266, 1266, 1266, 
	1266, 1136, 1266, 1136, 1136, 1136, 1136, 1136, 
	1136, 1136, 1136, 1136, 1136, 1136, 1136, 1136, 
	1136, 1136, 1136, 1136, 1136, 1136, 1136, 1136, 
	1136, 1136, 1136, 1136, 1136, 1266, 1266, 1266, 
	1266, 1266, 1136, 1266, 1266, 1266, 1266, 1266, 
	1266, 1266, 1266, 1266, 1266, 1266, 1266, 1266, 
	1266, 1266, 1266, 1266, 1266, 1266, 1266, 1266, 
	1266, 1266, 1266, 1266, 1266, 1266, 1266, 1266, 
	1266, 1266, 1266, 1266, 1162, 1266, 1266, 1266, 
	1266, 1266, 1266, 1266, 1266, 1266, 1266, 1266, 
	1266, 1266, 1266, 1136, 1136, 1136, 1136, 1136, 
	1136, 1136, 1136, 1136, 1136, 1264, 1266, 1266, 
	1266, 1266, 1162, 1266, 1136, 1136, 1136, 1136, 
	1136, 1136, 1136, 1136, 1136, 1136, 1136, 1136, 
	1136, 1270, 1136, 1136, 1136, 1136, 1136, 1136, 
	1136, 1136, 1136, 1136, 1136, 1136, 1266, 1266, 
	1266, 1266, 1136, 1266, 1136, 1136, 1136, 1136, 
	1136, 1136, 1136, 1136, 1136, 1136, 1136, 1136, 
	1136, 1136, 1136, 1136, 1136, 1136, 1136, 1136, 
	1136, 1136, 1136, 1136, 1136, 1136, 1266, 1266, 
	1266, 1266, 1266, 1136, 1266, 1266, 1266, 1266, 
	1266, 1266, 1266, 1266, 1266, 1266, 1266, 1266, 
	1266, 1266, 1266, 1266, 1266, 1266, 1266, 1266, 
	1266, 1266, 1266, 1266, 1266, 1266, 1266, 1266, 
	1266, 1266, 1266, 1266, 1266, 1162, 1266, 1266, 
	1266, 1266, 1266, 1266, 1266, 1266, 1266, 1266, 
	1266, 1266, 1266, 1266, 1136, 1136, 1136, 1136, 
	1136, 1136, 1136, 1136, 1136, 1136, 1264, 1266, 
	1266, 1266, 1266, 1162, 1266, 1136, 1136, 1136, 
	1136, 1136, 1136, 1136, 1136, 1136, 1136, 1136, 
	1136, 1136, 1271, 1136, 1136, 1136, 1136, 1136, 
	1136, 1136, 1136, 1136, 1136, 1136, 1136, 1266, 
	1266, 1266, 1266, 1136, 1266, 1136, 1136, 1136, 
	1136, 1136, 1136, 1136, 1136, 1136, 1136, 1136, 
	1136, 1136, 1136, 1136, 1136, 1136, 1136, 1136, 
	1136, 1136, 1136, 1136, 1136, 1136, 1136, 1266, 
	1266, 1266, 1266, 1266, 1136, 1266, 1266, 1266, 
	1266, 1266, 1266, 1266, 1266, 1266, 1266, 1266, 
	1266, 1266, 1266, 1266, 1266, 1266, 1266, 1266, 
	1266, 1266, 1266, 1266, 1266, 1266, 1266, 1266, 
	1266, 1266, 1266, 1266, 1266, 1266, 1162, 1266, 
	1266, 1266, 1266, 1266, 1266, 1266, 1266, 1266, 
	1266, 1266, 1266, 1266, 1266, 1136, 1136, 1136, 
	1136, 1136, 1136, 1136, 1136, 1136, 1136, 1264, 
	1266, 1266, 1266, 1266, 1162, 1266, 1136, 1136, 
	1136, 1270, 1136, 1136, 1136, 1136, 1136, 1136, 
	1136, 1136, 1136, 1136, 1136, 1136, 1136, 1136, 
	1136, 1136, 1136, 1136, 1136, 1136, 1136, 1136, 
	1266, 1266, 1266, 1266, 1136, 1266, 1136, 1136, 
	1136, 1136, 1136, 1136, 1136, 1136, 1136, 1136, 
	1136, 1136, 1136, 1136, 1136, 1136, 1136, 1136, 
	1136, 1136, 1136, 1136, 1136, 1136, 1136, 1136, 
	1266, 1266, 1266, 1266, 1266, 1136, 1272, 1274, 
	1273, 1275, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1162, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1276, 1276, 1276, 1276, 
	1276, 1162, 1276, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1276, 1276, 1276, 
	1276, 1277, 1276, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1276, 1276, 1276, 
	1276, 1276, 1110, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1162, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1276, 1276, 1276, 
	1276, 1276, 1162, 1276, 1110, 1110, 1110, 1110, 
	1278, 1279, 1110, 1110, 1110, 1110, 1110, 1280, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1276, 1276, 
	1276, 1276, 1110, 1276, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1276, 1276, 
	1276, 1276, 1276, 1110, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1162, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1276, 1276, 
	1276, 1276, 1276, 1162, 1276, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1281, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1276, 
	1276, 1276, 1276, 1110, 1276, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1276, 
	1276, 1276, 1276, 1276, 1110, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1162, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1276, 
	1276, 1276, 1276, 1276, 1162, 1276, 1110, 1110, 
	1282, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1276, 1276, 1276, 1276, 1110, 1276, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1276, 1276, 1276, 1276, 1276, 1110, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1162, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1276, 1276, 1276, 1276, 1276, 1162, 1276, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1283, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1276, 1276, 1276, 1276, 1110, 1276, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1276, 1276, 1276, 1276, 1276, 1110, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1162, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1276, 1276, 1276, 1276, 1276, 1162, 1276, 
	1110, 1110, 1110, 1284, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1276, 1276, 1276, 1276, 1110, 1276, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1276, 1276, 1276, 1276, 1276, 1110, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1162, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1276, 1276, 1276, 1276, 1276, 1162, 
	1276, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1285, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1276, 1276, 1276, 1276, 1110, 
	1276, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1276, 1276, 1276, 1276, 1276, 
	1110, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1162, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1276, 1276, 1276, 1276, 1276, 
	1162, 1276, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1286, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1276, 1276, 1276, 1276, 
	1110, 1276, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1276, 1276, 1276, 1276, 
	1276, 1110, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1162, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1276, 1276, 1276, 1276, 
	1276, 1162, 1276, 1110, 1110, 1110, 1110, 1110, 
	1110, 1287, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1276, 1276, 1276, 
	1276, 1110, 1276, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1276, 1276, 1276, 
	1276, 1276, 1110, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1162, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1276, 1276, 1276, 
	1276, 1276, 1162, 1276, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1276, 1276, 
	1276, 1276, 1288, 1276, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1276, 1276, 
	1276, 1276, 1276, 1110, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1162, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1276, 1276, 
	1276, 1276, 1276, 1162, 1276, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1276, 
	1276, 1276, 1276, 1289, 1276, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1276, 
	1276, 1276, 1276, 1276, 1110, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1162, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1276, 
	1276, 1276, 1276, 1276, 1162, 1276, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1290, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1276, 1276, 1276, 1276, 1110, 1276, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1276, 1276, 1276, 1276, 1276, 1110, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1162, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1276, 1276, 1276, 1276, 1276, 1162, 1276, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1291, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1276, 1276, 1276, 1276, 1110, 1276, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1276, 1276, 1276, 1276, 1276, 1110, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1162, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1276, 1276, 1276, 1276, 1276, 1162, 1276, 
	1110, 1110, 1110, 1110, 1292, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1276, 1276, 1276, 1276, 1110, 1276, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1276, 1276, 1276, 1276, 1276, 1110, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1162, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1276, 1276, 1276, 1276, 1276, 1162, 
	1276, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1276, 1276, 1276, 1276, 1293, 
	1276, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1276, 1276, 1276, 1276, 1276, 
	1110, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1162, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1276, 1276, 1276, 1276, 1276, 
	1162, 1276, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1276, 1276, 1276, 1276, 
	1294, 1276, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1276, 1276, 1276, 1276, 
	1276, 1110, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1162, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1276, 1276, 1276, 1276, 
	1276, 1162, 1276, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1295, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1276, 1276, 1276, 
	1276, 1110, 1276, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1276, 1276, 1276, 
	1276, 1276, 1110, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1162, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1276, 1276, 1276, 
	1276, 1276, 1162, 1276, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1291, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1276, 1276, 
	1276, 1276, 1110, 1276, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1276, 1276, 
	1276, 1276, 1276, 1110, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1162, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1276, 1276, 
	1276, 1276, 1276, 1162, 1276, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1276, 
	1276, 1276, 1276, 1110, 1276, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1296, 1110, 1297, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1276, 
	1276, 1276, 1276, 1276, 1110, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1162, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1276, 
	1276, 1276, 1276, 1276, 1162, 1276, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1276, 1276, 1276, 1276, 1110, 1276, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1298, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1276, 1276, 1276, 1276, 1276, 1110, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1162, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1276, 1276, 1276, 1276, 1276, 1162, 1276, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1276, 1276, 1276, 1276, 1110, 1276, 1299, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1276, 1276, 1276, 1276, 1276, 1110, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1162, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1276, 1276, 1276, 1276, 1276, 1162, 1276, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1276, 1276, 1276, 1276, 1110, 1276, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1300, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1276, 1276, 1276, 1276, 1276, 1110, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1162, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1276, 1276, 1276, 1276, 1276, 1162, 
	1276, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1276, 1276, 1276, 1276, 1110, 
	1276, 1110, 1110, 1110, 1301, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1276, 1276, 1276, 1276, 1276, 
	1110, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1162, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1276, 1276, 1276, 1276, 1276, 
	1162, 1276, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1276, 1276, 1276, 1276, 
	1110, 1276, 1110, 1110, 1110, 1110, 1302, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1303, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1276, 1276, 1276, 1276, 
	1276, 1110, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1162, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1276, 1276, 1276, 1276, 
	1276, 1162, 1276, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1276, 1276, 1276, 
	1276, 1110, 1276, 1110, 1110, 1110, 1110, 1110, 
	1110, 1304, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1276, 1276, 1276, 
	1276, 1276, 1110, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1162, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1276, 1276, 1276, 
	1276, 1276, 1162, 1276, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1276, 1276, 
	1276, 1276, 1110, 1276, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1305, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1276, 1276, 
	1276, 1276, 1276, 1110, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1162, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1276, 1276, 
	1276, 1276, 1276, 1162, 1276, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1276, 
	1276, 1276, 1276, 1110, 1276, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1301, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1276, 
	1276, 1276, 1276, 1276, 1110, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1162, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1276, 
	1276, 1276, 1276, 1276, 1162, 1276, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1276, 1276, 1276, 1276, 1110, 1276, 1110, 1110, 
	1110, 1110, 1306, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1276, 1276, 1276, 1276, 1276, 1110, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1162, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1276, 1276, 1276, 1276, 1276, 1162, 1276, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1276, 1276, 1276, 1276, 1110, 1276, 1307, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1276, 1276, 1276, 1276, 1276, 1110, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1162, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1276, 1276, 1276, 1276, 1276, 1162, 1276, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1276, 1276, 1276, 1276, 1110, 1276, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1308, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1276, 1276, 1276, 1276, 1276, 1110, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1162, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1276, 1276, 1276, 1276, 1276, 1162, 
	1276, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1276, 1276, 1276, 1276, 1110, 
	1276, 1309, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1310, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1276, 1276, 1276, 1276, 1276, 
	1110, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1162, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1276, 1276, 1276, 1276, 1276, 
	1162, 1276, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1276, 1276, 1276, 1276, 
	1110, 1276, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1311, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1276, 1276, 1276, 1276, 
	1276, 1110, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1162, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1276, 1276, 1276, 1276, 
	1276, 1162, 1276, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1276, 1276, 1276, 
	1276, 1110, 1276, 1110, 1110, 1110, 1110, 1301, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1276, 1276, 1276, 
	1276, 1276, 1110, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1162, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1276, 1276, 1276, 
	1276, 1276, 1162, 1276, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1276, 1276, 
	1276, 1276, 1110, 1276, 1312, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1276, 1276, 
	1276, 1276, 1276, 1110, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1162, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1276, 1276, 
	1276, 1276, 1276, 1162, 1276, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1276, 
	1276, 1276, 1276, 1110, 1276, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1313, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1276, 
	1276, 1276, 1276, 1276, 1110, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1162, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1276, 
	1276, 1276, 1276, 1276, 1162, 1276, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1276, 1276, 1276, 1276, 1110, 1276, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1314, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1276, 1276, 1276, 1276, 1276, 1110, 1315, 1315, 
	1315, 1315, 1315, 1315, 1315, 1315, 1315, 263, 
	264, 263, 263, 263, 1315, 1315, 1315, 1315, 
	1315, 1315, 1315, 1315, 1315, 1315, 1315, 1315, 
	1315, 1315, 1315, 1315, 1315, 1315, 263, 1162, 
	1315, 265, 1315, 1315, 1315, 1315, 1315, 1315, 
	1315, 1315, 1315, 1315, 1315, 1315, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1315, 1315, 266, 1315, 1315, 1162, 1315, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1315, 267, 1315, 1315, 1110, 1315, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1315, 1315, 1315, 1315, 1315, 1110, 280, 
	281, 280, 280, 280, 279, 279, 279, 279, 
	279, 279, 279, 279, 279, 279, 279, 279, 
	279, 279, 279, 279, 279, 279, 280, 279, 
	279, 282, 279, 279, 279, 279, 279, 279, 
	279, 279, 279, 279, 279, 279, 279, 279, 
	279, 279, 279, 279, 279, 279, 279, 279, 
	279, 279, 285, 279, 279, 279, 279, 279, 
	279, 279, 279, 279, 279, 279, 279, 279, 
	279, 279, 279, 279, 279, 279, 279, 279, 
	279, 279, 279, 279, 279, 279, 279, 279, 
	279, 279, 284, 279, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1162, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1276, 1276, 
	1276, 1276, 1276, 1162, 1276, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1276, 
	1276, 1276, 1276, 1110, 1276, 1110, 1110, 1110, 
	1110, 1317, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1318, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1276, 
	1276, 1276, 1276, 1276, 1110, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1162, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1276, 
	1276, 1276, 1276, 1276, 1162, 1276, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1276, 1276, 1276, 1276, 1110, 1276, 1110, 1110, 
	1110, 1110, 1110, 1319, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1276, 1276, 1276, 1276, 1276, 1110, 1320, 1320, 
	1320, 1320, 1320, 1320, 1320, 1320, 1320, 1320, 
	1320, 1320, 1320, 1320, 1320, 1320, 1320, 1320, 
	1320, 1320, 1320, 1320, 1320, 1320, 1320, 1320, 
	1320, 1320, 1320, 1320, 1320, 1320, 1320, 1162, 
	1320, 1320, 1320, 1320, 1320, 1320, 1320, 1320, 
	1320, 1320, 1320, 1320, 1320, 1320, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1320, 1320, 1320, 1320, 1320, 1162, 1320, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1320, 1320, 1320, 1320, 1110, 1320, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1321, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1320, 1320, 1320, 1320, 1320, 1110, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1162, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1276, 1276, 1276, 1276, 1276, 1162, 1276, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1276, 1276, 1276, 1276, 1110, 1276, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1322, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1276, 1276, 1276, 1276, 1276, 1110, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1162, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1276, 1276, 1276, 1276, 1276, 1162, 
	1276, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1276, 1276, 1276, 1276, 1110, 
	1276, 1110, 1110, 1110, 1110, 1323, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1276, 1276, 1276, 1276, 1276, 
	1110, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1162, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1276, 1276, 1276, 1276, 1276, 
	1162, 1276, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1276, 1276, 1276, 1276, 
	1110, 1276, 1110, 1110, 1110, 1324, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1276, 1276, 1276, 1276, 
	1276, 1110, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1162, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1276, 1276, 1276, 1276, 
	1276, 1325, 1276, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1276, 1276, 1276, 
	1276, 1110, 1276, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1276, 1276, 1276, 
	1276, 1276, 1110, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1162, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1276, 1276, 1276, 
	1276, 1276, 1162, 1276, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1276, 1276, 
	1276, 1276, 1110, 1276, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1326, 
	1110, 1327, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1276, 1276, 
	1276, 1276, 1276, 1110, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1162, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1276, 1276, 
	1276, 1276, 1276, 1162, 1276, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1276, 
	1276, 1276, 1276, 1110, 1276, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1328, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1276, 
	1276, 1276, 1276, 1276, 1110, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1162, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1276, 
	1276, 1276, 1276, 1276, 1162, 1276, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1276, 1276, 1276, 1276, 1110, 1276, 1110, 1110, 
	1110, 1110, 1301, 1110, 1110, 1110, 1329, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1276, 1276, 1276, 1276, 1276, 1110, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1162, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1276, 1276, 1276, 1276, 1276, 1162, 1276, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1276, 1276, 1276, 1276, 1110, 1276, 1110, 
	1110, 1110, 1110, 1110, 1301, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1276, 1276, 1276, 1276, 1276, 1110, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1162, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1276, 1276, 1276, 1276, 1276, 1162, 1276, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1276, 1276, 1276, 1276, 1110, 1276, 
	1110, 1110, 1110, 1294, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1330, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1276, 1276, 1276, 1276, 1276, 1110, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1162, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1276, 1276, 1276, 1276, 1276, 1162, 
	1276, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1276, 1276, 1276, 1276, 1110, 
	1276, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1331, 1110, 1110, 
	1110, 1110, 1110, 1276, 1276, 1276, 1276, 1276, 
	1110, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1162, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1276, 1276, 1276, 1276, 1276, 
	1162, 1276, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1276, 1276, 1276, 1276, 
	1110, 1276, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1311, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1276, 1276, 1276, 1276, 
	1276, 1110, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1162, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1276, 1276, 1276, 1276, 
	1276, 1162, 1276, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1276, 1276, 1276, 
	1276, 1110, 1276, 1332, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1152, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1276, 1276, 1276, 
	1276, 1276, 1110, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1162, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1276, 1276, 1276, 
	1276, 1276, 1162, 1276, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1276, 1276, 
	1276, 1276, 1110, 1276, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1333, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1276, 1276, 
	1276, 1276, 1276, 1110, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1162, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1276, 1276, 
	1276, 1276, 1276, 1162, 1276, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1276, 
	1276, 1276, 1276, 1110, 1276, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1334, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1276, 
	1276, 1276, 1276, 1276, 1110, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1162, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1276, 
	1276, 1276, 1276, 1276, 1162, 1276, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1276, 1276, 1276, 1276, 1110, 1276, 1110, 1110, 
	1110, 1110, 1294, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1276, 1276, 1276, 1276, 1276, 1110, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1162, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1276, 1276, 1276, 1276, 1276, 1162, 1276, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1276, 1276, 1276, 1276, 1110, 1276, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1301, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1276, 1276, 1276, 1276, 1276, 1110, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1162, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1276, 1276, 1276, 1276, 1276, 1162, 1276, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1276, 1276, 1276, 1276, 1110, 1276, 
	1110, 1110, 1110, 1110, 1110, 1335, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1301, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1276, 1276, 1276, 1276, 1276, 1110, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1162, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1276, 1276, 1276, 1276, 1276, 1162, 
	1276, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1276, 1276, 1276, 1276, 1110, 
	1276, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1336, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1276, 1276, 1276, 1276, 1276, 
	1110, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1162, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1276, 1276, 1276, 1276, 1276, 
	1162, 1276, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1276, 1276, 1276, 1276, 
	1110, 1276, 1110, 1110, 1110, 1337, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1276, 1276, 1276, 1276, 
	1276, 1110, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1162, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1276, 1276, 1276, 1276, 
	1276, 1162, 1276, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1276, 1276, 1276, 
	1276, 1110, 1276, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1338, 
	1110, 1110, 1110, 1110, 1110, 1276, 1276, 1276, 
	1276, 1276, 1110, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1162, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1276, 1276, 1276, 
	1276, 1276, 1162, 1276, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1276, 1276, 
	1276, 1276, 1110, 1276, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1311, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1276, 1276, 
	1276, 1276, 1276, 1110, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1162, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1276, 1276, 
	1276, 1276, 1276, 1162, 1276, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1276, 
	1276, 1276, 1276, 1110, 1276, 1110, 1110, 1110, 
	1110, 1339, 1110, 1110, 1110, 1340, 1110, 1110, 
	1110, 1110, 1110, 1341, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1276, 
	1276, 1276, 1276, 1276, 1110, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1162, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1276, 
	1276, 1276, 1276, 1276, 1162, 1276, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1276, 1276, 1276, 1276, 1110, 1276, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1342, 1110, 1110, 
	1276, 1276, 1276, 1276, 1276, 1110, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1162, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1276, 1276, 1276, 1276, 1276, 1162, 1276, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1276, 1276, 1276, 1276, 1110, 1276, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1308, 1110, 1110, 1110, 1110, 1110, 
	1110, 1276, 1276, 1276, 1276, 1276, 1110, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1162, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1276, 1276, 1276, 1276, 1276, 1162, 1276, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1276, 1276, 1276, 1276, 1110, 1276, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1294, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1276, 1276, 1276, 1276, 1276, 1110, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1162, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1276, 1276, 1276, 1276, 1276, 1162, 
	1276, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1276, 1276, 1276, 1276, 1110, 
	1276, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1343, 1110, 1110, 1110, 
	1110, 1110, 1110, 1276, 1276, 1276, 1276, 1276, 
	1110, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1162, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1276, 1276, 1276, 1276, 1276, 
	1162, 1276, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1276, 1276, 1276, 1276, 
	1110, 1276, 1110, 1110, 1110, 1110, 1344, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1276, 1276, 1276, 1276, 
	1276, 1110, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1162, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1276, 1276, 1276, 1276, 
	1276, 1162, 1276, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1276, 1276, 1276, 
	1276, 1110, 1276, 1110, 1110, 1110, 1345, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1346, 1347, 1110, 
	1110, 1110, 1110, 1110, 1110, 1276, 1276, 1276, 
	1276, 1276, 1110, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1162, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1276, 1276, 1276, 
	1276, 1276, 1162, 1276, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1276, 1276, 
	1276, 1276, 1110, 1276, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1294, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1276, 1276, 
	1276, 1276, 1276, 1110, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1162, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1276, 1276, 
	1276, 1276, 1276, 1162, 1276, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1276, 
	1276, 1276, 1276, 1110, 1276, 1110, 1110, 1348, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1276, 
	1276, 1276, 1276, 1276, 1110, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1162, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1276, 
	1276, 1276, 1276, 1276, 1162, 1276, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1276, 1276, 1276, 1276, 1110, 1276, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1349, 1110, 1110, 1110, 1110, 1110, 
	1276, 1276, 1276, 1276, 1276, 1110, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1162, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1276, 1276, 1276, 1276, 1276, 1162, 1276, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1276, 1276, 1276, 1276, 1110, 1276, 1110, 
	1110, 1110, 1110, 1335, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1276, 1276, 1276, 1276, 1276, 1110, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1162, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1276, 1276, 1276, 1276, 1276, 1162, 1276, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1276, 1276, 1276, 1276, 1110, 1276, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1350, 1110, 1110, 1351, 1110, 1110, 1110, 
	1110, 1110, 1276, 1276, 1276, 1276, 1276, 1110, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1162, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1276, 1276, 1276, 1276, 1276, 1162, 
	1276, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1276, 1276, 1276, 1276, 1110, 
	1276, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1294, 1110, 1276, 1276, 1276, 1276, 1276, 
	1110, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1162, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1276, 1276, 1276, 1276, 1276, 
	1162, 1276, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1276, 1276, 1276, 1276, 
	1110, 1276, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1352, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1276, 1276, 1276, 1276, 
	1276, 1110, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1162, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1276, 1276, 1276, 1276, 
	1276, 1162, 1276, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1276, 1276, 1276, 
	1276, 1110, 1276, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1308, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1276, 1276, 1276, 
	1276, 1276, 1110, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1162, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1276, 1276, 1276, 
	1276, 1276, 1162, 1276, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1276, 1276, 
	1276, 1276, 1110, 1276, 1110, 1110, 1110, 1110, 
	1353, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1354, 1110, 1110, 1110, 1110, 1110, 1276, 1276, 
	1276, 1276, 1276, 1110, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1162, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1276, 1276, 
	1276, 1276, 1276, 1162, 1276, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1276, 
	1276, 1276, 1276, 1110, 1276, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1355, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1276, 
	1276, 1276, 1276, 1276, 1110, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1162, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1276, 
	1276, 1276, 1276, 1276, 1162, 1276, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1276, 1276, 1276, 1276, 1110, 1276, 1110, 1110, 
	1110, 1110, 1110, 1294, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1276, 1276, 1276, 1276, 1276, 1110, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1162, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1276, 1276, 1276, 1276, 1276, 1162, 1276, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1276, 1276, 1276, 1276, 1110, 1276, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1356, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1276, 1276, 1276, 1276, 1276, 1110, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1162, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1276, 1276, 1276, 1276, 1276, 1162, 1276, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1276, 1276, 1276, 1276, 1110, 1276, 
	1110, 1110, 1110, 1110, 1357, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1276, 1276, 1276, 1276, 1276, 1110, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1162, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1276, 1276, 1276, 1276, 1276, 1162, 
	1276, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1276, 1276, 1276, 1276, 1110, 
	1276, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1343, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1276, 1276, 1276, 1276, 1276, 
	1110, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1162, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1276, 1276, 1276, 1276, 1276, 
	1162, 1276, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1276, 1276, 1276, 1276, 
	1110, 1276, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1358, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1359, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1276, 1276, 1276, 1276, 
	1276, 1110, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1162, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1276, 1276, 1276, 1276, 
	1276, 1162, 1276, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1276, 1276, 1276, 
	1276, 1110, 1276, 1110, 1110, 1110, 1110, 1305, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1276, 1276, 1276, 
	1276, 1276, 1110, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1162, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1276, 1276, 1276, 
	1276, 1276, 1162, 1276, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1276, 1276, 
	1276, 1276, 1110, 1276, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1334, 1110, 1110, 1110, 1110, 1110, 1276, 1276, 
	1276, 1276, 1276, 1110, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1162, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1276, 1276, 
	1276, 1276, 1276, 1162, 1276, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1276, 
	1276, 1276, 1276, 1110, 1276, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1360, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1276, 
	1276, 1276, 1276, 1276, 1110, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1162, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1276, 
	1276, 1276, 1276, 1276, 1162, 1276, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1276, 1276, 1276, 1276, 1110, 1276, 1110, 1110, 
	1110, 1361, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1362, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1363, 1110, 1110, 1110, 1110, 1110, 1110, 
	1276, 1276, 1276, 1276, 1276, 1110, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1162, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1276, 1276, 1276, 1276, 1276, 1162, 1276, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1276, 1276, 1276, 1276, 1110, 1276, 1110, 
	1110, 1110, 1110, 1364, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1276, 1276, 1276, 1276, 1276, 1110, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1162, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1276, 1276, 1276, 1276, 1276, 1162, 1276, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1276, 1276, 1276, 1276, 1110, 1276, 
	1110, 1110, 1110, 1110, 1110, 1300, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1276, 1276, 1276, 1276, 1276, 1110, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1162, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1276, 1276, 1276, 1276, 1276, 1162, 
	1276, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1276, 1276, 1276, 1276, 1110, 
	1276, 1110, 1110, 1110, 1110, 1365, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1276, 1276, 1276, 1276, 1276, 
	1110, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1162, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1276, 1276, 1276, 1276, 1276, 
	1162, 1276, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1276, 1276, 1276, 1276, 
	1110, 1276, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1366, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1276, 1276, 1276, 1276, 
	1276, 1110, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1162, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1276, 1276, 1276, 1276, 
	1276, 1162, 1276, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1276, 1276, 1276, 
	1276, 1110, 1276, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1335, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1276, 1276, 1276, 
	1276, 1276, 1110, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1162, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1276, 1276, 1276, 
	1276, 1276, 1162, 1276, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1276, 1276, 
	1276, 1276, 1110, 1276, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1367, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1276, 1276, 
	1276, 1276, 1276, 1110, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1162, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1276, 1276, 
	1276, 1276, 1276, 1162, 1276, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1276, 
	1276, 1276, 1276, 1110, 1276, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1335, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1276, 
	1276, 1276, 1276, 1276, 1110, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1162, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1276, 
	1276, 1276, 1276, 1276, 1162, 1276, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1276, 1276, 1276, 1276, 1110, 1276, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1368, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1276, 1276, 1276, 1276, 1276, 1110, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1162, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1276, 1276, 1276, 1276, 1276, 1162, 1276, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1276, 1276, 1276, 1276, 1110, 1276, 1110, 
	1110, 1110, 1110, 1305, 1110, 1110, 1110, 1369, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1276, 1276, 1276, 1276, 1276, 1110, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1162, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1276, 1276, 1276, 1276, 1276, 1162, 1276, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1276, 1276, 1276, 1276, 1110, 1276, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1370, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1276, 1276, 1276, 1276, 1276, 1110, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1162, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1276, 1276, 1276, 1276, 1276, 1162, 
	1276, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1276, 1276, 1276, 1276, 1110, 
	1276, 1110, 1110, 1110, 1110, 1335, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1276, 1276, 1276, 1276, 1276, 
	1110, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1162, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1276, 1276, 1276, 1276, 1276, 
	1162, 1276, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1276, 1276, 1276, 1276, 
	1110, 1276, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1371, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1276, 1276, 1276, 1276, 
	1276, 1110, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1162, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1276, 1276, 1276, 1276, 
	1276, 1162, 1276, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1276, 1276, 1276, 
	1276, 1110, 1276, 1110, 1110, 1110, 1110, 1372, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1276, 1276, 1276, 
	1276, 1276, 1110, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1162, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1276, 1276, 1276, 
	1276, 1276, 1162, 1276, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1276, 1276, 
	1276, 1276, 1110, 1276, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1373, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1276, 1276, 
	1276, 1276, 1276, 1110, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1162, 1276, 1276, 
	1276, 1276, 1276, 1276, 1276, 1276, 1276, 1276, 
	1276, 1276, 1276, 1276, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1276, 1276, 
	1276, 1276, 1276, 1162, 1276, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1276, 
	1276, 1276, 1276, 1110, 1276, 1110, 1110, 1110, 
	1343, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1110, 
	1110, 1110, 1110, 1110, 1110, 1110, 1110, 1276, 
	1276, 1276, 1276, 1276, 1110, 1374, 1178, 1168, 
	1168, 1168, 1168, 1168, 1168, 1168, 1168, 1168, 
	1168, 1168, 1168, 1168, 1168, 1168, 1168, 1168, 
	1168, 1168, 1168, 1168, 1168, 1168, 1168, 1168, 
	1168, 1168, 1168, 1168, 1168, 1168, 1168, 1168, 
	1168, 1168, 1168, 1168, 1168, 1168, 1168, 1168, 
	1168, 1168, 1168, 1168, 1168, 1168, 1168, 1168, 
	1168, 1168, 1168, 1168, 1168, 1168, 1168, 1168, 
	1168, 1168, 1168, 1168, 1168, 1119, 1168, 1375, 
	1377, 1376, 1377, 1377, 1377, 1376, 1376, 1376, 
	1376, 1376, 1376, 1376, 1376, 1376, 1376, 1376, 
	1376, 1376, 1376, 1376, 1376, 1376, 1376, 1377, 
	1376, 1376, 1376, 1376, 1376, 1378, 1376, 1376, 
	1376, 1376, 1376, 1376, 1376, 291, 1376, 289, 
	1379, 289, 289, 289, 1379, 1379, 1379, 1379, 
	1379, 1379, 1379, 1379, 1379, 1379, 1379, 1379, 
	1379, 1379, 1379, 1379, 1379, 1379, 289, 1379, 
	1379, 1379, 1379, 1379, 290, 1379, 1379, 1379, 
	1379, 1379, 1379, 1379, 291, 1379, 292, 1379, 
	1381, 1380, 1380, 1380, 1380, 1380, 1380, 1380, 
	1380, 1380, 1380, 1380, 1380, 1380, 1380, 1380, 
	1380, 1380, 1380, 1380, 1380, 1380, 1380, 1380, 
	1380, 1380, 1380, 1380, 1380, 1380, 1380, 1380, 
	1380, 1380, 1380, 1380, 1380, 1380, 1380, 1380, 
	1380, 1380, 1380, 1380, 1380, 1380, 1380, 1380, 
	1380, 1380, 1380, 1382, 1380, 1381, 1380, 1381, 
	1380, 1380, 1380, 1380, 1380, 1380, 1380, 1380, 
	1380, 1380, 1380, 1380, 1380, 1380, 1380, 1380, 
	1380, 1380, 1380, 1380, 1380, 1380, 1380, 1380, 
	1380, 1380, 1380, 1380, 1380, 1380, 1380, 1380, 
	1380, 1380, 1380, 1380, 1380, 1380, 1380, 1380, 
	1380, 1380, 1380, 1380, 1380, 1380, 1380, 1380, 
	1380, 1380, 1380, 1380, 1380, 1380, 1380, 1380, 
	1380, 1380, 1380, 1380, 1380, 1380, 1380, 1380, 
	1380, 1380, 1380, 1380, 1380, 1380, 1380, 1380, 
	1380, 1380, 1380, 1380, 1380, 1380, 1380, 1380, 
	1380, 1380, 1380, 1380, 1380, 1380, 1380, 1380, 
	1380, 1380, 1384, 1380, 1381, 1380, 1380, 1380, 
	1380, 1380, 1380, 1380, 1380, 1380, 1380, 1380, 
	1380, 1380, 1380, 1380, 1380, 1380, 1380, 1380, 
	1380, 1380, 1380, 1380, 1380, 1380, 1380, 1380, 
	1380, 1380, 1380, 1380, 1380, 1380, 1380, 1380, 
	1380, 1380, 1380, 1380, 1380, 1380, 1380, 1380, 
	1380, 1380, 1380, 1380, 1380, 1380, 1380, 1380, 
	1380, 1380, 1380, 1380, 1380, 1380, 1380, 1380, 
	1380, 1380, 1380, 1380, 1380, 1380, 1380, 1380, 
	1380, 1380, 1380, 1380, 1380, 1380, 1380, 1380, 
	1380, 1380, 1380, 1380, 1380, 1380, 1380, 1380, 
	1380, 1380, 1380, 1380, 1380, 1380, 1380, 1380, 
	1380, 1380, 1380, 1380, 1380, 1380, 1380, 1380, 
	1385, 1380, 1381, 1380, 1380, 1380, 1380, 1380, 
	1380, 1380, 1380, 1380, 1380, 1380, 1380, 1380, 
	1380, 1380, 1380, 1380, 1380, 1380, 1380, 1380, 
	1380, 1380, 1380, 1380, 1380, 1380, 1380, 1380, 
	1380, 1380, 1380, 1380, 1380, 1380, 1380, 1380, 
	1380, 1380, 1380, 1380, 1380, 1380, 1380, 1380, 
	1380, 1380, 1380, 1380, 1380, 1380, 1380, 1380, 
	1380, 1380, 1380, 1380, 1380, 1380, 1380, 1380, 
	1380, 1380, 1380, 1380, 1380, 1380, 1380, 1380, 
	1380, 1380, 1380, 1380, 1380, 1380, 1380, 1380, 
	1380, 1380, 1380, 1380, 1380, 1380, 1380, 1380, 
	1380, 1380, 1380, 1380, 1386, 1380, 1388, 1386, 
	0
]

class << self
	attr_accessor :_lex_trans_targs
	private :_lex_trans_targs, :_lex_trans_targs=
end
self._lex_trans_targs = [
	160, 2, 3, 4, 166, 6, 7, 8, 
	9, 10, 160, 160, 169, 172, 169, 12, 
	173, 14, 169, 180, 181, 184, 185, 191, 
	186, 187, 188, 19, 189, 190, 193, 195, 
	196, 197, 198, 199, 25, 18, 182, 183, 
	27, 231, 232, 234, 236, 234, 30, 237, 
	32, 234, 244, 245, 248, 249, 255, 250, 
	251, 252, 37, 253, 254, 257, 259, 260, 
	261, 262, 263, 43, 36, 246, 247, 45, 
	295, 296, 303, 305, 303, 48, 306, 50, 
	311, 310, 0, 52, 312, 313, 54, 314, 
	315, 315, 315, 315, 428, 57, 58, 428, 
	456, 59, 60, 456, 456, 460, 460, 64, 
	59, 65, 458, 459, 461, 462, 460, 456, 
	463, 464, 466, 61, 62, 467, 468, 63, 
	460, 66, 67, 72, 79, 470, 471, 65, 
	458, 459, 461, 462, 460, 456, 463, 464, 
	466, 61, 62, 467, 468, 63, 66, 67, 
	72, 79, 470, 471, 469, 68, 69, 70, 
	71, 73, 74, 77, 75, 76, 78, 80, 
	81, 456, 83, 84, 85, 87, 90, 88, 
	89, 91, 93, 487, 487, 487, 488, 95, 
	490, 96, 491, 97, 488, 95, 490, 96, 
	491, 525, 525, 525, 100, 101, 102, 103, 
	536, 525, 525, 541, 525, 525, 562, 525, 
	107, 563, 569, 110, 113, 115, 117, 118, 
	119, 113, 114, 573, 114, 573, 116, 525, 
	587, 588, 591, 592, 598, 593, 594, 595, 
	124, 596, 597, 600, 602, 603, 604, 605, 
	606, 130, 123, 589, 590, 132, 642, 643, 
	134, 527, 98, 529, 136, 137, 645, 746, 
	139, 140, 141, 746, 754, 754, 754, 144, 
	775, 774, 754, 777, 779, 764, 811, 150, 
	151, 152, 156, 157, 150, 151, 152, 156, 
	157, 153, 153, 151, 152, 154, 155, 153, 
	153, 151, 152, 154, 155, 858, 151, 754, 
	927, 158, 159, 927, 927, 160, 160, 161, 
	162, 163, 165, 167, 168, 160, 160, 160, 
	164, 160, 164, 160, 1, 160, 160, 160, 
	5, 169, 169, 170, 169, 171, 174, 169, 
	169, 11, 13, 169, 169, 169, 175, 176, 
	177, 15, 21, 26, 200, 28, 169, 169, 
	169, 178, 179, 169, 16, 169, 169, 169, 
	17, 169, 169, 169, 20, 192, 194, 22, 
	169, 169, 23, 24, 169, 201, 205, 209, 
	202, 203, 204, 206, 207, 208, 169, 169, 
	210, 214, 220, 211, 218, 219, 212, 216, 
	217, 213, 215, 169, 221, 230, 229, 222, 
	223, 224, 225, 226, 227, 228, 169, 169, 
	169, 233, 234, 234, 234, 235, 238, 234, 
	29, 31, 234, 234, 234, 239, 240, 241, 
	33, 39, 44, 264, 46, 234, 234, 234, 
	242, 243, 234, 34, 234, 234, 234, 35, 
	234, 234, 234, 38, 256, 258, 40, 234, 
	234, 41, 42, 234, 265, 269, 273, 266, 
	267, 268, 270, 271, 272, 234, 234, 274, 
	278, 284, 275, 282, 283, 276, 280, 281, 
	277, 279, 234, 285, 294, 293, 286, 287, 
	288, 289, 290, 291, 292, 234, 234, 234, 
	297, 298, 298, 299, 298, 300, 298, 298, 
	298, 301, 301, 301, 302, 301, 301, 301, 
	303, 303, 303, 304, 303, 47, 49, 303, 
	303, 307, 307, 307, 308, 309, 308, 51, 
	53, 310, 310, 310, 316, 315, 315, 317, 
	318, 319, 320, 322, 325, 326, 327, 328, 
	315, 329, 330, 332, 334, 335, 336, 340, 
	342, 343, 344, 360, 365, 372, 377, 384, 
	391, 394, 395, 399, 393, 403, 411, 415, 
	417, 422, 424, 427, 315, 315, 315, 315, 
	315, 315, 321, 315, 321, 315, 323, 55, 
	324, 315, 56, 315, 315, 331, 333, 315, 
	337, 338, 339, 335, 341, 315, 345, 346, 
	355, 358, 347, 348, 349, 350, 351, 352, 
	353, 354, 316, 356, 357, 359, 361, 364, 
	362, 363, 366, 369, 367, 368, 370, 371, 
	373, 375, 374, 376, 378, 379, 315, 380, 
	381, 382, 383, 315, 385, 388, 386, 387, 
	389, 390, 392, 396, 397, 398, 400, 402, 
	401, 404, 405, 406, 408, 407, 409, 410, 
	412, 413, 414, 416, 418, 419, 420, 421, 
	423, 425, 426, 429, 428, 428, 430, 431, 
	433, 428, 428, 428, 432, 428, 432, 434, 
	428, 436, 435, 435, 439, 440, 441, 442, 
	435, 444, 445, 446, 447, 449, 451, 452, 
	453, 454, 455, 435, 437, 435, 438, 435, 
	435, 435, 435, 435, 443, 435, 443, 448, 
	435, 450, 435, 456, 456, 457, 472, 473, 
	459, 475, 476, 463, 477, 478, 479, 480, 
	481, 483, 484, 485, 486, 456, 456, 456, 
	456, 456, 456, 460, 465, 456, 456, 456, 
	456, 456, 456, 456, 456, 456, 474, 456, 
	474, 456, 456, 456, 456, 482, 456, 82, 
	86, 92, 487, 489, 492, 94, 487, 487, 
	488, 493, 493, 494, 495, 497, 499, 500, 
	493, 493, 496, 493, 496, 493, 498, 493, 
	493, 493, 502, 501, 501, 503, 504, 505, 
	507, 509, 510, 515, 522, 501, 501, 501, 
	501, 506, 501, 506, 501, 508, 501, 501, 
	502, 511, 512, 513, 514, 516, 517, 520, 
	518, 519, 521, 523, 524, 526, 525, 534, 
	535, 537, 538, 540, 542, 543, 544, 546, 
	547, 548, 550, 551, 572, 575, 576, 577, 
	645, 646, 647, 648, 649, 545, 651, 667, 
	672, 679, 684, 686, 692, 695, 696, 700, 
	694, 704, 715, 719, 722, 730, 734, 737, 
	738, 525, 98, 528, 525, 525, 530, 532, 
	533, 525, 531, 525, 525, 525, 525, 525, 
	99, 525, 525, 525, 525, 525, 539, 525, 
	539, 525, 525, 104, 525, 525, 105, 525, 
	525, 545, 525, 549, 525, 552, 561, 525, 
	106, 564, 565, 566, 525, 567, 108, 570, 
	109, 111, 571, 525, 553, 555, 525, 554, 
	525, 525, 556, 559, 560, 525, 557, 558, 
	525, 525, 525, 525, 568, 112, 574, 525, 
	525, 578, 525, 525, 525, 579, 581, 525, 
	580, 525, 580, 525, 582, 583, 584, 120, 
	126, 131, 607, 133, 525, 525, 525, 585, 
	586, 525, 121, 525, 525, 525, 122, 525, 
	525, 525, 125, 599, 601, 127, 525, 525, 
	128, 129, 525, 608, 612, 616, 609, 610, 
	611, 613, 614, 615, 525, 525, 617, 621, 
	627, 618, 625, 626, 619, 623, 624, 620, 
	622, 525, 628, 636, 641, 629, 630, 631, 
	632, 633, 634, 635, 637, 638, 639, 640, 
	525, 525, 525, 644, 135, 138, 525, 650, 
	525, 652, 653, 662, 665, 654, 655, 656, 
	657, 658, 659, 660, 661, 526, 663, 664, 
	666, 668, 671, 669, 670, 673, 676, 674, 
	675, 677, 678, 680, 682, 681, 683, 685, 
	687, 689, 688, 690, 691, 693, 526, 697, 
	698, 699, 701, 703, 702, 705, 706, 707, 
	712, 708, 709, 710, 525, 526, 527, 98, 
	711, 532, 713, 714, 716, 717, 718, 720, 
	721, 723, 724, 725, 728, 726, 727, 729, 
	731, 732, 733, 735, 736, 525, 739, 739, 
	740, 741, 742, 744, 739, 739, 739, 743, 
	739, 743, 739, 745, 739, 747, 746, 746, 
	748, 749, 746, 750, 752, 746, 746, 746, 
	746, 751, 746, 751, 753, 746, 755, 754, 
	754, 758, 759, 760, 754, 761, 763, 766, 
	767, 768, 769, 770, 754, 771, 772, 776, 
	799, 803, 754, 804, 806, 808, 754, 809, 
	810, 812, 816, 818, 819, 754, 821, 839, 
	844, 851, 859, 866, 873, 878, 879, 883, 
	877, 888, 898, 904, 907, 916, 920, 924, 
	925, 926, 756, 754, 757, 754, 754, 754, 
	754, 754, 754, 762, 754, 762, 754, 142, 
	765, 754, 754, 754, 754, 754, 754, 754, 
	773, 754, 754, 143, 145, 754, 146, 783, 
	791, 794, 778, 795, 796, 784, 788, 789, 
	754, 778, 146, 780, 781, 147, 754, 780, 
	754, 754, 782, 754, 785, 787, 754, 785, 
	786, 788, 789, 787, 754, 754, 790, 754, 
	754, 792, 787, 788, 789, 792, 793, 754, 
	785, 787, 788, 789, 754, 785, 787, 788, 
	789, 754, 797, 787, 788, 789, 797, 798, 
	754, 146, 799, 778, 800, 788, 789, 801, 
	787, 146, 801, 778, 802, 805, 807, 148, 
	149, 754, 754, 813, 814, 815, 810, 817, 
	754, 754, 820, 754, 754, 822, 823, 832, 
	837, 824, 825, 826, 827, 828, 829, 830, 
	831, 755, 833, 834, 835, 836, 755, 838, 
	840, 843, 841, 842, 755, 755, 845, 848, 
	846, 847, 849, 850, 755, 852, 854, 853, 
	855, 856, 857, 754, 754, 860, 755, 861, 
	754, 862, 863, 864, 865, 756, 867, 870, 
	868, 869, 871, 872, 874, 875, 876, 755, 
	880, 881, 882, 884, 886, 887, 885, 755, 
	889, 890, 891, 894, 892, 893, 895, 896, 
	897, 899, 901, 900, 902, 903, 905, 906, 
	908, 909, 911, 914, 910, 912, 913, 915, 
	917, 918, 919, 921, 922, 923, 754, 754, 
	927, 928, 929, 927, 931, 930, 932, 930, 
	933, 934, 935, 930, 930
]

class << self
	attr_accessor :_lex_trans_actions
	private :_lex_trans_actions, :_lex_trans_actions=
end
self._lex_trans_actions = [
	1, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 2, 3, 4, 0, 5, 0, 
	0, 0, 6, 0, 7, 0, 8, 0, 
	7, 0, 0, 0, 0, 8, 7, 0, 
	8, 0, 7, 7, 0, 0, 0, 0, 
	0, 0, 0, 9, 0, 10, 0, 0, 
	0, 11, 0, 7, 0, 8, 0, 7, 
	0, 0, 0, 0, 8, 7, 0, 8, 
	0, 7, 7, 0, 0, 0, 0, 0, 
	0, 0, 12, 0, 13, 0, 0, 0, 
	0, 14, 0, 0, 0, 0, 0, 0, 
	15, 16, 17, 18, 19, 0, 0, 20, 
	21, 0, 0, 22, 23, 24, 25, 26, 
	27, 27, 28, 29, 27, 30, 29, 31, 
	29, 27, 27, 28, 27, 32, 27, 27, 
	33, 27, 27, 27, 27, 27, 27, 0, 
	34, 35, 0, 36, 35, 37, 35, 0, 
	0, 34, 0, 38, 0, 0, 0, 0, 
	0, 0, 0, 0, 25, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 39, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 40, 41, 42, 43, 0, 
	0, 0, 43, 26, 44, 27, 27, 27, 
	44, 45, 46, 47, 0, 0, 0, 0, 
	0, 48, 49, 0, 50, 51, 0, 52, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 53, 53, 54, 0, 26, 0, 55, 
	0, 7, 0, 8, 0, 7, 0, 0, 
	0, 0, 8, 7, 0, 8, 0, 7, 
	7, 0, 0, 0, 0, 0, 0, 0, 
	0, 56, 0, 0, 0, 0, 56, 57, 
	0, 0, 0, 58, 59, 60, 61, 0, 
	7, 7, 62, 63, 63, 0, 0, 0, 
	26, 0, 0, 0, 27, 64, 27, 27, 
	27, 65, 66, 67, 66, 66, 66, 0, 
	68, 69, 68, 68, 68, 70, 71, 72, 
	73, 0, 74, 75, 76, 79, 80, 0, 
	26, 0, 7, 0, 7, 81, 82, 83, 
	65, 84, 0, 85, 0, 86, 87, 88, 
	0, 89, 90, 0, 91, 7, 7, 92, 
	93, 0, 0, 94, 95, 96, 97, 97, 
	97, 97, 97, 97, 97, 97, 98, 99, 
	100, 0, 0, 101, 0, 102, 103, 104, 
	0, 105, 106, 107, 0, 7, 0, 0, 
	108, 109, 0, 26, 110, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 111, 112, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 113, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 114, 115, 
	116, 0, 117, 118, 119, 7, 7, 120, 
	0, 0, 121, 122, 123, 97, 97, 97, 
	97, 97, 97, 97, 97, 124, 125, 126, 
	0, 0, 127, 0, 128, 129, 130, 0, 
	131, 132, 133, 0, 7, 0, 0, 134, 
	135, 0, 26, 136, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 137, 138, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 139, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 140, 141, 142, 
	0, 143, 144, 0, 145, 0, 146, 147, 
	148, 149, 150, 151, 0, 152, 153, 154, 
	155, 156, 157, 7, 158, 0, 0, 159, 
	160, 161, 162, 163, 164, 0, 165, 0, 
	0, 166, 167, 168, 169, 170, 171, 0, 
	26, 0, 0, 7, 7, 0, 0, 0, 
	172, 0, 0, 0, 0, 173, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 174, 175, 176, 177, 
	178, 179, 65, 180, 0, 181, 0, 0, 
	0, 182, 0, 183, 184, 0, 0, 185, 
	0, 0, 0, 186, 0, 187, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 186, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 188, 0, 
	0, 0, 0, 189, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 7, 190, 191, 0, 0, 
	0, 192, 193, 194, 65, 195, 0, 26, 
	196, 0, 197, 198, 0, 26, 0, 0, 
	199, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 200, 0, 201, 0, 202, 
	203, 204, 205, 206, 65, 207, 0, 0, 
	208, 0, 209, 210, 211, 212, 26, 0, 
	25, 0, 0, 25, 0, 0, 0, 0, 
	0, 0, 7, 7, 7, 213, 214, 215, 
	216, 217, 218, 219, 0, 220, 221, 222, 
	223, 224, 225, 226, 227, 228, 65, 229, 
	0, 230, 231, 232, 233, 234, 235, 0, 
	0, 0, 236, 7, 7, 0, 237, 238, 
	239, 240, 241, 0, 0, 0, 0, 0, 
	242, 243, 65, 244, 0, 245, 26, 246, 
	247, 248, 249, 250, 251, 0, 26, 0, 
	0, 0, 0, 0, 0, 252, 253, 254, 
	255, 65, 256, 0, 257, 26, 258, 259, 
	260, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 261, 262, 0, 
	8, 0, 0, 7, 263, 0, 0, 0, 
	0, 0, 0, 7, 7, 0, 264, 0, 
	264, 264, 264, 0, 0, 264, 261, 261, 
	261, 261, 261, 261, 261, 261, 261, 261, 
	261, 261, 261, 261, 261, 261, 261, 0, 
	264, 265, 266, 0, 267, 268, 0, 56, 
	0, 269, 0, 270, 271, 272, 273, 274, 
	27, 275, 276, 277, 278, 279, 65, 280, 
	0, 281, 282, 0, 283, 284, 0, 285, 
	286, 263, 287, 0, 288, 0, 0, 289, 
	0, 0, 0, 0, 290, 0, 0, 0, 
	0, 0, 0, 291, 0, 0, 292, 0, 
	293, 294, 0, 0, 0, 295, 0, 0, 
	296, 297, 298, 299, 0, 0, 0, 300, 
	301, 0, 302, 303, 304, 0, 7, 305, 
	306, 307, 0, 308, 97, 97, 97, 97, 
	97, 97, 97, 97, 309, 310, 311, 0, 
	0, 312, 0, 313, 314, 315, 0, 316, 
	317, 318, 0, 7, 0, 0, 319, 320, 
	0, 26, 321, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 322, 323, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 324, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	325, 326, 327, 0, 0, 0, 328, 26, 
	329, 261, 261, 261, 261, 261, 261, 261, 
	261, 261, 261, 261, 261, 56, 261, 261, 
	261, 261, 261, 261, 261, 261, 261, 261, 
	261, 261, 261, 261, 261, 261, 261, 261, 
	261, 261, 261, 261, 261, 261, 330, 261, 
	261, 261, 261, 261, 261, 261, 261, 261, 
	261, 261, 261, 331, 332, 333, 334, 335, 
	335, 331, 261, 261, 261, 261, 261, 261, 
	261, 261, 261, 261, 261, 261, 261, 261, 
	261, 261, 261, 261, 261, 336, 337, 338, 
	0, 26, 0, 0, 339, 340, 341, 65, 
	342, 0, 343, 26, 344, 7, 345, 346, 
	0, 26, 347, 0, 0, 348, 349, 350, 
	351, 65, 352, 0, 26, 353, 354, 355, 
	356, 0, 26, 0, 357, 0, 7, 0, 
	0, 0, 0, 0, 358, 0, 0, 359, 
	359, 0, 360, 0, 0, 0, 361, 7, 
	362, 362, 362, 0, 0, 363, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 364, 365, 0, 366, 367, 368, 
	369, 370, 371, 65, 372, 0, 373, 0, 
	0, 374, 375, 376, 377, 378, 379, 380, 
	0, 381, 382, 0, 0, 383, 384, 385, 
	0, 0, 386, 0, 0, 385, 387, 387, 
	388, 389, 0, 390, 390, 0, 391, 392, 
	393, 394, 0, 395, 385, 385, 396, 0, 
	0, 397, 397, 0, 398, 399, 0, 400, 
	401, 402, 402, 403, 403, 0, 0, 404, 
	405, 405, 406, 406, 407, 408, 408, 409, 
	409, 410, 411, 411, 412, 412, 0, 0, 
	413, 414, 415, 416, 417, 418, 418, 415, 
	417, 419, 359, 420, 0, 0, 0, 0, 
	0, 421, 422, 362, 362, 362, 423, 362, 
	424, 425, 26, 426, 427, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 428, 0, 0, 0, 0, 423, 0, 
	0, 0, 0, 0, 429, 430, 0, 0, 
	0, 0, 0, 0, 431, 0, 0, 0, 
	0, 0, 430, 432, 433, 0, 434, 0, 
	435, 0, 0, 0, 0, 436, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 437, 
	0, 0, 0, 0, 0, 0, 0, 436, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 438, 439, 
	440, 7, 74, 441, 0, 442, 0, 443, 
	0, 0, 0, 444, 445
]

class << self
	attr_accessor :_lex_to_state_actions
	private :_lex_to_state_actions, :_lex_to_state_actions=
end
self._lex_to_state_actions = [
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	77, 0, 0, 0, 0, 0, 0, 0, 
	0, 77, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 77, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 77, 0, 0, 77, 0, 77, 
	0, 0, 0, 77, 77, 0, 77, 0, 
	0, 0, 0, 77, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 77, 0, 0, 0, 
	0, 0, 0, 77, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	77, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 77, 
	0, 0, 0, 0, 0, 77, 0, 0, 
	0, 0, 0, 0, 0, 77, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 77, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 77, 0, 0, 0, 0, 
	0, 0, 77, 0, 0, 0, 0, 0, 
	0, 0, 77, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 77, 
	0, 0, 77, 0, 0, 0, 0, 0
]

class << self
	attr_accessor :_lex_from_state_actions
	private :_lex_from_state_actions, :_lex_from_state_actions=
end
self._lex_from_state_actions = [
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	78, 0, 0, 0, 0, 0, 0, 0, 
	0, 78, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 78, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 78, 0, 0, 78, 0, 78, 
	0, 0, 0, 78, 78, 0, 78, 0, 
	0, 0, 0, 78, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 78, 0, 0, 0, 
	0, 0, 0, 78, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	78, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 78, 
	0, 0, 0, 0, 0, 78, 0, 0, 
	0, 0, 0, 0, 0, 78, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 78, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 78, 0, 0, 0, 0, 
	0, 0, 78, 0, 0, 0, 0, 0, 
	0, 0, 78, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 78, 
	0, 0, 78, 0, 0, 0, 0, 0
]

class << self
	attr_accessor :_lex_eof_trans
	private :_lex_eof_trans, :_lex_eof_trans=
end
self._lex_eof_trans = [
	0, 1, 1, 1, 1, 1, 1, 1, 
	1, 1, 1, 13, 13, 13, 13, 19, 
	19, 19, 19, 19, 19, 19, 19, 19, 
	19, 19, 19, 19, 19, 44, 44, 44, 
	44, 50, 50, 50, 50, 50, 50, 50, 
	50, 50, 50, 50, 50, 50, 50, 75, 
	75, 75, 75, 0, 0, 0, 0, 89, 
	91, 93, 93, 97, 97, 101, 101, 101, 
	101, 101, 101, 101, 101, 101, 101, 101, 
	101, 101, 101, 101, 101, 101, 101, 101, 
	101, 101, 162, 162, 162, 162, 162, 162, 
	162, 162, 162, 162, 162, 162, 172, 174, 
	174, 174, 186, 188, 188, 188, 188, 188, 
	194, 186, 198, 198, 198, 198, 198, 198, 
	198, 198, 198, 198, 198, 198, 198, 198, 
	216, 216, 216, 216, 216, 216, 216, 216, 
	216, 216, 216, 216, 216, 216, 186, 198, 
	198, 198, 198, 248, 248, 248, 253, 255, 
	255, 255, 259, 259, 253, 259, 259, 259, 
	259, 259, 259, 259, 259, 259, 289, 289, 
	0, 302, 303, 304, 306, 308, 310, 308, 
	308, 0, 320, 321, 325, 325, 326, 335, 
	336, 337, 337, 337, 340, 340, 342, 343, 
	344, 344, 344, 346, 347, 348, 348, 344, 
	340, 340, 353, 354, 354, 354, 354, 354, 
	357, 357, 357, 357, 357, 357, 357, 357, 
	367, 368, 368, 368, 368, 380, 380, 380, 
	368, 368, 368, 368, 380, 380, 380, 380, 
	380, 380, 380, 380, 380, 391, 380, 392, 
	393, 393, 0, 400, 404, 404, 405, 414, 
	415, 416, 416, 416, 419, 419, 421, 422, 
	423, 423, 423, 425, 426, 427, 427, 423, 
	419, 419, 432, 433, 433, 433, 433, 433, 
	436, 436, 436, 436, 436, 436, 436, 436, 
	446, 447, 447, 447, 447, 459, 459, 459, 
	447, 447, 447, 447, 459, 459, 459, 459, 
	459, 459, 459, 459, 459, 470, 459, 471, 
	472, 472, 0, 479, 480, 0, 486, 0, 
	493, 497, 497, 0, 0, 503, 0, 506, 
	506, 507, 508, 0, 549, 551, 552, 553, 
	554, 556, 558, 562, 562, 553, 553, 553, 
	553, 564, 553, 553, 558, 553, 553, 549, 
	568, 568, 568, 568, 568, 568, 558, 558, 
	574, 574, 574, 574, 574, 574, 574, 574, 
	574, 574, 574, 574, 574, 574, 574, 574, 
	574, 574, 574, 574, 574, 574, 574, 574, 
	574, 574, 574, 574, 574, 574, 574, 574, 
	574, 574, 574, 607, 574, 574, 574, 574, 
	574, 574, 574, 574, 574, 574, 574, 574, 
	574, 574, 574, 574, 574, 574, 574, 574, 
	574, 574, 574, 574, 574, 574, 574, 574, 
	574, 574, 574, 574, 574, 574, 574, 574, 
	574, 574, 574, 574, 574, 574, 574, 574, 
	574, 574, 574, 553, 0, 650, 651, 652, 
	654, 650, 657, 0, 676, 678, 680, 681, 
	682, 683, 684, 686, 683, 683, 683, 683, 
	683, 689, 683, 683, 691, 689, 689, 683, 
	0, 710, 711, 97, 97, 714, 715, 97, 
	711, 711, 718, 720, 723, 711, 724, 711, 
	725, 726, 728, 730, 711, 718, 731, 731, 
	720, 731, 735, 731, 731, 731, 731, 0, 
	172, 743, 744, 743, 743, 0, 753, 754, 
	756, 758, 760, 758, 762, 0, 774, 775, 
	776, 777, 779, 781, 783, 784, 784, 784, 
	784, 784, 784, 784, 784, 784, 784, 784, 
	784, 784, 784, 784, 784, 0, 186, 842, 
	845, 846, 850, 852, 853, 854, 855, 856, 
	858, 861, 862, 864, 866, 869, 870, 872, 
	873, 186, 870, 870, 861, 861, 861, 861, 
	892, 895, 897, 898, 902, 905, 906, 907, 
	908, 892, 892, 892, 892, 892, 892, 892, 
	892, 892, 892, 892, 861, 912, 861, 861, 
	861, 913, 920, 920, 922, 924, 933, 934, 
	935, 935, 935, 938, 938, 940, 941, 942, 
	942, 942, 944, 945, 946, 946, 942, 938, 
	938, 951, 952, 952, 952, 952, 952, 955, 
	955, 955, 955, 955, 955, 955, 955, 965, 
	966, 966, 966, 966, 978, 978, 978, 966, 
	966, 966, 966, 978, 978, 978, 978, 978, 
	978, 978, 978, 978, 978, 978, 978, 978, 
	978, 993, 994, 995, 995, 186, 861, 861, 
	999, 861, 856, 1001, 1001, 1001, 1001, 1001, 
	1001, 1001, 1001, 1001, 1001, 1001, 1001, 1001, 
	1001, 1001, 1001, 1001, 1001, 1001, 1001, 1001, 
	1001, 1001, 1001, 1001, 1001, 1001, 1001, 1001, 
	1001, 1001, 1001, 1001, 1001, 1001, 1001, 1001, 
	1001, 1001, 1001, 1001, 1001, 1001, 1001, 1001, 
	1001, 1001, 1001, 1001, 1001, 1001, 1001, 1001, 
	1001, 1001, 1001, 1001, 1001, 1001, 1053, 846, 
	1001, 1001, 1001, 1001, 1001, 1001, 1001, 1001, 
	1001, 1001, 1001, 1001, 1001, 1001, 1001, 1001, 
	1001, 1001, 1001, 1001, 1001, 1001, 1001, 1001, 
	1001, 1078, 861, 0, 1085, 1086, 1087, 1089, 
	1091, 1093, 0, 1102, 1103, 1104, 1105, 1107, 
	1102, 1110, 0, 259, 1164, 1166, 1167, 1168, 
	1169, 1171, 1173, 1175, 1178, 1178, 1169, 1169, 
	1181, 1182, 1169, 1169, 1184, 1186, 1187, 1187, 
	1190, 1201, 259, 1207, 1209, 1210, 1212, 1190, 
	1190, 1215, 1215, 1215, 1221, 1222, 1224, 1225, 
	1215, 1215, 1232, 1237, 1242, 1215, 1215, 1249, 
	1249, 1215, 1215, 1186, 1169, 1169, 1186, 1169, 
	1169, 1175, 259, 1266, 1267, 1267, 1267, 1267, 
	1267, 1267, 1273, 1175, 1276, 1277, 1277, 1277, 
	1277, 1277, 1277, 1277, 1277, 1277, 1277, 1277, 
	1277, 1277, 1277, 1277, 1277, 1277, 1277, 1277, 
	1277, 1277, 1277, 1277, 1277, 1277, 1277, 1277, 
	1277, 1277, 1277, 1277, 1277, 1277, 1277, 1277, 
	1277, 1316, 1317, 1277, 1277, 1321, 1277, 1277, 
	1277, 1277, 1277, 1277, 1277, 1277, 1277, 1277, 
	1277, 1277, 1277, 1277, 1277, 1277, 1277, 1277, 
	1277, 1277, 1277, 1277, 1277, 1277, 1277, 1277, 
	1277, 1277, 1277, 1277, 1277, 1277, 1277, 1277, 
	1277, 1277, 1277, 1277, 1277, 1277, 1277, 1277, 
	1277, 1277, 1277, 1277, 1277, 1277, 1277, 1277, 
	1277, 1277, 1277, 1277, 1277, 1277, 1277, 1277, 
	1277, 1277, 1277, 1277, 1375, 1169, 1376, 0, 
	1380, 1380, 0, 1384, 1384, 1384, 1384, 1388
]

class << self
	attr_accessor :lex_start
end
self.lex_start = 160;
class << self
	attr_accessor :lex_error
end
self.lex_error = 0;

class << self
	attr_accessor :lex_en_interp_words
end
self.lex_en_interp_words = 169;
class << self
	attr_accessor :lex_en_interp_string
end
self.lex_en_interp_string = 234;
class << self
	attr_accessor :lex_en_plain_words
end
self.lex_en_plain_words = 298;
class << self
	attr_accessor :lex_en_plain_string
end
self.lex_en_plain_string = 301;
class << self
	attr_accessor :lex_en_interp_backslash_delimited
end
self.lex_en_interp_backslash_delimited = 303;
class << self
	attr_accessor :lex_en_plain_backslash_delimited
end
self.lex_en_plain_backslash_delimited = 307;
class << self
	attr_accessor :lex_en_regexp_modifiers
end
self.lex_en_regexp_modifiers = 308;
class << self
	attr_accessor :lex_en_expr_variable
end
self.lex_en_expr_variable = 310;
class << self
	attr_accessor :lex_en_expr_fname
end
self.lex_en_expr_fname = 315;
class << self
	attr_accessor :lex_en_expr_endfn
end
self.lex_en_expr_endfn = 428;
class << self
	attr_accessor :lex_en_expr_dot
end
self.lex_en_expr_dot = 435;
class << self
	attr_accessor :lex_en_expr_arg
end
self.lex_en_expr_arg = 456;
class << self
	attr_accessor :lex_en_expr_cmdarg
end
self.lex_en_expr_cmdarg = 487;
class << self
	attr_accessor :lex_en_expr_endarg
end
self.lex_en_expr_endarg = 493;
class << self
	attr_accessor :lex_en_expr_mid
end
self.lex_en_expr_mid = 501;
class << self
	attr_accessor :lex_en_expr_beg
end
self.lex_en_expr_beg = 525;
class << self
	attr_accessor :lex_en_expr_labelarg
end
self.lex_en_expr_labelarg = 739;
class << self
	attr_accessor :lex_en_expr_value
end
self.lex_en_expr_value = 746;
class << self
	attr_accessor :lex_en_expr_end
end
self.lex_en_expr_end = 754;
class << self
	attr_accessor :lex_en_leading_dot
end
self.lex_en_leading_dot = 927;
class << self
	attr_accessor :lex_en_line_comment
end
self.lex_en_line_comment = 930;
class << self
	attr_accessor :lex_en_line_begin
end
self.lex_en_line_begin = 160;


# line 82 "lib/parser/lexer.rl"
  # %

  ESCAPES = {
    ?a.ord => "\a", ?b.ord  => "\b", ?e.ord => "\e", ?f.ord => "\f",
    ?n.ord => "\n", ?r.ord  => "\r", ?s.ord => "\s", ?t.ord => "\t",
    ?v.ord => "\v", ?\\.ord => "\\"
  }.freeze

  REGEXP_META_CHARACTERS = Regexp.union(*"\\$()*+.<>?[]^{|}".chars).freeze

  RBRACE_OR_RBRACK = %w"} ]".freeze

  attr_reader   :source_buffer

  attr_accessor :diagnostics
  attr_accessor :static_env
  attr_accessor :force_utf32

  attr_accessor :cond, :cmdarg, :in_kwarg

  attr_accessor :tokens, :comments

  def initialize(version)
    @version    = version
    @static_env = nil

    @tokens     = nil
    @comments   = nil

    reset
  end

  def reset(reset_state=true)
    # Ragel state:
    if reset_state
      # Unit tests set state prior to resetting lexer.
      @cs     = self.class.lex_en_line_begin

      @cond   = StackState.new('cond')
      @cmdarg = StackState.new('cmdarg')
      @cond_stack   = []
      @cmdarg_stack = []
    end

    @force_utf32   = false # Set to true by some tests

    @source_pts    = nil # @source as a codepoint array

    @p             = 0   # stream position (saved manually in #advance)
    @ts            = nil # token start
    @te            = nil # token end
    @act           = 0   # next action

    @stack         = []  # state stack
    @top           = 0   # state stack top pointer

    # Lexer state:
    @token_queue   = []
    @literal_stack = []

    @eq_begin_s    = nil # location of last encountered =begin
    @sharp_s       = nil # location of last encountered #

    @newline_s     = nil # location of last encountered newline

    @num_base      = nil # last numeric base
    @num_digits_s  = nil # starting position of numeric digits
    @num_suffix_s  = nil # starting position of numeric suffix
    @num_xfrm      = nil # numeric suffix-induced transformation

    @escape_s      = nil # starting position of current sequence
    @escape        = nil # last escaped sequence, as string

    @herebody_s    = nil # starting position of current heredoc line

    # Ruby 1.9 ->() lambdas emit a distinct token if do/{ is
    # encountered after a matching closing parenthesis.
    @paren_nest    = 0
    @lambda_stack  = []

    # After encountering the closing line of <<~SQUIGGLY_HEREDOC,
    # we store the indentation level and give it out to the parser
    # on request. It is not possible to infer indentation level just
    # from the AST because escape sequences such as `\ ` or `\t` are
    # expanded inside the lexer, but count as non-whitespace for
    # indentation purposes.
    @dedent_level  = nil

    # If the lexer is in `command state' (aka expr_value)
    # at the entry to #advance, it will transition to expr_cmdarg
    # instead of expr_arg at certain points.
    @command_state = false

    # True at the end of "def foo a:"
    @in_kwarg      = false
  end

  def source_buffer=(source_buffer)
    @source_buffer = source_buffer

    if @source_buffer
      source = @source_buffer.source

      if defined?(Encoding) && source.encoding == Encoding::UTF_8
        @source_pts = source.unpack('U*')
      else
        @source_pts = source.unpack('C*')
      end

      if @source_pts[0] == 0xfeff
        # Skip byte order mark.
        @p = 1
      end
    else
      @source_pts = nil
    end
  end

  def encoding
    @source_buffer.source.encoding
  end

  LEX_STATES = {
    :line_begin    => lex_en_line_begin,
    :expr_dot      => lex_en_expr_dot,
    :expr_fname    => lex_en_expr_fname,
    :expr_value    => lex_en_expr_value,
    :expr_beg      => lex_en_expr_beg,
    :expr_mid      => lex_en_expr_mid,
    :expr_arg      => lex_en_expr_arg,
    :expr_cmdarg   => lex_en_expr_cmdarg,
    :expr_end      => lex_en_expr_end,
    :expr_endarg   => lex_en_expr_endarg,
    :expr_endfn    => lex_en_expr_endfn,
    :expr_labelarg => lex_en_expr_labelarg,

    :interp_string => lex_en_interp_string,
    :interp_words  => lex_en_interp_words,
    :plain_string  => lex_en_plain_string,
    :plain_words   => lex_en_plain_string,
  }

  def state
    LEX_STATES.invert.fetch(@cs, @cs)
  end

  def state=(state)
    @cs = LEX_STATES.fetch(state)
  end

  def push_cmdarg
    @cmdarg_stack.push(@cmdarg)
    @cmdarg = StackState.new("cmdarg.#{@cmdarg_stack.count}")
  end

  def pop_cmdarg
    @cmdarg = @cmdarg_stack.pop
  end

  def push_cond
    @cond_stack.push(@cond)
    @cond = StackState.new("cond.#{@cond_stack.count}")
  end

  def pop_cond
    @cond = @cond_stack.pop
  end

  def dedent_level
    # We erase @dedent_level as a precaution to avoid accidentally
    # using a stale value.
    dedent_level, @dedent_level = @dedent_level, nil
    dedent_level
  end

  # Return next token: [type, value].
  def advance
    if @token_queue.any?
      return @token_queue.shift
    end

    # Ugly, but dependent on Ragel output. Consider refactoring it somehow.
    klass = self.class
    _lex_trans_keys         = klass.send :_lex_trans_keys
    _lex_key_spans          = klass.send :_lex_key_spans
    _lex_index_offsets      = klass.send :_lex_index_offsets
    _lex_indicies           = klass.send :_lex_indicies
    _lex_trans_targs        = klass.send :_lex_trans_targs
    _lex_trans_actions      = klass.send :_lex_trans_actions
    _lex_to_state_actions   = klass.send :_lex_to_state_actions
    _lex_from_state_actions = klass.send :_lex_from_state_actions
    _lex_eof_trans          = klass.send :_lex_eof_trans

    pe = @source_pts.size + 2
    p, eof = @p, pe

    @command_state = (@cs == klass.lex_en_expr_value ||
                      @cs == klass.lex_en_line_begin)

    
# line 10784 "lib/parser/lexer.rb"
begin
	testEof = false
	_slen, _trans, _keys, _inds, _acts, _nacts = nil
	_goto_level = 0
	_resume = 10
	_eof_trans = 15
	_again = 20
	_test_eof = 30
	_out = 40
	while true
	if _goto_level <= 0
	if p == pe
		_goto_level = _test_eof
		next
	end
	if  @cs == 0
		_goto_level = _out
		next
	end
	end
	if _goto_level <= _resume
	case _lex_from_state_actions[ @cs] 
	when 78 then
# line 1 "NONE"
		begin
 @ts = p
		end
# line 10812 "lib/parser/lexer.rb"
	end
	_keys =  @cs << 1
	_inds = _lex_index_offsets[ @cs]
	_slen = _lex_key_spans[ @cs]
	_trans = if (   _slen > 0 && 
			_lex_trans_keys[_keys] <= ( (@source_pts[p] || 0)) && 
			( (@source_pts[p] || 0)) <= _lex_trans_keys[_keys + 1] 
		    ) then
			_lex_indicies[ _inds + ( (@source_pts[p] || 0)) - _lex_trans_keys[_keys] ] 
		 else 
			_lex_indicies[ _inds + _slen ]
		 end
	end
	if _goto_level <= _eof_trans
	 @cs = _lex_trans_targs[_trans]
	if _lex_trans_actions[_trans] != 0
	case _lex_trans_actions[_trans]
	when 26 then
# line 486 "lib/parser/lexer.rl"
		begin

    # Record position of a newline for precise location reporting on tNL
    # tokens.
    #
    # This action is embedded directly into c_nl, as it is idempotent and
    # there are no cases when we need to skip it.
    @newline_s = p
  		end
	when 97 then
# line 766 "lib/parser/lexer.rl"
		begin

    @escape_s = p
    @escape   = nil
  		end
	when 27 then
# line 806 "lib/parser/lexer.rl"
		begin

    # After every heredoc was parsed, @herebody_s contains the
    # position of next token after all heredocs.
    if @herebody_s
      p = @herebody_s
      @herebody_s = nil
    end
  		end
	when 65 then
# line 1108 "lib/parser/lexer.rl"
		begin
 @sharp_s = p - 1 		end
	when 68 then
# line 1111 "lib/parser/lexer.rl"
		begin
 emit_comment(@sharp_s, p == pe ? p - 2 : p) 		end
	when 266 then
# line 1152 "lib/parser/lexer.rl"
		begin
 tm = p 		end
	when 34 then
# line 1421 "lib/parser/lexer.rl"
		begin
 tm = p 		end
	when 36 then
# line 1437 "lib/parser/lexer.rl"
		begin
 tm = p 		end
	when 38 then
# line 1465 "lib/parser/lexer.rl"
		begin
 tm = p 		end
	when 53 then
# line 1651 "lib/parser/lexer.rl"
		begin
 heredoc_e      = p 		end
	when 306 then
# line 1705 "lib/parser/lexer.rl"
		begin
 @escape = nil 		end
	when 335 then
# line 1778 "lib/parser/lexer.rl"
		begin
 tm = p 		end
	when 411 then
# line 1996 "lib/parser/lexer.rl"
		begin
 @num_base = 16; @num_digits_s = p 		end
	when 405 then
# line 1997 "lib/parser/lexer.rl"
		begin
 @num_base = 10; @num_digits_s = p 		end
	when 408 then
# line 1998 "lib/parser/lexer.rl"
		begin
 @num_base = 8;  @num_digits_s = p 		end
	when 402 then
# line 1999 "lib/parser/lexer.rl"
		begin
 @num_base = 2;  @num_digits_s = p 		end
	when 417 then
# line 2000 "lib/parser/lexer.rl"
		begin
 @num_base = 10; @num_digits_s = @ts 		end
	when 385 then
# line 2001 "lib/parser/lexer.rl"
		begin
 @num_base = 8;  @num_digits_s = @ts 		end
	when 397 then
# line 2002 "lib/parser/lexer.rl"
		begin
 @num_suffix_s = p 		end
	when 392 then
# line 2059 "lib/parser/lexer.rl"
		begin
 @num_suffix_s = p 		end
	when 390 then
# line 2060 "lib/parser/lexer.rl"
		begin
 @num_suffix_s = p 		end
	when 74 then
# line 2195 "lib/parser/lexer.rl"
		begin
 tm = p 		end
	when 7 then
# line 1 "NONE"
		begin
 @te = p+1
		end
	when 94 then
# line 1007 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
    current_literal = literal
    current_literal.flush_string
    current_literal.extend_content

    emit(:tSTRING_DBEG, '#{'.freeze)

    if current_literal.heredoc?
      current_literal.saved_herebody_s = @herebody_s
      @herebody_s = nil
    end

    current_literal.start_interp_brace
    	begin
		 @stack[ @top] =  @cs
		 @top+= 1
		 @cs = 746
		_goto_level = _again
		next
	end

   end
		end
	when 5 then
# line 952 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
    current_literal = literal
    current_literal.flush_string
    current_literal.extend_content

    emit(:tSTRING_DVAR, nil, @ts, @ts + 1)

    p = @ts
    	begin
		 @stack[ @top] =  @cs
		 @top+= 1
		 @cs = 310
		_goto_level = _again
		next
	end

   end
		end
	when 90 then
# line 877 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
    current_literal = literal
    if @te == pe
      diagnostic :fatal, :string_eof, nil,
                 range(current_literal.str_s, current_literal.str_s + 1)
    end

    if current_literal.heredoc?
      line = tok(@herebody_s, @ts).gsub(/\r+$/, ''.freeze)

      if version?(18, 19, 20)
        # See ruby:c48b4209c
        line = line.gsub(/\r.*$/, ''.freeze)
      end

      # Try ending the heredoc with the complete most recently
      # scanned line. @herebody_s always refers to the start of such line.
      if current_literal.nest_and_try_closing(line, @herebody_s, @ts)
        # Adjust @herebody_s to point to the next line.
        @herebody_s = @te

        # Continue regular lexing after the heredoc reference (<<END).
        p = current_literal.heredoc_e - 1
         @cs = (pop_literal); 	begin
		p += 1
		_goto_level = _out
		next
	end

      else
        # Calculate indentation level for <<~HEREDOCs.
        current_literal.infer_indent_level(line)

        # Ditto.
        @herebody_s = @te
      end
    else
      # Try ending the literal with a newline.
      if current_literal.nest_and_try_closing(tok, @ts, @te)
         @cs = (pop_literal); 	begin
		p += 1
		_goto_level = _out
		next
	end

      end

      if @herebody_s
        # This is a regular literal intertwined with a heredoc. Like:
        #
        #     p <<-foo+"1
        #     bar
        #     foo
        #     2"
        #
        # which, incidentally, evaluates to "bar\n1\n2".
        p = @herebody_s - 1
        @herebody_s = nil
      end
    end

    if current_literal.words? && !eof_codepoint?(@source_pts[p])
      current_literal.extend_space @ts, @te
    else
      # A literal newline is appended if the heredoc was _not_ closed
      # this time (see fbreak above). See also Literal#nest_and_try_closing
      # for rationale of calling #flush_string here.
      current_literal.extend_string tok, @ts, @te
      current_literal.flush_string
    end
   end
		end
	when 89 then
# line 815 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
    string = tok

    # tLABEL_END is only possible in non-cond context on >= 2.2
    if @version >= 22 && !@cond.active?
      lookahead = @source_buffer.slice(@te...@te+2)
    end

    current_literal = literal
    if !current_literal.heredoc? &&
          (token = current_literal.nest_and_try_closing(string, @ts, @te, lookahead))
      if token[0] == :tLABEL_END
        p += 1
        pop_literal
         @cs = 739;
      else
         @cs = (pop_literal);
      end
      	begin
		p += 1
		_goto_level = _out
		next
	end

    else
      current_literal.extend_string(string, @ts, @te)
    end
   end
		end
	when 95 then
# line 952 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    current_literal = literal
    current_literal.flush_string
    current_literal.extend_content

    emit(:tSTRING_DVAR, nil, @ts, @ts + 1)

    p = @ts
    	begin
		 @stack[ @top] =  @cs
		 @top+= 1
		 @cs = 310
		_goto_level = _again
		next
	end

   end
		end
	when 92 then
# line 939 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    literal.extend_space @ts, @te
   end
		end
	when 93 then
# line 815 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    string = tok

    # tLABEL_END is only possible in non-cond context on >= 2.2
    if @version >= 22 && !@cond.active?
      lookahead = @source_buffer.slice(@te...@te+2)
    end

    current_literal = literal
    if !current_literal.heredoc? &&
          (token = current_literal.nest_and_try_closing(string, @ts, @te, lookahead))
      if token[0] == :tLABEL_END
        p += 1
        pop_literal
         @cs = 739;
      else
         @cs = (pop_literal);
      end
      	begin
		p += 1
		_goto_level = _out
		next
	end

    else
      current_literal.extend_string(string, @ts, @te)
    end
   end
		end
	when 6 then
# line 839 "lib/parser/lexer.rl"
		begin
 begin p = (( @te))-1; end
 begin 
    current_literal = literal
    # Get the first character after the backslash.
    escaped_char = @source_buffer.slice(@escape_s).chr

    if current_literal.munge_escape? escaped_char
      # If this particular literal uses this character as an opening
      # or closing delimiter, it is an escape sequence for that
      # particular character. Write it without the backslash.

      if current_literal.regexp? && REGEXP_META_CHARACTERS.match(escaped_char)
        # Regular expressions should include escaped delimiters in their
        # escaped form, except when the escaped character is
        # a closing delimiter but not a regexp metacharacter.
        #
        # The backslash itself cannot be used as a closing delimiter
        # at the same time as an escape symbol, but it is always munged,
        # so this branch also executes for the non-closing-delimiter case
        # for the backslash.
        current_literal.extend_string(tok, @ts, @te)
      else
        current_literal.extend_string(escaped_char, @ts, @te)
      end
    else
      # It does not. So this is an actual escape sequence, yay!
      if current_literal.regexp?
        # Regular expressions should include escape sequences in their
        # escaped form. On the other hand, escaped newlines are removed.
        current_literal.extend_string(tok.gsub("\\\n".freeze, ''.freeze), @ts, @te)
      else
        current_literal.extend_string(@escape || tok, @ts, @te)
      end
    end
   end
		end
	when 4 then
# line 815 "lib/parser/lexer.rl"
		begin
 begin p = (( @te))-1; end
 begin 
    string = tok

    # tLABEL_END is only possible in non-cond context on >= 2.2
    if @version >= 22 && !@cond.active?
      lookahead = @source_buffer.slice(@te...@te+2)
    end

    current_literal = literal
    if !current_literal.heredoc? &&
          (token = current_literal.nest_and_try_closing(string, @ts, @te, lookahead))
      if token[0] == :tLABEL_END
        p += 1
        pop_literal
         @cs = 739;
      else
         @cs = (pop_literal);
      end
      	begin
		p += 1
		_goto_level = _out
		next
	end

    else
      current_literal.extend_string(string, @ts, @te)
    end
   end
		end
	when 121 then
# line 1007 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
    current_literal = literal
    current_literal.flush_string
    current_literal.extend_content

    emit(:tSTRING_DBEG, '#{'.freeze)

    if current_literal.heredoc?
      current_literal.saved_herebody_s = @herebody_s
      @herebody_s = nil
    end

    current_literal.start_interp_brace
    	begin
		 @stack[ @top] =  @cs
		 @top+= 1
		 @cs = 746
		_goto_level = _again
		next
	end

   end
		end
	when 10 then
# line 952 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
    current_literal = literal
    current_literal.flush_string
    current_literal.extend_content

    emit(:tSTRING_DVAR, nil, @ts, @ts + 1)

    p = @ts
    	begin
		 @stack[ @top] =  @cs
		 @top+= 1
		 @cs = 310
		_goto_level = _again
		next
	end

   end
		end
	when 118 then
# line 877 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
    current_literal = literal
    if @te == pe
      diagnostic :fatal, :string_eof, nil,
                 range(current_literal.str_s, current_literal.str_s + 1)
    end

    if current_literal.heredoc?
      line = tok(@herebody_s, @ts).gsub(/\r+$/, ''.freeze)

      if version?(18, 19, 20)
        # See ruby:c48b4209c
        line = line.gsub(/\r.*$/, ''.freeze)
      end

      # Try ending the heredoc with the complete most recently
      # scanned line. @herebody_s always refers to the start of such line.
      if current_literal.nest_and_try_closing(line, @herebody_s, @ts)
        # Adjust @herebody_s to point to the next line.
        @herebody_s = @te

        # Continue regular lexing after the heredoc reference (<<END).
        p = current_literal.heredoc_e - 1
         @cs = (pop_literal); 	begin
		p += 1
		_goto_level = _out
		next
	end

      else
        # Calculate indentation level for <<~HEREDOCs.
        current_literal.infer_indent_level(line)

        # Ditto.
        @herebody_s = @te
      end
    else
      # Try ending the literal with a newline.
      if current_literal.nest_and_try_closing(tok, @ts, @te)
         @cs = (pop_literal); 	begin
		p += 1
		_goto_level = _out
		next
	end

      end

      if @herebody_s
        # This is a regular literal intertwined with a heredoc. Like:
        #
        #     p <<-foo+"1
        #     bar
        #     foo
        #     2"
        #
        # which, incidentally, evaluates to "bar\n1\n2".
        p = @herebody_s - 1
        @herebody_s = nil
      end
    end

    if current_literal.words? && !eof_codepoint?(@source_pts[p])
      current_literal.extend_space @ts, @te
    else
      # A literal newline is appended if the heredoc was _not_ closed
      # this time (see fbreak above). See also Literal#nest_and_try_closing
      # for rationale of calling #flush_string here.
      current_literal.extend_string tok, @ts, @te
      current_literal.flush_string
    end
   end
		end
	when 117 then
# line 815 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
    string = tok

    # tLABEL_END is only possible in non-cond context on >= 2.2
    if @version >= 22 && !@cond.active?
      lookahead = @source_buffer.slice(@te...@te+2)
    end

    current_literal = literal
    if !current_literal.heredoc? &&
          (token = current_literal.nest_and_try_closing(string, @ts, @te, lookahead))
      if token[0] == :tLABEL_END
        p += 1
        pop_literal
         @cs = 739;
      else
         @cs = (pop_literal);
      end
      	begin
		p += 1
		_goto_level = _out
		next
	end

    else
      current_literal.extend_string(string, @ts, @te)
    end
   end
		end
	when 122 then
# line 952 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    current_literal = literal
    current_literal.flush_string
    current_literal.extend_content

    emit(:tSTRING_DVAR, nil, @ts, @ts + 1)

    p = @ts
    	begin
		 @stack[ @top] =  @cs
		 @top+= 1
		 @cs = 310
		_goto_level = _again
		next
	end

   end
		end
	when 120 then
# line 815 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    string = tok

    # tLABEL_END is only possible in non-cond context on >= 2.2
    if @version >= 22 && !@cond.active?
      lookahead = @source_buffer.slice(@te...@te+2)
    end

    current_literal = literal
    if !current_literal.heredoc? &&
          (token = current_literal.nest_and_try_closing(string, @ts, @te, lookahead))
      if token[0] == :tLABEL_END
        p += 1
        pop_literal
         @cs = 739;
      else
         @cs = (pop_literal);
      end
      	begin
		p += 1
		_goto_level = _out
		next
	end

    else
      current_literal.extend_string(string, @ts, @te)
    end
   end
		end
	when 11 then
# line 839 "lib/parser/lexer.rl"
		begin
 begin p = (( @te))-1; end
 begin 
    current_literal = literal
    # Get the first character after the backslash.
    escaped_char = @source_buffer.slice(@escape_s).chr

    if current_literal.munge_escape? escaped_char
      # If this particular literal uses this character as an opening
      # or closing delimiter, it is an escape sequence for that
      # particular character. Write it without the backslash.

      if current_literal.regexp? && REGEXP_META_CHARACTERS.match(escaped_char)
        # Regular expressions should include escaped delimiters in their
        # escaped form, except when the escaped character is
        # a closing delimiter but not a regexp metacharacter.
        #
        # The backslash itself cannot be used as a closing delimiter
        # at the same time as an escape symbol, but it is always munged,
        # so this branch also executes for the non-closing-delimiter case
        # for the backslash.
        current_literal.extend_string(tok, @ts, @te)
      else
        current_literal.extend_string(escaped_char, @ts, @te)
      end
    else
      # It does not. So this is an actual escape sequence, yay!
      if current_literal.regexp?
        # Regular expressions should include escape sequences in their
        # escaped form. On the other hand, escaped newlines are removed.
        current_literal.extend_string(tok.gsub("\\\n".freeze, ''.freeze), @ts, @te)
      else
        current_literal.extend_string(@escape || tok, @ts, @te)
      end
    end
   end
		end
	when 9 then
# line 815 "lib/parser/lexer.rl"
		begin
 begin p = (( @te))-1; end
 begin 
    string = tok

    # tLABEL_END is only possible in non-cond context on >= 2.2
    if @version >= 22 && !@cond.active?
      lookahead = @source_buffer.slice(@te...@te+2)
    end

    current_literal = literal
    if !current_literal.heredoc? &&
          (token = current_literal.nest_and_try_closing(string, @ts, @te, lookahead))
      if token[0] == :tLABEL_END
        p += 1
        pop_literal
         @cs = 739;
      else
         @cs = (pop_literal);
      end
      	begin
		p += 1
		_goto_level = _out
		next
	end

    else
      current_literal.extend_string(string, @ts, @te)
    end
   end
		end
	when 144 then
# line 877 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
    current_literal = literal
    if @te == pe
      diagnostic :fatal, :string_eof, nil,
                 range(current_literal.str_s, current_literal.str_s + 1)
    end

    if current_literal.heredoc?
      line = tok(@herebody_s, @ts).gsub(/\r+$/, ''.freeze)

      if version?(18, 19, 20)
        # See ruby:c48b4209c
        line = line.gsub(/\r.*$/, ''.freeze)
      end

      # Try ending the heredoc with the complete most recently
      # scanned line. @herebody_s always refers to the start of such line.
      if current_literal.nest_and_try_closing(line, @herebody_s, @ts)
        # Adjust @herebody_s to point to the next line.
        @herebody_s = @te

        # Continue regular lexing after the heredoc reference (<<END).
        p = current_literal.heredoc_e - 1
         @cs = (pop_literal); 	begin
		p += 1
		_goto_level = _out
		next
	end

      else
        # Calculate indentation level for <<~HEREDOCs.
        current_literal.infer_indent_level(line)

        # Ditto.
        @herebody_s = @te
      end
    else
      # Try ending the literal with a newline.
      if current_literal.nest_and_try_closing(tok, @ts, @te)
         @cs = (pop_literal); 	begin
		p += 1
		_goto_level = _out
		next
	end

      end

      if @herebody_s
        # This is a regular literal intertwined with a heredoc. Like:
        #
        #     p <<-foo+"1
        #     bar
        #     foo
        #     2"
        #
        # which, incidentally, evaluates to "bar\n1\n2".
        p = @herebody_s - 1
        @herebody_s = nil
      end
    end

    if current_literal.words? && !eof_codepoint?(@source_pts[p])
      current_literal.extend_space @ts, @te
    else
      # A literal newline is appended if the heredoc was _not_ closed
      # this time (see fbreak above). See also Literal#nest_and_try_closing
      # for rationale of calling #flush_string here.
      current_literal.extend_string tok, @ts, @te
      current_literal.flush_string
    end
   end
		end
	when 143 then
# line 815 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
    string = tok

    # tLABEL_END is only possible in non-cond context on >= 2.2
    if @version >= 22 && !@cond.active?
      lookahead = @source_buffer.slice(@te...@te+2)
    end

    current_literal = literal
    if !current_literal.heredoc? &&
          (token = current_literal.nest_and_try_closing(string, @ts, @te, lookahead))
      if token[0] == :tLABEL_END
        p += 1
        pop_literal
         @cs = 739;
      else
         @cs = (pop_literal);
      end
      	begin
		p += 1
		_goto_level = _out
		next
	end

    else
      current_literal.extend_string(string, @ts, @te)
    end
   end
		end
	when 146 then
# line 939 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    literal.extend_space @ts, @te
   end
		end
	when 147 then
# line 815 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    string = tok

    # tLABEL_END is only possible in non-cond context on >= 2.2
    if @version >= 22 && !@cond.active?
      lookahead = @source_buffer.slice(@te...@te+2)
    end

    current_literal = literal
    if !current_literal.heredoc? &&
          (token = current_literal.nest_and_try_closing(string, @ts, @te, lookahead))
      if token[0] == :tLABEL_END
        p += 1
        pop_literal
         @cs = 739;
      else
         @cs = (pop_literal);
      end
      	begin
		p += 1
		_goto_level = _out
		next
	end

    else
      current_literal.extend_string(string, @ts, @te)
    end
   end
		end
	when 150 then
# line 877 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
    current_literal = literal
    if @te == pe
      diagnostic :fatal, :string_eof, nil,
                 range(current_literal.str_s, current_literal.str_s + 1)
    end

    if current_literal.heredoc?
      line = tok(@herebody_s, @ts).gsub(/\r+$/, ''.freeze)

      if version?(18, 19, 20)
        # See ruby:c48b4209c
        line = line.gsub(/\r.*$/, ''.freeze)
      end

      # Try ending the heredoc with the complete most recently
      # scanned line. @herebody_s always refers to the start of such line.
      if current_literal.nest_and_try_closing(line, @herebody_s, @ts)
        # Adjust @herebody_s to point to the next line.
        @herebody_s = @te

        # Continue regular lexing after the heredoc reference (<<END).
        p = current_literal.heredoc_e - 1
         @cs = (pop_literal); 	begin
		p += 1
		_goto_level = _out
		next
	end

      else
        # Calculate indentation level for <<~HEREDOCs.
        current_literal.infer_indent_level(line)

        # Ditto.
        @herebody_s = @te
      end
    else
      # Try ending the literal with a newline.
      if current_literal.nest_and_try_closing(tok, @ts, @te)
         @cs = (pop_literal); 	begin
		p += 1
		_goto_level = _out
		next
	end

      end

      if @herebody_s
        # This is a regular literal intertwined with a heredoc. Like:
        #
        #     p <<-foo+"1
        #     bar
        #     foo
        #     2"
        #
        # which, incidentally, evaluates to "bar\n1\n2".
        p = @herebody_s - 1
        @herebody_s = nil
      end
    end

    if current_literal.words? && !eof_codepoint?(@source_pts[p])
      current_literal.extend_space @ts, @te
    else
      # A literal newline is appended if the heredoc was _not_ closed
      # this time (see fbreak above). See also Literal#nest_and_try_closing
      # for rationale of calling #flush_string here.
      current_literal.extend_string tok, @ts, @te
      current_literal.flush_string
    end
   end
		end
	when 149 then
# line 815 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
    string = tok

    # tLABEL_END is only possible in non-cond context on >= 2.2
    if @version >= 22 && !@cond.active?
      lookahead = @source_buffer.slice(@te...@te+2)
    end

    current_literal = literal
    if !current_literal.heredoc? &&
          (token = current_literal.nest_and_try_closing(string, @ts, @te, lookahead))
      if token[0] == :tLABEL_END
        p += 1
        pop_literal
         @cs = 739;
      else
         @cs = (pop_literal);
      end
      	begin
		p += 1
		_goto_level = _out
		next
	end

    else
      current_literal.extend_string(string, @ts, @te)
    end
   end
		end
	when 152 then
# line 815 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    string = tok

    # tLABEL_END is only possible in non-cond context on >= 2.2
    if @version >= 22 && !@cond.active?
      lookahead = @source_buffer.slice(@te...@te+2)
    end

    current_literal = literal
    if !current_literal.heredoc? &&
          (token = current_literal.nest_and_try_closing(string, @ts, @te, lookahead))
      if token[0] == :tLABEL_END
        p += 1
        pop_literal
         @cs = 739;
      else
         @cs = (pop_literal);
      end
      	begin
		p += 1
		_goto_level = _out
		next
	end

    else
      current_literal.extend_string(string, @ts, @te)
    end
   end
		end
	when 159 then
# line 1007 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
    current_literal = literal
    current_literal.flush_string
    current_literal.extend_content

    emit(:tSTRING_DBEG, '#{'.freeze)

    if current_literal.heredoc?
      current_literal.saved_herebody_s = @herebody_s
      @herebody_s = nil
    end

    current_literal.start_interp_brace
    	begin
		 @stack[ @top] =  @cs
		 @top+= 1
		 @cs = 746
		_goto_level = _again
		next
	end

   end
		end
	when 13 then
# line 952 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
    current_literal = literal
    current_literal.flush_string
    current_literal.extend_content

    emit(:tSTRING_DVAR, nil, @ts, @ts + 1)

    p = @ts
    	begin
		 @stack[ @top] =  @cs
		 @top+= 1
		 @cs = 310
		_goto_level = _again
		next
	end

   end
		end
	when 156 then
# line 877 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
    current_literal = literal
    if @te == pe
      diagnostic :fatal, :string_eof, nil,
                 range(current_literal.str_s, current_literal.str_s + 1)
    end

    if current_literal.heredoc?
      line = tok(@herebody_s, @ts).gsub(/\r+$/, ''.freeze)

      if version?(18, 19, 20)
        # See ruby:c48b4209c
        line = line.gsub(/\r.*$/, ''.freeze)
      end

      # Try ending the heredoc with the complete most recently
      # scanned line. @herebody_s always refers to the start of such line.
      if current_literal.nest_and_try_closing(line, @herebody_s, @ts)
        # Adjust @herebody_s to point to the next line.
        @herebody_s = @te

        # Continue regular lexing after the heredoc reference (<<END).
        p = current_literal.heredoc_e - 1
         @cs = (pop_literal); 	begin
		p += 1
		_goto_level = _out
		next
	end

      else
        # Calculate indentation level for <<~HEREDOCs.
        current_literal.infer_indent_level(line)

        # Ditto.
        @herebody_s = @te
      end
    else
      # Try ending the literal with a newline.
      if current_literal.nest_and_try_closing(tok, @ts, @te)
         @cs = (pop_literal); 	begin
		p += 1
		_goto_level = _out
		next
	end

      end

      if @herebody_s
        # This is a regular literal intertwined with a heredoc. Like:
        #
        #     p <<-foo+"1
        #     bar
        #     foo
        #     2"
        #
        # which, incidentally, evaluates to "bar\n1\n2".
        p = @herebody_s - 1
        @herebody_s = nil
      end
    end

    if current_literal.words? && !eof_codepoint?(@source_pts[p])
      current_literal.extend_space @ts, @te
    else
      # A literal newline is appended if the heredoc was _not_ closed
      # this time (see fbreak above). See also Literal#nest_and_try_closing
      # for rationale of calling #flush_string here.
      current_literal.extend_string tok, @ts, @te
      current_literal.flush_string
    end
   end
		end
	when 155 then
# line 815 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
    string = tok

    # tLABEL_END is only possible in non-cond context on >= 2.2
    if @version >= 22 && !@cond.active?
      lookahead = @source_buffer.slice(@te...@te+2)
    end

    current_literal = literal
    if !current_literal.heredoc? &&
          (token = current_literal.nest_and_try_closing(string, @ts, @te, lookahead))
      if token[0] == :tLABEL_END
        p += 1
        pop_literal
         @cs = 739;
      else
         @cs = (pop_literal);
      end
      	begin
		p += 1
		_goto_level = _out
		next
	end

    else
      current_literal.extend_string(string, @ts, @te)
    end
   end
		end
	when 160 then
# line 952 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    current_literal = literal
    current_literal.flush_string
    current_literal.extend_content

    emit(:tSTRING_DVAR, nil, @ts, @ts + 1)

    p = @ts
    	begin
		 @stack[ @top] =  @cs
		 @top+= 1
		 @cs = 310
		_goto_level = _again
		next
	end

   end
		end
	when 158 then
# line 815 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    string = tok

    # tLABEL_END is only possible in non-cond context on >= 2.2
    if @version >= 22 && !@cond.active?
      lookahead = @source_buffer.slice(@te...@te+2)
    end

    current_literal = literal
    if !current_literal.heredoc? &&
          (token = current_literal.nest_and_try_closing(string, @ts, @te, lookahead))
      if token[0] == :tLABEL_END
        p += 1
        pop_literal
         @cs = 739;
      else
         @cs = (pop_literal);
      end
      	begin
		p += 1
		_goto_level = _out
		next
	end

    else
      current_literal.extend_string(string, @ts, @te)
    end
   end
		end
	when 12 then
# line 815 "lib/parser/lexer.rl"
		begin
 begin p = (( @te))-1; end
 begin 
    string = tok

    # tLABEL_END is only possible in non-cond context on >= 2.2
    if @version >= 22 && !@cond.active?
      lookahead = @source_buffer.slice(@te...@te+2)
    end

    current_literal = literal
    if !current_literal.heredoc? &&
          (token = current_literal.nest_and_try_closing(string, @ts, @te, lookahead))
      if token[0] == :tLABEL_END
        p += 1
        pop_literal
         @cs = 739;
      else
         @cs = (pop_literal);
      end
      	begin
		p += 1
		_goto_level = _out
		next
	end

    else
      current_literal.extend_string(string, @ts, @te)
    end
   end
		end
	when 162 then
# line 877 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
    current_literal = literal
    if @te == pe
      diagnostic :fatal, :string_eof, nil,
                 range(current_literal.str_s, current_literal.str_s + 1)
    end

    if current_literal.heredoc?
      line = tok(@herebody_s, @ts).gsub(/\r+$/, ''.freeze)

      if version?(18, 19, 20)
        # See ruby:c48b4209c
        line = line.gsub(/\r.*$/, ''.freeze)
      end

      # Try ending the heredoc with the complete most recently
      # scanned line. @herebody_s always refers to the start of such line.
      if current_literal.nest_and_try_closing(line, @herebody_s, @ts)
        # Adjust @herebody_s to point to the next line.
        @herebody_s = @te

        # Continue regular lexing after the heredoc reference (<<END).
        p = current_literal.heredoc_e - 1
         @cs = (pop_literal); 	begin
		p += 1
		_goto_level = _out
		next
	end

      else
        # Calculate indentation level for <<~HEREDOCs.
        current_literal.infer_indent_level(line)

        # Ditto.
        @herebody_s = @te
      end
    else
      # Try ending the literal with a newline.
      if current_literal.nest_and_try_closing(tok, @ts, @te)
         @cs = (pop_literal); 	begin
		p += 1
		_goto_level = _out
		next
	end

      end

      if @herebody_s
        # This is a regular literal intertwined with a heredoc. Like:
        #
        #     p <<-foo+"1
        #     bar
        #     foo
        #     2"
        #
        # which, incidentally, evaluates to "bar\n1\n2".
        p = @herebody_s - 1
        @herebody_s = nil
      end
    end

    if current_literal.words? && !eof_codepoint?(@source_pts[p])
      current_literal.extend_space @ts, @te
    else
      # A literal newline is appended if the heredoc was _not_ closed
      # this time (see fbreak above). See also Literal#nest_and_try_closing
      # for rationale of calling #flush_string here.
      current_literal.extend_string tok, @ts, @te
      current_literal.flush_string
    end
   end
		end
	when 161 then
# line 815 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
    string = tok

    # tLABEL_END is only possible in non-cond context on >= 2.2
    if @version >= 22 && !@cond.active?
      lookahead = @source_buffer.slice(@te...@te+2)
    end

    current_literal = literal
    if !current_literal.heredoc? &&
          (token = current_literal.nest_and_try_closing(string, @ts, @te, lookahead))
      if token[0] == :tLABEL_END
        p += 1
        pop_literal
         @cs = 739;
      else
         @cs = (pop_literal);
      end
      	begin
		p += 1
		_goto_level = _out
		next
	end

    else
      current_literal.extend_string(string, @ts, @te)
    end
   end
		end
	when 164 then
# line 1083 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
        emit(:tREGEXP_OPT, tok(@ts, @te - 1), @ts, @te - 1)
        p = p - 1; 	begin
		 @cs = 754
		_goto_level = _again
		next
	end

       end
		end
	when 165 then
# line 1071 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        unknown_options = tok.scan(/[^imxouesn]/)
        if unknown_options.any?
          diagnostic :error, :regexp_options,
                     { :options => unknown_options.join }
        end

        emit(:tREGEXP_OPT)
         @cs = 754; 	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 14 then
# line 1211 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
        if    tok =~ /^\$([1-9][0-9]*)$/
          emit(:tNTH_REF, tok(@ts + 1).to_i)
        elsif tok =~ /^\$([&`'+])$/
          emit(:tBACK_REF)
        else
          emit(:tGVAR)
        end

         @cs = (stack_pop); 	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 166 then
# line 1211 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        if    tok =~ /^\$([1-9][0-9]*)$/
          emit(:tNTH_REF, tok(@ts + 1).to_i)
        elsif tok =~ /^\$([&`'+])$/
          emit(:tBACK_REF)
        else
          emit(:tGVAR)
        end

         @cs = (stack_pop); 	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 168 then
# line 1224 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        if tok =~ /^@@[0-9]/
          diagnostic :error, :cvar_name, { :name => tok }
        end

        emit(:tCVAR)
         @cs = (stack_pop); 	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 167 then
# line 1234 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        if tok =~ /^@[0-9]/
          diagnostic :error, :ivar_name, { :name => tok }
        end

        emit(:tIVAR)
         @cs = (stack_pop); 	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 189 then
# line 1255 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin  emit_table(KEYWORDS_BEGIN);
            @cs = 428; 	begin
		p += 1
		_goto_level = _out
		next
	end
  end
		end
	when 175 then
# line 1263 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin  emit(:tIDENTIFIER)
            @cs = 428; 	begin
		p += 1
		_goto_level = _out
		next
	end
  end
		end
	when 16 then
# line 1267 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin  p = @ts - 1
            @cs = 754; 	begin
		 @stack[ @top] =  @cs
		 @top+= 1
		 @cs = 310
		_goto_level = _again
		next
	end
  end
		end
	when 172 then
# line 1276 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin  emit_table(PUNCTUATION)
            @cs = 428; 	begin
		p += 1
		_goto_level = _out
		next
	end
  end
		end
	when 184 then
# line 1280 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin  p = p - 1; p = p - 1; 	begin
		 @cs = 754
		_goto_level = _again
		next
	end
  end
		end
	when 18 then
# line 1286 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
        if version?(23)
          type, delimiter = tok[0..-2], tok[-1].chr
          	begin
		 @cs = (push_literal(type, delimiter, @ts))
		_goto_level = _again
		next
	end

        else
          p = @ts - 1
          	begin
		 @cs = 754
		_goto_level = _again
		next
	end

        end
       end
		end
	when 171 then
# line 1299 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin  p = p - 1; 	begin
		 @cs = 754
		_goto_level = _again
		next
	end
  end
		end
	when 170 then
# line 512 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
    # Sit at EOF indefinitely. #advance would return $eof each time.
    # This allows to feed the lexer more data if needed; this is only used
    # in tests.
    #
    # Note that this action is not embedded into e_eof like e_heredoc_nl and e_bs
    # below. This is due to the fact that scanner state at EOF is observed
    # by tests, and encapsulating it in a rule would break the introspection.
    p = p - 1; 	begin
		p += 1
		_goto_level = _out
		next
	end

   end
		end
	when 188 then
# line 1255 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  emit_table(KEYWORDS_BEGIN);
            @cs = 428; 	begin
		p += 1
		_goto_level = _out
		next
	end
  end
		end
	when 185 then
# line 1259 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  emit(:tCONSTANT)
            @cs = 428; 	begin
		p += 1
		_goto_level = _out
		next
	end
  end
		end
	when 187 then
# line 1263 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  emit(:tIDENTIFIER)
            @cs = 428; 	begin
		p += 1
		_goto_level = _out
		next
	end
  end
		end
	when 182 then
# line 1267 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  p = @ts - 1
            @cs = 754; 	begin
		 @stack[ @top] =  @cs
		 @top+= 1
		 @cs = 310
		_goto_level = _again
		next
	end
  end
		end
	when 178 then
# line 1276 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  emit_table(PUNCTUATION)
            @cs = 428; 	begin
		p += 1
		_goto_level = _out
		next
	end
  end
		end
	when 183 then
# line 1283 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  p = p - 1; 	begin
		 @cs = 525
		_goto_level = _again
		next
	end
  end
		end
	when 176 then
# line 1296 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1;		end
	when 181 then
# line 1299 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  p = p - 1; 	begin
		 @cs = 754
		_goto_level = _again
		next
	end
  end
		end
	when 17 then
# line 1276 "lib/parser/lexer.rl"
		begin
 begin p = (( @te))-1; end
 begin  emit_table(PUNCTUATION)
            @cs = 428; 	begin
		p += 1
		_goto_level = _out
		next
	end
  end
		end
	when 15 then
# line 1299 "lib/parser/lexer.rl"
		begin
 begin p = (( @te))-1; end
 begin  p = p - 1; 	begin
		 @cs = 754
		_goto_level = _again
		next
	end
  end
		end
	when 174 then
# line 1 "NONE"
		begin
	case  @act
	when 31 then
	begin begin p = (( @te))-1; end
 emit_table(KEYWORDS_BEGIN);
            @cs = 428; 	begin
		p += 1
		_goto_level = _out
		next
	end
 end
	when 32 then
	begin begin p = (( @te))-1; end
 emit(:tCONSTANT)
            @cs = 428; 	begin
		p += 1
		_goto_level = _out
		next
	end
 end
	when 33 then
	begin begin p = (( @te))-1; end
 emit(:tIDENTIFIER)
            @cs = 428; 	begin
		p += 1
		_goto_level = _out
		next
	end
 end
end 
			end
	when 20 then
# line 1311 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin  emit(:tLABEL, tok(@ts, @te - 1))
            @cs = 739; 	begin
		p += 1
		_goto_level = _out
		next
	end
  end
		end
	when 191 then
# line 1317 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin  p = p - 1; 	begin
		 @cs = 754
		_goto_level = _again
		next
	end
  end
		end
	when 190 then
# line 512 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
    # Sit at EOF indefinitely. #advance would return $eof each time.
    # This allows to feed the lexer more data if needed; this is only used
    # in tests.
    #
    # Note that this action is not embedded into e_eof like e_heredoc_nl and e_bs
    # below. This is due to the fact that scanner state at EOF is observed
    # by tests, and encapsulating it in a rule would break the introspection.
    p = p - 1; 	begin
		p += 1
		_goto_level = _out
		next
	end

   end
		end
	when 193 then
# line 1314 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1;		end
	when 192 then
# line 1317 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  p = p - 1; 	begin
		 @cs = 754
		_goto_level = _again
		next
	end
  end
		end
	when 19 then
# line 1317 "lib/parser/lexer.rl"
		begin
 begin p = (( @te))-1; end
 begin  p = p - 1; 	begin
		 @cs = 754
		_goto_level = _again
		next
	end
  end
		end
	when 199 then
# line 1343 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin  emit_table(PUNCTUATION)
            @cs = 456; 	begin
		p += 1
		_goto_level = _out
		next
	end
  end
		end
	when 198 then
# line 1349 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin  p = p - 1; 	begin
		 @cs = 754
		_goto_level = _again
		next
	end
  end
		end
	when 197 then
# line 512 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
    # Sit at EOF indefinitely. #advance would return $eof each time.
    # This allows to feed the lexer more data if needed; this is only used
    # in tests.
    #
    # Note that this action is not embedded into e_eof like e_heredoc_nl and e_bs
    # below. This is due to the fact that scanner state at EOF is observed
    # by tests, and encapsulating it in a rule would break the introspection.
    p = p - 1; 	begin
		p += 1
		_goto_level = _out
		next
	end

   end
		end
	when 209 then
# line 1328 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  emit(:tCONSTANT)
            @cs = (arg_or_cmdarg); 	begin
		p += 1
		_goto_level = _out
		next
	end
  end
		end
	when 200 then
# line 1332 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  emit(:tIDENTIFIER)
            @cs = (arg_or_cmdarg); 	begin
		p += 1
		_goto_level = _out
		next
	end
  end
		end
	when 205 then
# line 1343 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  emit_table(PUNCTUATION)
            @cs = 456; 	begin
		p += 1
		_goto_level = _out
		next
	end
  end
		end
	when 203 then
# line 1346 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1;		end
	when 208 then
# line 1349 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  p = p - 1; 	begin
		 @cs = 754
		_goto_level = _again
		next
	end
  end
		end
	when 232 then
# line 1407 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
        # Unlike expr_beg as invoked in the next rule, do not warn
        p = @ts - 1
        	begin
		 @cs = 754
		_goto_level = _again
		next
	end

       end
		end
	when 215 then
# line 1425 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
        if tok(tm, tm + 1) == '/'.freeze
          # Ambiguous regexp literal.
          diagnostic :warning, :ambiguous_literal, nil, range(tm, tm + 1)
        end

        p = tm - 1
        	begin
		 @cs = 525
		_goto_level = _again
		next
	end

       end
		end
	when 221 then
# line 1449 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin  p = p - 1; p = p - 1; 	begin
		 @cs = 525
		_goto_level = _again
		next
	end
  end
		end
	when 22 then
# line 1457 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin  p = @ts - 1; 	begin
		 @cs = 525
		_goto_level = _again
		next
	end
  end
		end
	when 223 then
# line 1466 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin  p = tm - 1; 	begin
		 @cs = 754
		_goto_level = _again
		next
	end
  end
		end
	when 37 then
# line 1475 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
        p = @ts - 1
        	begin
		 @cs = 754
		_goto_level = _again
		next
	end

       end
		end
	when 210 then
# line 1489 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin  p = p - 1; 	begin
		 @cs = 525
		_goto_level = _again
		next
	end
  end
		end
	when 211 then
# line 512 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
    # Sit at EOF indefinitely. #advance would return $eof each time.
    # This allows to feed the lexer more data if needed; this is only used
    # in tests.
    #
    # Note that this action is not embedded into e_eof like e_heredoc_nl and e_bs
    # below. This is due to the fact that scanner state at EOF is observed
    # by tests, and encapsulating it in a rule would break the introspection.
    p = p - 1; 	begin
		p += 1
		_goto_level = _out
		next
	end

   end
		end
	when 222 then
# line 1416 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  p = p - 1; 	begin
		 @cs = 525
		_goto_level = _again
		next
	end
  end
		end
	when 218 then
# line 1438 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        diagnostic :warning, :ambiguous_prefix, { :prefix => tok(tm, @te) },
                   range(tm, @te)

        p = tm - 1
        	begin
		 @cs = 525
		_goto_level = _again
		next
	end

       end
		end
	when 220 then
# line 1454 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  p = p - 1; 	begin
		 @cs = 525
		_goto_level = _again
		next
	end
  end
		end
	when 214 then
# line 1475 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        p = @ts - 1
        	begin
		 @cs = 754
		_goto_level = _again
		next
	end

       end
		end
	when 213 then
# line 1480 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1;		end
	when 231 then
# line 1489 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  p = p - 1; 	begin
		 @cs = 525
		_goto_level = _again
		next
	end
  end
		end
	when 23 then
# line 1480 "lib/parser/lexer.rl"
		begin
 begin p = (( @te))-1; end
		end
	when 39 then
# line 1489 "lib/parser/lexer.rl"
		begin
 begin p = (( @te))-1; end
 begin  p = p - 1; 	begin
		 @cs = 525
		_goto_level = _again
		next
	end
  end
		end
	when 21 then
# line 1 "NONE"
		begin
	case  @act
	when 59 then
	begin begin p = (( @te))-1; end

        if tok(tm, tm + 1) == '/'.freeze
          # Ambiguous regexp literal.
          diagnostic :warning, :ambiguous_literal, nil, range(tm, tm + 1)
        end

        p = tm - 1
        	begin
		 @cs = 525
		_goto_level = _again
		next
	end

      end
	when 60 then
	begin begin p = (( @te))-1; end

        diagnostic :warning, :ambiguous_prefix, { :prefix => tok(tm, @te) },
                   range(tm, @te)

        p = tm - 1
        	begin
		 @cs = 525
		_goto_level = _again
		next
	end

      end
	when 65 then
	begin begin p = (( @te))-1; end

        p = @ts - 1
        	begin
		 @cs = 754
		_goto_level = _again
		next
	end

      end
	else
	begin begin p = (( @te))-1; end
end
end 
			end
	when 41 then
# line 1525 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin  p = @ts - 1
           	begin
		 @cs = 456
		_goto_level = _again
		next
	end
  end
		end
	when 236 then
# line 512 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
    # Sit at EOF indefinitely. #advance would return $eof each time.
    # This allows to feed the lexer more data if needed; this is only used
    # in tests.
    #
    # Note that this action is not embedded into e_eof like e_heredoc_nl and e_bs
    # below. This is due to the fact that scanner state at EOF is observed
    # by tests, and encapsulating it in a rule would break the introspection.
    p = p - 1; 	begin
		p += 1
		_goto_level = _out
		next
	end

   end
		end
	when 237 then
# line 1525 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  p = @ts - 1
           	begin
		 @cs = 456
		_goto_level = _again
		next
	end
  end
		end
	when 42 then
# line 1525 "lib/parser/lexer.rl"
		begin
 begin p = (( @te))-1; end
 begin  p = @ts - 1
           	begin
		 @cs = 456
		_goto_level = _again
		next
	end
  end
		end
	when 40 then
# line 1 "NONE"
		begin
	case  @act
	when 72 then
	begin begin p = (( @te))-1; end

        if @cond.active?
          emit(:kDO_COND, 'do'.freeze, @te - 2, @te)
        else
          emit(:kDO, 'do'.freeze, @te - 2, @te)
        end
         @cs = 746; 	begin
		p += 1
		_goto_level = _out
		next
	end

      end
	when 73 then
	begin begin p = (( @te))-1; end
 p = @ts - 1
           	begin
		 @cs = 456
		_goto_level = _again
		next
	end
 end
end 
			end
	when 247 then
# line 1552 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin  emit_do(true)
            @cs = 746; 	begin
		p += 1
		_goto_level = _out
		next
	end
  end
		end
	when 240 then
# line 1558 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin  p = p - 1; 	begin
		 @cs = 754
		_goto_level = _again
		next
	end
  end
		end
	when 241 then
# line 512 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
    # Sit at EOF indefinitely. #advance would return $eof each time.
    # This allows to feed the lexer more data if needed; this is only used
    # in tests.
    #
    # Note that this action is not embedded into e_eof like e_heredoc_nl and e_bs
    # below. This is due to the fact that scanner state at EOF is observed
    # by tests, and encapsulating it in a rule would break the introspection.
    p = p - 1; 	begin
		p += 1
		_goto_level = _out
		next
	end

   end
		end
	when 242 then
# line 1555 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1;		end
	when 245 then
# line 1558 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  p = p - 1; 	begin
		 @cs = 754
		_goto_level = _again
		next
	end
  end
		end
	when 251 then
# line 1582 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin  p = p - 1; 	begin
		 @cs = 525
		_goto_level = _again
		next
	end
  end
		end
	when 250 then
# line 512 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
    # Sit at EOF indefinitely. #advance would return $eof each time.
    # This allows to feed the lexer more data if needed; this is only used
    # in tests.
    #
    # Note that this action is not embedded into e_eof like e_heredoc_nl and e_bs
    # below. This is due to the fact that scanner state at EOF is observed
    # by tests, and encapsulating it in a rule would break the introspection.
    p = p - 1; 	begin
		p += 1
		_goto_level = _out
		next
	end

   end
		end
	when 259 then
# line 1574 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  p = @ts - 1; 	begin
		 @cs = 525
		_goto_level = _again
		next
	end
  end
		end
	when 253 then
# line 1576 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1;		end
	when 257 then
# line 1582 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  p = p - 1; 	begin
		 @cs = 525
		_goto_level = _again
		next
	end
  end
		end
	when 252 then
# line 1 "NONE"
		begin
	case  @act
	when 80 then
	begin begin p = (( @te))-1; end
 emit_table(KEYWORDS)
            @cs = 525; 	begin
		p += 1
		_goto_level = _out
		next
	end
 end
	when 81 then
	begin begin p = (( @te))-1; end
 p = @ts - 1; 	begin
		 @cs = 525
		_goto_level = _again
		next
	end
 end
end 
			end
	when 287 then
# line 1599 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
        p = p - 1;
        if tok.start_with? '-'.freeze
          emit(:tUMINUS_NUM, '-'.freeze, @ts, @ts + 1)
           @cs = 754; 	begin
		p += 1
		_goto_level = _out
		next
	end

        end
       end
		end
	when 288 then
# line 1619 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
        type = delimiter = tok[0].chr
        p = p - 1; 	begin
		 @cs = (push_literal(type, delimiter, @ts))
		_goto_level = _again
		next
	end

       end
		end
	when 282 then
# line 1626 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
        type, delimiter = @source_buffer.slice(@ts).chr, tok[-1].chr
        	begin
		 @cs = (push_literal(type, delimiter, @ts))
		_goto_level = _again
		next
	end

       end
		end
	when 49 then
# line 1633 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
        type, delimiter = tok[0..-2], tok[-1].chr
        	begin
		 @cs = (push_literal(type, delimiter, @ts))
		_goto_level = _again
		next
	end

       end
		end
	when 289 then
# line 1679 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
        type, delimiter = tok, tok[-1].chr
        	begin
		 @cs = (push_literal(type, delimiter, @ts))
		_goto_level = _again
		next
	end

       end
		end
	when 52 then
# line 1693 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
        emit(:tSYMBOL, tok(@ts + 1), @ts)
         @cs = 754; 	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 303 then
# line 1724 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
        escape = { " "  => '\s', "\r" => '\r', "\n" => '\n', "\t" => '\t',
                   "\v" => '\v', "\f" => '\f' }[@source_buffer.slice(@ts + 1)]
        diagnostic :warning, :invalid_escape_use, { :escape => escape }, range

        p = @ts - 1
        	begin
		 @cs = 754
		_goto_level = _again
		next
	end

       end
		end
	when 302 then
# line 1734 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
        diagnostic :fatal, :incomplete_escape, nil, range(@ts, @ts + 1)
       end
		end
	when 290 then
# line 1773 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin  emit_table(PUNCTUATION_BEGIN)
           	begin
		p += 1
		_goto_level = _out
		next
	end
  end
		end
	when 46 then
# line 1793 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
        p = p - 1;

        if version?(18)
          ident = tok(@ts, @te - 2)

          emit((@source_buffer.slice(@ts) =~ /[A-Z]/) ? :tCONSTANT : :tIDENTIFIER,
               ident, @ts, @te - 2)
          p = p - 1; # continue as a symbol

          if !@static_env.nil? && @static_env.declared?(ident)
             @cs = 754;
          else
             @cs = (arg_or_cmdarg);
          end
        else
          emit(:tLABEL, tok(@ts, @te - 2), @ts, @te - 1)
           @cs = 739;
        end

        	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 276 then
# line 1839 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin  p = @ts - 1
           	begin
		 @cs = 160
		_goto_level = _again
		next
	end
  end
		end
	when 50 then
# line 1852 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin  p = @ts - 1; 	begin
		 @cs = 754
		_goto_level = _again
		next
	end
  end
		end
	when 262 then
# line 512 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
    # Sit at EOF indefinitely. #advance would return $eof each time.
    # This allows to feed the lexer more data if needed; this is only used
    # in tests.
    #
    # Note that this action is not embedded into e_eof like e_heredoc_nl and e_bs
    # below. This is due to the fact that scanner state at EOF is observed
    # by tests, and encapsulating it in a rule would break the introspection.
    p = p - 1; 	begin
		p += 1
		_goto_level = _out
		next
	end

   end
		end
	when 286 then
# line 1609 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  emit(:tSTAR, '*'.freeze)
           	begin
		p += 1
		_goto_level = _out
		next
	end
  end
		end
	when 283 then
# line 1633 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        type, delimiter = tok[0..-2], tok[-1].chr
        	begin
		 @cs = (push_literal(type, delimiter, @ts))
		_goto_level = _again
		next
	end

       end
		end
	when 281 then
# line 1639 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        diagnostic :fatal, :string_eof, nil, range(@ts, @ts + 1)
       end
		end
	when 291 then
# line 1693 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        emit(:tSYMBOL, tok(@ts + 1), @ts)
         @cs = 754; 	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 301 then
# line 1734 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        diagnostic :fatal, :incomplete_escape, nil, range(@ts, @ts + 1)
       end
		end
	when 307 then
# line 1740 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        p = @ts - 1
        	begin
		 @cs = 754
		_goto_level = _again
		next
	end

       end
		end
	when 284 then
# line 1773 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  emit_table(PUNCTUATION_BEGIN)
           	begin
		p += 1
		_goto_level = _out
		next
	end
  end
		end
	when 329 then
# line 1196 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    emit(:tIDENTIFIER)

    if !@static_env.nil? && @static_env.declared?(tok)
       @cs = 428; 	begin
		p += 1
		_goto_level = _out
		next
	end

    else
       @cs = (arg_or_cmdarg); 	begin
		p += 1
		_goto_level = _out
		next
	end

    end
   end
		end
	when 273 then
# line 1836 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1;		end
	when 275 then
# line 1839 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  p = @ts - 1
           	begin
		 @cs = 160
		_goto_level = _again
		next
	end
  end
		end
	when 278 then
# line 1852 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  p = @ts - 1; 	begin
		 @cs = 754
		_goto_level = _again
		next
	end
  end
		end
	when 48 then
# line 1639 "lib/parser/lexer.rl"
		begin
 begin p = (( @te))-1; end
 begin 
        diagnostic :fatal, :string_eof, nil, range(@ts, @ts + 1)
       end
		end
	when 55 then
# line 1707 "lib/parser/lexer.rl"
		begin
 begin p = (( @te))-1; end
 begin 
        value = @escape || tok(@ts + 1)

        if version?(18)
          if defined?(Encoding)
            emit(:tINTEGER, value.dup.force_encoding(Encoding::BINARY)[0].ord)
          else
            emit(:tINTEGER, value[0].ord)
          end
        else
          emit(:tCHARACTER, value)
        end

         @cs = 754; 	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 47 then
# line 1836 "lib/parser/lexer.rl"
		begin
 begin p = (( @te))-1; end
		end
	when 51 then
# line 1852 "lib/parser/lexer.rl"
		begin
 begin p = (( @te))-1; end
 begin  p = @ts - 1; 	begin
		 @cs = 754
		_goto_level = _again
		next
	end
  end
		end
	when 45 then
# line 1 "NONE"
		begin
	case  @act
	when 103 then
	begin begin p = (( @te))-1; end
 emit_table(PUNCTUATION_BEGIN)
           	begin
		p += 1
		_goto_level = _out
		next
	end
 end
	when 104 then
	begin begin p = (( @te))-1; end
 emit(:kRESCUE, 'rescue'.freeze, @ts, tm)
           p = tm - 1
            @cs = 501; 	begin
		p += 1
		_goto_level = _out
		next
	end
 end
	when 105 then
	begin begin p = (( @te))-1; end
 emit_table(KEYWORDS_BEGIN)
            @cs = 746; 	begin
		p += 1
		_goto_level = _out
		next
	end
 end
	when 107 then
	begin begin p = (( @te))-1; end
 p = @ts - 1
           	begin
		 @cs = 754
		_goto_level = _again
		next
	end
 end
	when 108 then
	begin begin p = (( @te))-1; end

    emit(:tIDENTIFIER)

    if !@static_env.nil? && @static_env.declared?(tok)
       @cs = 428; 	begin
		p += 1
		_goto_level = _out
		next
	end

    else
       @cs = (arg_or_cmdarg); 	begin
		p += 1
		_goto_level = _out
		next
	end

    end
  end
	when 111 then
	begin begin p = (( @te))-1; end
 p = @ts - 1; 	begin
		 @cs = 754
		_goto_level = _again
		next
	end
 end
end 
			end
	when 337 then
# line 1872 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin  p = p - 1; 	begin
		 @cs = 525
		_goto_level = _again
		next
	end
  end
		end
	when 338 then
# line 512 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
    # Sit at EOF indefinitely. #advance would return $eof each time.
    # This allows to feed the lexer more data if needed; this is only used
    # in tests.
    #
    # Note that this action is not embedded into e_eof like e_heredoc_nl and e_bs
    # below. This is due to the fact that scanner state at EOF is observed
    # by tests, and encapsulating it in a rule would break the introspection.
    p = p - 1; 	begin
		p += 1
		_goto_level = _out
		next
	end

   end
		end
	when 339 then
# line 1860 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1;		end
	when 343 then
# line 1872 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  p = p - 1; 	begin
		 @cs = 525
		_goto_level = _again
		next
	end
  end
		end
	when 58 then
# line 1882 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin  p = @ts - 1
           	begin
		 @cs = 754
		_goto_level = _again
		next
	end
  end
		end
	when 347 then
# line 1887 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
        	begin
		 @cs = (push_literal(tok, tok, @ts))
		_goto_level = _again
		next
	end

       end
		end
	when 346 then
# line 1897 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin  p = p - 1; 	begin
		 @cs = 525
		_goto_level = _again
		next
	end
  end
		end
	when 345 then
# line 512 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
    # Sit at EOF indefinitely. #advance would return $eof each time.
    # This allows to feed the lexer more data if needed; this is only used
    # in tests.
    #
    # Note that this action is not embedded into e_eof like e_heredoc_nl and e_bs
    # below. This is due to the fact that scanner state at EOF is observed
    # by tests, and encapsulating it in a rule would break the introspection.
    p = p - 1; 	begin
		p += 1
		_goto_level = _out
		next
	end

   end
		end
	when 349 then
# line 1891 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1;		end
	when 348 then
# line 1897 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  p = p - 1; 	begin
		 @cs = 525
		_goto_level = _again
		next
	end
  end
		end
	when 57 then
# line 1897 "lib/parser/lexer.rl"
		begin
 begin p = (( @te))-1; end
 begin  p = p - 1; 	begin
		 @cs = 525
		_goto_level = _again
		next
	end
  end
		end
	when 379 then
# line 1908 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
        emit(:tLAMBDA, '->'.freeze, @ts, @ts + 2)

        @lambda_stack.push @paren_nest
         @cs = 428; 	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 72 then
# line 1945 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin  emit(:kCLASS, 'class'.freeze, @ts, @ts + 5)
           emit(:tLSHFT, '<<'.freeze,    @te - 2, @te)
            @cs = 746; 	begin
		p += 1
		_goto_level = _out
		next
	end
  end
		end
	when 357 then
# line 2080 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
        type, delimiter = tok, tok[-1].chr
        	begin
		 @cs = (push_literal(type, delimiter, @ts, nil, false, false, true))
		_goto_level = _again
		next
	end

       end
		end
	when 60 then
# line 2098 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin  p = @ts - 1; 	begin
		 @stack[ @top] =  @cs
		 @top+= 1
		 @cs = 310
		_goto_level = _again
		next
	end
  end
		end
	when 376 then
# line 2105 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin  emit_table(PUNCTUATION)
            @cs = 435; 	begin
		p += 1
		_goto_level = _out
		next
	end
  end
		end
	when 370 then
# line 2132 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin  emit_table(PUNCTUATION)
            @cs = 525; 	begin
		p += 1
		_goto_level = _out
		next
	end
  end
		end
	when 363 then
# line 2136 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
        emit_table(PUNCTUATION)
        @cond.lexpop; @cmdarg.lexpop

        if RBRACE_OR_RBRACK.include?(tok)
           @cs = 493;
        else # )
          # fnext expr_endfn; ?
        end

        	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 375 then
# line 2150 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin  emit(:tOP_ASGN, tok(@ts, @te - 1))
            @cs = 525; 	begin
		p += 1
		_goto_level = _out
		next
	end
  end
		end
	when 361 then
# line 2154 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin  emit(:tEH, '?'.freeze)
            @cs = 746; 	begin
		p += 1
		_goto_level = _out
		next
	end
  end
		end
	when 358 then
# line 2162 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin  emit_table(PUNCTUATION)
            @cs = 525; 	begin
		p += 1
		_goto_level = _out
		next
	end
  end
		end
	when 360 then
# line 2175 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin  emit(:tSEMI, ';'.freeze)
            @cs = 746; 	begin
		p += 1
		_goto_level = _out
		next
	end
  end
		end
	when 425 then
# line 2178 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
        diagnostic :error, :bare_backslash, nil, range(@ts, @ts + 1)
        p = p - 1;
       end
		end
	when 356 then
# line 2184 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
        diagnostic :fatal, :unexpected, { :character => tok.inspect[1..-2] }
       end
		end
	when 355 then
# line 512 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
    # Sit at EOF indefinitely. #advance would return $eof each time.
    # This allows to feed the lexer more data if needed; this is only used
    # in tests.
    #
    # Note that this action is not embedded into e_eof like e_heredoc_nl and e_bs
    # below. This is due to the fact that scanner state at EOF is observed
    # by tests, and encapsulating it in a rule would break the introspection.
    p = p - 1; 	begin
		p += 1
		_goto_level = _out
		next
	end

   end
		end
	when 435 then
# line 1941 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  emit_table(KEYWORDS)
            @cs = 315; 	begin
		p += 1
		_goto_level = _out
		next
	end
  end
		end
	when 433 then
# line 1945 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  emit(:kCLASS, 'class'.freeze, @ts, @ts + 5)
           emit(:tLSHFT, '<<'.freeze,    @te - 2, @te)
            @cs = 746; 	begin
		p += 1
		_goto_level = _out
		next
	end
  end
		end
	when 432 then
# line 1956 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  emit_table(KEYWORDS)
            @cs = 746; 	begin
		p += 1
		_goto_level = _out
		next
	end
  end
		end
	when 382 then
# line 2030 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        diagnostic :error, :no_dot_digit_literal
       end
		end
	when 422 then
# line 2090 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  emit(:tCONSTANT)
            @cs = (arg_or_cmdarg); 	begin
		p += 1
		_goto_level = _out
		next
	end
  end
		end
	when 374 then
# line 2098 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  p = @ts - 1; 	begin
		 @stack[ @top] =  @cs
		 @top+= 1
		 @cs = 310
		_goto_level = _again
		next
	end
  end
		end
	when 380 then
# line 2105 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  emit_table(PUNCTUATION)
            @cs = 435; 	begin
		p += 1
		_goto_level = _out
		next
	end
  end
		end
	when 427 then
# line 1196 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    emit(:tIDENTIFIER)

    if !@static_env.nil? && @static_env.declared?(tok)
       @cs = 428; 	begin
		p += 1
		_goto_level = _out
		next
	end

    else
       @cs = (arg_or_cmdarg); 	begin
		p += 1
		_goto_level = _out
		next
	end

    end
   end
		end
	when 369 then
# line 2132 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  emit_table(PUNCTUATION)
            @cs = 525; 	begin
		p += 1
		_goto_level = _out
		next
	end
  end
		end
	when 381 then
# line 2162 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  emit_table(PUNCTUATION)
            @cs = 525; 	begin
		p += 1
		_goto_level = _out
		next
	end
  end
		end
	when 367 then
# line 2169 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1;		end
	when 373 then
# line 2184 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        diagnostic :fatal, :unexpected, { :character => tok.inspect[1..-2] }
       end
		end
	when 61 then
# line 2030 "lib/parser/lexer.rl"
		begin
 begin p = (( @te))-1; end
 begin 
        diagnostic :error, :no_dot_digit_literal
       end
		end
	when 59 then
# line 2184 "lib/parser/lexer.rl"
		begin
 begin p = (( @te))-1; end
 begin 
        diagnostic :fatal, :unexpected, { :character => tok.inspect[1..-2] }
       end
		end
	when 62 then
# line 1 "NONE"
		begin
	case  @act
	when 124 then
	begin begin p = (( @te))-1; end

        if @lambda_stack.last == @paren_nest
          @lambda_stack.pop

          if tok == '{'.freeze
            emit(:tLAMBEG, '{'.freeze)
          else # 'do'
            emit(:kDO_LAMBDA, 'do'.freeze)
          end
        else
          if tok == '{'.freeze
            emit(:tLCURLY, '{'.freeze)
          else # 'do'
            emit_do
          end
        end

         @cs = 746; 	begin
		p += 1
		_goto_level = _out
		next
	end

      end
	when 125 then
	begin begin p = (( @te))-1; end
 emit_table(KEYWORDS)
            @cs = 315; 	begin
		p += 1
		_goto_level = _out
		next
	end
 end
	when 126 then
	begin begin p = (( @te))-1; end
 emit(:kCLASS, 'class'.freeze, @ts, @ts + 5)
           emit(:tLSHFT, '<<'.freeze,    @te - 2, @te)
            @cs = 746; 	begin
		p += 1
		_goto_level = _out
		next
	end
 end
	when 127 then
	begin begin p = (( @te))-1; end
 emit_table(KEYWORDS)
            @cs = 525; 	begin
		p += 1
		_goto_level = _out
		next
	end
 end
	when 128 then
	begin begin p = (( @te))-1; end
 emit_table(KEYWORDS)
            @cs = 746; 	begin
		p += 1
		_goto_level = _out
		next
	end
 end
	when 129 then
	begin begin p = (( @te))-1; end
 emit_table(KEYWORDS)
            @cs = 501; 	begin
		p += 1
		_goto_level = _out
		next
	end
 end
	when 130 then
	begin begin p = (( @te))-1; end

        emit_table(KEYWORDS)

        if version?(18) && tok == 'not'.freeze
           @cs = 525; 	begin
		p += 1
		_goto_level = _out
		next
	end

        else
           @cs = 456; 	begin
		p += 1
		_goto_level = _out
		next
	end

        end
      end
	when 131 then
	begin begin p = (( @te))-1; end

        if version?(18)
          emit(:tIDENTIFIER)

          unless !@static_env.nil? && @static_env.declared?(tok)
             @cs = (arg_or_cmdarg);
          end
        else
          emit(:k__ENCODING__, '__ENCODING__'.freeze)
        end
        	begin
		p += 1
		_goto_level = _out
		next
	end

      end
	when 132 then
	begin begin p = (( @te))-1; end
 emit_table(KEYWORDS)
           	begin
		p += 1
		_goto_level = _out
		next
	end
 end
	when 133 then
	begin begin p = (( @te))-1; end

        digits = tok(@num_digits_s, @num_suffix_s)

        if digits.end_with? '_'.freeze
          diagnostic :error, :trailing_in_number, { :character => '_'.freeze },
                     range(@te - 1, @te)
        elsif digits.empty? && @num_base == 8 && version?(18)
          # 1.8 did not raise an error on 0o.
          digits = '0'.freeze
        elsif digits.empty?
          diagnostic :error, :empty_numeric
        elsif @num_base == 8 && (invalid_idx = digits.index(/[89]/))
          invalid_s = @num_digits_s + invalid_idx
          diagnostic :error, :invalid_octal, nil,
                     range(invalid_s, invalid_s + 1)
        end

        if version?(18, 19, 20)
          emit(:tINTEGER, digits.to_i(@num_base), @ts, @num_suffix_s)
          p = @num_suffix_s - 1
        else
          @num_xfrm.call(digits.to_i(@num_base))
        end
        	begin
		p += 1
		_goto_level = _out
		next
	end

      end
	when 135 then
	begin begin p = (( @te))-1; end

        if version?(18, 19, 20)
          diagnostic :error,
                     :trailing_in_number, { :character => tok(@te - 1, @te) },
                     range(@te - 1, @te)
        else
          emit(:tINTEGER, tok(@ts, @te - 1).to_i, @ts, @te - 1)
          p = p - 1; 	begin
		p += 1
		_goto_level = _out
		next
	end

        end
      end
	when 136 then
	begin begin p = (( @te))-1; end

        if version?(18, 19, 20)
          diagnostic :error,
                     :trailing_in_number, { :character => tok(@te - 1, @te) },
                     range(@te - 1, @te)
        else
          emit(:tFLOAT, tok(@ts, @te - 1).to_f, @ts, @te - 1)
          p = p - 1; 	begin
		p += 1
		_goto_level = _out
		next
	end

        end
      end
	when 137 then
	begin begin p = (( @te))-1; end

        digits = tok(@ts, @num_suffix_s)

        if version?(18, 19, 20)
          emit(:tFLOAT, Float(digits), @ts, @num_suffix_s)
          p = @num_suffix_s - 1
        else
          @num_xfrm.call(digits)
        end
        	begin
		p += 1
		_goto_level = _out
		next
	end

      end
	when 139 then
	begin begin p = (( @te))-1; end
 emit(:tCONSTANT)
            @cs = (arg_or_cmdarg); 	begin
		p += 1
		_goto_level = _out
		next
	end
 end
	when 143 then
	begin begin p = (( @te))-1; end

    emit(:tIDENTIFIER)

    if !@static_env.nil? && @static_env.declared?(tok)
       @cs = 428; 	begin
		p += 1
		_goto_level = _out
		next
	end

    else
       @cs = (arg_or_cmdarg); 	begin
		p += 1
		_goto_level = _out
		next
	end

    end
  end
	when 144 then
	begin begin p = (( @te))-1; end

        if tm == @te
          # Suffix was consumed, e.g. foo!
          emit(:tFID)
        else
          # Suffix was not consumed, e.g. foo!=
          emit(:tIDENTIFIER, tok(@ts, tm), @ts, tm)
          p = tm - 1
        end
         @cs = 456; 	begin
		p += 1
		_goto_level = _out
		next
	end

      end
end 
			end
	when 76 then
# line 2196 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin  p = tm - 1; 	begin
		 @cs = 754
		_goto_level = _again
		next
	end
  end
		end
	when 440 then
# line 2199 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin  emit(:tNL, nil, @newline_s, @newline_s + 1)
           p = p - 1;  @cs = 160; 	begin
		p += 1
		_goto_level = _out
		next
	end
  end
		end
	when 441 then
# line 2199 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  emit(:tNL, nil, @newline_s, @newline_s + 1)
           p = p - 1;  @cs = 160; 	begin
		p += 1
		_goto_level = _out
		next
	end
  end
		end
	when 73 then
# line 2199 "lib/parser/lexer.rl"
		begin
 begin p = (( @te))-1; end
 begin  emit(:tNL, nil, @newline_s, @newline_s + 1)
           p = p - 1;  @cs = 160; 	begin
		p += 1
		_goto_level = _out
		next
	end
  end
		end
	when 444 then
# line 2209 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        emit_comment(@eq_begin_s, @te)
        	begin
		 @cs = 160
		_goto_level = _again
		next
	end

       end
		end
	when 443 then
# line 2217 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        diagnostic :fatal, :embedded_document, nil,
                   range(@eq_begin_s, @eq_begin_s + '=begin'.length)
       end
		end
	when 87 then
# line 2227 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin  @eq_begin_s = @ts
           	begin
		 @cs = 930
		_goto_level = _again
		next
	end
  end
		end
	when 2 then
# line 2231 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin  p = pe - 3  end
		end
	when 79 then
# line 2234 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin  p = p - 1; 	begin
		 @cs = 746
		_goto_level = _again
		next
	end
  end
		end
	when 80 then
# line 512 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
    # Sit at EOF indefinitely. #advance would return $eof each time.
    # This allows to feed the lexer more data if needed; this is only used
    # in tests.
    #
    # Note that this action is not embedded into e_eof like e_heredoc_nl and e_bs
    # below. This is due to the fact that scanner state at EOF is observed
    # by tests, and encapsulating it in a rule would break the introspection.
    p = p - 1; 	begin
		p += 1
		_goto_level = _out
		next
	end

   end
		end
	when 81 then
# line 2224 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1;		end
	when 86 then
# line 2227 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  @eq_begin_s = @ts
           	begin
		 @cs = 930
		_goto_level = _again
		next
	end
  end
		end
	when 85 then
# line 2234 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  p = p - 1; 	begin
		 @cs = 746
		_goto_level = _again
		next
	end
  end
		end
	when 1 then
# line 2234 "lib/parser/lexer.rl"
		begin
 begin p = (( @te))-1; end
 begin  p = p - 1; 	begin
		 @cs = 746
		_goto_level = _again
		next
	end
  end
		end
	when 71 then
# line 486 "lib/parser/lexer.rl"
		begin

    # Record position of a newline for precise location reporting on tNL
    # tokens.
    #
    # This action is embedded directly into c_nl, as it is idempotent and
    # there are no cases when we need to skip it.
    @newline_s = p
  		end
# line 1111 "lib/parser/lexer.rl"
		begin
 emit_comment(@sharp_s, p == pe ? p - 2 : p) 		end
	when 91 then
# line 486 "lib/parser/lexer.rl"
		begin

    # Record position of a newline for precise location reporting on tNL
    # tokens.
    #
    # This action is embedded directly into c_nl, as it is idempotent and
    # there are no cases when we need to skip it.
    @newline_s = p
  		end
# line 877 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
    current_literal = literal
    if @te == pe
      diagnostic :fatal, :string_eof, nil,
                 range(current_literal.str_s, current_literal.str_s + 1)
    end

    if current_literal.heredoc?
      line = tok(@herebody_s, @ts).gsub(/\r+$/, ''.freeze)

      if version?(18, 19, 20)
        # See ruby:c48b4209c
        line = line.gsub(/\r.*$/, ''.freeze)
      end

      # Try ending the heredoc with the complete most recently
      # scanned line. @herebody_s always refers to the start of such line.
      if current_literal.nest_and_try_closing(line, @herebody_s, @ts)
        # Adjust @herebody_s to point to the next line.
        @herebody_s = @te

        # Continue regular lexing after the heredoc reference (<<END).
        p = current_literal.heredoc_e - 1
         @cs = (pop_literal); 	begin
		p += 1
		_goto_level = _out
		next
	end

      else
        # Calculate indentation level for <<~HEREDOCs.
        current_literal.infer_indent_level(line)

        # Ditto.
        @herebody_s = @te
      end
    else
      # Try ending the literal with a newline.
      if current_literal.nest_and_try_closing(tok, @ts, @te)
         @cs = (pop_literal); 	begin
		p += 1
		_goto_level = _out
		next
	end

      end

      if @herebody_s
        # This is a regular literal intertwined with a heredoc. Like:
        #
        #     p <<-foo+"1
        #     bar
        #     foo
        #     2"
        #
        # which, incidentally, evaluates to "bar\n1\n2".
        p = @herebody_s - 1
        @herebody_s = nil
      end
    end

    if current_literal.words? && !eof_codepoint?(@source_pts[p])
      current_literal.extend_space @ts, @te
    else
      # A literal newline is appended if the heredoc was _not_ closed
      # this time (see fbreak above). See also Literal#nest_and_try_closing
      # for rationale of calling #flush_string here.
      current_literal.extend_string tok, @ts, @te
      current_literal.flush_string
    end
   end
		end
	when 119 then
# line 486 "lib/parser/lexer.rl"
		begin

    # Record position of a newline for precise location reporting on tNL
    # tokens.
    #
    # This action is embedded directly into c_nl, as it is idempotent and
    # there are no cases when we need to skip it.
    @newline_s = p
  		end
# line 877 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
    current_literal = literal
    if @te == pe
      diagnostic :fatal, :string_eof, nil,
                 range(current_literal.str_s, current_literal.str_s + 1)
    end

    if current_literal.heredoc?
      line = tok(@herebody_s, @ts).gsub(/\r+$/, ''.freeze)

      if version?(18, 19, 20)
        # See ruby:c48b4209c
        line = line.gsub(/\r.*$/, ''.freeze)
      end

      # Try ending the heredoc with the complete most recently
      # scanned line. @herebody_s always refers to the start of such line.
      if current_literal.nest_and_try_closing(line, @herebody_s, @ts)
        # Adjust @herebody_s to point to the next line.
        @herebody_s = @te

        # Continue regular lexing after the heredoc reference (<<END).
        p = current_literal.heredoc_e - 1
         @cs = (pop_literal); 	begin
		p += 1
		_goto_level = _out
		next
	end

      else
        # Calculate indentation level for <<~HEREDOCs.
        current_literal.infer_indent_level(line)

        # Ditto.
        @herebody_s = @te
      end
    else
      # Try ending the literal with a newline.
      if current_literal.nest_and_try_closing(tok, @ts, @te)
         @cs = (pop_literal); 	begin
		p += 1
		_goto_level = _out
		next
	end

      end

      if @herebody_s
        # This is a regular literal intertwined with a heredoc. Like:
        #
        #     p <<-foo+"1
        #     bar
        #     foo
        #     2"
        #
        # which, incidentally, evaluates to "bar\n1\n2".
        p = @herebody_s - 1
        @herebody_s = nil
      end
    end

    if current_literal.words? && !eof_codepoint?(@source_pts[p])
      current_literal.extend_space @ts, @te
    else
      # A literal newline is appended if the heredoc was _not_ closed
      # this time (see fbreak above). See also Literal#nest_and_try_closing
      # for rationale of calling #flush_string here.
      current_literal.extend_string tok, @ts, @te
      current_literal.flush_string
    end
   end
		end
	when 145 then
# line 486 "lib/parser/lexer.rl"
		begin

    # Record position of a newline for precise location reporting on tNL
    # tokens.
    #
    # This action is embedded directly into c_nl, as it is idempotent and
    # there are no cases when we need to skip it.
    @newline_s = p
  		end
# line 877 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
    current_literal = literal
    if @te == pe
      diagnostic :fatal, :string_eof, nil,
                 range(current_literal.str_s, current_literal.str_s + 1)
    end

    if current_literal.heredoc?
      line = tok(@herebody_s, @ts).gsub(/\r+$/, ''.freeze)

      if version?(18, 19, 20)
        # See ruby:c48b4209c
        line = line.gsub(/\r.*$/, ''.freeze)
      end

      # Try ending the heredoc with the complete most recently
      # scanned line. @herebody_s always refers to the start of such line.
      if current_literal.nest_and_try_closing(line, @herebody_s, @ts)
        # Adjust @herebody_s to point to the next line.
        @herebody_s = @te

        # Continue regular lexing after the heredoc reference (<<END).
        p = current_literal.heredoc_e - 1
         @cs = (pop_literal); 	begin
		p += 1
		_goto_level = _out
		next
	end

      else
        # Calculate indentation level for <<~HEREDOCs.
        current_literal.infer_indent_level(line)

        # Ditto.
        @herebody_s = @te
      end
    else
      # Try ending the literal with a newline.
      if current_literal.nest_and_try_closing(tok, @ts, @te)
         @cs = (pop_literal); 	begin
		p += 1
		_goto_level = _out
		next
	end

      end

      if @herebody_s
        # This is a regular literal intertwined with a heredoc. Like:
        #
        #     p <<-foo+"1
        #     bar
        #     foo
        #     2"
        #
        # which, incidentally, evaluates to "bar\n1\n2".
        p = @herebody_s - 1
        @herebody_s = nil
      end
    end

    if current_literal.words? && !eof_codepoint?(@source_pts[p])
      current_literal.extend_space @ts, @te
    else
      # A literal newline is appended if the heredoc was _not_ closed
      # this time (see fbreak above). See also Literal#nest_and_try_closing
      # for rationale of calling #flush_string here.
      current_literal.extend_string tok, @ts, @te
      current_literal.flush_string
    end
   end
		end
	when 151 then
# line 486 "lib/parser/lexer.rl"
		begin

    # Record position of a newline for precise location reporting on tNL
    # tokens.
    #
    # This action is embedded directly into c_nl, as it is idempotent and
    # there are no cases when we need to skip it.
    @newline_s = p
  		end
# line 877 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
    current_literal = literal
    if @te == pe
      diagnostic :fatal, :string_eof, nil,
                 range(current_literal.str_s, current_literal.str_s + 1)
    end

    if current_literal.heredoc?
      line = tok(@herebody_s, @ts).gsub(/\r+$/, ''.freeze)

      if version?(18, 19, 20)
        # See ruby:c48b4209c
        line = line.gsub(/\r.*$/, ''.freeze)
      end

      # Try ending the heredoc with the complete most recently
      # scanned line. @herebody_s always refers to the start of such line.
      if current_literal.nest_and_try_closing(line, @herebody_s, @ts)
        # Adjust @herebody_s to point to the next line.
        @herebody_s = @te

        # Continue regular lexing after the heredoc reference (<<END).
        p = current_literal.heredoc_e - 1
         @cs = (pop_literal); 	begin
		p += 1
		_goto_level = _out
		next
	end

      else
        # Calculate indentation level for <<~HEREDOCs.
        current_literal.infer_indent_level(line)

        # Ditto.
        @herebody_s = @te
      end
    else
      # Try ending the literal with a newline.
      if current_literal.nest_and_try_closing(tok, @ts, @te)
         @cs = (pop_literal); 	begin
		p += 1
		_goto_level = _out
		next
	end

      end

      if @herebody_s
        # This is a regular literal intertwined with a heredoc. Like:
        #
        #     p <<-foo+"1
        #     bar
        #     foo
        #     2"
        #
        # which, incidentally, evaluates to "bar\n1\n2".
        p = @herebody_s - 1
        @herebody_s = nil
      end
    end

    if current_literal.words? && !eof_codepoint?(@source_pts[p])
      current_literal.extend_space @ts, @te
    else
      # A literal newline is appended if the heredoc was _not_ closed
      # this time (see fbreak above). See also Literal#nest_and_try_closing
      # for rationale of calling #flush_string here.
      current_literal.extend_string tok, @ts, @te
      current_literal.flush_string
    end
   end
		end
	when 157 then
# line 486 "lib/parser/lexer.rl"
		begin

    # Record position of a newline for precise location reporting on tNL
    # tokens.
    #
    # This action is embedded directly into c_nl, as it is idempotent and
    # there are no cases when we need to skip it.
    @newline_s = p
  		end
# line 877 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
    current_literal = literal
    if @te == pe
      diagnostic :fatal, :string_eof, nil,
                 range(current_literal.str_s, current_literal.str_s + 1)
    end

    if current_literal.heredoc?
      line = tok(@herebody_s, @ts).gsub(/\r+$/, ''.freeze)

      if version?(18, 19, 20)
        # See ruby:c48b4209c
        line = line.gsub(/\r.*$/, ''.freeze)
      end

      # Try ending the heredoc with the complete most recently
      # scanned line. @herebody_s always refers to the start of such line.
      if current_literal.nest_and_try_closing(line, @herebody_s, @ts)
        # Adjust @herebody_s to point to the next line.
        @herebody_s = @te

        # Continue regular lexing after the heredoc reference (<<END).
        p = current_literal.heredoc_e - 1
         @cs = (pop_literal); 	begin
		p += 1
		_goto_level = _out
		next
	end

      else
        # Calculate indentation level for <<~HEREDOCs.
        current_literal.infer_indent_level(line)

        # Ditto.
        @herebody_s = @te
      end
    else
      # Try ending the literal with a newline.
      if current_literal.nest_and_try_closing(tok, @ts, @te)
         @cs = (pop_literal); 	begin
		p += 1
		_goto_level = _out
		next
	end

      end

      if @herebody_s
        # This is a regular literal intertwined with a heredoc. Like:
        #
        #     p <<-foo+"1
        #     bar
        #     foo
        #     2"
        #
        # which, incidentally, evaluates to "bar\n1\n2".
        p = @herebody_s - 1
        @herebody_s = nil
      end
    end

    if current_literal.words? && !eof_codepoint?(@source_pts[p])
      current_literal.extend_space @ts, @te
    else
      # A literal newline is appended if the heredoc was _not_ closed
      # this time (see fbreak above). See also Literal#nest_and_try_closing
      # for rationale of calling #flush_string here.
      current_literal.extend_string tok, @ts, @te
      current_literal.flush_string
    end
   end
		end
	when 163 then
# line 486 "lib/parser/lexer.rl"
		begin

    # Record position of a newline for precise location reporting on tNL
    # tokens.
    #
    # This action is embedded directly into c_nl, as it is idempotent and
    # there are no cases when we need to skip it.
    @newline_s = p
  		end
# line 877 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
    current_literal = literal
    if @te == pe
      diagnostic :fatal, :string_eof, nil,
                 range(current_literal.str_s, current_literal.str_s + 1)
    end

    if current_literal.heredoc?
      line = tok(@herebody_s, @ts).gsub(/\r+$/, ''.freeze)

      if version?(18, 19, 20)
        # See ruby:c48b4209c
        line = line.gsub(/\r.*$/, ''.freeze)
      end

      # Try ending the heredoc with the complete most recently
      # scanned line. @herebody_s always refers to the start of such line.
      if current_literal.nest_and_try_closing(line, @herebody_s, @ts)
        # Adjust @herebody_s to point to the next line.
        @herebody_s = @te

        # Continue regular lexing after the heredoc reference (<<END).
        p = current_literal.heredoc_e - 1
         @cs = (pop_literal); 	begin
		p += 1
		_goto_level = _out
		next
	end

      else
        # Calculate indentation level for <<~HEREDOCs.
        current_literal.infer_indent_level(line)

        # Ditto.
        @herebody_s = @te
      end
    else
      # Try ending the literal with a newline.
      if current_literal.nest_and_try_closing(tok, @ts, @te)
         @cs = (pop_literal); 	begin
		p += 1
		_goto_level = _out
		next
	end

      end

      if @herebody_s
        # This is a regular literal intertwined with a heredoc. Like:
        #
        #     p <<-foo+"1
        #     bar
        #     foo
        #     2"
        #
        # which, incidentally, evaluates to "bar\n1\n2".
        p = @herebody_s - 1
        @herebody_s = nil
      end
    end

    if current_literal.words? && !eof_codepoint?(@source_pts[p])
      current_literal.extend_space @ts, @te
    else
      # A literal newline is appended if the heredoc was _not_ closed
      # this time (see fbreak above). See also Literal#nest_and_try_closing
      # for rationale of calling #flush_string here.
      current_literal.extend_string tok, @ts, @te
      current_literal.flush_string
    end
   end
		end
	when 233 then
# line 486 "lib/parser/lexer.rl"
		begin

    # Record position of a newline for precise location reporting on tNL
    # tokens.
    #
    # This action is embedded directly into c_nl, as it is idempotent and
    # there are no cases when we need to skip it.
    @newline_s = p
  		end
# line 1407 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
        # Unlike expr_beg as invoked in the next rule, do not warn
        p = @ts - 1
        	begin
		 @cs = 754
		_goto_level = _again
		next
	end

       end
		end
	when 224 then
# line 486 "lib/parser/lexer.rl"
		begin

    # Record position of a newline for precise location reporting on tNL
    # tokens.
    #
    # This action is embedded directly into c_nl, as it is idempotent and
    # there are no cases when we need to skip it.
    @newline_s = p
  		end
# line 1466 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin  p = tm - 1; 	begin
		 @cs = 754
		_goto_level = _again
		next
	end
  end
		end
	when 216 then
# line 486 "lib/parser/lexer.rl"
		begin

    # Record position of a newline for precise location reporting on tNL
    # tokens.
    #
    # This action is embedded directly into c_nl, as it is idempotent and
    # there are no cases when we need to skip it.
    @newline_s = p
  		end
# line 1475 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
        p = @ts - 1
        	begin
		 @cs = 754
		_goto_level = _again
		next
	end

       end
		end
	when 304 then
# line 486 "lib/parser/lexer.rl"
		begin

    # Record position of a newline for precise location reporting on tNL
    # tokens.
    #
    # This action is embedded directly into c_nl, as it is idempotent and
    # there are no cases when we need to skip it.
    @newline_s = p
  		end
# line 1724 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
        escape = { " "  => '\s', "\r" => '\r', "\n" => '\n', "\t" => '\t',
                   "\v" => '\v', "\f" => '\f' }[@source_buffer.slice(@ts + 1)]
        diagnostic :warning, :invalid_escape_use, { :escape => escape }, range

        p = @ts - 1
        	begin
		 @cs = 754
		_goto_level = _again
		next
	end

       end
		end
	when 277 then
# line 486 "lib/parser/lexer.rl"
		begin

    # Record position of a newline for precise location reporting on tNL
    # tokens.
    #
    # This action is embedded directly into c_nl, as it is idempotent and
    # there are no cases when we need to skip it.
    @newline_s = p
  		end
# line 1839 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin  p = @ts - 1
           	begin
		 @cs = 160
		_goto_level = _again
		next
	end
  end
		end
	when 445 then
# line 486 "lib/parser/lexer.rl"
		begin

    # Record position of a newline for precise location reporting on tNL
    # tokens.
    #
    # This action is embedded directly into c_nl, as it is idempotent and
    # there are no cases when we need to skip it.
    @newline_s = p
  		end
# line 2209 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
        emit_comment(@eq_begin_s, @te)
        	begin
		 @cs = 160
		_goto_level = _again
		next
	end

       end
		end
	when 442 then
# line 486 "lib/parser/lexer.rl"
		begin

    # Record position of a newline for precise location reporting on tNL
    # tokens.
    #
    # This action is embedded directly into c_nl, as it is idempotent and
    # there are no cases when we need to skip it.
    @newline_s = p
  		end
# line 2214 "lib/parser/lexer.rl"
		begin
 @te = p+1
		end
	when 88 then
# line 486 "lib/parser/lexer.rl"
		begin

    # Record position of a newline for precise location reporting on tNL
    # tokens.
    #
    # This action is embedded directly into c_nl, as it is idempotent and
    # there are no cases when we need to skip it.
    @newline_s = p
  		end
# line 2227 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin  @eq_begin_s = @ts
           	begin
		 @cs = 930
		_goto_level = _again
		next
	end
  end
		end
	when 3 then
# line 486 "lib/parser/lexer.rl"
		begin

    # Record position of a newline for precise location reporting on tNL
    # tokens.
    #
    # This action is embedded directly into c_nl, as it is idempotent and
    # there are no cases when we need to skip it.
    @newline_s = p
  		end
# line 2231 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin  p = pe - 3  end
		end
	when 399 then
# line 620 "lib/parser/lexer.rl"
		begin
 @num_xfrm = lambda { |chars| emit(:tRATIONAL,  Rational(chars)) } 		end
# line 2003 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        digits = tok(@num_digits_s, @num_suffix_s)

        if digits.end_with? '_'.freeze
          diagnostic :error, :trailing_in_number, { :character => '_'.freeze },
                     range(@te - 1, @te)
        elsif digits.empty? && @num_base == 8 && version?(18)
          # 1.8 did not raise an error on 0o.
          digits = '0'.freeze
        elsif digits.empty?
          diagnostic :error, :empty_numeric
        elsif @num_base == 8 && (invalid_idx = digits.index(/[89]/))
          invalid_s = @num_digits_s + invalid_idx
          diagnostic :error, :invalid_octal, nil,
                     range(invalid_s, invalid_s + 1)
        end

        if version?(18, 19, 20)
          emit(:tINTEGER, digits.to_i(@num_base), @ts, @num_suffix_s)
          p = @num_suffix_s - 1
        else
          @num_xfrm.call(digits.to_i(@num_base))
        end
        	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 398 then
# line 621 "lib/parser/lexer.rl"
		begin
 @num_xfrm = lambda { |chars| emit(:tIMAGINARY, Complex(0, chars)) } 		end
# line 2003 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        digits = tok(@num_digits_s, @num_suffix_s)

        if digits.end_with? '_'.freeze
          diagnostic :error, :trailing_in_number, { :character => '_'.freeze },
                     range(@te - 1, @te)
        elsif digits.empty? && @num_base == 8 && version?(18)
          # 1.8 did not raise an error on 0o.
          digits = '0'.freeze
        elsif digits.empty?
          diagnostic :error, :empty_numeric
        elsif @num_base == 8 && (invalid_idx = digits.index(/[89]/))
          invalid_s = @num_digits_s + invalid_idx
          diagnostic :error, :invalid_octal, nil,
                     range(invalid_s, invalid_s + 1)
        end

        if version?(18, 19, 20)
          emit(:tINTEGER, digits.to_i(@num_base), @ts, @num_suffix_s)
          p = @num_suffix_s - 1
        else
          @num_xfrm.call(digits.to_i(@num_base))
        end
        	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 400 then
# line 622 "lib/parser/lexer.rl"
		begin
 @num_xfrm = lambda { |chars| emit(:tIMAGINARY, Complex(0, Rational(chars))) } 		end
# line 2003 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        digits = tok(@num_digits_s, @num_suffix_s)

        if digits.end_with? '_'.freeze
          diagnostic :error, :trailing_in_number, { :character => '_'.freeze },
                     range(@te - 1, @te)
        elsif digits.empty? && @num_base == 8 && version?(18)
          # 1.8 did not raise an error on 0o.
          digits = '0'.freeze
        elsif digits.empty?
          diagnostic :error, :empty_numeric
        elsif @num_base == 8 && (invalid_idx = digits.index(/[89]/))
          invalid_s = @num_digits_s + invalid_idx
          diagnostic :error, :invalid_octal, nil,
                     range(invalid_s, invalid_s + 1)
        end

        if version?(18, 19, 20)
          emit(:tINTEGER, digits.to_i(@num_base), @ts, @num_suffix_s)
          p = @num_suffix_s - 1
        else
          @num_xfrm.call(digits.to_i(@num_base))
        end
        	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 393 then
# line 626 "lib/parser/lexer.rl"
		begin
 @num_xfrm = lambda { |chars| emit(:tIMAGINARY, Complex(0, Float(chars))) } 		end
# line 2062 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        digits = tok(@ts, @num_suffix_s)

        if version?(18, 19, 20)
          emit(:tFLOAT, Float(digits), @ts, @num_suffix_s)
          p = @num_suffix_s - 1
        else
          @num_xfrm.call(digits)
        end
        	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 394 then
# line 630 "lib/parser/lexer.rl"
		begin
 @num_xfrm = lambda { |chars| emit(:tRATIONAL,  Rational(chars)) } 		end
# line 2062 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        digits = tok(@ts, @num_suffix_s)

        if version?(18, 19, 20)
          emit(:tFLOAT, Float(digits), @ts, @num_suffix_s)
          p = @num_suffix_s - 1
        else
          @num_xfrm.call(digits)
        end
        	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 395 then
# line 631 "lib/parser/lexer.rl"
		begin
 @num_xfrm = lambda { |chars| emit(:tIMAGINARY, Complex(0, Rational(chars))) } 		end
# line 2062 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        digits = tok(@ts, @num_suffix_s)

        if version?(18, 19, 20)
          emit(:tFLOAT, Float(digits), @ts, @num_suffix_s)
          p = @num_suffix_s - 1
        else
          @num_xfrm.call(digits)
        end
        	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 114 then
# line 646 "lib/parser/lexer.rl"
		begin

    @escape = ""

    codepoints  = tok(@escape_s + 2, p - 1)
    codepoint_s = @escape_s + 2

    codepoints.split(/[ \t]/).each do |codepoint_str|
      codepoint = codepoint_str.to_i(16)

      if codepoint >= 0x110000
        diagnostic :error, :unicode_point_too_large, nil,
                   range(codepoint_s, codepoint_s + codepoint_str.length)
        break
      end

      @escape     += codepoint.chr(Encoding::UTF_8)
      codepoint_s += codepoint_str.length + 1
    end
  		end
# line 839 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    current_literal = literal
    # Get the first character after the backslash.
    escaped_char = @source_buffer.slice(@escape_s).chr

    if current_literal.munge_escape? escaped_char
      # If this particular literal uses this character as an opening
      # or closing delimiter, it is an escape sequence for that
      # particular character. Write it without the backslash.

      if current_literal.regexp? && REGEXP_META_CHARACTERS.match(escaped_char)
        # Regular expressions should include escaped delimiters in their
        # escaped form, except when the escaped character is
        # a closing delimiter but not a regexp metacharacter.
        #
        # The backslash itself cannot be used as a closing delimiter
        # at the same time as an escape symbol, but it is always munged,
        # so this branch also executes for the non-closing-delimiter case
        # for the backslash.
        current_literal.extend_string(tok, @ts, @te)
      else
        current_literal.extend_string(escaped_char, @ts, @te)
      end
    else
      # It does not. So this is an actual escape sequence, yay!
      if current_literal.regexp?
        # Regular expressions should include escape sequences in their
        # escaped form. On the other hand, escaped newlines are removed.
        current_literal.extend_string(tok.gsub("\\\n".freeze, ''.freeze), @ts, @te)
      else
        current_literal.extend_string(@escape || tok, @ts, @te)
      end
    end
   end
		end
	when 140 then
# line 646 "lib/parser/lexer.rl"
		begin

    @escape = ""

    codepoints  = tok(@escape_s + 2, p - 1)
    codepoint_s = @escape_s + 2

    codepoints.split(/[ \t]/).each do |codepoint_str|
      codepoint = codepoint_str.to_i(16)

      if codepoint >= 0x110000
        diagnostic :error, :unicode_point_too_large, nil,
                   range(codepoint_s, codepoint_s + codepoint_str.length)
        break
      end

      @escape     += codepoint.chr(Encoding::UTF_8)
      codepoint_s += codepoint_str.length + 1
    end
  		end
# line 839 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    current_literal = literal
    # Get the first character after the backslash.
    escaped_char = @source_buffer.slice(@escape_s).chr

    if current_literal.munge_escape? escaped_char
      # If this particular literal uses this character as an opening
      # or closing delimiter, it is an escape sequence for that
      # particular character. Write it without the backslash.

      if current_literal.regexp? && REGEXP_META_CHARACTERS.match(escaped_char)
        # Regular expressions should include escaped delimiters in their
        # escaped form, except when the escaped character is
        # a closing delimiter but not a regexp metacharacter.
        #
        # The backslash itself cannot be used as a closing delimiter
        # at the same time as an escape symbol, but it is always munged,
        # so this branch also executes for the non-closing-delimiter case
        # for the backslash.
        current_literal.extend_string(tok, @ts, @te)
      else
        current_literal.extend_string(escaped_char, @ts, @te)
      end
    else
      # It does not. So this is an actual escape sequence, yay!
      if current_literal.regexp?
        # Regular expressions should include escape sequences in their
        # escaped form. On the other hand, escaped newlines are removed.
        current_literal.extend_string(tok.gsub("\\\n".freeze, ''.freeze), @ts, @te)
      else
        current_literal.extend_string(@escape || tok, @ts, @te)
      end
    end
   end
		end
	when 325 then
# line 646 "lib/parser/lexer.rl"
		begin

    @escape = ""

    codepoints  = tok(@escape_s + 2, p - 1)
    codepoint_s = @escape_s + 2

    codepoints.split(/[ \t]/).each do |codepoint_str|
      codepoint = codepoint_str.to_i(16)

      if codepoint >= 0x110000
        diagnostic :error, :unicode_point_too_large, nil,
                   range(codepoint_s, codepoint_s + codepoint_str.length)
        break
      end

      @escape     += codepoint.chr(Encoding::UTF_8)
      codepoint_s += codepoint_str.length + 1
    end
  		end
# line 1707 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        value = @escape || tok(@ts + 1)

        if version?(18)
          if defined?(Encoding)
            emit(:tINTEGER, value.dup.force_encoding(Encoding::BINARY)[0].ord)
          else
            emit(:tINTEGER, value[0].ord)
          end
        else
          emit(:tCHARACTER, value)
        end

         @cs = 754; 	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 98 then
# line 666 "lib/parser/lexer.rl"
		begin

    codepoint = @source_pts[p - 1]
    if (@escape = ESCAPES[codepoint]).nil?
      @escape = encode_escape(@source_buffer.slice(p - 1))
    end
  		end
# line 839 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    current_literal = literal
    # Get the first character after the backslash.
    escaped_char = @source_buffer.slice(@escape_s).chr

    if current_literal.munge_escape? escaped_char
      # If this particular literal uses this character as an opening
      # or closing delimiter, it is an escape sequence for that
      # particular character. Write it without the backslash.

      if current_literal.regexp? && REGEXP_META_CHARACTERS.match(escaped_char)
        # Regular expressions should include escaped delimiters in their
        # escaped form, except when the escaped character is
        # a closing delimiter but not a regexp metacharacter.
        #
        # The backslash itself cannot be used as a closing delimiter
        # at the same time as an escape symbol, but it is always munged,
        # so this branch also executes for the non-closing-delimiter case
        # for the backslash.
        current_literal.extend_string(tok, @ts, @te)
      else
        current_literal.extend_string(escaped_char, @ts, @te)
      end
    else
      # It does not. So this is an actual escape sequence, yay!
      if current_literal.regexp?
        # Regular expressions should include escape sequences in their
        # escaped form. On the other hand, escaped newlines are removed.
        current_literal.extend_string(tok.gsub("\\\n".freeze, ''.freeze), @ts, @te)
      else
        current_literal.extend_string(@escape || tok, @ts, @te)
      end
    end
   end
		end
	when 124 then
# line 666 "lib/parser/lexer.rl"
		begin

    codepoint = @source_pts[p - 1]
    if (@escape = ESCAPES[codepoint]).nil?
      @escape = encode_escape(@source_buffer.slice(p - 1))
    end
  		end
# line 839 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    current_literal = literal
    # Get the first character after the backslash.
    escaped_char = @source_buffer.slice(@escape_s).chr

    if current_literal.munge_escape? escaped_char
      # If this particular literal uses this character as an opening
      # or closing delimiter, it is an escape sequence for that
      # particular character. Write it without the backslash.

      if current_literal.regexp? && REGEXP_META_CHARACTERS.match(escaped_char)
        # Regular expressions should include escaped delimiters in their
        # escaped form, except when the escaped character is
        # a closing delimiter but not a regexp metacharacter.
        #
        # The backslash itself cannot be used as a closing delimiter
        # at the same time as an escape symbol, but it is always munged,
        # so this branch also executes for the non-closing-delimiter case
        # for the backslash.
        current_literal.extend_string(tok, @ts, @te)
      else
        current_literal.extend_string(escaped_char, @ts, @te)
      end
    else
      # It does not. So this is an actual escape sequence, yay!
      if current_literal.regexp?
        # Regular expressions should include escape sequences in their
        # escaped form. On the other hand, escaped newlines are removed.
        current_literal.extend_string(tok.gsub("\\\n".freeze, ''.freeze), @ts, @te)
      else
        current_literal.extend_string(@escape || tok, @ts, @te)
      end
    end
   end
		end
	when 309 then
# line 666 "lib/parser/lexer.rl"
		begin

    codepoint = @source_pts[p - 1]
    if (@escape = ESCAPES[codepoint]).nil?
      @escape = encode_escape(@source_buffer.slice(p - 1))
    end
  		end
# line 1707 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        value = @escape || tok(@ts + 1)

        if version?(18)
          if defined?(Encoding)
            emit(:tINTEGER, value.dup.force_encoding(Encoding::BINARY)[0].ord)
          else
            emit(:tINTEGER, value[0].ord)
          end
        else
          emit(:tCHARACTER, value)
        end

         @cs = 754; 	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 101 then
# line 673 "lib/parser/lexer.rl"
		begin

    diagnostic :fatal, :invalid_escape
  		end
# line 839 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    current_literal = literal
    # Get the first character after the backslash.
    escaped_char = @source_buffer.slice(@escape_s).chr

    if current_literal.munge_escape? escaped_char
      # If this particular literal uses this character as an opening
      # or closing delimiter, it is an escape sequence for that
      # particular character. Write it without the backslash.

      if current_literal.regexp? && REGEXP_META_CHARACTERS.match(escaped_char)
        # Regular expressions should include escaped delimiters in their
        # escaped form, except when the escaped character is
        # a closing delimiter but not a regexp metacharacter.
        #
        # The backslash itself cannot be used as a closing delimiter
        # at the same time as an escape symbol, but it is always munged,
        # so this branch also executes for the non-closing-delimiter case
        # for the backslash.
        current_literal.extend_string(tok, @ts, @te)
      else
        current_literal.extend_string(escaped_char, @ts, @te)
      end
    else
      # It does not. So this is an actual escape sequence, yay!
      if current_literal.regexp?
        # Regular expressions should include escape sequences in their
        # escaped form. On the other hand, escaped newlines are removed.
        current_literal.extend_string(tok.gsub("\\\n".freeze, ''.freeze), @ts, @te)
      else
        current_literal.extend_string(@escape || tok, @ts, @te)
      end
    end
   end
		end
	when 127 then
# line 673 "lib/parser/lexer.rl"
		begin

    diagnostic :fatal, :invalid_escape
  		end
# line 839 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    current_literal = literal
    # Get the first character after the backslash.
    escaped_char = @source_buffer.slice(@escape_s).chr

    if current_literal.munge_escape? escaped_char
      # If this particular literal uses this character as an opening
      # or closing delimiter, it is an escape sequence for that
      # particular character. Write it without the backslash.

      if current_literal.regexp? && REGEXP_META_CHARACTERS.match(escaped_char)
        # Regular expressions should include escaped delimiters in their
        # escaped form, except when the escaped character is
        # a closing delimiter but not a regexp metacharacter.
        #
        # The backslash itself cannot be used as a closing delimiter
        # at the same time as an escape symbol, but it is always munged,
        # so this branch also executes for the non-closing-delimiter case
        # for the backslash.
        current_literal.extend_string(tok, @ts, @te)
      else
        current_literal.extend_string(escaped_char, @ts, @te)
      end
    else
      # It does not. So this is an actual escape sequence, yay!
      if current_literal.regexp?
        # Regular expressions should include escape sequences in their
        # escaped form. On the other hand, escaped newlines are removed.
        current_literal.extend_string(tok.gsub("\\\n".freeze, ''.freeze), @ts, @te)
      else
        current_literal.extend_string(@escape || tok, @ts, @te)
      end
    end
   end
		end
	when 312 then
# line 673 "lib/parser/lexer.rl"
		begin

    diagnostic :fatal, :invalid_escape
  		end
# line 1707 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        value = @escape || tok(@ts + 1)

        if version?(18)
          if defined?(Encoding)
            emit(:tINTEGER, value.dup.force_encoding(Encoding::BINARY)[0].ord)
          else
            emit(:tINTEGER, value[0].ord)
          end
        else
          emit(:tCHARACTER, value)
        end

         @cs = 754; 	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 103 then
# line 692 "lib/parser/lexer.rl"
		begin
 @escape = "\x7f" 		end
# line 839 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    current_literal = literal
    # Get the first character after the backslash.
    escaped_char = @source_buffer.slice(@escape_s).chr

    if current_literal.munge_escape? escaped_char
      # If this particular literal uses this character as an opening
      # or closing delimiter, it is an escape sequence for that
      # particular character. Write it without the backslash.

      if current_literal.regexp? && REGEXP_META_CHARACTERS.match(escaped_char)
        # Regular expressions should include escaped delimiters in their
        # escaped form, except when the escaped character is
        # a closing delimiter but not a regexp metacharacter.
        #
        # The backslash itself cannot be used as a closing delimiter
        # at the same time as an escape symbol, but it is always munged,
        # so this branch also executes for the non-closing-delimiter case
        # for the backslash.
        current_literal.extend_string(tok, @ts, @te)
      else
        current_literal.extend_string(escaped_char, @ts, @te)
      end
    else
      # It does not. So this is an actual escape sequence, yay!
      if current_literal.regexp?
        # Regular expressions should include escape sequences in their
        # escaped form. On the other hand, escaped newlines are removed.
        current_literal.extend_string(tok.gsub("\\\n".freeze, ''.freeze), @ts, @te)
      else
        current_literal.extend_string(@escape || tok, @ts, @te)
      end
    end
   end
		end
	when 129 then
# line 692 "lib/parser/lexer.rl"
		begin
 @escape = "\x7f" 		end
# line 839 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    current_literal = literal
    # Get the first character after the backslash.
    escaped_char = @source_buffer.slice(@escape_s).chr

    if current_literal.munge_escape? escaped_char
      # If this particular literal uses this character as an opening
      # or closing delimiter, it is an escape sequence for that
      # particular character. Write it without the backslash.

      if current_literal.regexp? && REGEXP_META_CHARACTERS.match(escaped_char)
        # Regular expressions should include escaped delimiters in their
        # escaped form, except when the escaped character is
        # a closing delimiter but not a regexp metacharacter.
        #
        # The backslash itself cannot be used as a closing delimiter
        # at the same time as an escape symbol, but it is always munged,
        # so this branch also executes for the non-closing-delimiter case
        # for the backslash.
        current_literal.extend_string(tok, @ts, @te)
      else
        current_literal.extend_string(escaped_char, @ts, @te)
      end
    else
      # It does not. So this is an actual escape sequence, yay!
      if current_literal.regexp?
        # Regular expressions should include escape sequences in their
        # escaped form. On the other hand, escaped newlines are removed.
        current_literal.extend_string(tok.gsub("\\\n".freeze, ''.freeze), @ts, @te)
      else
        current_literal.extend_string(@escape || tok, @ts, @te)
      end
    end
   end
		end
	when 314 then
# line 692 "lib/parser/lexer.rl"
		begin
 @escape = "\x7f" 		end
# line 1707 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        value = @escape || tok(@ts + 1)

        if version?(18)
          if defined?(Encoding)
            emit(:tINTEGER, value.dup.force_encoding(Encoding::BINARY)[0].ord)
          else
            emit(:tINTEGER, value[0].ord)
          end
        else
          emit(:tCHARACTER, value)
        end

         @cs = 754; 	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 100 then
# line 699 "lib/parser/lexer.rl"
		begin
 @escape = encode_escape(tok(@escape_s, p).to_i(8) % 0x100) 		end
# line 839 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    current_literal = literal
    # Get the first character after the backslash.
    escaped_char = @source_buffer.slice(@escape_s).chr

    if current_literal.munge_escape? escaped_char
      # If this particular literal uses this character as an opening
      # or closing delimiter, it is an escape sequence for that
      # particular character. Write it without the backslash.

      if current_literal.regexp? && REGEXP_META_CHARACTERS.match(escaped_char)
        # Regular expressions should include escaped delimiters in their
        # escaped form, except when the escaped character is
        # a closing delimiter but not a regexp metacharacter.
        #
        # The backslash itself cannot be used as a closing delimiter
        # at the same time as an escape symbol, but it is always munged,
        # so this branch also executes for the non-closing-delimiter case
        # for the backslash.
        current_literal.extend_string(tok, @ts, @te)
      else
        current_literal.extend_string(escaped_char, @ts, @te)
      end
    else
      # It does not. So this is an actual escape sequence, yay!
      if current_literal.regexp?
        # Regular expressions should include escape sequences in their
        # escaped form. On the other hand, escaped newlines are removed.
        current_literal.extend_string(tok.gsub("\\\n".freeze, ''.freeze), @ts, @te)
      else
        current_literal.extend_string(@escape || tok, @ts, @te)
      end
    end
   end
		end
	when 126 then
# line 699 "lib/parser/lexer.rl"
		begin
 @escape = encode_escape(tok(@escape_s, p).to_i(8) % 0x100) 		end
# line 839 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    current_literal = literal
    # Get the first character after the backslash.
    escaped_char = @source_buffer.slice(@escape_s).chr

    if current_literal.munge_escape? escaped_char
      # If this particular literal uses this character as an opening
      # or closing delimiter, it is an escape sequence for that
      # particular character. Write it without the backslash.

      if current_literal.regexp? && REGEXP_META_CHARACTERS.match(escaped_char)
        # Regular expressions should include escaped delimiters in their
        # escaped form, except when the escaped character is
        # a closing delimiter but not a regexp metacharacter.
        #
        # The backslash itself cannot be used as a closing delimiter
        # at the same time as an escape symbol, but it is always munged,
        # so this branch also executes for the non-closing-delimiter case
        # for the backslash.
        current_literal.extend_string(tok, @ts, @te)
      else
        current_literal.extend_string(escaped_char, @ts, @te)
      end
    else
      # It does not. So this is an actual escape sequence, yay!
      if current_literal.regexp?
        # Regular expressions should include escape sequences in their
        # escaped form. On the other hand, escaped newlines are removed.
        current_literal.extend_string(tok.gsub("\\\n".freeze, ''.freeze), @ts, @te)
      else
        current_literal.extend_string(@escape || tok, @ts, @te)
      end
    end
   end
		end
	when 311 then
# line 699 "lib/parser/lexer.rl"
		begin
 @escape = encode_escape(tok(@escape_s, p).to_i(8) % 0x100) 		end
# line 1707 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        value = @escape || tok(@ts + 1)

        if version?(18)
          if defined?(Encoding)
            emit(:tINTEGER, value.dup.force_encoding(Encoding::BINARY)[0].ord)
          else
            emit(:tINTEGER, value[0].ord)
          end
        else
          emit(:tCHARACTER, value)
        end

         @cs = 754; 	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 116 then
# line 703 "lib/parser/lexer.rl"
		begin
 @escape = encode_escape(tok(@escape_s + 1, p).to_i(16)) 		end
# line 839 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    current_literal = literal
    # Get the first character after the backslash.
    escaped_char = @source_buffer.slice(@escape_s).chr

    if current_literal.munge_escape? escaped_char
      # If this particular literal uses this character as an opening
      # or closing delimiter, it is an escape sequence for that
      # particular character. Write it without the backslash.

      if current_literal.regexp? && REGEXP_META_CHARACTERS.match(escaped_char)
        # Regular expressions should include escaped delimiters in their
        # escaped form, except when the escaped character is
        # a closing delimiter but not a regexp metacharacter.
        #
        # The backslash itself cannot be used as a closing delimiter
        # at the same time as an escape symbol, but it is always munged,
        # so this branch also executes for the non-closing-delimiter case
        # for the backslash.
        current_literal.extend_string(tok, @ts, @te)
      else
        current_literal.extend_string(escaped_char, @ts, @te)
      end
    else
      # It does not. So this is an actual escape sequence, yay!
      if current_literal.regexp?
        # Regular expressions should include escape sequences in their
        # escaped form. On the other hand, escaped newlines are removed.
        current_literal.extend_string(tok.gsub("\\\n".freeze, ''.freeze), @ts, @te)
      else
        current_literal.extend_string(@escape || tok, @ts, @te)
      end
    end
   end
		end
	when 142 then
# line 703 "lib/parser/lexer.rl"
		begin
 @escape = encode_escape(tok(@escape_s + 1, p).to_i(16)) 		end
# line 839 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    current_literal = literal
    # Get the first character after the backslash.
    escaped_char = @source_buffer.slice(@escape_s).chr

    if current_literal.munge_escape? escaped_char
      # If this particular literal uses this character as an opening
      # or closing delimiter, it is an escape sequence for that
      # particular character. Write it without the backslash.

      if current_literal.regexp? && REGEXP_META_CHARACTERS.match(escaped_char)
        # Regular expressions should include escaped delimiters in their
        # escaped form, except when the escaped character is
        # a closing delimiter but not a regexp metacharacter.
        #
        # The backslash itself cannot be used as a closing delimiter
        # at the same time as an escape symbol, but it is always munged,
        # so this branch also executes for the non-closing-delimiter case
        # for the backslash.
        current_literal.extend_string(tok, @ts, @te)
      else
        current_literal.extend_string(escaped_char, @ts, @te)
      end
    else
      # It does not. So this is an actual escape sequence, yay!
      if current_literal.regexp?
        # Regular expressions should include escape sequences in their
        # escaped form. On the other hand, escaped newlines are removed.
        current_literal.extend_string(tok.gsub("\\\n".freeze, ''.freeze), @ts, @te)
      else
        current_literal.extend_string(@escape || tok, @ts, @te)
      end
    end
   end
		end
	when 327 then
# line 703 "lib/parser/lexer.rl"
		begin
 @escape = encode_escape(tok(@escape_s + 1, p).to_i(16)) 		end
# line 1707 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        value = @escape || tok(@ts + 1)

        if version?(18)
          if defined?(Encoding)
            emit(:tINTEGER, value.dup.force_encoding(Encoding::BINARY)[0].ord)
          else
            emit(:tINTEGER, value[0].ord)
          end
        else
          emit(:tCHARACTER, value)
        end

         @cs = 754; 	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 111 then
# line 707 "lib/parser/lexer.rl"
		begin
 @escape = tok(@escape_s + 1, p).to_i(16).chr(Encoding::UTF_8) 		end
# line 839 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    current_literal = literal
    # Get the first character after the backslash.
    escaped_char = @source_buffer.slice(@escape_s).chr

    if current_literal.munge_escape? escaped_char
      # If this particular literal uses this character as an opening
      # or closing delimiter, it is an escape sequence for that
      # particular character. Write it without the backslash.

      if current_literal.regexp? && REGEXP_META_CHARACTERS.match(escaped_char)
        # Regular expressions should include escaped delimiters in their
        # escaped form, except when the escaped character is
        # a closing delimiter but not a regexp metacharacter.
        #
        # The backslash itself cannot be used as a closing delimiter
        # at the same time as an escape symbol, but it is always munged,
        # so this branch also executes for the non-closing-delimiter case
        # for the backslash.
        current_literal.extend_string(tok, @ts, @te)
      else
        current_literal.extend_string(escaped_char, @ts, @te)
      end
    else
      # It does not. So this is an actual escape sequence, yay!
      if current_literal.regexp?
        # Regular expressions should include escape sequences in their
        # escaped form. On the other hand, escaped newlines are removed.
        current_literal.extend_string(tok.gsub("\\\n".freeze, ''.freeze), @ts, @te)
      else
        current_literal.extend_string(@escape || tok, @ts, @te)
      end
    end
   end
		end
	when 137 then
# line 707 "lib/parser/lexer.rl"
		begin
 @escape = tok(@escape_s + 1, p).to_i(16).chr(Encoding::UTF_8) 		end
# line 839 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    current_literal = literal
    # Get the first character after the backslash.
    escaped_char = @source_buffer.slice(@escape_s).chr

    if current_literal.munge_escape? escaped_char
      # If this particular literal uses this character as an opening
      # or closing delimiter, it is an escape sequence for that
      # particular character. Write it without the backslash.

      if current_literal.regexp? && REGEXP_META_CHARACTERS.match(escaped_char)
        # Regular expressions should include escaped delimiters in their
        # escaped form, except when the escaped character is
        # a closing delimiter but not a regexp metacharacter.
        #
        # The backslash itself cannot be used as a closing delimiter
        # at the same time as an escape symbol, but it is always munged,
        # so this branch also executes for the non-closing-delimiter case
        # for the backslash.
        current_literal.extend_string(tok, @ts, @te)
      else
        current_literal.extend_string(escaped_char, @ts, @te)
      end
    else
      # It does not. So this is an actual escape sequence, yay!
      if current_literal.regexp?
        # Regular expressions should include escape sequences in their
        # escaped form. On the other hand, escaped newlines are removed.
        current_literal.extend_string(tok.gsub("\\\n".freeze, ''.freeze), @ts, @te)
      else
        current_literal.extend_string(@escape || tok, @ts, @te)
      end
    end
   end
		end
	when 322 then
# line 707 "lib/parser/lexer.rl"
		begin
 @escape = tok(@escape_s + 1, p).to_i(16).chr(Encoding::UTF_8) 		end
# line 1707 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        value = @escape || tok(@ts + 1)

        if version?(18)
          if defined?(Encoding)
            emit(:tINTEGER, value.dup.force_encoding(Encoding::BINARY)[0].ord)
          else
            emit(:tINTEGER, value[0].ord)
          end
        else
          emit(:tCHARACTER, value)
        end

         @cs = 754; 	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 115 then
# line 711 "lib/parser/lexer.rl"
		begin

        diagnostic :fatal, :invalid_hex_escape, nil, range(@escape_s - 1, p + 2)
      		end
# line 839 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    current_literal = literal
    # Get the first character after the backslash.
    escaped_char = @source_buffer.slice(@escape_s).chr

    if current_literal.munge_escape? escaped_char
      # If this particular literal uses this character as an opening
      # or closing delimiter, it is an escape sequence for that
      # particular character. Write it without the backslash.

      if current_literal.regexp? && REGEXP_META_CHARACTERS.match(escaped_char)
        # Regular expressions should include escaped delimiters in their
        # escaped form, except when the escaped character is
        # a closing delimiter but not a regexp metacharacter.
        #
        # The backslash itself cannot be used as a closing delimiter
        # at the same time as an escape symbol, but it is always munged,
        # so this branch also executes for the non-closing-delimiter case
        # for the backslash.
        current_literal.extend_string(tok, @ts, @te)
      else
        current_literal.extend_string(escaped_char, @ts, @te)
      end
    else
      # It does not. So this is an actual escape sequence, yay!
      if current_literal.regexp?
        # Regular expressions should include escape sequences in their
        # escaped form. On the other hand, escaped newlines are removed.
        current_literal.extend_string(tok.gsub("\\\n".freeze, ''.freeze), @ts, @te)
      else
        current_literal.extend_string(@escape || tok, @ts, @te)
      end
    end
   end
		end
	when 141 then
# line 711 "lib/parser/lexer.rl"
		begin

        diagnostic :fatal, :invalid_hex_escape, nil, range(@escape_s - 1, p + 2)
      		end
# line 839 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    current_literal = literal
    # Get the first character after the backslash.
    escaped_char = @source_buffer.slice(@escape_s).chr

    if current_literal.munge_escape? escaped_char
      # If this particular literal uses this character as an opening
      # or closing delimiter, it is an escape sequence for that
      # particular character. Write it without the backslash.

      if current_literal.regexp? && REGEXP_META_CHARACTERS.match(escaped_char)
        # Regular expressions should include escaped delimiters in their
        # escaped form, except when the escaped character is
        # a closing delimiter but not a regexp metacharacter.
        #
        # The backslash itself cannot be used as a closing delimiter
        # at the same time as an escape symbol, but it is always munged,
        # so this branch also executes for the non-closing-delimiter case
        # for the backslash.
        current_literal.extend_string(tok, @ts, @te)
      else
        current_literal.extend_string(escaped_char, @ts, @te)
      end
    else
      # It does not. So this is an actual escape sequence, yay!
      if current_literal.regexp?
        # Regular expressions should include escape sequences in their
        # escaped form. On the other hand, escaped newlines are removed.
        current_literal.extend_string(tok.gsub("\\\n".freeze, ''.freeze), @ts, @te)
      else
        current_literal.extend_string(@escape || tok, @ts, @te)
      end
    end
   end
		end
	when 326 then
# line 711 "lib/parser/lexer.rl"
		begin

        diagnostic :fatal, :invalid_hex_escape, nil, range(@escape_s - 1, p + 2)
      		end
# line 1707 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        value = @escape || tok(@ts + 1)

        if version?(18)
          if defined?(Encoding)
            emit(:tINTEGER, value.dup.force_encoding(Encoding::BINARY)[0].ord)
          else
            emit(:tINTEGER, value[0].ord)
          end
        else
          emit(:tCHARACTER, value)
        end

         @cs = 754; 	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 110 then
# line 723 "lib/parser/lexer.rl"
		begin

        diagnostic :fatal, :invalid_unicode_escape, nil, range(@escape_s - 1, p)
      		end
# line 839 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    current_literal = literal
    # Get the first character after the backslash.
    escaped_char = @source_buffer.slice(@escape_s).chr

    if current_literal.munge_escape? escaped_char
      # If this particular literal uses this character as an opening
      # or closing delimiter, it is an escape sequence for that
      # particular character. Write it without the backslash.

      if current_literal.regexp? && REGEXP_META_CHARACTERS.match(escaped_char)
        # Regular expressions should include escaped delimiters in their
        # escaped form, except when the escaped character is
        # a closing delimiter but not a regexp metacharacter.
        #
        # The backslash itself cannot be used as a closing delimiter
        # at the same time as an escape symbol, but it is always munged,
        # so this branch also executes for the non-closing-delimiter case
        # for the backslash.
        current_literal.extend_string(tok, @ts, @te)
      else
        current_literal.extend_string(escaped_char, @ts, @te)
      end
    else
      # It does not. So this is an actual escape sequence, yay!
      if current_literal.regexp?
        # Regular expressions should include escape sequences in their
        # escaped form. On the other hand, escaped newlines are removed.
        current_literal.extend_string(tok.gsub("\\\n".freeze, ''.freeze), @ts, @te)
      else
        current_literal.extend_string(@escape || tok, @ts, @te)
      end
    end
   end
		end
	when 136 then
# line 723 "lib/parser/lexer.rl"
		begin

        diagnostic :fatal, :invalid_unicode_escape, nil, range(@escape_s - 1, p)
      		end
# line 839 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    current_literal = literal
    # Get the first character after the backslash.
    escaped_char = @source_buffer.slice(@escape_s).chr

    if current_literal.munge_escape? escaped_char
      # If this particular literal uses this character as an opening
      # or closing delimiter, it is an escape sequence for that
      # particular character. Write it without the backslash.

      if current_literal.regexp? && REGEXP_META_CHARACTERS.match(escaped_char)
        # Regular expressions should include escaped delimiters in their
        # escaped form, except when the escaped character is
        # a closing delimiter but not a regexp metacharacter.
        #
        # The backslash itself cannot be used as a closing delimiter
        # at the same time as an escape symbol, but it is always munged,
        # so this branch also executes for the non-closing-delimiter case
        # for the backslash.
        current_literal.extend_string(tok, @ts, @te)
      else
        current_literal.extend_string(escaped_char, @ts, @te)
      end
    else
      # It does not. So this is an actual escape sequence, yay!
      if current_literal.regexp?
        # Regular expressions should include escape sequences in their
        # escaped form. On the other hand, escaped newlines are removed.
        current_literal.extend_string(tok.gsub("\\\n".freeze, ''.freeze), @ts, @te)
      else
        current_literal.extend_string(@escape || tok, @ts, @te)
      end
    end
   end
		end
	when 321 then
# line 723 "lib/parser/lexer.rl"
		begin

        diagnostic :fatal, :invalid_unicode_escape, nil, range(@escape_s - 1, p)
      		end
# line 1707 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        value = @escape || tok(@ts + 1)

        if version?(18)
          if defined?(Encoding)
            emit(:tINTEGER, value.dup.force_encoding(Encoding::BINARY)[0].ord)
          else
            emit(:tINTEGER, value[0].ord)
          end
        else
          emit(:tCHARACTER, value)
        end

         @cs = 754; 	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 113 then
# line 734 "lib/parser/lexer.rl"
		begin

          diagnostic :fatal, :unterminated_unicode, nil, range(p - 1, p)
        		end
# line 839 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    current_literal = literal
    # Get the first character after the backslash.
    escaped_char = @source_buffer.slice(@escape_s).chr

    if current_literal.munge_escape? escaped_char
      # If this particular literal uses this character as an opening
      # or closing delimiter, it is an escape sequence for that
      # particular character. Write it without the backslash.

      if current_literal.regexp? && REGEXP_META_CHARACTERS.match(escaped_char)
        # Regular expressions should include escaped delimiters in their
        # escaped form, except when the escaped character is
        # a closing delimiter but not a regexp metacharacter.
        #
        # The backslash itself cannot be used as a closing delimiter
        # at the same time as an escape symbol, but it is always munged,
        # so this branch also executes for the non-closing-delimiter case
        # for the backslash.
        current_literal.extend_string(tok, @ts, @te)
      else
        current_literal.extend_string(escaped_char, @ts, @te)
      end
    else
      # It does not. So this is an actual escape sequence, yay!
      if current_literal.regexp?
        # Regular expressions should include escape sequences in their
        # escaped form. On the other hand, escaped newlines are removed.
        current_literal.extend_string(tok.gsub("\\\n".freeze, ''.freeze), @ts, @te)
      else
        current_literal.extend_string(@escape || tok, @ts, @te)
      end
    end
   end
		end
	when 139 then
# line 734 "lib/parser/lexer.rl"
		begin

          diagnostic :fatal, :unterminated_unicode, nil, range(p - 1, p)
        		end
# line 839 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    current_literal = literal
    # Get the first character after the backslash.
    escaped_char = @source_buffer.slice(@escape_s).chr

    if current_literal.munge_escape? escaped_char
      # If this particular literal uses this character as an opening
      # or closing delimiter, it is an escape sequence for that
      # particular character. Write it without the backslash.

      if current_literal.regexp? && REGEXP_META_CHARACTERS.match(escaped_char)
        # Regular expressions should include escaped delimiters in their
        # escaped form, except when the escaped character is
        # a closing delimiter but not a regexp metacharacter.
        #
        # The backslash itself cannot be used as a closing delimiter
        # at the same time as an escape symbol, but it is always munged,
        # so this branch also executes for the non-closing-delimiter case
        # for the backslash.
        current_literal.extend_string(tok, @ts, @te)
      else
        current_literal.extend_string(escaped_char, @ts, @te)
      end
    else
      # It does not. So this is an actual escape sequence, yay!
      if current_literal.regexp?
        # Regular expressions should include escape sequences in their
        # escaped form. On the other hand, escaped newlines are removed.
        current_literal.extend_string(tok.gsub("\\\n".freeze, ''.freeze), @ts, @te)
      else
        current_literal.extend_string(@escape || tok, @ts, @te)
      end
    end
   end
		end
	when 324 then
# line 734 "lib/parser/lexer.rl"
		begin

          diagnostic :fatal, :unterminated_unicode, nil, range(p - 1, p)
        		end
# line 1707 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        value = @escape || tok(@ts + 1)

        if version?(18)
          if defined?(Encoding)
            emit(:tINTEGER, value.dup.force_encoding(Encoding::BINARY)[0].ord)
          else
            emit(:tINTEGER, value[0].ord)
          end
        else
          emit(:tCHARACTER, value)
        end

         @cs = 754; 	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 99 then
# line 760 "lib/parser/lexer.rl"
		begin

      diagnostic :fatal, :escape_eof, nil, range(p - 1, p)
    		end
# line 839 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    current_literal = literal
    # Get the first character after the backslash.
    escaped_char = @source_buffer.slice(@escape_s).chr

    if current_literal.munge_escape? escaped_char
      # If this particular literal uses this character as an opening
      # or closing delimiter, it is an escape sequence for that
      # particular character. Write it without the backslash.

      if current_literal.regexp? && REGEXP_META_CHARACTERS.match(escaped_char)
        # Regular expressions should include escaped delimiters in their
        # escaped form, except when the escaped character is
        # a closing delimiter but not a regexp metacharacter.
        #
        # The backslash itself cannot be used as a closing delimiter
        # at the same time as an escape symbol, but it is always munged,
        # so this branch also executes for the non-closing-delimiter case
        # for the backslash.
        current_literal.extend_string(tok, @ts, @te)
      else
        current_literal.extend_string(escaped_char, @ts, @te)
      end
    else
      # It does not. So this is an actual escape sequence, yay!
      if current_literal.regexp?
        # Regular expressions should include escape sequences in their
        # escaped form. On the other hand, escaped newlines are removed.
        current_literal.extend_string(tok.gsub("\\\n".freeze, ''.freeze), @ts, @te)
      else
        current_literal.extend_string(@escape || tok, @ts, @te)
      end
    end
   end
		end
	when 125 then
# line 760 "lib/parser/lexer.rl"
		begin

      diagnostic :fatal, :escape_eof, nil, range(p - 1, p)
    		end
# line 839 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    current_literal = literal
    # Get the first character after the backslash.
    escaped_char = @source_buffer.slice(@escape_s).chr

    if current_literal.munge_escape? escaped_char
      # If this particular literal uses this character as an opening
      # or closing delimiter, it is an escape sequence for that
      # particular character. Write it without the backslash.

      if current_literal.regexp? && REGEXP_META_CHARACTERS.match(escaped_char)
        # Regular expressions should include escaped delimiters in their
        # escaped form, except when the escaped character is
        # a closing delimiter but not a regexp metacharacter.
        #
        # The backslash itself cannot be used as a closing delimiter
        # at the same time as an escape symbol, but it is always munged,
        # so this branch also executes for the non-closing-delimiter case
        # for the backslash.
        current_literal.extend_string(tok, @ts, @te)
      else
        current_literal.extend_string(escaped_char, @ts, @te)
      end
    else
      # It does not. So this is an actual escape sequence, yay!
      if current_literal.regexp?
        # Regular expressions should include escape sequences in their
        # escaped form. On the other hand, escaped newlines are removed.
        current_literal.extend_string(tok.gsub("\\\n".freeze, ''.freeze), @ts, @te)
      else
        current_literal.extend_string(@escape || tok, @ts, @te)
      end
    end
   end
		end
	when 310 then
# line 760 "lib/parser/lexer.rl"
		begin

      diagnostic :fatal, :escape_eof, nil, range(p - 1, p)
    		end
# line 1707 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        value = @escape || tok(@ts + 1)

        if version?(18)
          if defined?(Encoding)
            emit(:tINTEGER, value.dup.force_encoding(Encoding::BINARY)[0].ord)
          else
            emit(:tINTEGER, value[0].ord)
          end
        else
          emit(:tCHARACTER, value)
        end

         @cs = 754; 	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 148 then
# line 766 "lib/parser/lexer.rl"
		begin

    @escape_s = p
    @escape   = nil
  		end
# line 839 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
    current_literal = literal
    # Get the first character after the backslash.
    escaped_char = @source_buffer.slice(@escape_s).chr

    if current_literal.munge_escape? escaped_char
      # If this particular literal uses this character as an opening
      # or closing delimiter, it is an escape sequence for that
      # particular character. Write it without the backslash.

      if current_literal.regexp? && REGEXP_META_CHARACTERS.match(escaped_char)
        # Regular expressions should include escaped delimiters in their
        # escaped form, except when the escaped character is
        # a closing delimiter but not a regexp metacharacter.
        #
        # The backslash itself cannot be used as a closing delimiter
        # at the same time as an escape symbol, but it is always munged,
        # so this branch also executes for the non-closing-delimiter case
        # for the backslash.
        current_literal.extend_string(tok, @ts, @te)
      else
        current_literal.extend_string(escaped_char, @ts, @te)
      end
    else
      # It does not. So this is an actual escape sequence, yay!
      if current_literal.regexp?
        # Regular expressions should include escape sequences in their
        # escaped form. On the other hand, escaped newlines are removed.
        current_literal.extend_string(tok.gsub("\\\n".freeze, ''.freeze), @ts, @te)
      else
        current_literal.extend_string(@escape || tok, @ts, @te)
      end
    end
   end
		end
	when 153 then
# line 766 "lib/parser/lexer.rl"
		begin

    @escape_s = p
    @escape   = nil
  		end
# line 839 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
    current_literal = literal
    # Get the first character after the backslash.
    escaped_char = @source_buffer.slice(@escape_s).chr

    if current_literal.munge_escape? escaped_char
      # If this particular literal uses this character as an opening
      # or closing delimiter, it is an escape sequence for that
      # particular character. Write it without the backslash.

      if current_literal.regexp? && REGEXP_META_CHARACTERS.match(escaped_char)
        # Regular expressions should include escaped delimiters in their
        # escaped form, except when the escaped character is
        # a closing delimiter but not a regexp metacharacter.
        #
        # The backslash itself cannot be used as a closing delimiter
        # at the same time as an escape symbol, but it is always munged,
        # so this branch also executes for the non-closing-delimiter case
        # for the backslash.
        current_literal.extend_string(tok, @ts, @te)
      else
        current_literal.extend_string(escaped_char, @ts, @te)
      end
    else
      # It does not. So this is an actual escape sequence, yay!
      if current_literal.regexp?
        # Regular expressions should include escape sequences in their
        # escaped form. On the other hand, escaped newlines are removed.
        current_literal.extend_string(tok.gsub("\\\n".freeze, ''.freeze), @ts, @te)
      else
        current_literal.extend_string(@escape || tok, @ts, @te)
      end
    end
   end
		end
	when 64 then
# line 806 "lib/parser/lexer.rl"
		begin

    # After every heredoc was parsed, @herebody_s contains the
    # position of next token after all heredocs.
    if @herebody_s
      p = @herebody_s
      @herebody_s = nil
    end
  		end
# line 486 "lib/parser/lexer.rl"
		begin

    # Record position of a newline for precise location reporting on tNL
    # tokens.
    #
    # This action is embedded directly into c_nl, as it is idempotent and
    # there are no cases when we need to skip it.
    @newline_s = p
  		end
	when 28 then
# line 806 "lib/parser/lexer.rl"
		begin

    # After every heredoc was parsed, @herebody_s contains the
    # position of next token after all heredocs.
    if @herebody_s
      p = @herebody_s
      @herebody_s = nil
    end
  		end
# line 1421 "lib/parser/lexer.rl"
		begin
 tm = p 		end
	when 30 then
# line 806 "lib/parser/lexer.rl"
		begin

    # After every heredoc was parsed, @herebody_s contains the
    # position of next token after all heredocs.
    if @herebody_s
      p = @herebody_s
      @herebody_s = nil
    end
  		end
# line 1437 "lib/parser/lexer.rl"
		begin
 tm = p 		end
	when 32 then
# line 806 "lib/parser/lexer.rl"
		begin

    # After every heredoc was parsed, @herebody_s contains the
    # position of next token after all heredocs.
    if @herebody_s
      p = @herebody_s
      @herebody_s = nil
    end
  		end
# line 1465 "lib/parser/lexer.rl"
		begin
 tm = p 		end
	when 177 then
# line 806 "lib/parser/lexer.rl"
		begin

    # After every heredoc was parsed, @herebody_s contains the
    # position of next token after all heredocs.
    if @herebody_s
      p = @herebody_s
      @herebody_s = nil
    end
  		end
# line 1296 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1;		end
	when 196 then
# line 806 "lib/parser/lexer.rl"
		begin

    # After every heredoc was parsed, @herebody_s contains the
    # position of next token after all heredocs.
    if @herebody_s
      p = @herebody_s
      @herebody_s = nil
    end
  		end
# line 1314 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1;		end
	when 204 then
# line 806 "lib/parser/lexer.rl"
		begin

    # After every heredoc was parsed, @herebody_s contains the
    # position of next token after all heredocs.
    if @herebody_s
      p = @herebody_s
      @herebody_s = nil
    end
  		end
# line 1346 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1;		end
	when 31 then
# line 806 "lib/parser/lexer.rl"
		begin

    # After every heredoc was parsed, @herebody_s contains the
    # position of next token after all heredocs.
    if @herebody_s
      p = @herebody_s
      @herebody_s = nil
    end
  		end
# line 1475 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
        p = @ts - 1
        	begin
		 @cs = 754
		_goto_level = _again
		next
	end

       end
		end
	when 235 then
# line 806 "lib/parser/lexer.rl"
		begin

    # After every heredoc was parsed, @herebody_s contains the
    # position of next token after all heredocs.
    if @herebody_s
      p = @herebody_s
      @herebody_s = nil
    end
  		end
# line 1480 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1;		end
	when 227 then
# line 806 "lib/parser/lexer.rl"
		begin

    # After every heredoc was parsed, @herebody_s contains the
    # position of next token after all heredocs.
    if @herebody_s
      p = @herebody_s
      @herebody_s = nil
    end
  		end
# line 1486 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  p = p - 1; 	begin
		 @cs = 754
		_goto_level = _again
		next
	end
  end
		end
	when 246 then
# line 806 "lib/parser/lexer.rl"
		begin

    # After every heredoc was parsed, @herebody_s contains the
    # position of next token after all heredocs.
    if @herebody_s
      p = @herebody_s
      @herebody_s = nil
    end
  		end
# line 1555 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1;		end
	when 258 then
# line 806 "lib/parser/lexer.rl"
		begin

    # After every heredoc was parsed, @herebody_s contains the
    # position of next token after all heredocs.
    if @herebody_s
      p = @herebody_s
      @herebody_s = nil
    end
  		end
# line 1576 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1;		end
	when 254 then
# line 806 "lib/parser/lexer.rl"
		begin

    # After every heredoc was parsed, @herebody_s contains the
    # position of next token after all heredocs.
    if @herebody_s
      p = @herebody_s
      @herebody_s = nil
    end
  		end
# line 1579 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  p = p - 1; 	begin
		 @cs = 754
		_goto_level = _again
		next
	end
  end
		end
	when 274 then
# line 806 "lib/parser/lexer.rl"
		begin

    # After every heredoc was parsed, @herebody_s contains the
    # position of next token after all heredocs.
    if @herebody_s
      p = @herebody_s
      @herebody_s = nil
    end
  		end
# line 1836 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1;		end
	when 344 then
# line 806 "lib/parser/lexer.rl"
		begin

    # After every heredoc was parsed, @herebody_s contains the
    # position of next token after all heredocs.
    if @herebody_s
      p = @herebody_s
      @herebody_s = nil
    end
  		end
# line 1860 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1;		end
	when 340 then
# line 806 "lib/parser/lexer.rl"
		begin

    # After every heredoc was parsed, @herebody_s contains the
    # position of next token after all heredocs.
    if @herebody_s
      p = @herebody_s
      @herebody_s = nil
    end
  		end
# line 1863 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
      if @in_kwarg
        p = p - 1; 	begin
		 @cs = 754
		_goto_level = _again
		next
	end

      else
        	begin
		 @cs = 160
		_goto_level = _again
		next
	end

      end
     end
		end
	when 353 then
# line 806 "lib/parser/lexer.rl"
		begin

    # After every heredoc was parsed, @herebody_s contains the
    # position of next token after all heredocs.
    if @herebody_s
      p = @herebody_s
      @herebody_s = nil
    end
  		end
# line 1891 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1;		end
	when 350 then
# line 806 "lib/parser/lexer.rl"
		begin

    # After every heredoc was parsed, @herebody_s contains the
    # position of next token after all heredocs.
    if @herebody_s
      p = @herebody_s
      @herebody_s = nil
    end
  		end
# line 1894 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  	begin
		 @cs = 160
		_goto_level = _again
		next
	end
  end
		end
	when 426 then
# line 806 "lib/parser/lexer.rl"
		begin

    # After every heredoc was parsed, @herebody_s contains the
    # position of next token after all heredocs.
    if @herebody_s
      p = @herebody_s
      @herebody_s = nil
    end
  		end
# line 2169 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1;		end
	when 368 then
# line 806 "lib/parser/lexer.rl"
		begin

    # After every heredoc was parsed, @herebody_s contains the
    # position of next token after all heredocs.
    if @herebody_s
      p = @herebody_s
      @herebody_s = nil
    end
  		end
# line 2172 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  	begin
		 @cs = 927
		_goto_level = _again
		next
	end
  end
		end
	when 82 then
# line 806 "lib/parser/lexer.rl"
		begin

    # After every heredoc was parsed, @herebody_s contains the
    # position of next token after all heredocs.
    if @herebody_s
      p = @herebody_s
      @herebody_s = nil
    end
  		end
# line 2224 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1;		end
	when 226 then
# line 977 "lib/parser/lexer.rl"
		begin

    @cond.push(false); @cmdarg.push(false)

    current_literal = literal
    if current_literal
      current_literal.start_interp_brace
    end
  		end
# line 1390 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        if @lambda_stack.last == @paren_nest
          p = @ts - 1
          	begin
		 @cs = 754
		_goto_level = _again
		next
	end

        else
          emit(:tLCURLY, '{'.freeze, @te - 1, @te)
           @cs = 746; 	begin
		p += 1
		_goto_level = _out
		next
	end

        end
       end
		end
	when 248 then
# line 977 "lib/parser/lexer.rl"
		begin

    @cond.push(false); @cmdarg.push(false)

    current_literal = literal
    if current_literal
      current_literal.start_interp_brace
    end
  		end
# line 1548 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  emit(:tLBRACE_ARG, '{'.freeze)
            @cs = 746;  end
		end
	when 336 then
# line 977 "lib/parser/lexer.rl"
		begin

    @cond.push(false); @cmdarg.push(false)

    current_literal = literal
    if current_literal
      current_literal.start_interp_brace
    end
  		end
# line 1751 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        if @lambda_stack.last == @paren_nest
          @lambda_stack.pop
          emit(:tLAMBEG, '{'.freeze)
        else
          emit(:tLBRACE, '{'.freeze)
        end
        	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 438 then
# line 977 "lib/parser/lexer.rl"
		begin

    @cond.push(false); @cmdarg.push(false)

    current_literal = literal
    if current_literal
      current_literal.start_interp_brace
    end
  		end
# line 1916 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        if @lambda_stack.last == @paren_nest
          @lambda_stack.pop

          if tok == '{'.freeze
            emit(:tLAMBEG, '{'.freeze)
          else # 'do'
            emit(:kDO_LAMBDA, 'do'.freeze)
          end
        else
          if tok == '{'.freeze
            emit(:tLCURLY, '{'.freeze)
          else # 'do'
            emit_do
          end
        end

         @cs = 746; 	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 439 then
# line 986 "lib/parser/lexer.rl"
		begin

    current_literal = literal
    if current_literal
      if current_literal.end_interp_brace_and_try_closing
        if version?(18, 19)
          emit(:tRCURLY, '}'.freeze, p - 1, p)
        else
          emit(:tSTRING_DEND, '}'.freeze, p - 1, p)
        end

        if current_literal.saved_herebody_s
          @herebody_s = current_literal.saved_herebody_s
        end

        p = p - 1;
         @cs = (stack_pop);
        	begin
		p += 1
		_goto_level = _out
		next
	end

      end
    end
  		end
# line 2136 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        emit_table(PUNCTUATION)
        @cond.lexpop; @cmdarg.lexpop

        if RBRACE_OR_RBRACK.include?(tok)
           @cs = 493;
        else # )
          # fnext expr_endfn; ?
        end

        	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 66 then
# line 1108 "lib/parser/lexer.rl"
		begin
 @sharp_s = p - 1 		end
# line 1111 "lib/parser/lexer.rl"
		begin
 emit_comment(@sharp_s, p == pe ? p - 2 : p) 		end
	when 69 then
# line 1111 "lib/parser/lexer.rl"
		begin
 emit_comment(@sharp_s, p == pe ? p - 2 : p) 		end
# line 486 "lib/parser/lexer.rl"
		begin

    # Record position of a newline for precise location reporting on tNL
    # tokens.
    #
    # This action is embedded directly into c_nl, as it is idempotent and
    # there are no cases when we need to skip it.
    @newline_s = p
  		end
	when 180 then
# line 1111 "lib/parser/lexer.rl"
		begin
 emit_comment(@sharp_s, p == pe ? p - 2 : p) 		end
# line 1296 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1;		end
	when 195 then
# line 1111 "lib/parser/lexer.rl"
		begin
 emit_comment(@sharp_s, p == pe ? p - 2 : p) 		end
# line 1314 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1;		end
	when 207 then
# line 1111 "lib/parser/lexer.rl"
		begin
 emit_comment(@sharp_s, p == pe ? p - 2 : p) 		end
# line 1346 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1;		end
	when 229 then
# line 1111 "lib/parser/lexer.rl"
		begin
 emit_comment(@sharp_s, p == pe ? p - 2 : p) 		end
# line 1483 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  	begin
		 @cs = 754
		_goto_level = _again
		next
	end
  end
		end
	when 244 then
# line 1111 "lib/parser/lexer.rl"
		begin
 emit_comment(@sharp_s, p == pe ? p - 2 : p) 		end
# line 1555 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1;		end
	when 256 then
# line 1111 "lib/parser/lexer.rl"
		begin
 emit_comment(@sharp_s, p == pe ? p - 2 : p) 		end
# line 1576 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1;		end
	when 280 then
# line 1111 "lib/parser/lexer.rl"
		begin
 emit_comment(@sharp_s, p == pe ? p - 2 : p) 		end
# line 1836 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1;		end
	when 342 then
# line 1111 "lib/parser/lexer.rl"
		begin
 emit_comment(@sharp_s, p == pe ? p - 2 : p) 		end
# line 1860 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1;		end
	when 352 then
# line 1111 "lib/parser/lexer.rl"
		begin
 emit_comment(@sharp_s, p == pe ? p - 2 : p) 		end
# line 1891 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1;		end
	when 372 then
# line 1111 "lib/parser/lexer.rl"
		begin
 emit_comment(@sharp_s, p == pe ? p - 2 : p) 		end
# line 2169 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1;		end
	when 84 then
# line 1111 "lib/parser/lexer.rl"
		begin
 emit_comment(@sharp_s, p == pe ? p - 2 : p) 		end
# line 2224 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1;		end
	when 201 then
# line 1152 "lib/parser/lexer.rl"
		begin
 tm = p 		end
# line 1336 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  emit(:tFID, tok(@ts, tm), @ts, tm)
            @cs = (arg_or_cmdarg); p = tm - 1; 	begin
		p += 1
		_goto_level = _out
		next
	end
  end
		end
	when 292 then
# line 1152 "lib/parser/lexer.rl"
		begin
 tm = p 		end
# line 1685 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        emit(:tSYMBOL, tok(@ts + 1, tm), @ts, tm)
        p = tm - 1
         @cs = 754; 	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 265 then
# line 1152 "lib/parser/lexer.rl"
		begin
 tm = p 		end
# line 1824 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  p = @ts - 1
           	begin
		 @cs = 754
		_goto_level = _again
		next
	end
  end
		end
	when 365 then
# line 1152 "lib/parser/lexer.rl"
		begin
 tm = p 		end
# line 1 "NONE"
		begin
	case  @act
	when 124 then
	begin begin p = (( @te))-1; end

        if @lambda_stack.last == @paren_nest
          @lambda_stack.pop

          if tok == '{'.freeze
            emit(:tLAMBEG, '{'.freeze)
          else # 'do'
            emit(:kDO_LAMBDA, 'do'.freeze)
          end
        else
          if tok == '{'.freeze
            emit(:tLCURLY, '{'.freeze)
          else # 'do'
            emit_do
          end
        end

         @cs = 746; 	begin
		p += 1
		_goto_level = _out
		next
	end

      end
	when 125 then
	begin begin p = (( @te))-1; end
 emit_table(KEYWORDS)
            @cs = 315; 	begin
		p += 1
		_goto_level = _out
		next
	end
 end
	when 126 then
	begin begin p = (( @te))-1; end
 emit(:kCLASS, 'class'.freeze, @ts, @ts + 5)
           emit(:tLSHFT, '<<'.freeze,    @te - 2, @te)
            @cs = 746; 	begin
		p += 1
		_goto_level = _out
		next
	end
 end
	when 127 then
	begin begin p = (( @te))-1; end
 emit_table(KEYWORDS)
            @cs = 525; 	begin
		p += 1
		_goto_level = _out
		next
	end
 end
	when 128 then
	begin begin p = (( @te))-1; end
 emit_table(KEYWORDS)
            @cs = 746; 	begin
		p += 1
		_goto_level = _out
		next
	end
 end
	when 129 then
	begin begin p = (( @te))-1; end
 emit_table(KEYWORDS)
            @cs = 501; 	begin
		p += 1
		_goto_level = _out
		next
	end
 end
	when 130 then
	begin begin p = (( @te))-1; end

        emit_table(KEYWORDS)

        if version?(18) && tok == 'not'.freeze
           @cs = 525; 	begin
		p += 1
		_goto_level = _out
		next
	end

        else
           @cs = 456; 	begin
		p += 1
		_goto_level = _out
		next
	end

        end
      end
	when 131 then
	begin begin p = (( @te))-1; end

        if version?(18)
          emit(:tIDENTIFIER)

          unless !@static_env.nil? && @static_env.declared?(tok)
             @cs = (arg_or_cmdarg);
          end
        else
          emit(:k__ENCODING__, '__ENCODING__'.freeze)
        end
        	begin
		p += 1
		_goto_level = _out
		next
	end

      end
	when 132 then
	begin begin p = (( @te))-1; end
 emit_table(KEYWORDS)
           	begin
		p += 1
		_goto_level = _out
		next
	end
 end
	when 133 then
	begin begin p = (( @te))-1; end

        digits = tok(@num_digits_s, @num_suffix_s)

        if digits.end_with? '_'.freeze
          diagnostic :error, :trailing_in_number, { :character => '_'.freeze },
                     range(@te - 1, @te)
        elsif digits.empty? && @num_base == 8 && version?(18)
          # 1.8 did not raise an error on 0o.
          digits = '0'.freeze
        elsif digits.empty?
          diagnostic :error, :empty_numeric
        elsif @num_base == 8 && (invalid_idx = digits.index(/[89]/))
          invalid_s = @num_digits_s + invalid_idx
          diagnostic :error, :invalid_octal, nil,
                     range(invalid_s, invalid_s + 1)
        end

        if version?(18, 19, 20)
          emit(:tINTEGER, digits.to_i(@num_base), @ts, @num_suffix_s)
          p = @num_suffix_s - 1
        else
          @num_xfrm.call(digits.to_i(@num_base))
        end
        	begin
		p += 1
		_goto_level = _out
		next
	end

      end
	when 135 then
	begin begin p = (( @te))-1; end

        if version?(18, 19, 20)
          diagnostic :error,
                     :trailing_in_number, { :character => tok(@te - 1, @te) },
                     range(@te - 1, @te)
        else
          emit(:tINTEGER, tok(@ts, @te - 1).to_i, @ts, @te - 1)
          p = p - 1; 	begin
		p += 1
		_goto_level = _out
		next
	end

        end
      end
	when 136 then
	begin begin p = (( @te))-1; end

        if version?(18, 19, 20)
          diagnostic :error,
                     :trailing_in_number, { :character => tok(@te - 1, @te) },
                     range(@te - 1, @te)
        else
          emit(:tFLOAT, tok(@ts, @te - 1).to_f, @ts, @te - 1)
          p = p - 1; 	begin
		p += 1
		_goto_level = _out
		next
	end

        end
      end
	when 137 then
	begin begin p = (( @te))-1; end

        digits = tok(@ts, @num_suffix_s)

        if version?(18, 19, 20)
          emit(:tFLOAT, Float(digits), @ts, @num_suffix_s)
          p = @num_suffix_s - 1
        else
          @num_xfrm.call(digits)
        end
        	begin
		p += 1
		_goto_level = _out
		next
	end

      end
	when 139 then
	begin begin p = (( @te))-1; end
 emit(:tCONSTANT)
            @cs = (arg_or_cmdarg); 	begin
		p += 1
		_goto_level = _out
		next
	end
 end
	when 143 then
	begin begin p = (( @te))-1; end

    emit(:tIDENTIFIER)

    if !@static_env.nil? && @static_env.declared?(tok)
       @cs = 428; 	begin
		p += 1
		_goto_level = _out
		next
	end

    else
       @cs = (arg_or_cmdarg); 	begin
		p += 1
		_goto_level = _out
		next
	end

    end
  end
	when 144 then
	begin begin p = (( @te))-1; end

        if tm == @te
          # Suffix was consumed, e.g. foo!
          emit(:tFID)
        else
          # Suffix was not consumed, e.g. foo!=
          emit(:tIDENTIFIER, tok(@ts, tm), @ts, tm)
          p = tm - 1
        end
         @cs = 456; 	begin
		p += 1
		_goto_level = _out
		next
	end

      end
end 
			end
	when 202 then
# line 1153 "lib/parser/lexer.rl"
		begin
 tm = p - 2 		end
# line 1336 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  emit(:tFID, tok(@ts, tm), @ts, tm)
            @cs = (arg_or_cmdarg); p = tm - 1; 	begin
		p += 1
		_goto_level = _out
		next
	end
  end
		end
	when 293 then
# line 1153 "lib/parser/lexer.rl"
		begin
 tm = p - 2 		end
# line 1685 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        emit(:tSYMBOL, tok(@ts + 1, tm), @ts, tm)
        p = tm - 1
         @cs = 754; 	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 267 then
# line 1153 "lib/parser/lexer.rl"
		begin
 tm = p - 2 		end
# line 1824 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  p = @ts - 1
           	begin
		 @cs = 754
		_goto_level = _again
		next
	end
  end
		end
	when 366 then
# line 1153 "lib/parser/lexer.rl"
		begin
 tm = p - 2 		end
# line 2112 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        if tm == @te
          # Suffix was consumed, e.g. foo!
          emit(:tFID)
        else
          # Suffix was not consumed, e.g. foo!=
          emit(:tIDENTIFIER, tok(@ts, tm), @ts, tm)
          p = tm - 1
        end
         @cs = 456; 	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 294 then
# line 1158 "lib/parser/lexer.rl"
		begin
 tm = p 		end
# line 1685 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        emit(:tSYMBOL, tok(@ts + 1, tm), @ts, tm)
        p = tm - 1
         @cs = 754; 	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 268 then
# line 1158 "lib/parser/lexer.rl"
		begin
 tm = p 		end
# line 1824 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  p = @ts - 1
           	begin
		 @cs = 754
		_goto_level = _again
		next
	end
  end
		end
	when 295 then
# line 1159 "lib/parser/lexer.rl"
		begin
 tm = p - 2 		end
# line 1685 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        emit(:tSYMBOL, tok(@ts + 1, tm), @ts, tm)
        p = tm - 1
         @cs = 754; 	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 269 then
# line 1159 "lib/parser/lexer.rl"
		begin
 tm = p - 2 		end
# line 1824 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  p = @ts - 1
           	begin
		 @cs = 754
		_goto_level = _again
		next
	end
  end
		end
	when 299 then
# line 1160 "lib/parser/lexer.rl"
		begin
 tm = p - 2 		end
# line 1685 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        emit(:tSYMBOL, tok(@ts + 1, tm), @ts, tm)
        p = tm - 1
         @cs = 754; 	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 272 then
# line 1160 "lib/parser/lexer.rl"
		begin
 tm = p - 2 		end
# line 1824 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  p = @ts - 1
           	begin
		 @cs = 754
		_goto_level = _again
		next
	end
  end
		end
	when 298 then
# line 1161 "lib/parser/lexer.rl"
		begin
 tm = p - 2 		end
# line 1685 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        emit(:tSYMBOL, tok(@ts + 1, tm), @ts, tm)
        p = tm - 1
         @cs = 754; 	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 271 then
# line 1161 "lib/parser/lexer.rl"
		begin
 tm = p - 2 		end
# line 1 "NONE"
		begin
	case  @act
	when 103 then
	begin begin p = (( @te))-1; end
 emit_table(PUNCTUATION_BEGIN)
           	begin
		p += 1
		_goto_level = _out
		next
	end
 end
	when 104 then
	begin begin p = (( @te))-1; end
 emit(:kRESCUE, 'rescue'.freeze, @ts, tm)
           p = tm - 1
            @cs = 501; 	begin
		p += 1
		_goto_level = _out
		next
	end
 end
	when 105 then
	begin begin p = (( @te))-1; end
 emit_table(KEYWORDS_BEGIN)
            @cs = 746; 	begin
		p += 1
		_goto_level = _out
		next
	end
 end
	when 107 then
	begin begin p = (( @te))-1; end
 p = @ts - 1
           	begin
		 @cs = 754
		_goto_level = _again
		next
	end
 end
	when 108 then
	begin begin p = (( @te))-1; end

    emit(:tIDENTIFIER)

    if !@static_env.nil? && @static_env.declared?(tok)
       @cs = 428; 	begin
		p += 1
		_goto_level = _out
		next
	end

    else
       @cs = (arg_or_cmdarg); 	begin
		p += 1
		_goto_level = _out
		next
	end

    end
  end
	when 111 then
	begin begin p = (( @te))-1; end
 p = @ts - 1; 	begin
		 @cs = 754
		_goto_level = _again
		next
	end
 end
end 
			end
	when 296 then
# line 1162 "lib/parser/lexer.rl"
		begin
 tm = p - 3 		end
# line 1685 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        emit(:tSYMBOL, tok(@ts + 1, tm), @ts, tm)
        p = tm - 1
         @cs = 754; 	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 270 then
# line 1162 "lib/parser/lexer.rl"
		begin
 tm = p - 3 		end
# line 1824 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  p = @ts - 1
           	begin
		 @cs = 754
		_goto_level = _again
		next
	end
  end
		end
	when 297 then
# line 1167 "lib/parser/lexer.rl"
		begin
 tm = p - 2 		end
# line 1685 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        emit(:tSYMBOL, tok(@ts + 1, tm), @ts, tm)
        p = tm - 1
         @cs = 754; 	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 421 then
# line 1172 "lib/parser/lexer.rl"
		begin
 tm = p - 2 		end
# line 2094 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  emit(:tCONSTANT, tok(@ts, tm), @ts, tm)
           p = tm - 1; 	begin
		p += 1
		_goto_level = _out
		next
	end
  end
		end
	when 225 then
# line 1178 "lib/parser/lexer.rl"
		begin

    @cond.push(false); @cmdarg.push(false)
  		end
# line 1384 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  emit(:tLBRACK, '['.freeze, @te - 1, @te)
            @cs = 525; 	begin
		p += 1
		_goto_level = _out
		next
	end
  end
		end
	when 328 then
# line 1178 "lib/parser/lexer.rl"
		begin

    @cond.push(false); @cmdarg.push(false)
  		end
# line 1763 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  emit(:tLBRACK, '['.freeze)
           	begin
		p += 1
		_goto_level = _out
		next
	end
  end
		end
	when 424 then
# line 1178 "lib/parser/lexer.rl"
		begin

    @cond.push(false); @cmdarg.push(false)
  		end
# line 2158 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  emit(:tLBRACK2, '['.freeze)
            @cs = 525; 	begin
		p += 1
		_goto_level = _out
		next
	end
  end
		end
	when 217 then
# line 1185 "lib/parser/lexer.rl"
		begin

    @cond.push(false); @cmdarg.push(false)

    @paren_nest += 1
  		end
# line 1365 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        if version?(18)
          emit(:tLPAREN2, '('.freeze, @te - 1, @te)
           @cs = 746; 	begin
		p += 1
		_goto_level = _out
		next
	end

        else
          emit(:tLPAREN_ARG, '('.freeze, @te - 1, @te)
           @cs = 525; 	begin
		p += 1
		_goto_level = _out
		next
	end

        end
       end
		end
	when 230 then
# line 1185 "lib/parser/lexer.rl"
		begin

    @cond.push(false); @cmdarg.push(false)

    @paren_nest += 1
  		end
# line 1378 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  emit(:tLPAREN2, '('.freeze)
            @cs = 525; 	begin
		p += 1
		_goto_level = _out
		next
	end
  end
		end
	when 238 then
# line 1185 "lib/parser/lexer.rl"
		begin

    @cond.push(false); @cmdarg.push(false)

    @paren_nest += 1
  		end
# line 1502 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        emit(:tLPAREN_ARG, '('.freeze, @te - 1, @te)
        if version?(18)
           @cs = 746; 	begin
		p += 1
		_goto_level = _out
		next
	end

        else
           @cs = 525; 	begin
		p += 1
		_goto_level = _out
		next
	end

        end
       end
		end
	when 285 then
# line 1185 "lib/parser/lexer.rl"
		begin

    @cond.push(false); @cmdarg.push(false)

    @paren_nest += 1
  		end
# line 1768 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  emit(:tLPAREN, '('.freeze)
           	begin
		p += 1
		_goto_level = _out
		next
	end
  end
		end
	when 377 then
# line 1185 "lib/parser/lexer.rl"
		begin

    @cond.push(false); @cmdarg.push(false)

    @paren_nest += 1
  		end
# line 2132 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  emit_table(PUNCTUATION)
            @cs = 525; 	begin
		p += 1
		_goto_level = _out
		next
	end
  end
		end
	when 378 then
# line 1191 "lib/parser/lexer.rl"
		begin

    @paren_nest -= 1
  		end
# line 2136 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        emit_table(PUNCTUATION)
        @cond.lexpop; @cmdarg.lexpop

        if RBRACE_OR_RBRACK.include?(tok)
           @cs = 493;
        else # )
          # fnext expr_endfn; ?
        end

        	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 54 then
# line 1651 "lib/parser/lexer.rl"
		begin
 heredoc_e      = p 		end
# line 486 "lib/parser/lexer.rl"
		begin

    # Record position of a newline for precise location reporting on tNL
    # tokens.
    #
    # This action is embedded directly into c_nl, as it is idempotent and
    # there are no cases when we need to skip it.
    @newline_s = p
  		end
	when 300 then
# line 1652 "lib/parser/lexer.rl"
		begin
 new_herebody_s = p 		end
# line 1653 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        tok(@ts, heredoc_e) =~ /^<<(-?)(~?)(["'`]?)(.*)\3$/

        indent      = !$1.empty? || !$2.empty?
        dedent_body = !$2.empty?
        type        =  $3.empty? ? '<<"'.freeze : ('<<'.freeze + $3)
        delimiter   =  $4

        if dedent_body && version?(18, 19, 20, 21, 22)
          emit(:tLSHFT, '<<'.freeze, @ts, @ts + 2)
          p = @ts + 1
           @cs = 525; 	begin
		p += 1
		_goto_level = _out
		next
	end

        else
           @cs = (push_literal(type, delimiter, @ts, heredoc_e, indent, dedent_body));

          @herebody_s ||= new_herebody_s
          p = @herebody_s - 1
        end
       end
		end
	when 305 then
# line 1705 "lib/parser/lexer.rl"
		begin
 @escape = nil 		end
# line 1707 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        value = @escape || tok(@ts + 1)

        if version?(18)
          if defined?(Encoding)
            emit(:tINTEGER, value.dup.force_encoding(Encoding::BINARY)[0].ord)
          else
            emit(:tINTEGER, value[0].ord)
          end
        else
          emit(:tCHARACTER, value)
        end

         @cs = 754; 	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 332 then
# line 1778 "lib/parser/lexer.rl"
		begin
 tm = p 		end
# line 1779 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  emit(:kRESCUE, 'rescue'.freeze, @ts, tm)
           p = tm - 1
            @cs = 501; 	begin
		p += 1
		_goto_level = _out
		next
	end
  end
		end
	when 412 then
# line 1996 "lib/parser/lexer.rl"
		begin
 @num_base = 16; @num_digits_s = p 		end
# line 2002 "lib/parser/lexer.rl"
		begin
 @num_suffix_s = p 		end
	when 406 then
# line 1997 "lib/parser/lexer.rl"
		begin
 @num_base = 10; @num_digits_s = p 		end
# line 2002 "lib/parser/lexer.rl"
		begin
 @num_suffix_s = p 		end
	when 409 then
# line 1998 "lib/parser/lexer.rl"
		begin
 @num_base = 8;  @num_digits_s = p 		end
# line 2002 "lib/parser/lexer.rl"
		begin
 @num_suffix_s = p 		end
	when 403 then
# line 1999 "lib/parser/lexer.rl"
		begin
 @num_base = 2;  @num_digits_s = p 		end
# line 2002 "lib/parser/lexer.rl"
		begin
 @num_suffix_s = p 		end
	when 418 then
# line 2000 "lib/parser/lexer.rl"
		begin
 @num_base = 10; @num_digits_s = @ts 		end
# line 2002 "lib/parser/lexer.rl"
		begin
 @num_suffix_s = p 		end
	when 387 then
# line 2001 "lib/parser/lexer.rl"
		begin
 @num_base = 8;  @num_digits_s = @ts 		end
# line 2002 "lib/parser/lexer.rl"
		begin
 @num_suffix_s = p 		end
	when 419 then
# line 2002 "lib/parser/lexer.rl"
		begin
 @num_suffix_s = p 		end
# line 619 "lib/parser/lexer.rl"
		begin
 @num_xfrm = lambda { |chars| emit(:tINTEGER,   chars) } 		end
	when 75 then
# line 2195 "lib/parser/lexer.rl"
		begin
 tm = p 		end
# line 2196 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin  p = tm - 1; 	begin
		 @cs = 754
		_goto_level = _again
		next
	end
  end
		end
	when 8 then
# line 1 "NONE"
		begin
 @te = p+1
		end
# line 486 "lib/parser/lexer.rl"
		begin

    # Record position of a newline for precise location reporting on tNL
    # tokens.
    #
    # This action is embedded directly into c_nl, as it is idempotent and
    # there are no cases when we need to skip it.
    @newline_s = p
  		end
	when 186 then
# line 1 "NONE"
		begin
 @te = p+1
		end
# line 1255 "lib/parser/lexer.rl"
		begin
 @act = 31;		end
	when 173 then
# line 1 "NONE"
		begin
 @te = p+1
		end
# line 1259 "lib/parser/lexer.rl"
		begin
 @act = 32;		end
	when 169 then
# line 1 "NONE"
		begin
 @te = p+1
		end
# line 1263 "lib/parser/lexer.rl"
		begin
 @act = 33;		end
	when 24 then
# line 1 "NONE"
		begin
 @te = p+1
		end
# line 1425 "lib/parser/lexer.rl"
		begin
 @act = 59;		end
	when 219 then
# line 1 "NONE"
		begin
 @te = p+1
		end
# line 1438 "lib/parser/lexer.rl"
		begin
 @act = 60;		end
	when 25 then
# line 1 "NONE"
		begin
 @te = p+1
		end
# line 1475 "lib/parser/lexer.rl"
		begin
 @act = 65;		end
	when 212 then
# line 1 "NONE"
		begin
 @te = p+1
		end
# line 1480 "lib/parser/lexer.rl"
		begin
 @act = 66;		end
	when 239 then
# line 1 "NONE"
		begin
 @te = p+1
		end
# line 1512 "lib/parser/lexer.rl"
		begin
 @act = 72;		end
	when 43 then
# line 1 "NONE"
		begin
 @te = p+1
		end
# line 1525 "lib/parser/lexer.rl"
		begin
 @act = 73;		end
	when 260 then
# line 1 "NONE"
		begin
 @te = p+1
		end
# line 1570 "lib/parser/lexer.rl"
		begin
 @act = 80;		end
	when 249 then
# line 1 "NONE"
		begin
 @te = p+1
		end
# line 1574 "lib/parser/lexer.rl"
		begin
 @act = 81;		end
	when 263 then
# line 1 "NONE"
		begin
 @te = p+1
		end
# line 1773 "lib/parser/lexer.rl"
		begin
 @act = 103;		end
	when 331 then
# line 1 "NONE"
		begin
 @te = p+1
		end
# line 1779 "lib/parser/lexer.rl"
		begin
 @act = 104;		end
	when 330 then
# line 1 "NONE"
		begin
 @te = p+1
		end
# line 1785 "lib/parser/lexer.rl"
		begin
 @act = 105;		end
	when 56 then
# line 1 "NONE"
		begin
 @te = p+1
		end
# line 1824 "lib/parser/lexer.rl"
		begin
 @act = 107;		end
	when 261 then
# line 1 "NONE"
		begin
 @te = p+1
		end
# line 1196 "lib/parser/lexer.rl"
		begin
 @act = 108;		end
	when 264 then
# line 1 "NONE"
		begin
 @te = p+1
		end
# line 1852 "lib/parser/lexer.rl"
		begin
 @act = 111;		end
	when 434 then
# line 1 "NONE"
		begin
 @te = p+1
		end
# line 1916 "lib/parser/lexer.rl"
		begin
 @act = 124;		end
	when 429 then
# line 1 "NONE"
		begin
 @te = p+1
		end
# line 1941 "lib/parser/lexer.rl"
		begin
 @act = 125;		end
	when 437 then
# line 1 "NONE"
		begin
 @te = p+1
		end
# line 1951 "lib/parser/lexer.rl"
		begin
 @act = 127;		end
	when 430 then
# line 1 "NONE"
		begin
 @te = p+1
		end
# line 1956 "lib/parser/lexer.rl"
		begin
 @act = 128;		end
	when 431 then
# line 1 "NONE"
		begin
 @te = p+1
		end
# line 1960 "lib/parser/lexer.rl"
		begin
 @act = 129;		end
	when 436 then
# line 1 "NONE"
		begin
 @te = p+1
		end
# line 1964 "lib/parser/lexer.rl"
		begin
 @act = 130;		end
	when 428 then
# line 1 "NONE"
		begin
 @te = p+1
		end
# line 1975 "lib/parser/lexer.rl"
		begin
 @act = 131;		end
	when 423 then
# line 1 "NONE"
		begin
 @te = p+1
		end
# line 1989 "lib/parser/lexer.rl"
		begin
 @act = 132;		end
	when 359 then
# line 1 "NONE"
		begin
 @te = p+1
		end
# line 2003 "lib/parser/lexer.rl"
		begin
 @act = 133;		end
	when 389 then
# line 1 "NONE"
		begin
 @te = p+1
		end
# line 2047 "lib/parser/lexer.rl"
		begin
 @act = 136;		end
	when 63 then
# line 1 "NONE"
		begin
 @te = p+1
		end
# line 2062 "lib/parser/lexer.rl"
		begin
 @act = 137;		end
	when 362 then
# line 1 "NONE"
		begin
 @te = p+1
		end
# line 2090 "lib/parser/lexer.rl"
		begin
 @act = 139;		end
	when 354 then
# line 1 "NONE"
		begin
 @te = p+1
		end
# line 1196 "lib/parser/lexer.rl"
		begin
 @act = 143;		end
	when 364 then
# line 1 "NONE"
		begin
 @te = p+1
		end
# line 2112 "lib/parser/lexer.rl"
		begin
 @act = 144;		end
	when 154 then
# line 486 "lib/parser/lexer.rl"
		begin

    # Record position of a newline for precise location reporting on tNL
    # tokens.
    #
    # This action is embedded directly into c_nl, as it is idempotent and
    # there are no cases when we need to skip it.
    @newline_s = p
  		end
# line 877 "lib/parser/lexer.rl"
		begin
 @te = p+1
 begin 
    current_literal = literal
    if @te == pe
      diagnostic :fatal, :string_eof, nil,
                 range(current_literal.str_s, current_literal.str_s + 1)
    end

    if current_literal.heredoc?
      line = tok(@herebody_s, @ts).gsub(/\r+$/, ''.freeze)

      if version?(18, 19, 20)
        # See ruby:c48b4209c
        line = line.gsub(/\r.*$/, ''.freeze)
      end

      # Try ending the heredoc with the complete most recently
      # scanned line. @herebody_s always refers to the start of such line.
      if current_literal.nest_and_try_closing(line, @herebody_s, @ts)
        # Adjust @herebody_s to point to the next line.
        @herebody_s = @te

        # Continue regular lexing after the heredoc reference (<<END).
        p = current_literal.heredoc_e - 1
         @cs = (pop_literal); 	begin
		p += 1
		_goto_level = _out
		next
	end

      else
        # Calculate indentation level for <<~HEREDOCs.
        current_literal.infer_indent_level(line)

        # Ditto.
        @herebody_s = @te
      end
    else
      # Try ending the literal with a newline.
      if current_literal.nest_and_try_closing(tok, @ts, @te)
         @cs = (pop_literal); 	begin
		p += 1
		_goto_level = _out
		next
	end

      end

      if @herebody_s
        # This is a regular literal intertwined with a heredoc. Like:
        #
        #     p <<-foo+"1
        #     bar
        #     foo
        #     2"
        #
        # which, incidentally, evaluates to "bar\n1\n2".
        p = @herebody_s - 1
        @herebody_s = nil
      end
    end

    if current_literal.words? && !eof_codepoint?(@source_pts[p])
      current_literal.extend_space @ts, @te
    else
      # A literal newline is appended if the heredoc was _not_ closed
      # this time (see fbreak above). See also Literal#nest_and_try_closing
      # for rationale of calling #flush_string here.
      current_literal.extend_string tok, @ts, @te
      current_literal.flush_string
    end
   end
		end
# line 766 "lib/parser/lexer.rl"
		begin

    @escape_s = p
    @escape   = nil
  		end
	when 104 then
# line 666 "lib/parser/lexer.rl"
		begin

    codepoint = @source_pts[p - 1]
    if (@escape = ESCAPES[codepoint]).nil?
      @escape = encode_escape(@source_buffer.slice(p - 1))
    end
  		end
# line 677 "lib/parser/lexer.rl"
		begin

    @escape = encode_escape(@escape[0].ord & 0x9f)
  		end
# line 839 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    current_literal = literal
    # Get the first character after the backslash.
    escaped_char = @source_buffer.slice(@escape_s).chr

    if current_literal.munge_escape? escaped_char
      # If this particular literal uses this character as an opening
      # or closing delimiter, it is an escape sequence for that
      # particular character. Write it without the backslash.

      if current_literal.regexp? && REGEXP_META_CHARACTERS.match(escaped_char)
        # Regular expressions should include escaped delimiters in their
        # escaped form, except when the escaped character is
        # a closing delimiter but not a regexp metacharacter.
        #
        # The backslash itself cannot be used as a closing delimiter
        # at the same time as an escape symbol, but it is always munged,
        # so this branch also executes for the non-closing-delimiter case
        # for the backslash.
        current_literal.extend_string(tok, @ts, @te)
      else
        current_literal.extend_string(escaped_char, @ts, @te)
      end
    else
      # It does not. So this is an actual escape sequence, yay!
      if current_literal.regexp?
        # Regular expressions should include escape sequences in their
        # escaped form. On the other hand, escaped newlines are removed.
        current_literal.extend_string(tok.gsub("\\\n".freeze, ''.freeze), @ts, @te)
      else
        current_literal.extend_string(@escape || tok, @ts, @te)
      end
    end
   end
		end
	when 130 then
# line 666 "lib/parser/lexer.rl"
		begin

    codepoint = @source_pts[p - 1]
    if (@escape = ESCAPES[codepoint]).nil?
      @escape = encode_escape(@source_buffer.slice(p - 1))
    end
  		end
# line 677 "lib/parser/lexer.rl"
		begin

    @escape = encode_escape(@escape[0].ord & 0x9f)
  		end
# line 839 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    current_literal = literal
    # Get the first character after the backslash.
    escaped_char = @source_buffer.slice(@escape_s).chr

    if current_literal.munge_escape? escaped_char
      # If this particular literal uses this character as an opening
      # or closing delimiter, it is an escape sequence for that
      # particular character. Write it without the backslash.

      if current_literal.regexp? && REGEXP_META_CHARACTERS.match(escaped_char)
        # Regular expressions should include escaped delimiters in their
        # escaped form, except when the escaped character is
        # a closing delimiter but not a regexp metacharacter.
        #
        # The backslash itself cannot be used as a closing delimiter
        # at the same time as an escape symbol, but it is always munged,
        # so this branch also executes for the non-closing-delimiter case
        # for the backslash.
        current_literal.extend_string(tok, @ts, @te)
      else
        current_literal.extend_string(escaped_char, @ts, @te)
      end
    else
      # It does not. So this is an actual escape sequence, yay!
      if current_literal.regexp?
        # Regular expressions should include escape sequences in their
        # escaped form. On the other hand, escaped newlines are removed.
        current_literal.extend_string(tok.gsub("\\\n".freeze, ''.freeze), @ts, @te)
      else
        current_literal.extend_string(@escape || tok, @ts, @te)
      end
    end
   end
		end
	when 315 then
# line 666 "lib/parser/lexer.rl"
		begin

    codepoint = @source_pts[p - 1]
    if (@escape = ESCAPES[codepoint]).nil?
      @escape = encode_escape(@source_buffer.slice(p - 1))
    end
  		end
# line 677 "lib/parser/lexer.rl"
		begin

    @escape = encode_escape(@escape[0].ord & 0x9f)
  		end
# line 1707 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        value = @escape || tok(@ts + 1)

        if version?(18)
          if defined?(Encoding)
            emit(:tINTEGER, value.dup.force_encoding(Encoding::BINARY)[0].ord)
          else
            emit(:tINTEGER, value[0].ord)
          end
        else
          emit(:tCHARACTER, value)
        end

         @cs = 754; 	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 109 then
# line 666 "lib/parser/lexer.rl"
		begin

    codepoint = @source_pts[p - 1]
    if (@escape = ESCAPES[codepoint]).nil?
      @escape = encode_escape(@source_buffer.slice(p - 1))
    end
  		end
# line 681 "lib/parser/lexer.rl"
		begin

    @escape = encode_escape(@escape[0].ord | 0x80)
  		end
# line 839 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    current_literal = literal
    # Get the first character after the backslash.
    escaped_char = @source_buffer.slice(@escape_s).chr

    if current_literal.munge_escape? escaped_char
      # If this particular literal uses this character as an opening
      # or closing delimiter, it is an escape sequence for that
      # particular character. Write it without the backslash.

      if current_literal.regexp? && REGEXP_META_CHARACTERS.match(escaped_char)
        # Regular expressions should include escaped delimiters in their
        # escaped form, except when the escaped character is
        # a closing delimiter but not a regexp metacharacter.
        #
        # The backslash itself cannot be used as a closing delimiter
        # at the same time as an escape symbol, but it is always munged,
        # so this branch also executes for the non-closing-delimiter case
        # for the backslash.
        current_literal.extend_string(tok, @ts, @te)
      else
        current_literal.extend_string(escaped_char, @ts, @te)
      end
    else
      # It does not. So this is an actual escape sequence, yay!
      if current_literal.regexp?
        # Regular expressions should include escape sequences in their
        # escaped form. On the other hand, escaped newlines are removed.
        current_literal.extend_string(tok.gsub("\\\n".freeze, ''.freeze), @ts, @te)
      else
        current_literal.extend_string(@escape || tok, @ts, @te)
      end
    end
   end
		end
	when 135 then
# line 666 "lib/parser/lexer.rl"
		begin

    codepoint = @source_pts[p - 1]
    if (@escape = ESCAPES[codepoint]).nil?
      @escape = encode_escape(@source_buffer.slice(p - 1))
    end
  		end
# line 681 "lib/parser/lexer.rl"
		begin

    @escape = encode_escape(@escape[0].ord | 0x80)
  		end
# line 839 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    current_literal = literal
    # Get the first character after the backslash.
    escaped_char = @source_buffer.slice(@escape_s).chr

    if current_literal.munge_escape? escaped_char
      # If this particular literal uses this character as an opening
      # or closing delimiter, it is an escape sequence for that
      # particular character. Write it without the backslash.

      if current_literal.regexp? && REGEXP_META_CHARACTERS.match(escaped_char)
        # Regular expressions should include escaped delimiters in their
        # escaped form, except when the escaped character is
        # a closing delimiter but not a regexp metacharacter.
        #
        # The backslash itself cannot be used as a closing delimiter
        # at the same time as an escape symbol, but it is always munged,
        # so this branch also executes for the non-closing-delimiter case
        # for the backslash.
        current_literal.extend_string(tok, @ts, @te)
      else
        current_literal.extend_string(escaped_char, @ts, @te)
      end
    else
      # It does not. So this is an actual escape sequence, yay!
      if current_literal.regexp?
        # Regular expressions should include escape sequences in their
        # escaped form. On the other hand, escaped newlines are removed.
        current_literal.extend_string(tok.gsub("\\\n".freeze, ''.freeze), @ts, @te)
      else
        current_literal.extend_string(@escape || tok, @ts, @te)
      end
    end
   end
		end
	when 320 then
# line 666 "lib/parser/lexer.rl"
		begin

    codepoint = @source_pts[p - 1]
    if (@escape = ESCAPES[codepoint]).nil?
      @escape = encode_escape(@source_buffer.slice(p - 1))
    end
  		end
# line 681 "lib/parser/lexer.rl"
		begin

    @escape = encode_escape(@escape[0].ord | 0x80)
  		end
# line 1707 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        value = @escape || tok(@ts + 1)

        if version?(18)
          if defined?(Encoding)
            emit(:tINTEGER, value.dup.force_encoding(Encoding::BINARY)[0].ord)
          else
            emit(:tINTEGER, value[0].ord)
          end
        else
          emit(:tCHARACTER, value)
        end

         @cs = 754; 	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 108 then
# line 687 "lib/parser/lexer.rl"
		begin
 @escape = @source_buffer.slice(p - 1).chr 		end
# line 681 "lib/parser/lexer.rl"
		begin

    @escape = encode_escape(@escape[0].ord | 0x80)
  		end
# line 839 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    current_literal = literal
    # Get the first character after the backslash.
    escaped_char = @source_buffer.slice(@escape_s).chr

    if current_literal.munge_escape? escaped_char
      # If this particular literal uses this character as an opening
      # or closing delimiter, it is an escape sequence for that
      # particular character. Write it without the backslash.

      if current_literal.regexp? && REGEXP_META_CHARACTERS.match(escaped_char)
        # Regular expressions should include escaped delimiters in their
        # escaped form, except when the escaped character is
        # a closing delimiter but not a regexp metacharacter.
        #
        # The backslash itself cannot be used as a closing delimiter
        # at the same time as an escape symbol, but it is always munged,
        # so this branch also executes for the non-closing-delimiter case
        # for the backslash.
        current_literal.extend_string(tok, @ts, @te)
      else
        current_literal.extend_string(escaped_char, @ts, @te)
      end
    else
      # It does not. So this is an actual escape sequence, yay!
      if current_literal.regexp?
        # Regular expressions should include escape sequences in their
        # escaped form. On the other hand, escaped newlines are removed.
        current_literal.extend_string(tok.gsub("\\\n".freeze, ''.freeze), @ts, @te)
      else
        current_literal.extend_string(@escape || tok, @ts, @te)
      end
    end
   end
		end
	when 134 then
# line 687 "lib/parser/lexer.rl"
		begin
 @escape = @source_buffer.slice(p - 1).chr 		end
# line 681 "lib/parser/lexer.rl"
		begin

    @escape = encode_escape(@escape[0].ord | 0x80)
  		end
# line 839 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    current_literal = literal
    # Get the first character after the backslash.
    escaped_char = @source_buffer.slice(@escape_s).chr

    if current_literal.munge_escape? escaped_char
      # If this particular literal uses this character as an opening
      # or closing delimiter, it is an escape sequence for that
      # particular character. Write it without the backslash.

      if current_literal.regexp? && REGEXP_META_CHARACTERS.match(escaped_char)
        # Regular expressions should include escaped delimiters in their
        # escaped form, except when the escaped character is
        # a closing delimiter but not a regexp metacharacter.
        #
        # The backslash itself cannot be used as a closing delimiter
        # at the same time as an escape symbol, but it is always munged,
        # so this branch also executes for the non-closing-delimiter case
        # for the backslash.
        current_literal.extend_string(tok, @ts, @te)
      else
        current_literal.extend_string(escaped_char, @ts, @te)
      end
    else
      # It does not. So this is an actual escape sequence, yay!
      if current_literal.regexp?
        # Regular expressions should include escape sequences in their
        # escaped form. On the other hand, escaped newlines are removed.
        current_literal.extend_string(tok.gsub("\\\n".freeze, ''.freeze), @ts, @te)
      else
        current_literal.extend_string(@escape || tok, @ts, @te)
      end
    end
   end
		end
	when 319 then
# line 687 "lib/parser/lexer.rl"
		begin
 @escape = @source_buffer.slice(p - 1).chr 		end
# line 681 "lib/parser/lexer.rl"
		begin

    @escape = encode_escape(@escape[0].ord | 0x80)
  		end
# line 1707 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        value = @escape || tok(@ts + 1)

        if version?(18)
          if defined?(Encoding)
            emit(:tINTEGER, value.dup.force_encoding(Encoding::BINARY)[0].ord)
          else
            emit(:tINTEGER, value[0].ord)
          end
        else
          emit(:tCHARACTER, value)
        end

         @cs = 754; 	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 106 then
# line 692 "lib/parser/lexer.rl"
		begin
 @escape = "\x7f" 		end
# line 681 "lib/parser/lexer.rl"
		begin

    @escape = encode_escape(@escape[0].ord | 0x80)
  		end
# line 839 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    current_literal = literal
    # Get the first character after the backslash.
    escaped_char = @source_buffer.slice(@escape_s).chr

    if current_literal.munge_escape? escaped_char
      # If this particular literal uses this character as an opening
      # or closing delimiter, it is an escape sequence for that
      # particular character. Write it without the backslash.

      if current_literal.regexp? && REGEXP_META_CHARACTERS.match(escaped_char)
        # Regular expressions should include escaped delimiters in their
        # escaped form, except when the escaped character is
        # a closing delimiter but not a regexp metacharacter.
        #
        # The backslash itself cannot be used as a closing delimiter
        # at the same time as an escape symbol, but it is always munged,
        # so this branch also executes for the non-closing-delimiter case
        # for the backslash.
        current_literal.extend_string(tok, @ts, @te)
      else
        current_literal.extend_string(escaped_char, @ts, @te)
      end
    else
      # It does not. So this is an actual escape sequence, yay!
      if current_literal.regexp?
        # Regular expressions should include escape sequences in their
        # escaped form. On the other hand, escaped newlines are removed.
        current_literal.extend_string(tok.gsub("\\\n".freeze, ''.freeze), @ts, @te)
      else
        current_literal.extend_string(@escape || tok, @ts, @te)
      end
    end
   end
		end
	when 132 then
# line 692 "lib/parser/lexer.rl"
		begin
 @escape = "\x7f" 		end
# line 681 "lib/parser/lexer.rl"
		begin

    @escape = encode_escape(@escape[0].ord | 0x80)
  		end
# line 839 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    current_literal = literal
    # Get the first character after the backslash.
    escaped_char = @source_buffer.slice(@escape_s).chr

    if current_literal.munge_escape? escaped_char
      # If this particular literal uses this character as an opening
      # or closing delimiter, it is an escape sequence for that
      # particular character. Write it without the backslash.

      if current_literal.regexp? && REGEXP_META_CHARACTERS.match(escaped_char)
        # Regular expressions should include escaped delimiters in their
        # escaped form, except when the escaped character is
        # a closing delimiter but not a regexp metacharacter.
        #
        # The backslash itself cannot be used as a closing delimiter
        # at the same time as an escape symbol, but it is always munged,
        # so this branch also executes for the non-closing-delimiter case
        # for the backslash.
        current_literal.extend_string(tok, @ts, @te)
      else
        current_literal.extend_string(escaped_char, @ts, @te)
      end
    else
      # It does not. So this is an actual escape sequence, yay!
      if current_literal.regexp?
        # Regular expressions should include escape sequences in their
        # escaped form. On the other hand, escaped newlines are removed.
        current_literal.extend_string(tok.gsub("\\\n".freeze, ''.freeze), @ts, @te)
      else
        current_literal.extend_string(@escape || tok, @ts, @te)
      end
    end
   end
		end
	when 317 then
# line 692 "lib/parser/lexer.rl"
		begin
 @escape = "\x7f" 		end
# line 681 "lib/parser/lexer.rl"
		begin

    @escape = encode_escape(@escape[0].ord | 0x80)
  		end
# line 1707 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        value = @escape || tok(@ts + 1)

        if version?(18)
          if defined?(Encoding)
            emit(:tINTEGER, value.dup.force_encoding(Encoding::BINARY)[0].ord)
          else
            emit(:tINTEGER, value[0].ord)
          end
        else
          emit(:tCHARACTER, value)
        end

         @cs = 754; 	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 102 then
# line 693 "lib/parser/lexer.rl"
		begin
 @escape = @source_buffer.slice(p - 1).chr 		end
# line 677 "lib/parser/lexer.rl"
		begin

    @escape = encode_escape(@escape[0].ord & 0x9f)
  		end
# line 839 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    current_literal = literal
    # Get the first character after the backslash.
    escaped_char = @source_buffer.slice(@escape_s).chr

    if current_literal.munge_escape? escaped_char
      # If this particular literal uses this character as an opening
      # or closing delimiter, it is an escape sequence for that
      # particular character. Write it without the backslash.

      if current_literal.regexp? && REGEXP_META_CHARACTERS.match(escaped_char)
        # Regular expressions should include escaped delimiters in their
        # escaped form, except when the escaped character is
        # a closing delimiter but not a regexp metacharacter.
        #
        # The backslash itself cannot be used as a closing delimiter
        # at the same time as an escape symbol, but it is always munged,
        # so this branch also executes for the non-closing-delimiter case
        # for the backslash.
        current_literal.extend_string(tok, @ts, @te)
      else
        current_literal.extend_string(escaped_char, @ts, @te)
      end
    else
      # It does not. So this is an actual escape sequence, yay!
      if current_literal.regexp?
        # Regular expressions should include escape sequences in their
        # escaped form. On the other hand, escaped newlines are removed.
        current_literal.extend_string(tok.gsub("\\\n".freeze, ''.freeze), @ts, @te)
      else
        current_literal.extend_string(@escape || tok, @ts, @te)
      end
    end
   end
		end
	when 128 then
# line 693 "lib/parser/lexer.rl"
		begin
 @escape = @source_buffer.slice(p - 1).chr 		end
# line 677 "lib/parser/lexer.rl"
		begin

    @escape = encode_escape(@escape[0].ord & 0x9f)
  		end
# line 839 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    current_literal = literal
    # Get the first character after the backslash.
    escaped_char = @source_buffer.slice(@escape_s).chr

    if current_literal.munge_escape? escaped_char
      # If this particular literal uses this character as an opening
      # or closing delimiter, it is an escape sequence for that
      # particular character. Write it without the backslash.

      if current_literal.regexp? && REGEXP_META_CHARACTERS.match(escaped_char)
        # Regular expressions should include escaped delimiters in their
        # escaped form, except when the escaped character is
        # a closing delimiter but not a regexp metacharacter.
        #
        # The backslash itself cannot be used as a closing delimiter
        # at the same time as an escape symbol, but it is always munged,
        # so this branch also executes for the non-closing-delimiter case
        # for the backslash.
        current_literal.extend_string(tok, @ts, @te)
      else
        current_literal.extend_string(escaped_char, @ts, @te)
      end
    else
      # It does not. So this is an actual escape sequence, yay!
      if current_literal.regexp?
        # Regular expressions should include escape sequences in their
        # escaped form. On the other hand, escaped newlines are removed.
        current_literal.extend_string(tok.gsub("\\\n".freeze, ''.freeze), @ts, @te)
      else
        current_literal.extend_string(@escape || tok, @ts, @te)
      end
    end
   end
		end
	when 313 then
# line 693 "lib/parser/lexer.rl"
		begin
 @escape = @source_buffer.slice(p - 1).chr 		end
# line 677 "lib/parser/lexer.rl"
		begin

    @escape = encode_escape(@escape[0].ord & 0x9f)
  		end
# line 1707 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        value = @escape || tok(@ts + 1)

        if version?(18)
          if defined?(Encoding)
            emit(:tINTEGER, value.dup.force_encoding(Encoding::BINARY)[0].ord)
          else
            emit(:tINTEGER, value[0].ord)
          end
        else
          emit(:tCHARACTER, value)
        end

         @cs = 754; 	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 112 then
# line 723 "lib/parser/lexer.rl"
		begin

        diagnostic :fatal, :invalid_unicode_escape, nil, range(@escape_s - 1, p)
      		end
# line 734 "lib/parser/lexer.rl"
		begin

          diagnostic :fatal, :unterminated_unicode, nil, range(p - 1, p)
        		end
# line 839 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    current_literal = literal
    # Get the first character after the backslash.
    escaped_char = @source_buffer.slice(@escape_s).chr

    if current_literal.munge_escape? escaped_char
      # If this particular literal uses this character as an opening
      # or closing delimiter, it is an escape sequence for that
      # particular character. Write it without the backslash.

      if current_literal.regexp? && REGEXP_META_CHARACTERS.match(escaped_char)
        # Regular expressions should include escaped delimiters in their
        # escaped form, except when the escaped character is
        # a closing delimiter but not a regexp metacharacter.
        #
        # The backslash itself cannot be used as a closing delimiter
        # at the same time as an escape symbol, but it is always munged,
        # so this branch also executes for the non-closing-delimiter case
        # for the backslash.
        current_literal.extend_string(tok, @ts, @te)
      else
        current_literal.extend_string(escaped_char, @ts, @te)
      end
    else
      # It does not. So this is an actual escape sequence, yay!
      if current_literal.regexp?
        # Regular expressions should include escape sequences in their
        # escaped form. On the other hand, escaped newlines are removed.
        current_literal.extend_string(tok.gsub("\\\n".freeze, ''.freeze), @ts, @te)
      else
        current_literal.extend_string(@escape || tok, @ts, @te)
      end
    end
   end
		end
	when 138 then
# line 723 "lib/parser/lexer.rl"
		begin

        diagnostic :fatal, :invalid_unicode_escape, nil, range(@escape_s - 1, p)
      		end
# line 734 "lib/parser/lexer.rl"
		begin

          diagnostic :fatal, :unterminated_unicode, nil, range(p - 1, p)
        		end
# line 839 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    current_literal = literal
    # Get the first character after the backslash.
    escaped_char = @source_buffer.slice(@escape_s).chr

    if current_literal.munge_escape? escaped_char
      # If this particular literal uses this character as an opening
      # or closing delimiter, it is an escape sequence for that
      # particular character. Write it without the backslash.

      if current_literal.regexp? && REGEXP_META_CHARACTERS.match(escaped_char)
        # Regular expressions should include escaped delimiters in their
        # escaped form, except when the escaped character is
        # a closing delimiter but not a regexp metacharacter.
        #
        # The backslash itself cannot be used as a closing delimiter
        # at the same time as an escape symbol, but it is always munged,
        # so this branch also executes for the non-closing-delimiter case
        # for the backslash.
        current_literal.extend_string(tok, @ts, @te)
      else
        current_literal.extend_string(escaped_char, @ts, @te)
      end
    else
      # It does not. So this is an actual escape sequence, yay!
      if current_literal.regexp?
        # Regular expressions should include escape sequences in their
        # escaped form. On the other hand, escaped newlines are removed.
        current_literal.extend_string(tok.gsub("\\\n".freeze, ''.freeze), @ts, @te)
      else
        current_literal.extend_string(@escape || tok, @ts, @te)
      end
    end
   end
		end
	when 323 then
# line 723 "lib/parser/lexer.rl"
		begin

        diagnostic :fatal, :invalid_unicode_escape, nil, range(@escape_s - 1, p)
      		end
# line 734 "lib/parser/lexer.rl"
		begin

          diagnostic :fatal, :unterminated_unicode, nil, range(p - 1, p)
        		end
# line 1707 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        value = @escape || tok(@ts + 1)

        if version?(18)
          if defined?(Encoding)
            emit(:tINTEGER, value.dup.force_encoding(Encoding::BINARY)[0].ord)
          else
            emit(:tINTEGER, value[0].ord)
          end
        else
          emit(:tCHARACTER, value)
        end

         @cs = 754; 	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 96 then
# line 766 "lib/parser/lexer.rl"
		begin

    @escape_s = p
    @escape   = nil
  		end
# line 760 "lib/parser/lexer.rl"
		begin

      diagnostic :fatal, :escape_eof, nil, range(p - 1, p)
    		end
# line 839 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    current_literal = literal
    # Get the first character after the backslash.
    escaped_char = @source_buffer.slice(@escape_s).chr

    if current_literal.munge_escape? escaped_char
      # If this particular literal uses this character as an opening
      # or closing delimiter, it is an escape sequence for that
      # particular character. Write it without the backslash.

      if current_literal.regexp? && REGEXP_META_CHARACTERS.match(escaped_char)
        # Regular expressions should include escaped delimiters in their
        # escaped form, except when the escaped character is
        # a closing delimiter but not a regexp metacharacter.
        #
        # The backslash itself cannot be used as a closing delimiter
        # at the same time as an escape symbol, but it is always munged,
        # so this branch also executes for the non-closing-delimiter case
        # for the backslash.
        current_literal.extend_string(tok, @ts, @te)
      else
        current_literal.extend_string(escaped_char, @ts, @te)
      end
    else
      # It does not. So this is an actual escape sequence, yay!
      if current_literal.regexp?
        # Regular expressions should include escape sequences in their
        # escaped form. On the other hand, escaped newlines are removed.
        current_literal.extend_string(tok.gsub("\\\n".freeze, ''.freeze), @ts, @te)
      else
        current_literal.extend_string(@escape || tok, @ts, @te)
      end
    end
   end
		end
	when 123 then
# line 766 "lib/parser/lexer.rl"
		begin

    @escape_s = p
    @escape   = nil
  		end
# line 760 "lib/parser/lexer.rl"
		begin

      diagnostic :fatal, :escape_eof, nil, range(p - 1, p)
    		end
# line 839 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    current_literal = literal
    # Get the first character after the backslash.
    escaped_char = @source_buffer.slice(@escape_s).chr

    if current_literal.munge_escape? escaped_char
      # If this particular literal uses this character as an opening
      # or closing delimiter, it is an escape sequence for that
      # particular character. Write it without the backslash.

      if current_literal.regexp? && REGEXP_META_CHARACTERS.match(escaped_char)
        # Regular expressions should include escaped delimiters in their
        # escaped form, except when the escaped character is
        # a closing delimiter but not a regexp metacharacter.
        #
        # The backslash itself cannot be used as a closing delimiter
        # at the same time as an escape symbol, but it is always munged,
        # so this branch also executes for the non-closing-delimiter case
        # for the backslash.
        current_literal.extend_string(tok, @ts, @te)
      else
        current_literal.extend_string(escaped_char, @ts, @te)
      end
    else
      # It does not. So this is an actual escape sequence, yay!
      if current_literal.regexp?
        # Regular expressions should include escape sequences in their
        # escaped form. On the other hand, escaped newlines are removed.
        current_literal.extend_string(tok.gsub("\\\n".freeze, ''.freeze), @ts, @te)
      else
        current_literal.extend_string(@escape || tok, @ts, @te)
      end
    end
   end
		end
	when 308 then
# line 766 "lib/parser/lexer.rl"
		begin

    @escape_s = p
    @escape   = nil
  		end
# line 760 "lib/parser/lexer.rl"
		begin

      diagnostic :fatal, :escape_eof, nil, range(p - 1, p)
    		end
# line 1707 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        value = @escape || tok(@ts + 1)

        if version?(18)
          if defined?(Encoding)
            emit(:tINTEGER, value.dup.force_encoding(Encoding::BINARY)[0].ord)
          else
            emit(:tINTEGER, value[0].ord)
          end
        else
          emit(:tCHARACTER, value)
        end

         @cs = 754; 	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 67 then
# line 1108 "lib/parser/lexer.rl"
		begin
 @sharp_s = p - 1 		end
# line 1111 "lib/parser/lexer.rl"
		begin
 emit_comment(@sharp_s, p == pe ? p - 2 : p) 		end
# line 486 "lib/parser/lexer.rl"
		begin

    # Record position of a newline for precise location reporting on tNL
    # tokens.
    #
    # This action is embedded directly into c_nl, as it is idempotent and
    # there are no cases when we need to skip it.
    @newline_s = p
  		end
	when 179 then
# line 1108 "lib/parser/lexer.rl"
		begin
 @sharp_s = p - 1 		end
# line 1111 "lib/parser/lexer.rl"
		begin
 emit_comment(@sharp_s, p == pe ? p - 2 : p) 		end
# line 1296 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1;		end
	when 194 then
# line 1108 "lib/parser/lexer.rl"
		begin
 @sharp_s = p - 1 		end
# line 1111 "lib/parser/lexer.rl"
		begin
 emit_comment(@sharp_s, p == pe ? p - 2 : p) 		end
# line 1314 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1;		end
	when 206 then
# line 1108 "lib/parser/lexer.rl"
		begin
 @sharp_s = p - 1 		end
# line 1111 "lib/parser/lexer.rl"
		begin
 emit_comment(@sharp_s, p == pe ? p - 2 : p) 		end
# line 1346 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1;		end
	when 228 then
# line 1108 "lib/parser/lexer.rl"
		begin
 @sharp_s = p - 1 		end
# line 1111 "lib/parser/lexer.rl"
		begin
 emit_comment(@sharp_s, p == pe ? p - 2 : p) 		end
# line 1483 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin  	begin
		 @cs = 754
		_goto_level = _again
		next
	end
  end
		end
	when 243 then
# line 1108 "lib/parser/lexer.rl"
		begin
 @sharp_s = p - 1 		end
# line 1111 "lib/parser/lexer.rl"
		begin
 emit_comment(@sharp_s, p == pe ? p - 2 : p) 		end
# line 1555 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1;		end
	when 255 then
# line 1108 "lib/parser/lexer.rl"
		begin
 @sharp_s = p - 1 		end
# line 1111 "lib/parser/lexer.rl"
		begin
 emit_comment(@sharp_s, p == pe ? p - 2 : p) 		end
# line 1576 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1;		end
	when 279 then
# line 1108 "lib/parser/lexer.rl"
		begin
 @sharp_s = p - 1 		end
# line 1111 "lib/parser/lexer.rl"
		begin
 emit_comment(@sharp_s, p == pe ? p - 2 : p) 		end
# line 1836 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1;		end
	when 341 then
# line 1108 "lib/parser/lexer.rl"
		begin
 @sharp_s = p - 1 		end
# line 1111 "lib/parser/lexer.rl"
		begin
 emit_comment(@sharp_s, p == pe ? p - 2 : p) 		end
# line 1860 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1;		end
	when 351 then
# line 1108 "lib/parser/lexer.rl"
		begin
 @sharp_s = p - 1 		end
# line 1111 "lib/parser/lexer.rl"
		begin
 emit_comment(@sharp_s, p == pe ? p - 2 : p) 		end
# line 1891 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1;		end
	when 371 then
# line 1108 "lib/parser/lexer.rl"
		begin
 @sharp_s = p - 1 		end
# line 1111 "lib/parser/lexer.rl"
		begin
 emit_comment(@sharp_s, p == pe ? p - 2 : p) 		end
# line 2169 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1;		end
	when 83 then
# line 1108 "lib/parser/lexer.rl"
		begin
 @sharp_s = p - 1 		end
# line 1111 "lib/parser/lexer.rl"
		begin
 emit_comment(@sharp_s, p == pe ? p - 2 : p) 		end
# line 2224 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1;		end
	when 414 then
# line 2000 "lib/parser/lexer.rl"
		begin
 @num_base = 10; @num_digits_s = @ts 		end
# line 2002 "lib/parser/lexer.rl"
		begin
 @num_suffix_s = p 		end
# line 619 "lib/parser/lexer.rl"
		begin
 @num_xfrm = lambda { |chars| emit(:tINTEGER,   chars) } 		end
	when 384 then
# line 2001 "lib/parser/lexer.rl"
		begin
 @num_base = 8;  @num_digits_s = @ts 		end
# line 2002 "lib/parser/lexer.rl"
		begin
 @num_suffix_s = p 		end
# line 619 "lib/parser/lexer.rl"
		begin
 @num_xfrm = lambda { |chars| emit(:tINTEGER,   chars) } 		end
	when 396 then
# line 2002 "lib/parser/lexer.rl"
		begin
 @num_suffix_s = p 		end
# line 619 "lib/parser/lexer.rl"
		begin
 @num_xfrm = lambda { |chars| emit(:tINTEGER,   chars) } 		end
# line 2003 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        digits = tok(@num_digits_s, @num_suffix_s)

        if digits.end_with? '_'.freeze
          diagnostic :error, :trailing_in_number, { :character => '_'.freeze },
                     range(@te - 1, @te)
        elsif digits.empty? && @num_base == 8 && version?(18)
          # 1.8 did not raise an error on 0o.
          digits = '0'.freeze
        elsif digits.empty?
          diagnostic :error, :empty_numeric
        elsif @num_base == 8 && (invalid_idx = digits.index(/[89]/))
          invalid_s = @num_digits_s + invalid_idx
          diagnostic :error, :invalid_octal, nil,
                     range(invalid_s, invalid_s + 1)
        end

        if version?(18, 19, 20)
          emit(:tINTEGER, digits.to_i(@num_base), @ts, @num_suffix_s)
          p = @num_suffix_s - 1
        else
          @num_xfrm.call(digits.to_i(@num_base))
        end
        	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 391 then
# line 2059 "lib/parser/lexer.rl"
		begin
 @num_suffix_s = p 		end
# line 625 "lib/parser/lexer.rl"
		begin
 @num_xfrm = lambda { |chars| emit(:tFLOAT,     Float(chars)) } 		end
# line 2062 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        digits = tok(@ts, @num_suffix_s)

        if version?(18, 19, 20)
          emit(:tFLOAT, Float(digits), @ts, @num_suffix_s)
          p = @num_suffix_s - 1
        else
          @num_xfrm.call(digits)
        end
        	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 388 then
# line 2060 "lib/parser/lexer.rl"
		begin
 @num_suffix_s = p 		end
# line 625 "lib/parser/lexer.rl"
		begin
 @num_xfrm = lambda { |chars| emit(:tFLOAT,     Float(chars)) } 		end
# line 2062 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        digits = tok(@ts, @num_suffix_s)

        if version?(18, 19, 20)
          emit(:tFLOAT, Float(digits), @ts, @num_suffix_s)
          p = @num_suffix_s - 1
        else
          @num_xfrm.call(digits)
        end
        	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 234 then
# line 1 "NONE"
		begin
 @te = p+1
		end
# line 486 "lib/parser/lexer.rl"
		begin

    # Record position of a newline for precise location reporting on tNL
    # tokens.
    #
    # This action is embedded directly into c_nl, as it is idempotent and
    # there are no cases when we need to skip it.
    @newline_s = p
  		end
# line 1480 "lib/parser/lexer.rl"
		begin
 @act = 66;		end
	when 33 then
# line 1 "NONE"
		begin
 @te = p+1
		end
# line 806 "lib/parser/lexer.rl"
		begin

    # After every heredoc was parsed, @herebody_s contains the
    # position of next token after all heredocs.
    if @herebody_s
      p = @herebody_s
      @herebody_s = nil
    end
  		end
# line 1475 "lib/parser/lexer.rl"
		begin
 @act = 65;		end
	when 44 then
# line 1 "NONE"
		begin
 @te = p+1
		end
# line 806 "lib/parser/lexer.rl"
		begin

    # After every heredoc was parsed, @herebody_s contains the
    # position of next token after all heredocs.
    if @herebody_s
      p = @herebody_s
      @herebody_s = nil
    end
  		end
# line 1525 "lib/parser/lexer.rl"
		begin
 @act = 73;		end
	when 70 then
# line 1 "NONE"
		begin
 @te = p+1
		end
# line 1111 "lib/parser/lexer.rl"
		begin
 emit_comment(@sharp_s, p == pe ? p - 2 : p) 		end
# line 1945 "lib/parser/lexer.rl"
		begin
 @act = 126;		end
	when 35 then
# line 1 "NONE"
		begin
 @te = p+1
		end
# line 1437 "lib/parser/lexer.rl"
		begin
 tm = p 		end
# line 1438 "lib/parser/lexer.rl"
		begin
 @act = 60;		end
	when 334 then
# line 1 "NONE"
		begin
 @te = p+1
		end
# line 1778 "lib/parser/lexer.rl"
		begin
 tm = p 		end
# line 1824 "lib/parser/lexer.rl"
		begin
 @act = 107;		end
	when 333 then
# line 1 "NONE"
		begin
 @te = p+1
		end
# line 1778 "lib/parser/lexer.rl"
		begin
 tm = p 		end
# line 1196 "lib/parser/lexer.rl"
		begin
 @act = 108;		end
	when 415 then
# line 1 "NONE"
		begin
 @te = p+1
		end
# line 2000 "lib/parser/lexer.rl"
		begin
 @num_base = 10; @num_digits_s = @ts 		end
# line 2003 "lib/parser/lexer.rl"
		begin
 @act = 133;		end
	when 107 then
# line 666 "lib/parser/lexer.rl"
		begin

    codepoint = @source_pts[p - 1]
    if (@escape = ESCAPES[codepoint]).nil?
      @escape = encode_escape(@source_buffer.slice(p - 1))
    end
  		end
# line 677 "lib/parser/lexer.rl"
		begin

    @escape = encode_escape(@escape[0].ord & 0x9f)
  		end
# line 681 "lib/parser/lexer.rl"
		begin

    @escape = encode_escape(@escape[0].ord | 0x80)
  		end
# line 839 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    current_literal = literal
    # Get the first character after the backslash.
    escaped_char = @source_buffer.slice(@escape_s).chr

    if current_literal.munge_escape? escaped_char
      # If this particular literal uses this character as an opening
      # or closing delimiter, it is an escape sequence for that
      # particular character. Write it without the backslash.

      if current_literal.regexp? && REGEXP_META_CHARACTERS.match(escaped_char)
        # Regular expressions should include escaped delimiters in their
        # escaped form, except when the escaped character is
        # a closing delimiter but not a regexp metacharacter.
        #
        # The backslash itself cannot be used as a closing delimiter
        # at the same time as an escape symbol, but it is always munged,
        # so this branch also executes for the non-closing-delimiter case
        # for the backslash.
        current_literal.extend_string(tok, @ts, @te)
      else
        current_literal.extend_string(escaped_char, @ts, @te)
      end
    else
      # It does not. So this is an actual escape sequence, yay!
      if current_literal.regexp?
        # Regular expressions should include escape sequences in their
        # escaped form. On the other hand, escaped newlines are removed.
        current_literal.extend_string(tok.gsub("\\\n".freeze, ''.freeze), @ts, @te)
      else
        current_literal.extend_string(@escape || tok, @ts, @te)
      end
    end
   end
		end
	when 133 then
# line 666 "lib/parser/lexer.rl"
		begin

    codepoint = @source_pts[p - 1]
    if (@escape = ESCAPES[codepoint]).nil?
      @escape = encode_escape(@source_buffer.slice(p - 1))
    end
  		end
# line 677 "lib/parser/lexer.rl"
		begin

    @escape = encode_escape(@escape[0].ord & 0x9f)
  		end
# line 681 "lib/parser/lexer.rl"
		begin

    @escape = encode_escape(@escape[0].ord | 0x80)
  		end
# line 839 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    current_literal = literal
    # Get the first character after the backslash.
    escaped_char = @source_buffer.slice(@escape_s).chr

    if current_literal.munge_escape? escaped_char
      # If this particular literal uses this character as an opening
      # or closing delimiter, it is an escape sequence for that
      # particular character. Write it without the backslash.

      if current_literal.regexp? && REGEXP_META_CHARACTERS.match(escaped_char)
        # Regular expressions should include escaped delimiters in their
        # escaped form, except when the escaped character is
        # a closing delimiter but not a regexp metacharacter.
        #
        # The backslash itself cannot be used as a closing delimiter
        # at the same time as an escape symbol, but it is always munged,
        # so this branch also executes for the non-closing-delimiter case
        # for the backslash.
        current_literal.extend_string(tok, @ts, @te)
      else
        current_literal.extend_string(escaped_char, @ts, @te)
      end
    else
      # It does not. So this is an actual escape sequence, yay!
      if current_literal.regexp?
        # Regular expressions should include escape sequences in their
        # escaped form. On the other hand, escaped newlines are removed.
        current_literal.extend_string(tok.gsub("\\\n".freeze, ''.freeze), @ts, @te)
      else
        current_literal.extend_string(@escape || tok, @ts, @te)
      end
    end
   end
		end
	when 318 then
# line 666 "lib/parser/lexer.rl"
		begin

    codepoint = @source_pts[p - 1]
    if (@escape = ESCAPES[codepoint]).nil?
      @escape = encode_escape(@source_buffer.slice(p - 1))
    end
  		end
# line 677 "lib/parser/lexer.rl"
		begin

    @escape = encode_escape(@escape[0].ord & 0x9f)
  		end
# line 681 "lib/parser/lexer.rl"
		begin

    @escape = encode_escape(@escape[0].ord | 0x80)
  		end
# line 1707 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        value = @escape || tok(@ts + 1)

        if version?(18)
          if defined?(Encoding)
            emit(:tINTEGER, value.dup.force_encoding(Encoding::BINARY)[0].ord)
          else
            emit(:tINTEGER, value[0].ord)
          end
        else
          emit(:tCHARACTER, value)
        end

         @cs = 754; 	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 105 then
# line 693 "lib/parser/lexer.rl"
		begin
 @escape = @source_buffer.slice(p - 1).chr 		end
# line 677 "lib/parser/lexer.rl"
		begin

    @escape = encode_escape(@escape[0].ord & 0x9f)
  		end
# line 681 "lib/parser/lexer.rl"
		begin

    @escape = encode_escape(@escape[0].ord | 0x80)
  		end
# line 839 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    current_literal = literal
    # Get the first character after the backslash.
    escaped_char = @source_buffer.slice(@escape_s).chr

    if current_literal.munge_escape? escaped_char
      # If this particular literal uses this character as an opening
      # or closing delimiter, it is an escape sequence for that
      # particular character. Write it without the backslash.

      if current_literal.regexp? && REGEXP_META_CHARACTERS.match(escaped_char)
        # Regular expressions should include escaped delimiters in their
        # escaped form, except when the escaped character is
        # a closing delimiter but not a regexp metacharacter.
        #
        # The backslash itself cannot be used as a closing delimiter
        # at the same time as an escape symbol, but it is always munged,
        # so this branch also executes for the non-closing-delimiter case
        # for the backslash.
        current_literal.extend_string(tok, @ts, @te)
      else
        current_literal.extend_string(escaped_char, @ts, @te)
      end
    else
      # It does not. So this is an actual escape sequence, yay!
      if current_literal.regexp?
        # Regular expressions should include escape sequences in their
        # escaped form. On the other hand, escaped newlines are removed.
        current_literal.extend_string(tok.gsub("\\\n".freeze, ''.freeze), @ts, @te)
      else
        current_literal.extend_string(@escape || tok, @ts, @te)
      end
    end
   end
		end
	when 131 then
# line 693 "lib/parser/lexer.rl"
		begin
 @escape = @source_buffer.slice(p - 1).chr 		end
# line 677 "lib/parser/lexer.rl"
		begin

    @escape = encode_escape(@escape[0].ord & 0x9f)
  		end
# line 681 "lib/parser/lexer.rl"
		begin

    @escape = encode_escape(@escape[0].ord | 0x80)
  		end
# line 839 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
    current_literal = literal
    # Get the first character after the backslash.
    escaped_char = @source_buffer.slice(@escape_s).chr

    if current_literal.munge_escape? escaped_char
      # If this particular literal uses this character as an opening
      # or closing delimiter, it is an escape sequence for that
      # particular character. Write it without the backslash.

      if current_literal.regexp? && REGEXP_META_CHARACTERS.match(escaped_char)
        # Regular expressions should include escaped delimiters in their
        # escaped form, except when the escaped character is
        # a closing delimiter but not a regexp metacharacter.
        #
        # The backslash itself cannot be used as a closing delimiter
        # at the same time as an escape symbol, but it is always munged,
        # so this branch also executes for the non-closing-delimiter case
        # for the backslash.
        current_literal.extend_string(tok, @ts, @te)
      else
        current_literal.extend_string(escaped_char, @ts, @te)
      end
    else
      # It does not. So this is an actual escape sequence, yay!
      if current_literal.regexp?
        # Regular expressions should include escape sequences in their
        # escaped form. On the other hand, escaped newlines are removed.
        current_literal.extend_string(tok.gsub("\\\n".freeze, ''.freeze), @ts, @te)
      else
        current_literal.extend_string(@escape || tok, @ts, @te)
      end
    end
   end
		end
	when 316 then
# line 693 "lib/parser/lexer.rl"
		begin
 @escape = @source_buffer.slice(p - 1).chr 		end
# line 677 "lib/parser/lexer.rl"
		begin

    @escape = encode_escape(@escape[0].ord & 0x9f)
  		end
# line 681 "lib/parser/lexer.rl"
		begin

    @escape = encode_escape(@escape[0].ord | 0x80)
  		end
# line 1707 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        value = @escape || tok(@ts + 1)

        if version?(18)
          if defined?(Encoding)
            emit(:tINTEGER, value.dup.force_encoding(Encoding::BINARY)[0].ord)
          else
            emit(:tINTEGER, value[0].ord)
          end
        else
          emit(:tCHARACTER, value)
        end

         @cs = 754; 	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 410 then
# line 1996 "lib/parser/lexer.rl"
		begin
 @num_base = 16; @num_digits_s = p 		end
# line 2002 "lib/parser/lexer.rl"
		begin
 @num_suffix_s = p 		end
# line 619 "lib/parser/lexer.rl"
		begin
 @num_xfrm = lambda { |chars| emit(:tINTEGER,   chars) } 		end
# line 2003 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        digits = tok(@num_digits_s, @num_suffix_s)

        if digits.end_with? '_'.freeze
          diagnostic :error, :trailing_in_number, { :character => '_'.freeze },
                     range(@te - 1, @te)
        elsif digits.empty? && @num_base == 8 && version?(18)
          # 1.8 did not raise an error on 0o.
          digits = '0'.freeze
        elsif digits.empty?
          diagnostic :error, :empty_numeric
        elsif @num_base == 8 && (invalid_idx = digits.index(/[89]/))
          invalid_s = @num_digits_s + invalid_idx
          diagnostic :error, :invalid_octal, nil,
                     range(invalid_s, invalid_s + 1)
        end

        if version?(18, 19, 20)
          emit(:tINTEGER, digits.to_i(@num_base), @ts, @num_suffix_s)
          p = @num_suffix_s - 1
        else
          @num_xfrm.call(digits.to_i(@num_base))
        end
        	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 404 then
# line 1997 "lib/parser/lexer.rl"
		begin
 @num_base = 10; @num_digits_s = p 		end
# line 2002 "lib/parser/lexer.rl"
		begin
 @num_suffix_s = p 		end
# line 619 "lib/parser/lexer.rl"
		begin
 @num_xfrm = lambda { |chars| emit(:tINTEGER,   chars) } 		end
# line 2003 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        digits = tok(@num_digits_s, @num_suffix_s)

        if digits.end_with? '_'.freeze
          diagnostic :error, :trailing_in_number, { :character => '_'.freeze },
                     range(@te - 1, @te)
        elsif digits.empty? && @num_base == 8 && version?(18)
          # 1.8 did not raise an error on 0o.
          digits = '0'.freeze
        elsif digits.empty?
          diagnostic :error, :empty_numeric
        elsif @num_base == 8 && (invalid_idx = digits.index(/[89]/))
          invalid_s = @num_digits_s + invalid_idx
          diagnostic :error, :invalid_octal, nil,
                     range(invalid_s, invalid_s + 1)
        end

        if version?(18, 19, 20)
          emit(:tINTEGER, digits.to_i(@num_base), @ts, @num_suffix_s)
          p = @num_suffix_s - 1
        else
          @num_xfrm.call(digits.to_i(@num_base))
        end
        	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 407 then
# line 1998 "lib/parser/lexer.rl"
		begin
 @num_base = 8;  @num_digits_s = p 		end
# line 2002 "lib/parser/lexer.rl"
		begin
 @num_suffix_s = p 		end
# line 619 "lib/parser/lexer.rl"
		begin
 @num_xfrm = lambda { |chars| emit(:tINTEGER,   chars) } 		end
# line 2003 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        digits = tok(@num_digits_s, @num_suffix_s)

        if digits.end_with? '_'.freeze
          diagnostic :error, :trailing_in_number, { :character => '_'.freeze },
                     range(@te - 1, @te)
        elsif digits.empty? && @num_base == 8 && version?(18)
          # 1.8 did not raise an error on 0o.
          digits = '0'.freeze
        elsif digits.empty?
          diagnostic :error, :empty_numeric
        elsif @num_base == 8 && (invalid_idx = digits.index(/[89]/))
          invalid_s = @num_digits_s + invalid_idx
          diagnostic :error, :invalid_octal, nil,
                     range(invalid_s, invalid_s + 1)
        end

        if version?(18, 19, 20)
          emit(:tINTEGER, digits.to_i(@num_base), @ts, @num_suffix_s)
          p = @num_suffix_s - 1
        else
          @num_xfrm.call(digits.to_i(@num_base))
        end
        	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 401 then
# line 1999 "lib/parser/lexer.rl"
		begin
 @num_base = 2;  @num_digits_s = p 		end
# line 2002 "lib/parser/lexer.rl"
		begin
 @num_suffix_s = p 		end
# line 619 "lib/parser/lexer.rl"
		begin
 @num_xfrm = lambda { |chars| emit(:tINTEGER,   chars) } 		end
# line 2003 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        digits = tok(@num_digits_s, @num_suffix_s)

        if digits.end_with? '_'.freeze
          diagnostic :error, :trailing_in_number, { :character => '_'.freeze },
                     range(@te - 1, @te)
        elsif digits.empty? && @num_base == 8 && version?(18)
          # 1.8 did not raise an error on 0o.
          digits = '0'.freeze
        elsif digits.empty?
          diagnostic :error, :empty_numeric
        elsif @num_base == 8 && (invalid_idx = digits.index(/[89]/))
          invalid_s = @num_digits_s + invalid_idx
          diagnostic :error, :invalid_octal, nil,
                     range(invalid_s, invalid_s + 1)
        end

        if version?(18, 19, 20)
          emit(:tINTEGER, digits.to_i(@num_base), @ts, @num_suffix_s)
          p = @num_suffix_s - 1
        else
          @num_xfrm.call(digits.to_i(@num_base))
        end
        	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 413 then
# line 2000 "lib/parser/lexer.rl"
		begin
 @num_base = 10; @num_digits_s = @ts 		end
# line 2002 "lib/parser/lexer.rl"
		begin
 @num_suffix_s = p 		end
# line 619 "lib/parser/lexer.rl"
		begin
 @num_xfrm = lambda { |chars| emit(:tINTEGER,   chars) } 		end
# line 2003 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        digits = tok(@num_digits_s, @num_suffix_s)

        if digits.end_with? '_'.freeze
          diagnostic :error, :trailing_in_number, { :character => '_'.freeze },
                     range(@te - 1, @te)
        elsif digits.empty? && @num_base == 8 && version?(18)
          # 1.8 did not raise an error on 0o.
          digits = '0'.freeze
        elsif digits.empty?
          diagnostic :error, :empty_numeric
        elsif @num_base == 8 && (invalid_idx = digits.index(/[89]/))
          invalid_s = @num_digits_s + invalid_idx
          diagnostic :error, :invalid_octal, nil,
                     range(invalid_s, invalid_s + 1)
        end

        if version?(18, 19, 20)
          emit(:tINTEGER, digits.to_i(@num_base), @ts, @num_suffix_s)
          p = @num_suffix_s - 1
        else
          @num_xfrm.call(digits.to_i(@num_base))
        end
        	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 383 then
# line 2001 "lib/parser/lexer.rl"
		begin
 @num_base = 8;  @num_digits_s = @ts 		end
# line 2002 "lib/parser/lexer.rl"
		begin
 @num_suffix_s = p 		end
# line 619 "lib/parser/lexer.rl"
		begin
 @num_xfrm = lambda { |chars| emit(:tINTEGER,   chars) } 		end
# line 2003 "lib/parser/lexer.rl"
		begin
 @te = p
p = p - 1; begin 
        digits = tok(@num_digits_s, @num_suffix_s)

        if digits.end_with? '_'.freeze
          diagnostic :error, :trailing_in_number, { :character => '_'.freeze },
                     range(@te - 1, @te)
        elsif digits.empty? && @num_base == 8 && version?(18)
          # 1.8 did not raise an error on 0o.
          digits = '0'.freeze
        elsif digits.empty?
          diagnostic :error, :empty_numeric
        elsif @num_base == 8 && (invalid_idx = digits.index(/[89]/))
          invalid_s = @num_digits_s + invalid_idx
          diagnostic :error, :invalid_octal, nil,
                     range(invalid_s, invalid_s + 1)
        end

        if version?(18, 19, 20)
          emit(:tINTEGER, digits.to_i(@num_base), @ts, @num_suffix_s)
          p = @num_suffix_s - 1
        else
          @num_xfrm.call(digits.to_i(@num_base))
        end
        	begin
		p += 1
		_goto_level = _out
		next
	end

       end
		end
	when 29 then
# line 1 "NONE"
		begin
 @te = p+1
		end
# line 806 "lib/parser/lexer.rl"
		begin

    # After every heredoc was parsed, @herebody_s contains the
    # position of next token after all heredocs.
    if @herebody_s
      p = @herebody_s
      @herebody_s = nil
    end
  		end
# line 1437 "lib/parser/lexer.rl"
		begin
 tm = p 		end
# line 1438 "lib/parser/lexer.rl"
		begin
 @act = 60;		end
	when 420 then
# line 1 "NONE"
		begin
 @te = p+1
		end
# line 2002 "lib/parser/lexer.rl"
		begin
 @num_suffix_s = p 		end
# line 619 "lib/parser/lexer.rl"
		begin
 @num_xfrm = lambda { |chars| emit(:tINTEGER,   chars) } 		end
# line 2035 "lib/parser/lexer.rl"
		begin
 @act = 135;		end
	when 416 then
# line 1 "NONE"
		begin
 @te = p+1
		end
# line 2000 "lib/parser/lexer.rl"
		begin
 @num_base = 10; @num_digits_s = @ts 		end
# line 2002 "lib/parser/lexer.rl"
		begin
 @num_suffix_s = p 		end
# line 619 "lib/parser/lexer.rl"
		begin
 @num_xfrm = lambda { |chars| emit(:tINTEGER,   chars) } 		end
# line 2035 "lib/parser/lexer.rl"
		begin
 @act = 135;		end
	when 386 then
# line 1 "NONE"
		begin
 @te = p+1
		end
# line 2001 "lib/parser/lexer.rl"
		begin
 @num_base = 8;  @num_digits_s = @ts 		end
# line 2002 "lib/parser/lexer.rl"
		begin
 @num_suffix_s = p 		end
# line 619 "lib/parser/lexer.rl"
		begin
 @num_xfrm = lambda { |chars| emit(:tINTEGER,   chars) } 		end
# line 2035 "lib/parser/lexer.rl"
		begin
 @act = 135;		end
# line 20787 "lib/parser/lexer.rb"
	end
	end
	end
	if _goto_level <= _again
	case _lex_to_state_actions[ @cs] 
	when 77 then
# line 1 "NONE"
		begin
 @ts = nil;		end
# line 20797 "lib/parser/lexer.rb"
	end

	if  @cs == 0
		_goto_level = _out
		next
	end
	p += 1
	if p != pe
		_goto_level = _resume
		next
	end
	end
	if _goto_level <= _test_eof
	if p == eof
	if _lex_eof_trans[ @cs] > 0
		_trans = _lex_eof_trans[ @cs] - 1;
		_goto_level = _eof_trans
		next;
	end
	end

	end
	if _goto_level <= _out
		break
	end
end
	end

# line 282 "lib/parser/lexer.rl"
    # %

    @p = p

    if @token_queue.any?
      @token_queue.shift
    elsif @cs == klass.lex_error
      [ false, [ '$error'.freeze, range(p - 1, p) ] ]
    else
      eof = @source_pts.size
      [ false, [ '$eof'.freeze,   range(eof, eof) ] ]
    end
  end

  protected

  def eof_codepoint?(point)
    [0x04, 0x1a, 0x00].include? point
  end

  def version?(*versions)
    versions.include?(@version)
  end

  def stack_pop
    @top -= 1
    @stack[@top]
  end

  if defined?(Encoding)
    def encode_escape(ord)
      ord.chr.force_encoding(@source_buffer.source.encoding)
    end
  else
    def encode_escape(ord)
      ord.chr
    end
  end

  def tok(s = @ts, e = @te)
    @source_buffer.slice(s...e)
  end

  def range(s = @ts, e = @te)
    Parser::Source::Range.new(@source_buffer, s, e)
  end

  def emit(type, value = tok, s = @ts, e = @te)
    token = [ type, [ value, range(s, e) ] ]

    @token_queue.push(token)

    @tokens.push(token) if @tokens

    token
  end

  def emit_table(table, s = @ts, e = @te)
    value = tok(s, e)

    emit(table[value], value, s, e)
  end

  def emit_do(do_block=false)
    if @cond.active?
      emit(:kDO_COND, 'do'.freeze)
    elsif @cmdarg.active? || do_block
      emit(:kDO_BLOCK, 'do'.freeze)
    else
      emit(:kDO, 'do'.freeze)
    end
  end

  def arg_or_cmdarg
    if @command_state
      self.class.lex_en_expr_cmdarg
    else
      self.class.lex_en_expr_arg
    end
  end

  def emit_comment(s = @ts, e = @te)
    if @comments
      @comments.push(Parser::Source::Comment.new(range(s, e)))
    end

    if @tokens
      @tokens.push([ :tCOMMENT, [ tok(s, e), range(s, e) ] ])
    end

    nil
  end

  def diagnostic(type, reason, arguments=nil, location=range, highlights=[])
    @diagnostics.process(
        Parser::Diagnostic.new(type, reason, arguments, location, highlights))
  end

  #
  # === LITERAL STACK ===
  #

  def push_literal(*args)
    new_literal = Literal.new(self, *args)
    @literal_stack.push(new_literal)

    if new_literal.words?
      if new_literal.interpolate?
        self.class.lex_en_interp_words
      else
        self.class.lex_en_plain_words
      end
    elsif new_literal.backslash_delimited?
      if new_literal.interpolate?
        self.class.lex_en_interp_backslash_delimited
      else
        self.class.lex_en_plain_backslash_delimited
      end
    else
      if new_literal.interpolate?
        self.class.lex_en_interp_string
      else
        self.class.lex_en_plain_string
      end
    end
  end

  def literal
    @literal_stack.last
  end

  def pop_literal
    old_literal = @literal_stack.pop

    @dedent_level = old_literal.dedent_level

    if old_literal.type == :tREGEXP_BEG
      # Fetch modifiers.
      self.class.lex_en_regexp_modifiers
    else
      self.class.lex_en_expr_end
    end
  end

  # Mapping of strings to parser tokens.

  PUNCTUATION = {
    '='   => :tEQL,     '&'   => :tAMPER2,  '|'   => :tPIPE,
    '!'   => :tBANG,    '^'   => :tCARET,   '+'   => :tPLUS,
    '-'   => :tMINUS,   '*'   => :tSTAR2,   '/'   => :tDIVIDE,
    '%'   => :tPERCENT, '~'   => :tTILDE,   ','   => :tCOMMA,
    ';'   => :tSEMI,    '.'   => :tDOT,     '..'  => :tDOT2,
    '...' => :tDOT3,    '['   => :tLBRACK2, ']'   => :tRBRACK,
    '('   => :tLPAREN2, ')'   => :tRPAREN,  '?'   => :tEH,
    ':'   => :tCOLON,   '&&'  => :tANDOP,   '||'  => :tOROP,
    '-@'  => :tUMINUS,  '+@'  => :tUPLUS,   '~@'  => :tTILDE,
    '**'  => :tPOW,     '->'  => :tLAMBDA,  '=~'  => :tMATCH,
    '!~'  => :tNMATCH,  '=='  => :tEQ,      '!='  => :tNEQ,
    '>'   => :tGT,      '>>'  => :tRSHFT,   '>='  => :tGEQ,
    '<'   => :tLT,      '<<'  => :tLSHFT,   '<='  => :tLEQ,
    '=>'  => :tASSOC,   '::'  => :tCOLON2,  '===' => :tEQQ,
    '<=>' => :tCMP,     '[]'  => :tAREF,    '[]=' => :tASET,
    '{'   => :tLCURLY,  '}'   => :tRCURLY,  '`'   => :tBACK_REF2,
    '!@'  => :tBANG,    '&.'  => :tANDDOT,
  }

  PUNCTUATION_BEGIN = {
    '&'   => :tAMPER,   '*'   => :tSTAR,    '**'  => :tDSTAR,
    '+'   => :tUPLUS,   '-'   => :tUMINUS,  '::'  => :tCOLON3,
    '('   => :tLPAREN,  '{'   => :tLBRACE,  '['   => :tLBRACK,
  }

  KEYWORDS = {
    'if'     => :kIF_MOD,      'unless'   => :kUNLESS_MOD,
    'while'  => :kWHILE_MOD,   'until'    => :kUNTIL_MOD,
    'rescue' => :kRESCUE_MOD,  'defined?' => :kDEFINED,
    'BEGIN'  => :klBEGIN,      'END'      => :klEND,
  }

  KEYWORDS_BEGIN = {
    'if'     => :kIF,          'unless'   => :kUNLESS,
    'while'  => :kWHILE,       'until'    => :kUNTIL,
    'rescue' => :kRESCUE,      'defined?' => :kDEFINED,
  }

  %w(class module def undef begin end then elsif else ensure case when
     for break next redo retry in do return yield super self nil true
     false and or not alias __FILE__ __LINE__ __ENCODING__).each do |keyword|
    KEYWORDS_BEGIN[keyword] = KEYWORDS[keyword] = :"k#{keyword.upcase}"
  end

  
# line 2239 "lib/parser/lexer.rl"

  # %
end
